Package: BFpack
Type: Package
Title: Flexible Bayes Factor Testing of Scientific Expectations
Version: 1.6.0
Authors@R: 
    c(person(given = c("Joris"),
             family = "Mulder",
             role = c("aut", "cre"),
             email = "j.mulder3@tilburguniversity.edu"),
      person(given = "Caspar",
             family = "van Lissa",
             role = c("aut", "ctb"),
             email = "c.j.vanlissa@uu.nl"),
      person(given = "Donald R.",
             family = "Williams",
             role = c("aut", "ctb"),
             email = "drwwilliams@ucdavis.edu"),
      person(given = "Xin",
             family = "Gu",
             role = c("aut", "ctb"),
             email = "guxin57@hotmail.com"),
      person(given = "Anton",
             family = "Olsson-Collentine",
             role = c("aut", "ctb"),
             email = "J.A.E.OlssonCollentine@tilburguniversity.edu"),
      person(given = "Florian",
             family = "Boeing-Messing",
             role = c("aut", "ctb"),
             email = "F.Boeing-Messing@uvt.nl"),
      person(given = "Jean-Paul",
             family = "Fox",
             role = c("aut", "ctb"),
             email = "g.j.a.fox@utwente.nl"),
      person(given = "Janosch",
             family = "Menke",
             role = c("ctb"),
             email = "janosch.menke@uni-muenster.de"),
      person(given = "Robbie",
             family = "van Aert",
             role = c("ctb"),
             email = "R.C.M.vanAert@tilburguniversity.edu"),
      person(given = "Barry",
             family = "Brown",
             role = "ctb"),
      person(given = "James",
             family = "Lovato",
             role = "ctb"),
      person(given = "Kathy",
             family = "Russell",
             role = "ctb"),
      person("Lapack 3.8",
             role = "ctb"),
      person(given = "Jack",
             family = "Dongarra",
             role = "ctb"),
      person(given = "Jim",
             family = "Bunch",
             role = "ctb"),
      person(given = "Cleve",
             family = "Moler",
             role = "ctb"),
      person(given = "Gilbert",
             family = "Stewart",
             role = "ctb"),
      person(given = "John",
             family = "Burkandt",
             role = "ctb"),
      person(given = "Ashwith",
             family = "Rego",
             role = "ctb"),
      person(given = "Alexander",
             family = "Godunov",
             role = "ctb"),
      person(given = "Alan",
             family = "Miller",
             role = "ctb"),
      person(given = "Jean-Pierre",
             family = "Moreau",
             role = "ctb"),
      person("The R Core Team", role = "cph")
             )
Maintainer: Joris Mulder <j.mulder3@tilburguniversity.edu>
Description: Implementation of default Bayes factors
    for testing statistical hypotheses under various statistical models. The package is
    intended for applied quantitative researchers in the
    social and behavioral sciences, medical research,
    and related fields. The Bayes factor tests can be
    executed for statistical models such as 
    univariate and multivariate normal linear models,
    correlation analysis, generalized linear models, special cases of 
    linear mixed models, survival models, relational
    event models. Parameters that can be tested are
    location parameters (e.g., group means, regression coefficients),
    variances (e.g., group variances), and measures of 
    association (e.g,. polychoric/polyserial/biserial/tetrachoric/product
    moments correlations), among others.
    Relevant references on the methodology The statistical underpinnings are
    described in
    O'Hagan (1995) <DOI:10.1111/j.2517-6161.1995.tb02017.x>,
    Mulder and Xin (2022) <DOI:10.1080/00273171.2021.1904809>,
    Mulder and Gelissen (2019) <DOI:10.1080/02664763.2021.1992360>,
    Mulder and Fox (2019) <DOI:10.1214/18-BA1115>,
    Boeing-Messing, van Assen, Hofman, Hoijtink, and Mulder (2017) <DOI:10.1037/met0000116>,
    Hoijtink, Mulder, van Lissa, and Gu (2018) <DOI:10.1037/met0000201>,
    Gu, Mulder, and Hoijtink (2018) <DOI:10.1111/bmsp.12110>,
    Hoijtink, Gu, and Mulder (2018) <DOI:10.1111/bmsp.12145>, and
    Hoijtink, Gu, Mulder, and Rosseel (2018) <DOI:10.1037/met0000187>. When using the
    packages, please refer to the package Mulder et al. (2021) <DOI:10.18637/jss.v100.i18>
    and the relevant methodological papers.
License: GPL (>= 3)
URL: https://github.com/jomulder/BFpack
BugReports: https://github.com/jomulder/BFpack/issues
Encoding: UTF-8
LazyData: true
NeedsCompilation: yes
RoxygenNote: 7.3.3
Depends: R (>= 3.0.0), bain
Imports: stats, utils, MASS, mvtnorm, pracma, lme4, sandwich, QRM,
        coda, metaBMA, methods, berryFunctions, ergm, Bergm
Suggests: testthat, polycor, survival, pscl, metafor, knitr, rmarkdown,
        lmtest
VignetteBuilder: knitr
Packaged: 2026-04-13 20:54:28 UTC; jorismulder
Author: Joris Mulder [aut, cre],
  Caspar van Lissa [aut, ctb],
  Donald R. Williams [aut, ctb],
  Xin Gu [aut, ctb],
  Anton Olsson-Collentine [aut, ctb],
  Florian Boeing-Messing [aut, ctb],
  Jean-Paul Fox [aut, ctb],
  Janosch Menke [ctb],
  Robbie van Aert [ctb],
  Barry Brown [ctb],
  James Lovato [ctb],
  Kathy Russell [ctb],
  Lapack 3.8 [ctb],
  Jack Dongarra [ctb],
  Jim Bunch [ctb],
  Cleve Moler [ctb],
  Gilbert Stewart [ctb],
  John Burkandt [ctb],
  Ashwith Rego [ctb],
  Alexander Godunov [ctb],
  Alan Miller [ctb],
  Jean-Pierre Moreau [ctb],
  The R Core Team [cph]
Repository: CRAN
Date/Publication: 2026-04-13 21:40:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-25 13:34:24 UTC; unix
Archs: BFpack.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(BF,bartlett_htest)
S3method(BF,bergm)
S3method(BF,coeftest)
S3method(BF,cor_test)
S3method(BF,coxph)
S3method(BF,default)
S3method(BF,ergm)
S3method(BF,glm)
S3method(BF,hetcor)
S3method(BF,htest)
S3method(BF,lm)
S3method(BF,lmerMod)
S3method(BF,mvt_test)
S3method(BF,polr)
S3method(BF,rma.uni)
S3method(BF,survreg)
S3method(BF,t_test)
S3method(BF,zeroinfl)
S3method(bartlett_test,default)
S3method(get_estimates,bartlett_htest)
S3method(get_estimates,bergm)
S3method(get_estimates,cor_test)
S3method(get_estimates,coxph)
S3method(get_estimates,ergm)
S3method(get_estimates,glm)
S3method(get_estimates,hetcor)
S3method(get_estimates,lm)
S3method(get_estimates,mvt_test)
S3method(get_estimates,polr)
S3method(get_estimates,survreg)
S3method(get_estimates,t_test)
S3method(get_estimates,zeroinfl)
S3method(plot,cor_test)
S3method(print,BF)
S3method(print,cor_test)
S3method(summary,BF)
S3method(summary,cor_test)
export(BF)
export(bartlett_test)
export(cor_test)
export(mvt_test)
import(bain)
importFrom(Bergm,bergm)
importFrom(MASS,fitdistr)
importFrom(MASS,ginv)
importFrom(QRM,fit.st)
importFrom(berryFunctions,is.error)
importFrom(coda,as.mcmc)
importFrom(coda,gelman.diag)
importFrom(coda,mcmc)
importFrom(coda,mcmc.list)
importFrom(ergm,ergmMPLE)
importFrom(lme4,VarCorr)
importFrom(lme4,getME)
importFrom(metaBMA,prior)
importFrom(methods,is)
importFrom(mvtnorm,dmvnorm)
importFrom(mvtnorm,dmvt)
importFrom(mvtnorm,pmvnorm)
importFrom(mvtnorm,pmvt)
importFrom(mvtnorm,rmvnorm)
importFrom(mvtnorm,rmvt)
importFrom(pracma,Rank)
importFrom(pracma,rref)
importFrom(sandwich,sandwich)
importFrom(stats,approxfun)
importFrom(stats,as.formula)
importFrom(stats,bartlett.test)
importFrom(stats,coef)
importFrom(stats,complete.cases)
importFrom(stats,cov)
importFrom(stats,dbeta)
importFrom(stats,density)
importFrom(stats,dlnorm)
importFrom(stats,dlogis)
importFrom(stats,dnorm)
importFrom(stats,dt)
importFrom(stats,lm)
importFrom(stats,median)
importFrom(stats,model.matrix)
importFrom(stats,nobs)
importFrom(stats,pchisq)
importFrom(stats,pnorm)
importFrom(stats,pt)
importFrom(stats,qgamma)
importFrom(stats,qnorm)
importFrom(stats,qt)
importFrom(stats,quantile)
importFrom(stats,rWishart)
importFrom(stats,rbeta)
importFrom(stats,rchisq)
importFrom(stats,rgamma)
importFrom(stats,rlnorm)
importFrom(stats,rnorm)
importFrom(stats,rt)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,terms)
importFrom(stats,var)
importFrom(stats,vcov)
importFrom(utils,getFromNamespace)
importFrom(utils,globalVariables)
useDynLib(BFpack, .registration = TRUE)

<img src="man/figures/logo_BFpack.png" width = 850 />

# 

[![CRAN
Version](http://www.r-pkg.org/badges/version/BFpack)](https://cran.r-project.org/package=BFpack)
[![Downloads](https://cranlogs.r-pkg.org/badges/BFpack)](https://cran.r-project.org/package=BFpack)
[![R-CMD-check](https://github.com/jomulder/BFpack/workflows/R-CMD-check/badge.svg)](https://github.com/jomulder/BFpack/actions)
<!-- Insert codecov badge here -->

[![Contributor
Covenant](https://img.shields.io/badge/Contributor%20Covenant-v2.0%20adopted-ff69b4.svg)](https://www.contributor-covenant.org/version/2/0/code_of_conduct.html)

The `R` package **BFpack** contains a set of functions for exploratory
hypothesis testing (e.g., equal vs negative vs postive) and confirmatory
hypothesis testing (with equality and/or order constraints) using Bayes
factors and posterior probabilities under commonly used statistical
models, including (but not limited to) Bayesian t testing, (M)AN(C)OVA,
multivariate/univariate linear regression, correlation analysis,
multilevel analysis, or generalized linear models (e.g., logistic
regression). The main function `BF` needs a fitted model (e.g., an
object of class `lm` for a linear regression model) and (optionally) the
argument `hypothesis`, a string which specifies a set of equality/order
constraints on the parameters. By applying the function
`get_estimates`on a fitted model, the names of the parameters are
returned on which constrained hypotheses can be formulated. Bayes
factors and posterior probabilities are computed for the hypotheses of
interest.

## Installation

Install the latest release version of `BFpack` from CRAN:

``` r
install.packages("BFpack")
```

The current developmental version can be installed with

``` r
if (!requireNamespace("remotes")) { 
  install.packages("remotes")   
}   
remotes::install_github("jomulder/BFpack")
```

## Example analyses

Below several example analyses are provided using **BFpack**. As input the main function `BF()` requires
a fitted model which which the necessary elements are extracted to compute Bayes factors and posterior
probabilities for the hypotheses.

### Bayesian t testing

#### Univariate t testing

First a classical one sample t test needs executed on the test value
\(mu = 5\) on the `therapeutic` data (part of `BFpack`). Here a right one-tailed classical test is executed:

``` r
ttest1 <- bain::t_test(therapeutic, alternative = "greater", mu = 5)
```

The `t_test` function is part of the **bain** package. The function is
equivalent to the standard `t.test` function with the addition that the
returned object contains additional output than the standard `t.test`
function.

Two default Bayes factors are implemented in **BFpack** to execute a t test: the fractional Bayes factor
(O'Hagan, 1995) and the prior adjusted fractional Bayes factor (Mulder, 2014). Both do not require a
prior to be manually specified as a default prior is implicitly constructed using a minimal fraction
of the data. The remaining fraction is used for hypothesis testing. The fractional Bayes factor
behaves similar as the JZS Bayes factor (Rouder et al., 2009, as implemented in the **BayesFactor** package)
for standard null hypothesis testing and the prior adjusted Bayes factor was specifically designed for
testing one-sided hypotheses. When setting the argument `BF.type=1` or `BF.type=2`, the fractional
Bayes factor and the prior adjusted fractional Bayes factor is used, respectively. The default choice
is `BF.type=2`.

To perform a Bayesian t test, the `BF` function is run on the fitted object. 

``` r
library(BFpack)
BF1 <- BF(ttest1)
```

This executes an exploratory test around the null value: `H1: mu = 5`
versus `H2: mu < 5` versus `H3: mu > 5` assuming equal prior
probabilities for `H1`, `H2`, and `H3` of 1/3. The output presents the
posterior probabilities for the three hypotheses.

The same test would be executed when the same hypotheses are explicitly
specified using the `hypothesis` argument.

``` r
hypothesis <- "mu = 5; mu < 5; mu > 5"
BF(ttest1, hypothesis = hypothesis)
```

When testing hypotheses via the `hypothesis` argument, the output also
presents an `Evidence matrix` containing the Bayes factors between the
hypotheses.

The argument `prior.hyp` can be used to specify different prior probabilities
for the hypotheses. For example, when the left one-tailed hypothesis is not possible
based on prior considerations (e.g., see [preprint](https://arxiv.org/abs/1911.07728)) while the precise (null) hypothesis and the right
one-tailed hypothesis are equally likely, the argument `prior.hyp` should be a vector
specifying the prior probabilities of the respective hypotheses
``` r
BF(ttest1, hypothesis = "mu = 5; mu < 5; mu > 5", prior.hyp = c(.5,0,.5))
```

#### Multivariate t testing

Bayesian multivariate t tests can be executed by first fitting a multivariate (regression) model using the `lm` function, and subsequently, the means of the dependent variables (or other coefficients) in the model can be tested using the `BF()` function. Users have to be aware that the (adjusted) means are modeled using intercepts which are named `(Intercept)` by default by `lm` while the `hypothesis` argument in `BF()` does not allow effect names that include brackets (i.e., `(` or `)`). To circumvent this, one can create a vector of 1s, with name (say) `ones`, which can replace the intercept but results in an equivalent model.

For example, let us consider a multivariate normal model for the dependent variables `Superficial`, `Middle`, and `Deep` in the `fmri` data set:

``` r
fmri1 <- cbind(fmri,ones=1)
mlm1 <- lm(cbind(Superficial,Middle,Deep) ~ -1 + ones, data = fmri1)
```

Next, we can (for instance) test whether all means are equal to 0 (`H1`), whether all means are positive (`H2`), or neither (`complement`):

``` r
BFmlm1 <- BF(mlm1,
             hypothesis="ones_on_Superficial=ones_on_Middle=ones_on_Deep=0;
                        (ones_on_Superficial,ones_on_Middle,ones_on_Deep)>0",
             complement = TRUE)
```

### Analysis of variance

First an analysis of variance (ANOVA) model is fitted using the `aov`
fuction in `R`.

``` r
aov1 <- aov(price ~ anchor * motivation, data = tvprices)
```

Next a Bayesian test can be performed on the fitted object.

``` r
BF(aov1)
```

By default posterior probabilities are computed of whether main effects
and interaction effects are present. Alternative constrained hypotheses
can be tested on the model parameters `get_estimates(aov1)`.

Similar as for the Bayesian t test, two default Bayes factors are implemented in **BFpack** for (multivariate)
analysis of (co)variance: the fractional Bayes factor (O'Hagan, 1995) and the prior adjusted fractional Bayes factor (Mulder, 2014, Mulder and Gu, 2022). The default choice is `BF.type=2`, which uses the prior adjusted fractional
Bayes factor.


### Univariate/Multivariate multiple regression

As an example we consider the `fmri` data (McGuigin et al, 2020) as discussed in Mulder et al. (2021). First, a classical linear regression is fitted with dependent variable `Deep` and predictor variables `Face` and `Vehicle`:

``` r
lm1 <- lm(Deep ~ Face + Vehicle, data = fmri)
```

When not using the `hypothesis` argument, Bayes factors and posterior probabilities are given of whether each predictor
has a zero, negative, or positive effect (assuming equal prior probabilities) against the full model:

``` r
BF(lm1)
```

As motivated in Mulder et al. (2021), it was expected that `Face` had a negative effect on `Deep` and `Vehicle` had a positive effect on `Deep`. This (combined) one-sided hypothesis can be tested against its complement according to

``` r
BF(lm1, hypothesis = "Face < 0 < Vehicle")
```

The hypothesis of interest receives clear support from the data.

In a multivariate multiple regression model, hypotheses can be tested on the
effects on the same dependent variable, and on effects across different
dependent variables. The name of an effect is constructed as the name of
the predictor variable and the dependent variable separated by `_on_`.
Testing hypotheses with both constraints within a dependent variable and
across dependent variables makes use of a Monte Carlo estimate which may
take a few seconds.

``` r
lm2 <- lm(cbind(Superficial, Middle, Deep) ~ Face + Vehicle,
              data = fmri)
constraint2 <- "Face_on_Deep = Face_on_Superficial = Face_on_Middle < 0 <
     Vehicle_on_Deep = Vehicle_on_Superficial = Vehicle_on_Middle;
     Face_on_Deep < Face_on_Superficial = Face_on_Middle < 0 < Vehicle_on_Deep =
     Vehicle_on_Superficial = Vehicle_on_Middle"
set.seed(123)
BF3 <- BF(lm2, hypothesis = constraint2)
summary(BF3)
```

Finally note that for (multivariate) multiple regression again note that two default Bayes factors are implemented in **BFpack**: the fractional Bayes factor (O'Hagan, 1995) and the prior adjusted fractional Bayes factor (Mulder, 2014; Mulder and Gu, 2022) which can be chosen using the argument `BF.type=1` and `BF.type=2`, respectively. The default
choice is the prior adjusted fractional Bayes factor. This criterion was specifically designed for testing one-sided
and order constrained hypotheses.


### Logistic regression

An example hypothesis test is consdered under a logistic regression
model. First a logistic regression model is fitted using the `glm`
function

``` r
fit_glm <- glm(sent ~ ztrust + zfWHR + zAfro + glasses + attract + maturity +
               tattoos, family = binomial(), data = wilson)
```

The names of the regression coefficients on which constrained hypotheses
can be formualted can be extracted using the `get_estimates` function.

``` r
get_estimates(fit_glm)
```

Two different hypotheses are formulated with competing equality and/or
order constraints on the parameters of interest. These hypotheses are
motivated in Mulder et al. (2019)

``` r
BF_glm <- BF(fit_glm, hypothesis = "ztrust > (zfWHR, zAfro) > 0;
             ztrust > zfWHR = zAfro = 0")
summary(BF_glm)
```

By calling the `summary` function on the output object of class `BF`,
the results of the exploratory tests are presented of whether each
separate parameter is zero, negative, or positive, and the results of
the confirmatory test of the hypotheses under the `hypothesis` argument
are presented. When the hypotheses do not cover the complete parameter
space, by default the complement hypothesis is added which covers the
remaining parameter space that is not covered by the constraints under
the hypotheses of interest. In the above example, the complement
hypothesis covers the parameter space where neither `"ztrust > (zfWHR,
zAfro) > 0"` holds, nor where `"ztrust > zfWHR = zAfro = 0"` holds.

The Bayes factors and posterior posterior probabilities are based on the
approximated adjusted default Bayes factor (Gu et al., 2018).


### Correlation analysis

Bayes factors and posterior posterior probabilities among constrained hypotheses on measures of association
are computed using uniform prior for the correlations (Mulder and Gelissen, 2023).
By default `BF` performs exhaustice tests of whether the separate
correlations are zero, negative, or positive. The name of the
correlations is constructed using the names of the variables separated
by `_with_`. To compute Bayes factors and posterior probabilities, first
the unconstrained model needs to be fit using the `cor_test()` function. The
resulting object can be added to the `BF()` function:

``` r
set.seed(123)
cor1 <- cor_test(memory[,1:3])
BF1 <- BF(cor1)
print(BF1)
```

Constraints can also be tested between correlations, e.g., whether all
correlations are equal and positive versus an unconstrained complement. The function
`get_estimates()` gives the names of the correlations on which constrained hypotheses
can be formulated:

``` r
get_estimates(cor1)
BF2 <- BF(cor1, hypothesis = "Del_with_Im = Wmn_with_Im = Wmn_with_Del > 0")
print(BF2)
```

Depending on the class of the variables (`numeric`, `ordered`, `factor` with 2 levels), biserial,
polyserial, polychoric, tetrachoric, or product-moment correlations are tested. As an illustration
of these other types of measures of association, we change the measurement levels of a subset of
the `mtcars` data:

``` r
mtcars_test <- mtcars[,c(1,2,9,10)]
mtcars_test[,2] <- as.ordered(mtcars_test[,2])
mtcars_test[,3] <- as.factor(mtcars_test[,3])
mtcars_test[,4] <- as.integer(mtcars_test[,4])
```

To compute the Bayes factors and posterior probabilities, again we first the full unconstrained model
using the `cor_test()` function. The resulting object is placed in the `BF()` function to obtain the
Bayes factors and posterior proabilities:

``` r
cor2 <- cor_test(mtcars_test)
BF2 <- BF(cor2, hypothesis = "0 < am_with_mpg = gear_with_mpg")
print(BF2)
```


### Running `BF` on a named vector

The input for the `BF` function can also be a named vector containing
the estimates of the parameters of interest. In this case the error
covariance matrix of the estimates is also needed via the `Sigma`
argument, as well as the sample size that was used for obtaining the
estimates via the `n` argument. Bayes factors are then computed using
Gaussian approximations of the likelihood (and posterior), similar as in
classical Wald test.

We illustrate this for a Poisson regression model

``` r
poisson1 <- glm(formula = breaks ~ wool + tension, data = datasets::warpbreaks,
             family = poisson)
```

The estimates, the error covariance matrix, and the sample size are
extracted from the fitted model

``` r
estimates <- poisson1$coefficients
covmatrix <- vcov(poisson1)
samplesize <- nobs(poisson1)
```

Constrained hypotheses on the parameters `names(estimates)` can then be
tested as follows

``` r
BF1 <- BF(estimates, Sigma = covmatrix, n = samplesize, hypothesis = 
  "woolB > tensionM > tensionH; woolB = tensionM = tensionH")
```

Note that the same hypothesis test would be executed when calling

``` r
BF2 <- BF(poisson1, hypothesis = "woolB > tensionM > tensionH;
          woolB = tensionM = tensionH")
```

because the same Bayes factor is used when running `BF` on an object of
class `glm` (see `Method: Bayes factor using Gaussian approximations`
when calling `print(BF11)` and `print(BF2)`).

The Bayes factors and posterior posterior probabilities on named vectors
are based on the adjusted default Bayes factor using Gaussian approximations
(Gu et al., 2018).


## Citing **BFpack**

You can cite the package and the paper using the following reference

**Software paper**
> Mulder, J., Williams, D. R., Gu, X., Olsson-Collentine, A., Tomarken,
> A., Böing-Messing, F., Hoijtink, H., . . . van Lissa  (2021).
> BFpack: Flexible Bayes factor testing of scientific theories in R.
> Journal of Statistical Software, 100(18).
> Retrieved from <https://arxiv.org/abs/1911.07728>

**Software package**
> Mulder, J., van Lissa, C., Gu, X., Olsson-Collentine, A.,
> Boeing-Messing, F., Williams, D. R., Fox, J.-P., Menke, J., et
> al. (2020). BFpack: Flexible Bayes Factor Testing of Scientific
> Expectations. (Version 1.3.0) \[R package\].
> <https://CRAN.R-project.org/package=BFpack>

Other references with technical details of the methodology, please see

**Bayes factors for (multivariate) t tests, (M)AN(C)OVA, (multivariate) regression**
> Mulder, J. (2014). Prior adjusted default Bayes factors for testing (in)equality
> constrained hypotheses. Computational Statistics and Data Analysis, 71, 448–463.
> <https://doi.org/10.1016/j.csda.2013.07.017>

> Mulder, J. & Gu, X. (2022) Bayesian Testing of Scientific Expectations
> under Multivariate Normal Linear Models, Multivariate Behavioral Research,
57:5, 767-783.
> <https://doi.org/10.1080/00273171.2021.1904809>

**Bayes factors for testing measures of association (e.g., correlations)** 
> Mulder, J., & Gelissen, J. P. (2023). Bayes factor testing of equality
> and order constraints on measures of association in social research.
> Journal of Applied Statistics, 50(2), 315-351.
> <https://doi.org/10.1080/02664763.2021.1992360>

**Default Bayes factors using Gaussian approximations**
> Gu. X., Mulder, J., & Hoijtink, J. (2018). Approximated adjusted
> fractional Bayes factors: A general method for testing informative
> hypotheses. British Journal of Mathematical and Statistical
> Psychology.
> <https://doi.org/10.1111/bmsp.12110>

**Bayes factors under exponential random graphs**
> Mulder, J., Friel, N., & Leifeld, P. (2023). Bayesian Testing of Scientific
> Expectations Under Exponential Random Graph Models.
> <https://doi.org/10.48550/arXiv.2304.14750>

**Bayes factors of intraclass correlations**
> Mulder, J., & Fox, J.-P. (2019). Bayes Factor Testing of Multiple
> Intraclass Correlations. Bayesian Analysis. 14(2), 521-552.
> <https://doi.org/10.1214/18-BA1115>

**Bayes factors for testing group variances**
> Böing-Messing, F., van Assen, M. A. L. M., Hofman, A. D., Hoijtink, H., 
> & Mulder, J. (2017). Bayesian evaluation of constrained hypotheses on
> variances of multiple independent groups. Psychological Methods, 22(2), 
> 262–287.
> <https://doi.org/10.1037/met0000116>

> Böing-Messing, F. & Mulder, J. (2018). Automatic Bayes factors for testing
> equality-and inequality-constrained hypotheses on variances. Psychometrika,
> 83, 586–617.
> <https://link.springer.com/article/10.1007/s11336-018-9615-z>

**Bayes factors for meta-analyses**
> Van Aert R.C.M. & Mulder, J. (2022). Bayesian hypothesis testing and 
> estimation under the marginalized random-effects meta-analysis model.
> Psychonomic Bulletin & Review, 29, 55–69.
> <https://doi.org/10.3758/s13423-021-01918-9>


## Contributing and Contact Information

If you have ideas, please get involved. You can contribute by opening an
issue on GitHub, or sending a pull request with proposed features.

  - File a GitHub issue [here](https://github.com/jomulder/BFpack)
  - Make a pull request [here](https://github.com/jomulder/BFpack/pulls)

By participating in this project, you agree to abide by the Contributor
Code of Conduct v2.0.

#' Student t approximations of Fisher transformed correlations
#'
#' Approximated degrees of freedom and approximated scale of the Fisher transformed
#' correlations depending on the dimension of the vector of dependent variables P
#' based on a joint uniform prior.
#'
#' \tabular{lll}{
#'    \strong{nu} \tab \code{numeric} \tab Approximated degrees of freedom\cr
#'    \strong{sigma} \tab \code{numeric} \tab Approximated scale\cr
#'    \strong{P} \tab \code{integer} \tab Dimension of vector of dependent variables\cr
#' }
#' @docType data
#' @keywords datasets
#' @name Fcor
#' @usage data(Fcor)
#' @format A data.frame with 3 columns.
NULL

check_vcov <- function(x){
  if (!isTRUE(all.equal(x, t(x))) || any(diag(x) < 0)){
    saveRDS(x, "c:/git_repositories/BFpack/erordump.RData")
    stop(sQuote("sigma"), " is not a covariance matrix")
  }
}

process.prior.hyp.explo <- function(prior_hyp_explo, model){

  if(sum(class(model)=="bartlett_htest")>0){
    if(is.null(prior_hyp_explo)){
      prior_hyp_explo <- list(rep(.5,2))
    }else{
      if(!is.list(prior_hyp_explo)){
        prior_hyp_explo <- list(prior_hyp_explo)
      }else{
        prior_hyp_explo <- list(prior_hyp_explo[[1]])
      }
      if(length(prior_hyp_explo[[1]])!=2){
        stop("For an object of class 'bartlett_htest', the argument 'prior_hyp_explo' should be a vector
             of length 2 of the prior probabilities for the hypotheses of homogeneity of variances and an
             unconstrained alternative. Or use the default ('NULL') which implies equal prior probabilities.")
      }
    }
  }else{
    if(is.null(prior_hyp_explo)){ # then equal prior probabilities
      if(sum(class(model)=="aov")>0){
        prior_hyp_explo <- list(c(.5,.25,.25),rep(1,2),rep(1,2))
      }else{
        prior_hyp_explo <- list(c(.5,.25,.25))
      }
    }else{
      if(!is.list(prior_hyp_explo)){#then no error if it is a vector of length 3 for testing individual parameters
        if(length(prior_hyp_explo)==3){
          if(sum(class(model)=="aov")>0){
            prior_hyp_explo <- list(prior_hyp_explo,rep(1,2),rep(1,2))
          }else{
            prior_hyp_explo <- list(prior_hyp_explo)
          }
        }else{
          stop("The argument 'prior_hyp_explo' must be a vector of length three specifying the
             prior probabilities of a zero, negative, or positive parameter. In the case of an object of
             class 'aov' then 'prior_hyp_explo' can be a list of three elements containing the three prior
             probabilities of the exploratory test of the parameters, the two prior
             probabilities for testing the main effects, and the two prior probabilities for testing the
             interaction effects (check the documentation: ?BF).")
        }
      }else{
        if(length(prior_hyp_explo)==1){
          if(is.null(prior_hyp_explo[[1]])){
            prior_hyp_explo <- list(c(.5,.25,.25),rep(1,2),rep(1,2))
          }else{
            if(length(prior_hyp_explo[[1]]!=3)){
              stop("Specify three prior probabilities for the exploratory test of a zero, negative, or positive
                 effect or use the default 'NULL' implying equal prior probabilities (check the documentation: ?BF).")
            }
            prior_hyp_explo <- lapply(prior_hyp_explo,function(x){x/sum(x)})
          }
        }else{
          if(sum(class(model)=="aov")>0){

            if(length(prior_hyp_explo)==2){
              if(is.null(prior_hyp_explo[[1]])){
                prior_hyp_explo[[1]] <- c(.5,.25,.25)
              }else{
                if(length(prior_hyp_explo[[1]]!=3)){
                  stop("Specify three prior probabilities for the exploratory test of a zero, negative, or positive
                 effect or use the default 'NULL' implying equal prior probabilities (check the documentation: ?BF).")
                }
              }
              if(is.null(prior_hyp_explo[[2]])){
                prior_hyp_explo[[2]] <- rep(1,2)
              }else{
                if(length(prior_hyp_explo[[2]]!=2)){
                  stop("Specify two prior probabilities for the exploratory testing of main effects
                   or use the default 'NULL' implying equal prior probabilities
                   (check the documentation: ?BF).")
                }
              }
              prior_hyp_explo[[3]] <- rep(1,2)
            }
            if(length(prior_hyp_explo)==3){
              if(is.null(prior_hyp_explo[[1]])){
                prior_hyp_explo[[1]] <- c(.5,.25,.25)
              }else{
                if(length(prior_hyp_explo[[1]])!=3){
                  stop("Specify three prior probabilities for the exploratory test of a zero, negative, or positive
                 effect or use the default 'NULL' implying equal prior probabilities
                   (check the documentation: ?BF).")
                }
              }
              if(is.null(prior_hyp_explo[[2]])){
                prior_hyp_explo[[2]] <- rep(1,2)
              }else{
                if(length(prior_hyp_explo[[2]])!=2){
                  stop("Specify two prior probabilities for the exploratory testing of main effects
                   or use the default 'NULL' implying equal prior probabilities
                   (check the documentation: ?BF).")
                }
              }
              if(is.null(prior_hyp_explo[[3]])){
                prior_hyp_explo[[3]] <- rep(1,2)
              }else{
                if(length(prior_hyp_explo[[3]])!=2){
                  stop("Specify two prior probabilities for the exploratory testing of interaction effects
                   or use the default 'NULL' implying equal prior probabilities
                   (check the documentation: ?BF).")
                }
              }
            }
            if(length(prior_hyp_explo)>3){
              stop("Use a list of three vectors or use the default 'NULL' implying equal prior probabilities
               (check the documentation: ?BF).")
            }
          }else{
            if(is.null(prior_hyp_explo[[1]])){
              prior_hyp_explo <- list(c(.5,.25,.25))
            }else{
              if(length(prior_hyp_explo[[1]])!=3){
                stop("Specify three prior probabilities for the exploratory test of a zero, negative, or positive
                 effect or use the default 'NULL' implying equal prior probabilities (check the documentation: ?BF).")
              }
              prior_hyp_explo <- list(prior_hyp_explo[[1]]/sum(prior_hyp_explo[[1]]))
            }
          }
        }
      }
    }
    prior_hyp_explo <- lapply(prior_hyp_explo,function(x)x/sum(x))
  }

  return(prior_hyp_explo)

}

# from the output of the constraints in 'parse_hypothesis' create lists for the equality and order matrices
make_RrList <- function(parse_hyp){
  numhyp <- length(parse_hyp$hyp_mat)
  RrE <- lapply(1:numhyp,function(h){
    qE <- parse_hyp$n_constraints[h*2-1]
    if(qE==1){
      RrE_h <- t(as.matrix(parse_hyp$hyp_mat[[h]][1:qE,]))
    }else if(qE>1){
      RrE_h <- parse_hyp$hyp_mat[[h]][1:qE,]
    }else {RrE_h=NULL}
    RrE_h
  })
  RrO <- lapply(1:numhyp,function(h){
    qE <- parse_hyp$n_constraints[h*2-1]
    qO <- parse_hyp$n_constraints[h*2]
    if(qO==1){
      RrO_h <- t(as.matrix(parse_hyp$hyp_mat[[h]][qE+1:qO,]))
    }else if(qO>1){
      RrO_h <- parse_hyp$hyp_mat[[h]][qE+1:qO,]
    }else {RrO_h=NULL}
    RrO_h
  })
  return(list(RrE,RrO))
}

# from the output of the constraints in 'parse_hypothesis' create lists for the equality and order matrices
# different format parse_hyp object
make_RrList2 <- function(parse_hyp2){
  numhyp <- length(parse_hyp2$original_hypothesis)
  qE <- parse_hyp2$n_constraints[(0:(numhyp-1))*2+1]
  qO <- parse_hyp2$n_constraints[(1:numhyp)*2]
  RrE <- lapply(1:numhyp,function(h){
    startcon <- sum(qE[1:h]+qO[1:h])-qE[h]-qO[h]
    if(qE[h]==1){
      RrE_h <- t(as.matrix(parse_hyp2$hyp_mat[startcon+1:qE[h],]))
    }else if(qE[h]>1){
      RrE_h <- parse_hyp2$hyp_mat[startcon+1:qE[h],]
    }else {RrE_h=NULL}
    RrE_h
  })
  RrO <- lapply(1:numhyp,function(h){
    startcon <- sum(qE[1:h]+qO[1:h])-qE[h]-qO[h]
    if(qO[h]==1){
      RrO_h <- t(as.matrix(parse_hyp2$hyp_mat[startcon+qE[h]+1:qO[h],]))
    }else if(qO[h]>1){
      RrO_h <- parse_hyp2$hyp_mat[startcon+qE[h]+1:qO[h],]
    }else {RrO_h=NULL}
    RrO_h
  })
  return(list(RrE,RrO))
}

#for checking whether constraints are conflicting replace interval constraints by equality constraints
interval_RrStack <- function(RrStack){
  q1 <- nrow(RrStack)
  q2 <- ncol(RrStack)
  RrStack_out <- RrStack
  if(q1 > 1){
    row1 <- 1
    while(row1 < q1){
      for(row2 in (row1+1):q1){
        #        print(row2)
        if(sum(abs(RrStack_out[row1,-q2] + RrStack_out[row2,-q2]))==0){ # && RrStack_out[row1,q2]!=RrStack_out[row2,q2] ){
          #together row1 and row2 imply an interval constraint
          whichcol <- abs(RrStack_out[row1,-q2])!=0
          whichcol1 <- which(whichcol)
          if(sum(whichcol)==1){
            welkpos <- ifelse(RrStack_out[row1,c(whichcol,F)]>0,row1,row2)
            welkneg <- ifelse(RrStack_out[row1,c(whichcol,F)]<0,row1,row2)
            lb <- RrStack_out[welkpos,q2]
            ub <- -RrStack_out[welkneg,q2]
            RrStack_out[row1,] <- RrStack_out[welkpos,]
            RrStack_out[row1,q2] <- (ub+lb)/2
            RrStack_out <- RrStack_out[-row2,]
            q1 <- q1 - 1
          }else{
            RrStack_out[row1,q2] <- 0
            RrStack_out <- RrStack_out[-row2,]
            q1 <- q1 - 1
          }
          break
        }
      }
      row1 <- row1 + 1
    }
  }
  if(is.matrix(RrStack_out)==F){
    RrStack_out <- t(RrStack_out)
  }
  return(RrStack_out)
}

params_in_hyp <- function(hyp){
  params_in_hyp <- trimws(unique(strsplit(hyp, split = "[ =<>,\\(\\);&\\*+-]+", perl = TRUE)[[1]]))
  params_in_hyp <- params_in_hyp[!sapply(params_in_hyp, grepl, pattern = "^[0-9]*\\.?[0-9]+$")]
  params_in_hyp[grepl("^[a-zA-Z]", params_in_hyp)]
}

! rngfuncs.f90
module rngfuncs

implicit none

public

! See the R header `R_ext/Random.h` for details
interface
    ! double unif_rand(void);
    function unif_rand() bind(c,name="unif_rand")
        use, intrinsic :: iso_c_binding, only: c_double
        real(c_double) :: unif_rand
    end function

    ! void GetRNGstate(void);
    subroutine getrngstate() bind(c,name="GetRNGstate")
    end subroutine

    ! void PutRNGstate(void);
    subroutine putrngstate() bind(c,name="PutRNGstate")
    end subroutine

  end interface

end module


module rkinds0
   use, intrinsic :: iso_c_binding !c_int c_double
   ! use, intrinsic :: iso_fortran_env !int32 real64
   private
   integer, parameter, public :: rint = c_int   ! Using int32 from iso_fortran_env
   integer, parameter, public :: rdp = c_double   ! Using real64 from iso_fortran_env
   ! Using real64 from iso_fortran_env
end module


module corr_slice_mod
  use, intrinsic :: iso_c_binding, only: c_double
  use rkinds0, only: rdp
  use rngfuncs, only: unif_rand
  implicit none
  private
  public :: slice_update_corr_LKJ, init_y_from_R

  interface
    subroutine dpotrf(uplo, n, a, lda, info)
      import :: c_double
      character(len=1), intent(in) :: uplo
      integer, intent(in) :: n, lda
      integer, intent(out) :: info
      real(c_double), intent(inout) :: a(lda,*)
    end subroutine dpotrf

    subroutine dpotrs(uplo, n, nrhs, a, lda, b, ldb, info)
      import :: c_double
      character(len=1), intent(in) :: uplo
      integer, intent(in) :: n, nrhs, lda, ldb
      integer, intent(out) :: info
      real(c_double), intent(in) :: a(lda,*)
      real(c_double), intent(inout) :: b(ldb,*)
    end subroutine dpotrs

    subroutine dpotri(uplo, n, a, lda, info)
      import :: c_double
      character(len=1), intent(in) :: uplo
      integer, intent(in) :: n, lda
      integer, intent(out) :: info
      real(c_double), intent(inout) :: a(lda,*)
    end subroutine dpotri
  end interface

contains

subroutine init_y_from_R(P, R, y, info)
  implicit none
  integer, intent(in) :: P
  real(kind=rdp), intent(in) :: R(P,P)
  real(kind=rdp), intent(out) :: y(P*(P-1)/2)
  integer, intent(out) :: info

  real(kind=rdp) :: A(P,P)
  real(kind=rdp) :: w, z, denom
  real(kind=rdp), parameter :: eps = 1.0e-12_rdp
  integer :: i, j, idx

  ! copy; DPOTRF overwrites
  A(:,:) = R(:,:)

  call dpotrf('L', P, A, P, info)
  if (info /= 0) then
    ! fallback: identity (y=0) if Cholesky fails
    y(:) = 0.0_rdp
    return
  end if

  idx = 0
  do i = 2, P
    w = 1.0_rdp
    do j = 1, i-1
      idx = idx + 1
      denom = sqrt(max(w, 1.0e-300_rdp))
      z = A(i,j) / denom

      ! clamp to avoid atanh blowups
      if (z >  1.0_rdp - eps) z =  1.0_rdp - eps
      if (z < -1.0_rdp + eps) z = -1.0_rdp + eps

      ! atanh(z) = 0.5*log((1+z)/(1-z))
      y(idx) = 0.5_rdp * log( (1.0_rdp + z) / (1.0_rdp - z) )

      w = w * (1.0_rdp - z*z)
    end do
  end do
end subroutine

subroutine slice_update_corr_LKJ(P, n, Sz, eta, y, C, Cinv)
    implicit none
    integer, intent(in) :: P, n
    real(kind=rdp), intent(in) :: Sz(P,P), eta
    real(kind=rdp), intent(inout) :: y(P*(P-1)/2)
    real(kind=rdp), intent(inout) :: C(P,P), Cinv(P,P)
    real(kind=rdp) :: y_old(P*(P-1)/2)


    integer :: k, npar, errorflag, iterrr
    real(kind=rdp) :: logf0, logy, u, w, Lb, Rb
    real(kind=rdp), parameter :: slice_w = 1.0_rdp
    integer, parameter :: max_steps = 50

    npar = P*(P-1)/2
    y_old(:) = y(:)

    ! current log density
    call logpost_corr_LKJ_lapack(P, n, Sz, eta, y, logf0)

    do k = 1, npar

        ! vertical slice level
        u = unif_rand()
        logy = logf0 + log(u)

        ! stepping-out interval y(m)
        w  = slice_w
        Lb = y(k) - w*unif_rand()
        Rb = Lb + w

        call step_out(k, Lb, Rb, logy, y, P, n, Sz, eta, max_steps)

        ! shrinkage
        iterrr = 0
        do
          iterrr = iterrr + 1
          y(k) = Lb + (Rb-Lb)*unif_rand()
          y(k) = max(-7.0_rdp, min(7.0_rdp, y(k)))
          call logpost_corr_LKJ_lapack(P,n,Sz,eta,y,logf0)

          if (logf0 >= logy) exit

          if (y(k) < 0.5_rdp*(Lb+Rb)) then
          Lb = y(k)
          else
            Rb = y(k)
          end if

          if (iterrr >= 1000) then
            y(:) = y_old(:)
            call build_R_from_y(P, y, C)
            call invert_spd_lapack(P, C, Cinv, errorflag)
            return
          end if
        end do

    end do

    ! build final correlation matrix and inverse
    call build_R_from_y(P, y, C)
    call invert_spd_lapack(P, C, Cinv, errorflag)

    if (errorflag /= 0) then
      y(:) = y_old(:)
      call build_R_from_y(P, y, C)
      call invert_spd_lapack(P, C, Cinv, errorflag)
      if (errorflag /= 0) then
          ! keep incoming C/Cinv; just revert y and exit
          y(:) = y_old(:)
        return
      end if
    end if

end subroutine

subroutine logpost_corr_LKJ_lapack(P, n, Sz, eta, y, logpost)
    implicit none
    integer, intent(in) :: P, n
    real(kind=rdp), intent(in) :: Sz(P,P), eta
    real(kind=rdp), intent(in) :: y(P*(P-1)/2)
    real(kind=rdp), intent(out) :: logpost

    real(kind=rdp) :: R(P,P), A(P,P), B(P,P)
    real(kind=rdp) :: logdetR, traceTerm, logJac
    integer :: info, i

    call build_R_from_y(P, y, R)

    ! Copy R into A because DPOTRF overwrites input
    A(:,:) = R(:,:)

    ! Cholesky factorization: A = L*L' (lower triangle)
    call dpotrf('L', P, A, P, info)
    if (info /= 0) then
        logpost = -huge(1.0_rdp)
        return
    end if

    ! logdet(R) = 2 * sum(log(diag(L)))
    logdetR = 0.0_rdp
    do i = 1, P
        logdetR = logdetR + log( max(A(i,i), 1.0e-300_rdp) )
    end do
    logdetR = 2.0_rdp * logdetR

    ! Solve R * X = Sz for X using the Cholesky factors in A
    ! DPOTRS overwrites RHS with the solution.
    B(:,:) = Sz(:,:)
    call dpotrs('L', P, P, A, P, B, P, info)
    if (info /= 0) then
        logpost = -huge(1.0_rdp)
        return
    end if

    ! trace(Sz * R^{-1}) = trace(X) because X = R^{-1} Sz
    traceTerm = 0.0_rdp
    do i = 1, P
        traceTerm = traceTerm + B(i,i)
    end do

    call lkj_log_jacobian(P, y, logJac)

    logpost = -0.5_rdp * real(n,rdp) * logdetR   &
              -0.5_rdp * traceTerm               &
              + (eta - 1.0_rdp) * logdetR        &
              + logJac
end subroutine

subroutine lkj_log_jacobian(P, y, logJac)
    implicit none
    integer, intent(in) :: P
    real(kind=rdp), intent(in) :: y(P*(P-1)/2)
    real(kind=rdp), intent(out) :: logJac

    integer :: i, j, idx
    real(kind=rdp) :: z, log1mz2, expo
    real(kind=rdp), parameter :: tiny = 1.0e-300_rdp

    logJac = 0.0_rdp
    idx = 0

    ! y is ordered as: i=2..P, j=1..i-1
    do i = 2, P
        do j = 1, i-1
            idx = idx + 1
            z = tanh(y(idx))
            log1mz2 = log(max(1.0_rdp - z*z, tiny))

            ! dz/dy term
            expo = 1.0_rdp

            ! LKJ-Cholesky transform term (depends on column j)
            expo = expo + 0.5_rdp * real(P - j - 1, rdp)

            logJac = logJac + expo * log1mz2
        end do
    end do
end subroutine

subroutine build_R_from_y(P, y, R)
    implicit none
    integer, intent(in) :: P
    real(kind=rdp), intent(in) :: y(P*(P-1)/2)
    real(kind=rdp), intent(out) :: R(P,P)

    real(kind=rdp) :: L(P,P)
    real(kind=rdp) :: z, w
    integer :: i, j, idx

    L = 0.0_rdp
    idx = 0

    L(1,1) = 1.0_rdp

    do i = 2, P
        w = 1.0_rdp
        do j = 1, i-1
            idx = idx + 1
            z = tanh(y(idx))
            L(i,j) = z * sqrt(w)
            w = w * (1.0_rdp - z*z)
        end do
        L(i,i) = sqrt(w)
    end do

    R = matmul(L, transpose(L))

    ! enforce symmetry (numerical hygiene)
    R = 0.5_rdp * (R + transpose(R))

    ! enforce unit diagonal
    do i = 1, P
        R(i,i) = 1.0_rdp
    end do

end subroutine

subroutine step_out(k, Lb, Rb, logy, y, P, n, Sz, eta, max_steps)
    implicit none
    integer, intent(in) :: k, P, n, max_steps
    real(kind=rdp), intent(inout) :: Lb, Rb
    real(kind=rdp), intent(inout) :: y(:)
    real(kind=rdp), intent(in) :: logy, Sz(P,P), eta

    real(kind=rdp) :: logf, ysave, width
    integer :: j

    ysave = y(k)
    width = Rb - Lb
    if (width <= 0.0_rdp) width = 1.0_rdp

    ! step out on the left
    do j = 1, max_steps
        y(k) = Lb
        call logpost_corr_LKJ_lapack(P, n, Sz, eta, y, logf)
        if (logf <= logy) exit
        Lb = Lb - width
    end do

    ! step out on the right
    do j = 1, max_steps
        y(k) = Rb
        call logpost_corr_LKJ_lapack(P, n, Sz, eta, y, logf)
        if (logf <= logy) exit
        Rb = Rb + width
    end do

    y(k) = ysave
end subroutine

subroutine invert_spd_lapack(P, A, Ainv, info)
  implicit none
  integer, intent(in) :: P
  real(kind=rdp), intent(in) :: A(P,P)
  real(kind=rdp), intent(out) :: Ainv(P,P)
  integer, intent(out) :: info

  integer :: i, j
  real(kind=rdp) :: T(P,P)

  T(:,:) = A(:,:)

  call dpotrf('L', P, T, P, info)
  if (info /= 0) return

  call dpotri('L', P, T, P, info)
  if (info /= 0) return

  ! DPOTRI returns only the triangle it worked on; symmetrize
  Ainv(:,:) = 0.0_rdp
  do i = 1, P
    do j = 1, i
      Ainv(i,j) = T(i,j)
      Ainv(j,i) = T(i,j)
    end do
  end do
end subroutine

end module corr_slice_mod


! also support marginally uniform prior?

subroutine estimate_bct_ordinal(postZmean, postZcov, P, numcorr, K, numG, BHat, sdHat, CHat, XtXi, samsize0, &
    burnin, Ntot, Njs_in, Xgroups, Ygroups, C_quantiles, sigma_quantiles, B_quantiles, BDrawsStore, &
    sigmaDrawsStore, CDrawsStore, sdMH, ordinal_in, Cat_in, maxCat, gLiuSab, nuggetscale, corrAlg)
    ! corrAlg: 1 = Liu & Daniels MH, 2 = slice on R
!
!    WgroupsStore, meanMatMeanStore, SigmaMatDrawStore, CheckStore)
!
    use rkinds0, only: rint, rdp
    use rngfuncs
    use corr_slice_mod, only: slice_update_corr_LKJ, init_y_from_R
!
    implicit none
!
    integer(rint), intent(in) ::P, numcorr, K, numG, samsize0, burnin, Ntot, maxCat, corrAlg
    real(rdp), intent(in) ::  BHat(numG,K,P), sdHat(numG,P), CHat(numG,P,P), XtXi(numG,K,K), Cat_in(numG,P), &
                              Xgroups(numG,Ntot,K), Ygroups(numG,Ntot,P), ordinal_in(numG,P), &
                              sdMH(numG,P), nuggetscale, Njs_in(numG,1)
    real(rdp), intent(out)::  postZmean(numcorr,1), postZcov(numcorr,numcorr), B_quantiles(numG,K,P,3), &
                              C_quantiles(numG,P,P,3), sigma_quantiles(numG,P,3), BDrawsStore(samsize0,numG,K,P), &
                              sigmaDrawsStore(samsize0,numG,P), CDrawsStore(samsize0,numG,P,P), &
                              gLiuSab(samsize0,numG,P) !, WgroupsStore(samsize0,numG,Ntot,P), &
                              !meanMatMeanStore(samsize0,Ntot,P), SigmaMatDrawStore(samsize0,P,P), &
                              !CheckStore(samsize0,numG,10,P,P)
    real(rdp) ::  BDraws(numG,K,P), CDraws(numG,P,P), sigmaDraws(numG,P), meanMat(Ntot,P), SigmaMatDraw(P,P), &
                  R_MH, covBeta(K*P,K*P), Ds(P,P), Ccan(P,P), CcanInv(P,P), Ccurr(P,P), epsteps(P,P), &
                  SS1(P,P), SS1inv(P,P), rnunif, errorMatj(P,P), sigma_can(P), aa, bb, CcurrInv(numG,P,P), &
                  betaDrawj(1,P*K), acceptSigma(numG,P), dummyPP(P,P), dummyPPinv(P,P), &
                  varz1, varz2, varz1z2Plus, varz1z2Min, Cnugget(P,P), SigmaInv(P,P), sdMHg(numG,P), gLiuSab_can, &
                  Wgroups(numG,Ntot,P), alphaMin, alphaMax, Cinv(P,P), Bmean(K,P), acceptLS(numG,P), &
                  alphaMat(numG,maxCat+1,P), Wdummy(numG,P,Ntot,maxCat), condMean, condVar, &
                  Zcorr_sample(samsize0,numcorr), dummy3(samsize0), dummy2(samsize0), priorScale(P,P), &
                  diffmat(Ntot,P), meanO(P*K), para((P*K)*((P*K)+3)/2 + 1), randraw, gLiuSab_curr(numG,P), &
                  logR_MH, lkj_eta, epsmat(Ntot,P), yCorr(numG,P*(P-1)/2), Sz(P,P) !, thresh1, thresh2
    integer(rint) ::s1, g1, i1, corrteller, Cat(numG,P), ordinal(numG,P), Njs(numG), &
                  c1, c2, p1, Yi1Categorie, tellers(numG,maxCat,P), k1, p2, errorflag, &
                  lower_int, median_int, upper_int, priorDf, M
!
    ! check

    !initial posterior draws
    BDraws = BHat
    sigmaDraws = sdHat
    CDraws = CHat
    meanO = 0.0_rdp
    gLiuSab_curr = 1.0_rdp
    do g1=1,numG
        CDraws(g1,:,:) = 0.5_rdp*(CDraws(g1,:,:) + transpose(CDraws(g1,:,:)))
        do p1 = 1, P
            CDraws(g1,p1,p1) = 1.0_rdp
        end do
        call FINDInv(CDraws(g1,:,:),CcurrInv(g1,:,:),P,errorflag)
    end do
!
    do g1=1,numG
        do p1=1,P
            ordinal(g1,p1) = int(ordinal_in(g1,p1),kind=rint)
            Cat(g1,p1) = int(Cat_in(g1,p1),kind=rint)
        end do
        Njs(g1) = int(Njs_in(g1,1))
    end do
    do p1=1,P
        do g1=1,numG
            !initial values
            if(ordinal(g1,p1)==1) then
                sigmaDraws(g1,p1) = 1.0_rdp
                sigma_quantiles(g1,p1,1) = 1.0_rdp
                sigma_quantiles(g1,p1,2) = 1.0_rdp
                sigma_quantiles(g1,p1,3) = 1.0_rdp
            end if
        end do
    end do
!
    !set prior hyperparameters. Only supprt for joint uniform prior
    priorDf = 0 ! target prior as in Liu & Daniels
    priorScale = 0.0_rdp
!
    ! set prior hyper parameter eta
    lkj_eta = 1.0_rdp   ! uniform on R
    !lkj_eta = 2.0_rdp  ! shrink toward identity
    M = P*(P-1)/2
    !yCorr(:,:) = 0.0_rdp
    do g1 = 1, numG
        call init_y_from_R(P, CDraws(g1,:,:), yCorr(g1,1:M), errorflag)
    end do
!
!    if(priorchoice == 1) then
!        priorDf = -P-1
!        priorScale = 0
!    else
!        priorDf = P+1
!        priorScale = eye(P)
!    end if
!
    !define nugget matrix to avoid approximate nonpositive definite correlation matrices for candidates
    Cnugget = nuggetscale
    do p1=1,P
        Cnugget(p1,p1) = 1.0_rdp
    end do
    !thresh1 = -0.85_rdp
    !thresh2 = 0.85_rdp
!
    !count number of accepted draws for R (over all groups)
    acceptSigma = 0.0_rdp
    acceptLS = 0.0_rdp
    sdMHg = .1_rdp !for gLiuBanhatti parameter
!
    !initial values for latent W's corresponding to ordinal DVs
    Wgroups = Ygroups
    Wdummy = 0.0_rdp
!
    !initial values of boundary values alpha to link between ordinal Y and continuous latent W
    if(maxCat>1) then
        alphaMat = 0.0_rdp
        alphaMat(:,1,:) = -1e10_rdp    !alpha0
        alphaMat(:,2,:) = 0.0_rdp      !alpha1
        do p1=1,P
            do g1=1,numG
                if(ordinal(g1,p1)>0) then
                    do c1=3,Cat(g1,p1)
                        alphaMat(g1,c1,p1) = .3_rdp*(real(c1,kind=rdp)-2.0_rdp)
                    end do
                    alphaMat(g1,Cat(g1,p1)+1,p1) = 1e10_rdp
                end if
            end do
        end do
    end if
!
    call getrngstate()
!
    !start Gibbs sampler: burn-in period
    do s1 = 1,burnin
        corrteller = 0_rint
        tellers = 0_rint
        do g1 = 1,numG

            !compute means of latent W's for all observations
            meanMat(1:Njs(g1),1:P) = matmul(Xgroups(g1,1:Njs(g1),1:K),BDraws(g1,1:K,1:P))

            Ccurr = CDraws(g1,:,:)
            SigmaMatDraw = matmul(matmul(diag(sigmaDraws(g1,:),P),Ccurr),diag(sigmaDraws(g1,:),P))
!
            !draw latent W's for the ordinal Y's
            !compute mean vector for
!
            do p1=1,P
!
                if(ordinal(g1,p1)>0) then
                    do i1=1,Njs(g1)
                        Yi1Categorie = int(Ygroups(g1,i1,p1))
                        call compute_condMeanVar(p1,P,meanMat(i1,1:P),SigmaMatDraw, &
                            Wgroups(g1,i1,1:P),condMean,condVar)
                        select case (Yi1Categorie)
                            case(1)
                                call inverse_prob_sampling(condMean,condVar,alphaMat(g1,1,p1), &
                                    alphaMat(g1,2,p1),Wgroups(g1,i1,p1))
                                tellers(g1,1,p1) = tellers(g1,1,p1) + 1_rint
                                Wdummy(g1,p1,tellers(g1,1,p1),1) = Wgroups(g1,i1,p1)
                            case(2)
                                call inverse_prob_sampling(condMean,condVar,alphaMat(g1,2,p1), &
                                    alphaMat(g1,3,p1),Wgroups(g1,i1,p1))
                                tellers(g1,2,p1) = tellers(g1,2,p1) + 1_rint
                                Wdummy(g1,p1,tellers(g1,2,p1),2) = Wgroups(g1,i1,p1)
                            case(3)
                                call inverse_prob_sampling(condMean,condVar,alphaMat(g1,3,p1), &
                                    alphaMat(g1,4,p1),Wgroups(g1,i1,p1))
                                tellers(g1,3,p1) = tellers(g1,3,p1) + 1_rint
                                Wdummy(g1,p1,tellers(g1,3,p1),3) = Wgroups(g1,i1,p1)
                            case(4)
                                call inverse_prob_sampling(condMean,condVar,alphaMat(g1,4,p1), &
                                    alphaMat(g1,5,p1),Wgroups(g1,i1,p1))
                                tellers(g1,4,p1) = tellers(g1,4,p1) + 1_rint
                                Wdummy(g1,p1,tellers(g1,4,p1),4) = Wgroups(g1,i1,p1)
                            case(5)
                                call inverse_prob_sampling(condMean,condVar,alphaMat(g1,5,p1), &
                                    alphaMat(g1,6,p1),Wgroups(g1,i1,p1))
                                tellers(g1,5,p1) = tellers(g1,5,p1) + 1_rint
                                Wdummy(g1,p1,tellers(g1,5,p1),5) = Wgroups(g1,i1,p1)
                            case(6)
                                call inverse_prob_sampling(condMean,condVar,alphaMat(g1,6,p1), &
                                    alphaMat(g1,7,p1),Wgroups(g1,i1,p1))
                                tellers(g1,6,p1) = tellers(g1,6,p1) + 1_rint
                                Wdummy(g1,p1,tellers(g1,6,p1),6) = Wgroups(g1,i1,p1)
                            case(7)
                                call inverse_prob_sampling(condMean,condVar,alphaMat(g1,7,p1), &
                                    alphaMat(g1,8,p1),Wgroups(g1,i1,p1))
                                tellers(g1,7,p1) = tellers(g1,7,p1) + 1_rint
                                Wdummy(g1,p1,tellers(g1,7,p1),7) = Wgroups(g1,i1,p1)
                            case(8)
                                call inverse_prob_sampling(condMean,condVar,alphaMat(g1,8,p1), &
                                    alphaMat(g1,9,p1),Wgroups(g1,i1,p1))
                                tellers(g1,8,p1) = tellers(g1,8,p1) + 1_rint
                                Wdummy(g1,p1,tellers(g1,8,p1),8) = Wgroups(g1,i1,p1)
                            case(9)
                                call inverse_prob_sampling(condMean,condVar,alphaMat(g1,9,p1), &
                                    alphaMat(g1,10,p1),Wgroups(g1,i1,p1))
                                tellers(g1,9,p1) = tellers(g1,9,p1) + 1_rint
                                Wdummy(g1,p1,tellers(g1,9,p1),9) = Wgroups(g1,i1,p1)
                            case(10)
                                call inverse_prob_sampling(condMean,condVar,alphaMat(g1,10,p1), &
                                    alphaMat(g1,11,p1),Wgroups(g1,i1,p1))
                                tellers(g1,10,p1) = tellers(g1,10,p1) + 1_rint
                                Wdummy(g1,p1,tellers(g1,10,p1),10) = Wgroups(g1,i1,p1)
                            case(11)
                                call inverse_prob_sampling(condMean,condVar,alphaMat(g1,11,p1), &
                                    alphaMat(g1,12,p1),Wgroups(g1,i1,p1))
                                tellers(g1,11,p1) = tellers(g1,11,p1) + 1_rint
                                Wdummy(g1,p1,tellers(g1,11,p1),11) = Wgroups(g1,i1,p1)
                         end select
                    end do
!
                    !draw boundary's in alphaMat
                    if(Cat(g1,p1)>2) then
                        do c1=3,Cat(g1,p1)
                            alphaMin = maxval(Wdummy(g1,p1,1:tellers(g1,c1-1,p1),c1-1))
                            alphaMax = minval(Wdummy(g1,p1,1:tellers(g1,c1,p1),c1))
                            randraw = unif_rand() * .999998 + .000001 !avoid approx boundary values
                            alphaMat(g1,c1,p1) = randraw * (alphaMax-alphaMin) + alphaMin
                        end do
                    end if
                end if
            end do
!
            Bmean(1:K,1:P) = matmul(matmul(XtXi(g1,:,:),transpose(Xgroups(g1,1:Njs(g1),1:K))), &
                Wgroups(g1,1:Njs(g1),1:P))
            call kronecker(K,P,XtXi(g1,:,:),SigmaMatDraw,covBeta)
!
            call setgmn(meanO,covBeta,P*K,para)
            call GENMN(para,betaDrawj(1,1:(P*K)),P*K)
            do p1 = 1,P
                BDraws(g1,:,p1) = betaDrawj(1,((p1-1)*K+1):(p1*K)) + Bmean(1:K,p1)
            end do
!
            !------------------------------------------------------------
            ! Draw correlation matrix
            ! corrMethod = 1 : Liu & Daniels MH (existing code)
            ! corrMethod = 2 : Slice sampler on R (uniform prior)
            !------------------------------------------------------------

            ! residuals
            diffmat(1:Njs(g1),1:P) = Wgroups(g1,1:Njs(g1),1:P) - &
              matmul(Xgroups(g1,1:Njs(g1),1:K), BDraws(g1,1:K,1:P))
!
            errorMatj = matmul(transpose(diffmat(1:Njs(g1),1:P)),diffmat(1:Njs(g1),1:P))

            if (corrAlg == 1) then

              !========================================================
              ! Method 1: Liu & Daniels (2006)
              !========================================================
!
              Ds = diag(1/sqrt(diagonals(errorMatj,P)),P)
              epsmat(1:Njs(g1),1:P) = matmul(diffmat(1:Njs(g1),1:P),Ds) !diffmat is now epsilon in LD
              epsteps = matmul(transpose(epsmat(1:Njs(g1),1:P)), epsmat(1:Njs(g1),1:P))
              SS1 = matmul(matmul(diag(1/sigmaDraws(g1,:),P),epsteps),diag(1/sigmaDraws(g1,:),P))
!            write(*,*)
!            write(*,*)SS1
              call FINDInv(SS1+priorScale,SS1inv,P,errorflag)
              call gen_wish(SS1inv,Njs(g1)+priorDf,dummyPP,P) !!!!!
              call FINDInv(dummyPP,dummyPPinv,P,errorflag)
              Ccan = matmul(matmul(diag(1/sqrt(diagonals(dummyPPinv,P)),P),dummyPPinv), &
                  diag(1/sqrt(diagonals(dummyPPinv,P)),P))
              Ccan = Ccan * Cnugget
            !do p1 = 1,P-1
            !    do p2 = p1+1,P
            !        Ccan(p1,p2) = scale_function(Ccan(p1,p2), nuggetscale, thresh2, thresh1)
            !        Ccan(p2,p1) = Ccan(p1,p2)
            !    end do
            !end do
            ! check acceptance of candidate via MH
              logR_MH = 0.5*real(P+1,kind=rdp) * (log(det(Ccan,P,-1)) - log(det(Ccurr,P,-1)))
              R_MH = exp(logR_MH)
              rnunif = unif_rand()
              if(rnunif < R_MH) then
                    CDraws(g1,:,:) = Ccan(:,:)
                  call FINDInv(Ccan,CcanInv,P,errorflag)
                  CcurrInv(g1,:,:) = CcanInv(:,:)
                  !acceptC(g1) = acceptC(g1) + 1
              end if
              Cinv(:,:) = CcurrInv(g1,:,:)
              !
            else if (corrAlg == 2) then

              !========================================================
              ! Method 2: Onion sampler based on:
              ! Lewandowski, D., Kurowicka, D., & Joe, H. (2009).
              ! Generating random correlation matrices based on vines and extended onion method.
              ! Journal of Multivariate Analysis, 100(9), 1989–2001.
              !========================================================

              ! Standardize residuals
              do i1 = 1, P
                  do p1 = 1, P
                      Sz(i1,p1) = errorMatj(i1,p1) / (sigmaDraws(g1,i1)*sigmaDraws(g1,p1))
                  end do
              end do

              ! One full slice sweep over LKJ coordinates
              call slice_update_corr_LKJ(P, Njs(g1), Sz, lkj_eta, yCorr(g1,1:M), &
                CDraws(g1,:,:), CcurrInv(g1,:,:))

              Cinv(:,:) = CcurrInv(g1,:,:)

            end if

            !draw sigma's
            do p1 = 1,P
                if(ordinal(g1,p1)==0) then
                    bb = sum(errorMatj(p1,:)*Cinv(p1,:)/sigmaDraws(g1,:)) - &
                        errorMatj(p1,p1)*Cinv(p1,p1)/sigmaDraws(g1,p1)
                    aa = Cinv(p1,p1)*errorMatj(p1,p1)
                    sigma_can(:) = sigmaDraws(g1,:)
                    sigma_can(p1) = rnormal()
                    sigma_can(p1) = sigma_can(p1)*sdMH(g1,p1) + sigmaDraws(g1,p1) !random walk
                    R_MH = exp((-real(Njs(g1))+1.0)*(log(sigma_can(p1))-log(sigmaDraws(g1,p1)) ) &
                           -.5*aa*(sigma_can(p1)**(-2) - sigmaDraws(g1,p1)**(-2)) &
                           -bb*(sigma_can(p1)**(-1) - sigmaDraws(g1,p1)**(-1)) )
                    rnunif = unif_rand()
                    if(rnunif < R_MH .and. sigma_can(p1)>0.0) then
                        sigmaDraws(g1,p1) = sigma_can(p1)
                        acceptSigma(g1,p1) = acceptSigma(g1,p1) + 1.0
                    end if
                end if
            end do

            !Draw parameter extended parameter by Liu and Sabatti (2001) via random walk
            SigmaInv = matmul(matmul(diag(1/sigmaDraws(g1,:),P),Cinv),diag(1/sigmaDraws(g1,:),P))
            do p1 = 1,P
                if(ordinal(g1,p1)>0) then !draw gLiuSab_curr(g1,p1)
                    aa = errorMatj(p1,p1)*SigmaInv(p1,p1)/2.0
                    bb = sum(errorMatj(p1,:)*SigmaInv(p1,:)*gLiuSab_curr(g1,:)) - errorMatj(p1,p1) * &
                        SigmaInv(p1,p1)*gLiuSab_curr(g1,p1)
                    gLiuSab_can = rnormal()
                    gLiuSab_can = gLiuSab_can * sdMHg(g1,p1) + gLiuSab_curr(g1,p1) ! random (moon) walk
                    R_MH = exp((K + Cat(g1,p1) - 2.0 + Njs(g1) - 1)*(log(gLiuSab_can) - log(gLiuSab_curr(g1,p1))) &
                                -aa*(gLiuSab_can**2 - gLiuSab_curr(g1,p1)**2) - bb*(gLiuSab_can - gLiuSab_curr(g1,p1)))
                    rnunif = unif_rand()
                    if(rnunif < R_MH .and. gLiuSab_can>0) then
                        gLiuSab_curr(g1,p1) = gLiuSab_can
                        acceptLS(g1,p1) = acceptLS(g1,p1) + 1.0
                        !update the other parameter through the parameter transformation g(x) = g * x
                        BDraws(g1,1:K,p1) = BDraws(g1,1:K,p1)*gLiuSab_curr(g1,p1)
                        alphaMat(g1,3:Cat(g1,p1),p1) = alphaMat(g1,3:Cat(g1,p1),p1)*gLiuSab_curr(g1,p1)
                        Wgroups(g1,1:Njs(g1),p1) = Wgroups(g1,1:Njs(g1),p1)*gLiuSab_curr(g1,p1)
                    end if
                end if
            end do

        end do
!
    end do

    do s1 = 1,samsize0
        corrteller = 0_rint
        tellers = 0_rint

        do g1 = 1,numG
!
            !compute means of latent W's for all observations
            meanMat(1:Njs(g1),1:P) = matmul(Xgroups(g1,1:Njs(g1),1:K),BDraws(g1,1:K,1:P))
!            meanMatMeanStore(s1,1:Njs(g1),1:P) = meanMat(1:Njs(g1),1:P)
            Ccurr = CDraws(g1,:,:)
            SigmaMatDraw = matmul(matmul(diag(sigmaDraws(g1,:),P),Ccurr),diag(sigmaDraws(g1,:),P))
!            SigmaMatDrawStore(s1,:,:) = SigmaMatDraw(:,:)
!
            !draw latent W's for the ordinal Y's
            !compute mean vector for
!
!compute_condMeanVar(welke,dimIn,meanIn,covmIn,obsIn,condMean,condVar)

            do p1=1,P
                if(ordinal(g1,p1)>0) then
                    do i1=1,Njs(g1)
                        Yi1Categorie = int(Ygroups(g1,i1,p1))
                        call compute_condMeanVar(p1,P,meanMat(i1,1:P),SigmaMatDraw, &
                            Wgroups(g1,i1,1:P),condMean,condVar)
!                        CheckStore(s1,g1,i1,p1,1:P) = meanMat(i1,1:P)
!                        CheckStore(s1,g1,i1,p1,(P+1):(P+P)) = SigmaMatDraw(1,1:P)
!                        CheckStore(s1,g1,i1,p1,(P+P+1):(P+P+P)) = Wgroups(g1,i1,1:P)
!                        CheckStore(s1,g1,i1,p1,(P+P+P+1):(P+P+P+2)) = (/condMean,condVar/)
                        select case (Yi1Categorie)
                            case(1)
                                call inverse_prob_sampling(condMean,condVar,alphaMat(g1,1,p1), &
                                    alphaMat(g1,2,p1),Wgroups(g1,i1,p1))
                                tellers(g1,1,p1) = tellers(g1,1,p1) + 1_rint
                                Wdummy(g1,p1,tellers(g1,1,p1),1) = Wgroups(g1,i1,p1)
!                                CheckStore(s1,g1,i1,p1,(P+P+P+3):(P+P+P+5)) = (/alphaMat(g1,1,p1), &
!                                    alphaMat(g1,2,p1),Wgroups(g1,i1,p1)/)
                            case(2)
                                call inverse_prob_sampling(condMean,condVar,alphaMat(g1,2,p1), &
                                    alphaMat(g1,3,p1),Wgroups(g1,i1,p1))
                                tellers(g1,2,p1) = tellers(g1,2,p1) + 1_rint
                                Wdummy(g1,p1,tellers(g1,2,p1),2) = Wgroups(g1,i1,p1)
!                                CheckStore(s1,g1,i1,p1,(P+P+P+3):(P+P+P+5)) = (/alphaMat(g1,2,p1), &
!                                    alphaMat(g1,3,p1),Wgroups(g1,i1,p1)/)
                            case(3)
                                call inverse_prob_sampling(condMean,condVar,alphaMat(g1,3,p1), &
                                    alphaMat(g1,4,p1),Wgroups(g1,i1,p1))
                                tellers(g1,3,p1) = tellers(g1,3,p1) + 1_rint
                                Wdummy(g1,p1,tellers(g1,3,p1),3) = Wgroups(g1,i1,p1)
                            case(4)
                                call inverse_prob_sampling(condMean,condVar,alphaMat(g1,4,p1), &
                                    alphaMat(g1,5,p1),Wgroups(g1,i1,p1))
                                tellers(g1,4,p1) = tellers(g1,4,p1) + 1_rint
                                Wdummy(g1,p1,tellers(g1,4,p1),4) = Wgroups(g1,i1,p1)
                            case(5)
                                call inverse_prob_sampling(condMean,condVar,alphaMat(g1,5,p1), &
                                    alphaMat(g1,6,p1),Wgroups(g1,i1,p1))
                                tellers(g1,5,p1) = tellers(g1,5,p1) + 1_rint
                                Wdummy(g1,p1,tellers(g1,5,p1),5) = Wgroups(g1,i1,p1)
                            case(6)
                                call inverse_prob_sampling(condMean,condVar,alphaMat(g1,6,p1), &
                                    alphaMat(g1,7,p1),Wgroups(g1,i1,p1))
                                tellers(g1,6,p1) = tellers(g1,6,p1) + 1_rint
                                Wdummy(g1,p1,tellers(g1,6,p1),6) = Wgroups(g1,i1,p1)
                            case(7)
                                call inverse_prob_sampling(condMean,condVar,alphaMat(g1,7,p1), &
                                    alphaMat(g1,8,p1),Wgroups(g1,i1,p1))
                                tellers(g1,7,p1) = tellers(g1,7,p1) + 1_rint
                                Wdummy(g1,p1,tellers(g1,7,p1),7) = Wgroups(g1,i1,p1)
                            case(8)
                                call inverse_prob_sampling(condMean,condVar,alphaMat(g1,8,p1), &
                                    alphaMat(g1,9,p1),Wgroups(g1,i1,p1))
                                tellers(g1,8,p1) = tellers(g1,8,p1) + 1_rint
                                Wdummy(g1,p1,tellers(g1,8,p1),8) = Wgroups(g1,i1,p1)
                            case(9)
                                call inverse_prob_sampling(condMean,condVar,alphaMat(g1,9,p1), &
                                    alphaMat(g1,10,p1),Wgroups(g1,i1,p1))
                                tellers(g1,9,p1) = tellers(g1,9,p1) + 1_rint
                                Wdummy(g1,p1,tellers(g1,9,p1),9) = Wgroups(g1,i1,p1)
                            case(10)
                                call inverse_prob_sampling(condMean,condVar,alphaMat(g1,10,p1), &
                                    alphaMat(g1,11,p1),Wgroups(g1,i1,p1))
                                tellers(g1,10,p1) = tellers(g1,10,p1) + 1_rint
                                Wdummy(g1,p1,tellers(g1,10,p1),10) = Wgroups(g1,i1,p1)
                            case(11)
                                call inverse_prob_sampling(condMean,condVar,alphaMat(g1,11,p1), &
                                    alphaMat(g1,12,p1),Wgroups(g1,i1,p1))
                                tellers(g1,11,p1) = tellers(g1,11,p1) + 1_rint
                                Wdummy(g1,p1,tellers(g1,11,p1),11) = Wgroups(g1,i1,p1)
                         end select
                    end do
!
                    !draw boundary's in alphaMat
                    if(Cat(g1,p1)>2_rint) then
                        do c1=3,Cat(g1,p1)
                            alphaMin = maxval(Wdummy(g1,p1,1:tellers(g1,c1-1,p1),c1-1))
                            alphaMax = minval(Wdummy(g1,p1,1:tellers(g1,c1,p1),c1))
                            randraw = unif_rand() * .999998 + .000001
                            alphaMat(g1,c1,p1) = randraw * (alphaMax-alphaMin) + alphaMin
                        end do
                    end if
                end if
!
            end do
!            WgroupsStore(s1,g1,:,:) = Wgroups(g1,:,:)
!
            Bmean(1:K,1:P) = matmul(matmul(XtXi(g1,:,:),transpose(Xgroups(g1,1:Njs(g1),1:K))), &
                Wgroups(g1,1:Njs(g1),1:P))
            call kronecker(K,P,XtXi(g1,:,:),SigmaMatDraw,covBeta)
!
            call setgmn(meanO,covBeta,P*K,para)
            call GENMN(para,betaDrawj(1,1:(P*K)),P*K)
            do p1 = 1,P
                BDraws(g1,:,p1) = betaDrawj(1,((p1-1)*K+1):(p1*K)) + Bmean(1:K,p1)
            end do
!            BDraws = BHat
            !CheckStore(s1,g1,1,1,1:P) = BDraws(g1,1,:)
!

            !------------------------------------------------------------
            ! Draw correlation matrix
            ! corrMethod = 1 : Liu & Daniels MH (existing code)
            ! corrMethod = 2 : Slice sampler on R (uniform prior)
            !------------------------------------------------------------

            ! residuals
            diffmat(1:Njs(g1),1:P) = Wgroups(g1,1:Njs(g1),1:P) - &
              matmul(Xgroups(g1,1:Njs(g1),1:K), BDraws(g1,1:K,1:P))
!
            errorMatj = matmul(transpose(diffmat(1:Njs(g1),1:P)),diffmat(1:Njs(g1),1:P))

            if (corrAlg == 1) then

              !========================================================
              ! Method 1: Liu & Daniels
              !========================================================
!
              Ds = diag(1/sqrt(diagonals(errorMatj,P)),P)
              epsmat(1:Njs(g1),1:P) = matmul(diffmat(1:Njs(g1),1:P),Ds) !diffmat is now epsilon in LD
              epsteps = matmul(transpose(epsmat(1:Njs(g1),1:P)), epsmat(1:Njs(g1),1:P))
              SS1 = matmul(matmul(diag(1/sigmaDraws(g1,:),P),epsteps),diag(1/sigmaDraws(g1,:),P))
!            write(*,*)
!            write(*,*)SS1
              call FINDInv(SS1+priorScale,SS1inv,P,errorflag)
              call gen_wish(SS1inv,Njs(g1)+priorDf,dummyPP,P) !!!!!
              call FINDInv(dummyPP,dummyPPinv,P,errorflag)
              Ccan = matmul(matmul(diag(1/sqrt(diagonals(dummyPPinv,P)),P),dummyPPinv), &
                  diag(1/sqrt(diagonals(dummyPPinv,P)),P))
              Ccan = Ccan * Cnugget
            !do p1 = 1,P-1
            !    do p2 = p1+1,P
            !        Ccan(p1,p2) = scale_function(Ccan(p1,p2), nuggetscale, thresh2, thresh1)
            !        Ccan(p2,p1) = Ccan(p1,p2)
            !    end do
            !end do
            ! check acceptance of candidate via MH
              logR_MH = 0.5*real(P+1,kind=rdp) * (log(det(Ccan,P,-1)) - log(det(Ccurr,P,-1)))
              R_MH = exp(logR_MH)
              rnunif = unif_rand()
              if(rnunif < R_MH) then
                    CDraws(g1,:,:) = Ccan(:,:)
                  call FINDInv(Ccan,CcanInv,P,errorflag)
                  CcurrInv(g1,:,:) = CcanInv(:,:)
                  !acceptC(g1) = acceptC(g1) + 1
              end if
              Cinv(:,:) = CcurrInv(g1,:,:)
              !
            else if (corrAlg == 2) then

              !========================================================
              ! Method 2: Slice sampler for R (with LKJ prior)
              !========================================================

              ! Standardize residuals using *model* sigmas (Jeffreys kept)
              do i1 = 1, P
                  do p1 = 1, P
                      Sz(i1,p1) = errorMatj(i1,p1) / (sigmaDraws(g1,i1)*sigmaDraws(g1,p1))
                  end do
              end do

              ! One full slice sweep over LKJ coordinates
              call slice_update_corr_LKJ(P, Njs(g1), Sz, lkj_eta, yCorr(g1,1:M), &
                CDraws(g1,:,:), CcurrInv(g1,:,:))

              Cinv(:,:) = CcurrInv(g1,:,:)

            end if

            do i1 = 1,P-1 !keep Fisher z transformed posterior draws of rho's
                Zcorr_sample(s1,(corrteller+1):(corrteller+P-i1)) = .5*log((1+CDraws(g1,(1+i1):P,i1))/ &
                    (1-CDraws(g1,(1+i1):P,i1)))
                corrteller = corrteller + (P - i1)
            end do

            !draw sigma's
            do p1 = 1,P
                if(ordinal(g1,p1)==0) then
                    bb = sum(errorMatj(p1,:)*Cinv(p1,:)/sigmaDraws(g1,:)) - &
                        errorMatj(p1,p1)*Cinv(p1,p1)/sigmaDraws(g1,p1)
                    aa = Cinv(p1,p1)*errorMatj(p1,p1)
                    sigma_can(:) = sigmaDraws(g1,:)
                    sigma_can(p1) = rnormal()
                    sigma_can(p1) = sigma_can(p1)*sdMH(g1,p1) + sigmaDraws(g1,p1) !random walk
                    R_MH = exp((-real(Njs(g1))+1.0_rdp)*(log(sigma_can(p1))-log(sigmaDraws(g1,p1)) ) &
                           -.5_rdp*aa*(sigma_can(p1)**(-2.0_rdp) - sigmaDraws(g1,p1)**(-2.0_rdp)) &
                           -bb*(sigma_can(p1)**(-1.0_rdp) - sigmaDraws(g1,p1)**(-1.0_rdp)) )
                    rnunif = unif_rand()
                    if(rnunif < R_MH .and. sigma_can(p1) > 0.0_rdp) then
                        sigmaDraws(g1,p1) = sigma_can(p1)
                        acceptSigma(g1,p1) = acceptSigma(g1,p1) + 1.0_rdp
                    end if
                end if
            end do
!
            !Draw parameter extended parameter by Liu and Sabatti (2001) via random walk
            SigmaInv = matmul(matmul(diag(1/sigmaDraws(g1,:),P),Cinv),diag(1/sigmaDraws(g1,:),P))
            do p1 = 1,P
                if(ordinal(g1,p1)>0) then !draw gLiuSab_curr(g1,p1)
                    aa = errorMatj(p1,p1)*SigmaInv(p1,p1)/2.0_rdp
                    bb = sum(errorMatj(p1,:)*SigmaInv(p1,:)*gLiuSab_curr(g1,:)) - errorMatj(p1,p1) * &
                        SigmaInv(p1,p1)*gLiuSab_curr(g1,p1)
                    gLiuSab_can = rnormal()
                    gLiuSab_can = gLiuSab_can*sdMHg(g1,p1) + gLiuSab_curr(g1,p1) ! random (moon) walk
                    R_MH = exp((K + Cat(g1,p1) - 2.0_rdp + real(Njs(g1),kind=rdp) - 1)*(log(gLiuSab_can) - &
                                log(gLiuSab_curr(g1,p1))) -aa*(gLiuSab_can**2.0_rdp - gLiuSab_curr(g1,p1)**2.0_rdp) &
                                - bb*(gLiuSab_can - gLiuSab_curr(g1,p1)))
                    rnunif = unif_rand()
                    if(rnunif < R_MH .and. gLiuSab_can>0.0_rdp) then
                        gLiuSab_curr(g1,p1) = gLiuSab_can
                        acceptLS(g1,p1) = acceptLS(g1,p1) + 1.0_rdp
                        !update the other parameter through the parameter transformation g(x) = g * x
                        BDraws(g1,1:K,p1) = BDraws(g1,1:K,p1)*gLiuSab_curr(g1,p1)
                        alphaMat(g1,3:Cat(g1,p1),p1) = alphaMat(g1,3:Cat(g1,p1),p1)*gLiuSab_curr(g1,p1)
                        Wgroups(g1,1:Njs(g1),p1) = Wgroups(g1,1:Njs(g1),p1)*gLiuSab_curr(g1,p1)
                    end if
                end if
            end do
!
        end do

        BDrawsStore(s1,1:numG,1:K,1:P) = BDraws(1:numG,1:K,1:P)
        sigmaDrawsStore(s1,1:numG,1:P) = sigmaDraws(1:numG,1:P)
        CDrawsStore(s1,1:numG,1:P,1:P) = CDraws(1:numG,1:P,1:P)
        gLiuSab(s1,1:numG,1:P) = gLiuSab_curr(1:numG,1:P)
!
    end do
!
    call putrngstate()
!
    ! compute posterior mean
    do c1=1,numcorr
        dummy2(:) = Zcorr_sample(:,c1)
        dummy3 = dummy2
        call piksrt(samsize0,dummy3)
        postZmean(c1,1) = dummy3(int(samsize0*.5))
    end do
!
    ! compute posterior covariance matrix
    do c1=1,numcorr
        do c2=c1,numcorr
            call robust_covest(samsize0, Zcorr_sample(1:samsize0,c1), Zcorr_sample(1:samsize0,c2), &
                postZmean(c1,1), postZmean(c2,1), varz1, varz2, varz1z2Plus, varz1z2Min)
            postZcov(c1,c2) = (varz1*varz2)**.5 * (varz1z2Plus - varz1z2Min)/(varz1z2Plus + varz1z2Min)
            postZcov(c2,c1) = postZcov(c1,c2)
        end do
    end do
!
    ! compute posterior quantiles
    lower_int = int(samsize0*.025_rdp,kind=rint)
    median_int = int(samsize0*.5_rdp,kind=rint)
    upper_int = int(samsize0*.975_rdp,kind=rint)
    C_quantiles = 0
    do g1=1,numG
        do p1=1,P
            !for the sigma's
            dummy2(:) = sigmaDrawsStore(:,g1,p1)
            dummy3=dummy2
            call piksrt(samsize0,dummy3)
            sigma_quantiles(g1,p1,1) = dummy3(lower_int)
            sigma_quantiles(g1,p1,2) = dummy3(median_int)
            sigma_quantiles(g1,p1,3) = dummy3(upper_int)
            !for the beta coefficients
            do k1=1,K
                dummy2(:) = BDrawsStore(:,g1,k1,p1)
                dummy3 = dummy2
                call piksrt(samsize0,dummy3)
                B_quantiles(g1,k1,p1,1) = dummy3(lower_int)
                B_quantiles(g1,k1,p1,2) = dummy3(median_int)
                B_quantiles(g1,k1,p1,3) = dummy3(upper_int)
            end do
            if(p1>1)then
                do p2=1,p1-1
                    dummy2(:) = CDrawsStore(:,g1,p1,p2)
                    dummy3 = dummy2
                    call piksrt(samsize0,dummy3)
                    C_quantiles(g1,p1,p2,1) = dummy3(lower_int)
                    C_quantiles(g1,p1,p2,2) = dummy3(median_int)
                    C_quantiles(g1,p1,p2,3) = dummy3(upper_int)
                end do
            end if
        end do
    end do

    !write(*,*)'Cmedians'
    !write(*,*)C_quantiles(1,1:3,1:3,2)
    !write(*,*)B_quantiles(1,1,1,1:3)
    !write(*,*)sigma_quantiles(1,1,1:3)

    !write(*,*)'end'

contains


subroutine robust_covest(m, betas1, betas2, mn1, mn2, varb1, varb2, varb1b2Plus, varb1b2Min)

    implicit none
!
!    integer, parameter :: rdp = selected_real_kind(15)
!    integer, parameter :: rint = selected_int_kind(6)

    !Declare local variables
    integer(rint), intent(in)  :: m
    real(rdp), intent(in)    :: betas1(m), betas2(m), mn1, mn2
    real(rdp), intent(out)   :: varb1, varb2, varb1b2Plus, varb1b2Min

    real(rdp)                :: dummy1(m), dummy2(m), Phi075, xxx
    integer(rint)              :: mmin, i
!
    xxx=0.75_rdp
    Phi075 = dinvnr(xxx)
    mmin = 0_rint
!
    !robust variance estimators of beta1 and beta2

    dummy1=abs(betas1 - mn1)
    call piksrt(m,dummy1)
    do i=1,m
        if(dummy1(i)>0) then
            mmin = i
            exit
        end if
    end do
    varb1 = ((dummy1(int(mmin+(m-mmin)*.5)) + dummy1(int(mmin+(m-mmin)*.5+1)))*.5/Phi075)**2.0
    dummy1=abs(betas2 - mn2)
    call piksrt(m,dummy1)
    do i=1,m
        if(dummy1(i)>0) then
            mmin = i
            exit
        end if
    end do
    varb2 = ((dummy1(int(mmin+(m-mmin)*.5)) + dummy1(int(mmin+(m-mmin)*.5+1)))*.5/Phi075)**2.0
!
    !robust variance estimators of beta1 + beta2
    dummy2 = betas1 + betas2
    dummy1=abs(dummy2 - mn1 - mn2)
    call piksrt (m,dummy1)
    do i=1,m
        if(dummy1(i)>0) then
            mmin = i
            exit
        end if
    end do
    varb1b2Plus = ((dummy1(int(mmin+(m-mmin)*.5)) + dummy1(int(mmin+(m-mmin)*.5+1)))*.5/Phi075)**2.0
!
    !robust variance estimators of beta1 - beta2
    dummy2 = betas1 - betas2
    dummy1= abs(dummy2 - mn1 + mn2)
    call piksrt(m,dummy1)
    do i=1,m
        if(dummy1(i)>0) then
            mmin = i
            exit
        end if
    end do
    varb1b2Min = ((dummy1(int(mmin+(m-mmin)*.5)) + dummy1(int(mmin+(m-mmin)*.5+1)))*.5/Phi075)**2.0
!
end subroutine


SUBROUTINE piksrt(n,arr)

  implicit none

!  integer, parameter :: rdp = selected_real_kind(15)
!  integer, parameter :: rint = selected_int_kind(6)

  integer(rint) :: n, i,j
  real(rdp)   :: arr(n), a

  do j=2, n
    a=arr(j)
    do i=j-1,1,-1
      if (arr(i)<=a) goto 10
      arr(i+1)=arr(i)
    end do
  i=0
10  arr(i+1)=a
  end do
  return
END SUBROUTINE


function eye(n)

    implicit none

!    integer, parameter :: rdp = selected_real_kind(15)
!    integer, parameter :: rint = selected_int_kind(6)

    integer(rint):: i,n
    real(rdp):: eye(n,n)
    real(rdp):: check(n,n)

    check=0
    do i=1,n
        check(i,i)= 1
    enddo

    eye(:,:)=check(:,:)
    return

end function eye


!elemental function scale_function(x, s, x1, x2) result(y)
!
!  implicit none
!
!  real(rdp), intent(in) :: x, s, x1, x2
!  real(rdp) :: y, m_left, m_right
!  real(rdp) :: cond1, cond2, cond3
!
!  m_left  = (-x1 - s) / (-1d0 - x1)
!  m_right = (s - x2) / (1d0 - x2)
!
!  cond1 = merge(1.0d0, 0.0d0, x < x1)
!  cond2 = merge(1.0d0, 0.0d0, x >= x1 .and. x <= x2)
!  cond3 = merge(1.0d0, 0.0d0, x > x2)
!
!  y = cond1 * (m_left * (x - x1) + x1) + &
!      cond2 * x + &
!      cond3 * (m_right * (x - x2) + x2)
!
!end function scale_function


subroutine kronecker(dimA,dimB,A,B,AB)
!
    implicit none
!
!    integer, parameter :: rdp = selected_real_kind(15)
!    integer, parameter :: rint = selected_int_kind(6)
!
    integer(rint), intent(in) :: dimA, dimB
    real(rdp), intent(in)   :: A(dimA,dimA), B(dimB,dimB) !dummy arguments
    real(rdp), intent(out)  :: AB(dimA*dimB,dimA*dimB) !output matrix of the kronecker product
    integer(rint)             :: i,j !loop counters
!
    do i=1,dimA
        do j=1,dimA
            AB((1+dimB*(i-1)):(dimB+dimB*(i-1)),(1+dimB*(j-1)):(dimB+dimB*(j-1))) = A(i,j)*B(:,:)
        end do
    end do
!
end subroutine kronecker


function diag(A, n)

    implicit none
!
!    integer, parameter :: rdp = selected_real_kind(15)
!    integer, parameter :: rint = selected_int_kind(6)

    integer(rint) :: n,i
    real(rdp)   :: A(n), check(n,n)
    real(rdp)   :: diag(n,n)

    check = 0
    do i=1,n
        check(i,i)=A(i)
    end do
    diag(:,:)=check(:,:)

    return

end function diag


function diagonals(A, n)

    implicit none
!
!    integer, parameter :: rdp = selected_real_kind(15)
!    integer, parameter :: rint = selected_int_kind(6)

    integer(rint) :: n,i
    real(rdp)   :: A(n,n), diagonals(n), check(n)

    do i=1,n
        check(i)= A(i,i)
    enddo
    diagonals(:)=check(:)

    return

end function diagonals



function rnormal ()

!*****************************************************************************80
!
!! RNORMAL returns a unit pseudonormal R8.
!
!  Discussion:
!
!    The standard normal probability distribution function (PDF) has
!    mean 0 and standard deviation 1.
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    06 August 2013
!
!  Author:
!
!    John Burkardt
!
!  Parameters:
!
!
!    Output, real ( kind = 8 ) RNORMAL, a normally distributed
!    random value.
!

  implicit none
!
!  integer, parameter :: rdp = selected_real_kind(15)
!  integer, parameter :: rint = selected_int_kind(6)

  real ( kind = rdp ) r1
  real ( kind = rdp ) r2
  real ( kind = rdp ) r3
  real ( kind = rdp ) rnormal
  real ( kind = rdp ), parameter :: pi = 3.141592653589793D+00
  !real ( kind = rdp ) GG
  real ( kind = rdp ) x
  !integer ( kind = rint ) iseed

  !nseed = 1344
  r1 = unif_rand()
  r2 = unif_rand()
  r3 = unif_rand()
  !PRINT*, iseed, r1, r2, r3
  !call random_number(GG)
  !r1 = GG
  !call random_number(GG)
  !r2 = GG
  x = sqrt ( - 2.0D+00 * log ( r3 ) ) * cos ( 2.0D+00 * pi * r2 )

  rnormal = x

  return
end function rnormal


recursive function det(a,n,permanent) result(accumulation)
    ! setting permanent to 1 computes the permanent.
    ! setting permanent to -1 computes the determinant.

    implicit none
!
!    integer, parameter :: rdp = selected_real_kind(15)
!    integer, parameter :: rint = selected_int_kind(6)

    integer(rint), intent(in) :: n, permanent
    real(rdp), dimension(n,n), intent(in) :: a
    real(rdp), dimension(n-1, n-1) :: b
    real(rdp) :: accumulation
    integer(rint) :: i, sgn

    if (n .eq. 1) then
      accumulation = a(1,1)
    else
      accumulation = 0
      sgn = 1
      do i=1, n
        b(:, :(i-1)) = a(2:, :i-1)
        b(:, i:) = a(2:, i+1:)
        accumulation = accumulation + sgn * a(1, i) * det(b, n-1, permanent)
        sgn = sgn * permanent
      enddo
    endif
end function det


subroutine gen_wish(A,nu,B,P)
!
    implicit none
!
    integer(rint), intent (in) :: nu,P
    real(rdp), intent (in)     :: A(P,P)
    real(rdp), intent (out)    :: B(P,P)
    real(rdp)                  :: RNmat(nu,P),para((P*(P+3)/2) + 1), m0(P)
    integer(rint)              :: i
!
    m0=0

    call setgmn(m0,A,P,para)

    do i=1,nu

        call GENMN(para,RNmat(i,:),P)

    end do

    B = matmul(transpose(RNmat),RNmat)
!
end subroutine gen_wish


SUBROUTINE setgmn(meanv,covm,p,parm)
!**********************************************************************
!
!     SUBROUTINE SETGMN( MEANV, COVM, P, PARM)
!            SET Generate Multivariate Normal random deviate
!
!
!                              Function
!
!
!      Places P, MEANV, and the Cholesky factoriztion of COVM
!      in GENMN.
!
!
!                              Arguments
!
!
!     MEANV --> Mean vector of multivariate normal distribution.
!                                        REAL MEANV(P)
!
!     COVM   <--> (Input) Covariance   matrix    of  the  multivariate
!                 normal distribution
!                 (Output) Destroyed on output
!                                        REAL !OVM(P,P)
!
!     P     --> Dimension of the normal, or length of MEANV.
!                                        INTEGER P
!
!     PARM <-- Array of parameters needed to generate multivariate norma
!                deviates (P, MEANV and !holesky decomposition of
!                !OVM).
!                1 : 1                - P
!                2 : P + 1            - MEANV
!                P+2 : P*(P+3)/2 + 1  - !holesky decomposition of !OVM
!                                             REAL PARM(P*(P+3)/2 + 1)
!
!**********************************************************************
!     .. Scalar Arguments ..

      implicit none
!
!      integer, parameter :: rdp = selected_real_kind(15)
!      integer, parameter :: rint = selected_int_kind(6)

      INTEGER(rint) p
!     ..
!     .. Array Arguments ..
      REAL(rdp) covm(p,p),meanv(p),parm(p*(p+3)/2+1)
!     ..
!     .. Local Scalars ..
      INTEGER(rint) i,icount,info,j
!
      external :: dpotrf
!
!
!     TEST THE INPUT
!
      IF (.NOT. (p.LE.0)) GO TO 10
!      WRITE (*,*) 'P nonpositive in SETGMN'
!      WRITE (*,*) 'Value of P: ',p
!      STOP 'P nonpositive in SETGMN'

   10 parm(1) = p
!
!     PUT P AND MEANV INTO PARM
!
      DO 20,i = 2,p + 1
          parm(i) = meanv(i-1)
   20 CONTINUE
!
!      Cholesky decomposition to find A s.t. trans(A)*(A) = COVM
!
!      CALL spofa(covm,p,p,info)
      CALL dpotrf('U',p,covm,p,info)
      IF (.NOT. (info.NE.0)) GO TO 30
!      WRITE (*,*) ' !OVM not positive definite in SETGMN'
!      STOP ' COVM not positive definite in SETGMN'

   30 icount = p + 1
!
!     PUT UPPER HALF OF A, WHICH IS NOW THE !HOLESKY FA!TOR, INTO PARM
!          !OVM(1,1) = PARM(P+2)
!          !OVM(1,2) = PARM(P+3)
!                    :
!          !OVM(1,P) = PARM(2P+1)
!          !OVM(2,2) = PARM(2P+2)  ...
!
      DO 50,i = 1,p
          DO 40,j = i,p
              icount = icount + 1
              parm(icount) = covm(i,j)
   40     CONTINUE
   50 CONTINUE
      RETURN
!
END SUBROUTINE setgmn



SUBROUTINE genmn(parm,x,p)
  !**********************************************************************
  !
  !     SUBROUTINE GENMN(PARM,X,WORK)
  !              GENerate Multivariate Normal random deviate
   !
   !
   !                              Arguments
   !
  !
  !     PARM --> Parameters needed to generate multivariate normal
  !               deviates (MEANV and Cholesky decomposition of
  !               COVM). Set by a previous call to SETGMN.
  !               1 : 1                - size of deviate, P
  !               2 : P + 1            - mean vector
  !               P+2 : P*(P+3)/2 + 1  - upper half of cholesky
  !                                       decomposition of cov matrix
  !                                             DOUBLE PRECISION PARM(*)
  !
  !     X    <-- Vector deviate generated.
  !                                             DOUBLE PRECISION X(P)
  !
  !     WORK <--> Scratch array
  !                                             DOUBLE PRECISION WORK(P)
  !
  !
  !                              Method
  !
  !
  !     1) Generate P independent standard normal deviates - Ei ~ N(0,1)
  !
  !     2) Using Cholesky decomposition find A s.t. trans(A)*A = COVM
  !
  !     3) trans(A)E + MEANV ~ N(MEANV,!OVM)
  !
  !**********************************************************************
  !     .. Array Arguments ..

        implicit none
!
!        integer, parameter :: rdp = selected_real_kind(15)
!        integer, parameter :: rint = selected_int_kind(6)

        integer(rint), intent(in) :: p
        real(rdp), intent(in) :: parm(p*(p+3)/2 + 1)
        real(rdp)             :: work(p)
        real(rdp), intent(out):: x(p)
  !     ..
  !     .. Local Scalars ..
        real(rdp) ae
        integer(rint) :: i,icount,j
  !    ..
  !     .. External Functions ..
  !     ..
  !     .. Executable Statements ..
  !
  !     Generate P independent normal deviates - WORK ~ N(0,1)
  !
        DO 10,i = 1,p
            work(i) = rnormal()
     10 CONTINUE
        DO 30,i = 1,p
  !
  !     PARM (P+2 : P*(P+3)/2 + 1) contains A, the Cholesky
  !      decomposition of the desired covariance matrix.
  !          trans(A)(1,1) = PARM(P+2)
  !          trans(A)(2,1) = PARM(P+3)
  !          trans(A)(2,2) = PARM(P+2+P)
  !          trans(A)(3,1) = PARM(P+4)
  !          trans(A)(3,2) = PARM(P+3+P)
  !          trans(A)(3,3) = PARM(P+2-1+2P)  ...
  !
  !     trans(A)*WORK + MEANV ~ N(MEANV,COVM)
  !
            icount = 0
            ae = 0.0
            DO 20,j = 1,i
                icount = icount + j - 1
                ae = ae + parm(i+ (j-1)*p-icount+p+1)*work(j)
     20     CONTINUE
            x(i) = ae + parm(i+1)
     30 CONTINUE
        RETURN
  !
END SUBROUTINE genmn



FUNCTION sdot(N,SX,INCX,SY,INCY)
 !
 !  -- Reference BLAS level1 routine (version 3.8.0) --
 !  -- Reference BLAS is a software package provided by Univ. of Tennessee,    --
 !  -- Univ. of California Berkeley, Univ. of Colorado Denver and NAG Ltd..--
 !     November 2017
 !
 !     .. Scalar Arguments ..

    implicit none
!
!    integer, parameter :: rdp = selected_real_kind(15)
!    integer, parameter :: rint = selected_int_kind(6)

    INTEGER(rint) :: INCX,INCY,N
 !     ..
  !    .. Array Arguments ..
    REAL(rdp) :: SX(*),SY(*)
    real ( kind = rdp ) sdot
 !    ..
 !
 !  =====================================================================
 !
 !     .. Local Scalars ..
    REAL(rdp) :: STEMP
    INTEGER(rint) :: I,IX,IY,M,MP1
 !     ..
 !     .. Intrinsic Functions ..
    INTRINSIC mod
 !     ..
    stemp = 0.0e0
    sdot = 0.0e0
    IF (n.LE.0) RETURN
        IF (incx.EQ.1 .AND. incy.EQ.1) THEN
 !
 !        code for both increments equal to 1
 !
 !
 !        clean-up loop
 !
        m = mod(n,5)
        IF (m.NE.0) THEN
            DO i = 1,m
                stemp = stemp + sx(i)*sy(i)
            END DO
            IF (n.LT.5) THEN
                sdot = stemp
                RETURN
            END IF
        END IF
        mp1 = m + 1
        DO i = mp1,n,5
            stemp = stemp + sx(i)*sy(i) + sx(i+1)*sy(i+1) + sx(i+2)*sy(i+2) + sx(i+3)*sy(i+3) + sx(i+4)*sy(i+4)
        END DO
    ELSE
 !
 !        code for unequal increments or equal increments
 !          not equal to 1
 !
        ix = 1
        iy = 1
        IF (incx.LT.0) ix = (-n+1)*incx + 1
            IF (incy.LT.0) iy = (-n+1)*incy + 1
                DO i = 1,n
                    stemp = stemp + sx(ix)*sy(iy)
                    ix = ix + incx
                    iy = iy + incy
                END DO
            END IF
            sdot = stemp
        RETURN
END FUNCTION sdot



subroutine compute_condMeanVar(welke,dimIn,meanIn,covmIn,obsIn,condMean,condVar)

    implicit none
!
!    integer, parameter :: rdp = selected_real_kind(15)
!    integer, parameter :: rint = selected_int_kind(6)

    integer(rint), intent(in)         :: welke, dimIn
    real (kind = rdp), intent(in)   :: meanIn(dimIn), covmIn(dimIn,dimIn), obsIn(dimIn)
    real (kind = rdp), intent(out)  :: condMean, condVar
    real (kind = rdp)               :: dummy3(1,1), dummy2(dimIn-1,1), S12(1,dimIn-1), S22(dimIn-1,dimIn-1), &
                                       S22inv(dimIn-1,dimIn-1), meanLocal(dimIn,1)
    integer(rint)                     :: errorflag
!
    meanLocal(1:dimIn,1) = meanIn(1:dimIn)
!
    if(welke > 1 .and. welke < dimIn) then

        S12(1,1:(welke-1)) = covmIn(welke,1:(welke-1))
        S12(1,welke:(dimIn-1)) = covmIn(welke,(welke+1):dimIn)
        S22(1:(welke-1),1:(welke-1)) = covmIn(1:(welke-1),1:(welke-1))
        S22(1:(welke-1),welke:(dimIn-1)) = covmIn(1:(welke-1),(welke+1):dimIn)
        S22(welke:(dimIn-1),1:(welke-1)) = covmIn((welke+1):dimIn,1:(welke-1))
        S22(welke:(dimIn-1),welke:(dimIn-1)) = covmIn((welke+1):dimIn,(welke+1):dimIn)
        call FINDInv(S22,S22inv,dimIn-1,errorflag)
        dummy2(1:(welke-1),1) = obsIn(1:(welke-1)) - meanLocal(1:(welke-1),1)
        dummy2(welke:(dimIn-1),1) = obsIn((welke+1):dimIn) - meanLocal((welke+1):dimIn,1)

    else if(welke == 1) then

        S12(1,welke:(dimIn-1)) = covmIn(welke,(welke+1):dimIn)
        S22(welke:(dimIn-1),welke:(dimIn-1)) = covmIn((welke+1):dimIn,(welke+1):dimIn)
        dummy2(welke:(dimIn-1),1) = obsIn((welke+1):dimIn) - meanLocal((welke+1):dimIn,1)

    else !welke == dimIn

        S12(1,1:(welke-1)) = covmIn(welke,1:(welke-1))
        S22(1:(welke-1),1:(welke-1)) = covmIn(1:(welke-1),1:(welke-1))
        dummy2(1:(welke-1),1) = obsIn(1:(welke-1)) - meanLocal(1:(welke-1),1)

    end if

    call FINDInv(S22,S22inv,dimIn-1,errorflag)
    dummy3 = matmul(matmul(S12,S22inv),dummy2)
    condMean = meanLocal(welke,1) + dummy3(1,1) !conditional mean
    dummy3 = matmul(matmul(S12,S22inv),transpose(S12))
    condVar = covmIn(welke,welke) - dummy3(1,1) !conditional variance
!
end subroutine compute_condMeanVar


subroutine inverse_prob_sampling(condMean,condVar,LB,UB,condDraw)

    implicit none
!
    real ( kind = rdp ), intent(in)   :: condMean, condVar, LB, UB
    real ( kind = rdp ), intent(out)  :: condDraw
    real ( kind = rdp )               :: xdraw
    real ( kind = rdp )               :: LBstand, UBstand, yUB, yLB, rnIPS, &
                                         pi, machPres, rrand
    integer(rint)                          teller
    logical                           :: uppie

    parameter(pi=3.141592653)
    uppie = .false.
!
    machPres = 1e-6
    !normalize bounds
    UBstand = (UB - condMean)/sqrt(condVar)
    LBstand = (LB - condMean)/sqrt(condVar)
    !yUB = anordf (UBstand)
    !yUB = alnorm ( UBstand, uppie )
    yUB = cumnor(UBstand)
    !yLB = anordf (LBstand)
    !yLB = alnorm ( LBstand, uppie )
    yLB = cumnor (LBstand)
    teller = 0
    rrand = unif_rand()
    rrand = rrand * .999998_rdp + .000001_rdp
    rnIPS = rrand*(yUB - yLB) + yLB
    if(rnIPS .le. machPres) then
        rnIPS = machPres
    end if
    if(rnIPS .ge. (1.0 - machPres)) then
        rnIPS = 1.0 - machPres
    end if
    xdraw = dinvnr ( rnIPS )
    condDraw = xdraw * sqrt(condVar) + condMean

!    if(abs(rnIPS) > machPres .and. abs(rnIPS-1) > machPres) then
!        !inverse probability sampling
!        xdraw = dinvnr ( rnIPS )
!        condDraw = xdraw * sqrt(condVar) + condMean
!    else if(UBstand>-4.0 .and. LBstand<4.0) then !IPS must be redone
!        teller = teller + 1
!        go to 601
!    else
!        if(condMean > UB) then
!            condDraw = UB
!        else if(condMean < LB) then
!            condDraw = LB
!        end if
!    end if
!
end subroutine inverse_prob_sampling



function alnorm ( x, upper )

!*****************************************************************************80
!
!! ALNORM computes the cumulative density of the standard normal distribution.
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    13 January 2008
!
!  Author:
!
!    Original FORTRAN77 version by David Hill.
!    FORTRAN90 version by John Burkardt.
!
!  Reference:
!
!    David Hill,
!    Algorithm AS 66:
!    The Normal Integral,
!    Applied Statistics,
!    Volume 22, Number 3, 1973, pages 424-427.
!
!  Parameters:
!
!    Input, real ( kind = 8 ) X, is one endpoint of the semi-infinite interval
!    over which the integration takes place.
!
!    Input, logical UPPER, determines whether the upper or lower
!    interval is to be integrated:
!    .TRUE.  => integrate from X to + Infinity;
!    .FALSE. => integrate from - Infinity to X.
!
!    Output, real ( kind = 8 ) ALNORM, the integral of the standard normal
!    distribution over the desired interval.
!

  implicit none
!
!  integer, parameter :: rdp = selected_real_kind(15)
!  integer, parameter :: rint = selected_int_kind(6)

  real ( kind = rdp ), parameter :: a1 = 5.75885480458D+00
  real ( kind = rdp ), parameter :: a2 = 2.62433121679D+00
  real ( kind = rdp ), parameter :: a3 = 5.92885724438D+00
  real ( kind = rdp ) alnorm
  real ( kind = rdp ), parameter :: b1 = -29.8213557807D+00
  real ( kind = rdp ), parameter :: b2 = 48.6959930692D+00
  real ( kind = rdp ), parameter :: c1 = -0.000000038052D+00
  real ( kind = rdp ), parameter :: c2 = 0.000398064794D+00
  real ( kind = rdp ), parameter :: c3 = -0.151679116635D+00
  real ( kind = rdp ), parameter :: c4 = 4.8385912808D+00
  real ( kind = rdp ), parameter :: c5 = 0.742380924027D+00
  real ( kind = rdp ), parameter :: c6 = 3.99019417011D+00
  real ( kind = rdp ), parameter :: con = 1.28D+00
  real ( kind = rdp ), parameter :: d1 = 1.00000615302D+00
  real ( kind = rdp ), parameter :: d2 = 1.98615381364D+00
  real ( kind = rdp ), parameter :: d3 = 5.29330324926D+00
  real ( kind = rdp ), parameter :: d4 = -15.1508972451D+00
  real ( kind = rdp ), parameter :: d5 = 30.789933034D+00
  real ( kind = rdp ), parameter :: ltone = 7.0D+00
  real ( kind = rdp ), parameter :: p = 0.398942280444D+00
  real ( kind = rdp ), parameter :: q = 0.39990348504D+00
  real ( kind = rdp ), parameter :: r = 0.398942280385D+00
  logical up
  logical upper
  real ( kind = rdp ), parameter :: utzero = 18.66D+00
  real ( kind = rdp ) x
  real ( kind = rdp ) y
  real ( kind = rdp ) z

  up = upper
  z = x

  if ( z < 0.0D+00 ) then
    up = .not. up
    z = - z
  end if

  if ( ltone < z .and. ( ( .not. up ) .or. utzero < z ) ) then

    if ( up ) then
      alnorm = 0.0D+00
    else
      alnorm = 1.0D+00
    end if

    return

  end if

  y = 0.5D+00 * z * z

  if ( z <= con ) then

    alnorm = 0.5D+00 - z * ( p - q * y &
      / ( y + a1 + b1 &
      / ( y + a2 + b2 &
      / ( y + a3 ))))

  else

    alnorm = r * exp ( - y ) &
      / ( z + c1 + d1 &
      / ( z + c2 + d2 &
      / ( z + c3 + d3 &
      / ( z + c4 + d4 &
      / ( z + c5 + d5 &
      / ( z + c6 ))))))

  end if

  if ( .not. up ) then
    alnorm = 1.0D+00 - alnorm
  end if

  return
end function alnorm



function dinvnr ( p )

!*****************************************************************************80
!
!! DINVNR computes the inverse of the normal distribution.
!
!  Discussion:
!
!    This routine returns X such that
!
!      CUMNOR(X) = P,
!
!    that is, so that
!
!      P = integral ( -oo <= T <= X ) exp(-U*U/2)/sqrt(2*PI) dU
!
!    The rational function is used as a
!    starting value for the Newton method of finding roots.
!
!  Reference:
!
!    William Kennedy, James Gentle,
!    Statistical Computing,
!    Marcel Dekker, NY, 1980,
!    QA276.4 K46
!
!  Parameters:
!
!    Input, real ( kind = 8 ) P.
!
!    Output, real ( kind = 8 ) DINVNR, the argument X for which the
!    Normal CDF has the value P.
!

  implicit none
!
!  integer, parameter :: rdp = selected_real_kind(15)
!  integer, parameter :: rint = selected_int_kind(6)

  real ( kind = rdp ) cum
  real ( kind = rdp ) dinvnr
  real ( kind = rdp ) dx
  real ( kind = rdp ), parameter :: eps = 1.0D-13
  integer ( kind = rint ) i
  integer ( kind = rint ), parameter :: maxit = 100
  real ( kind = rdp ) p
  real ( kind = rdp ) pp
  real ( kind = rdp ), parameter :: r2pi = 0.3989422804014326D+00
  real ( kind = rdp ) strtx
  !real ( kind = rdp ) stvaln
  real ( kind = rdp ) xcur

  pp = min ( p, 1-p )
  strtx = stvaln(pp)
  xcur = strtx
!
!  Newton iterations.
!
  do i = 1, maxit

    cum = cumnor(xcur)
    dx = ( cum - pp ) / ( r2pi * exp ( -0.5D+00 * xcur * xcur ) )
    xcur = xcur - dx

    if ( abs ( dx / xcur ) < eps ) then
      if ( p <= 1-p ) then
        dinvnr = xcur
      else
        dinvnr = -xcur
      end if
      return
    end if

  end do

  if ( p <= 1-p ) then
    dinvnr = strtx
  else
    dinvnr = -strtx
  end if

  return
end function dinvnr



function stvaln ( p )

!*****************************************************************************80
!
!! STVALN provides starting values for the inverse of the normal distribution.
!
!  Discussion:
!
!    The routine returns an X for which it is approximately true that
!      P = CUMNOR(X),
!    that is,
!      P = Integral ( -infinity < U <= X ) exp(-U*U/2)/sqrt(2*PI) dU.
!
!  Reference:
!
!    William Kennedy, James Gentle,
!    Statistical Computing,
!    Marcel Dekker, NY, 1980, page 95,
!    QA276.4 K46
!
!  Parameters:
!
!    Input, real ( kind = 8 ) P, the probability whose normal deviate
!    is sought.
!
!    Output, real ( kind = 8 ) STVALN, the normal deviate whose probability
!    is approximately P.
!

  implicit none
!
!  integer, parameter :: rdp = selected_real_kind(15)
!  integer, parameter :: rint = selected_int_kind(6)

  !real ( kind = rdp ) eval_pol
  real ( kind = rdp ) p
  real ( kind = rdp ) sgn
  real ( kind = rdp ) stvaln
  real ( kind = rdp ), parameter, dimension(0:4) :: xden = (/ &
    0.993484626060D-01, &
    0.588581570495D+00, &
    0.531103462366D+00, &
    0.103537752850D+00, &
    0.38560700634D-02 /)
  real ( kind = rdp ), parameter, dimension(0:4) :: xnum = (/ &
    -0.322232431088D+00, &
    -1.000000000000D+00, &
    -0.342242088547D+00, &
    -0.204231210245D-01, &
    -0.453642210148D-04 /)
  real ( kind = rdp ) y
  real ( kind = rdp ) z

  if ( p <= 0.5D+00 ) then

    sgn = -1.0D+00
    z = p

  else

    sgn = 1.0D+00
    z = 1.0D+00 - p

  end if

  y = sqrt ( -2.0D+00 * log ( z ) )
  stvaln = y + eval_pol ( xnum, 4, y ) / eval_pol ( xden, 4, y )
  stvaln = sgn * stvaln

  return
end function stvaln


function eval_pol ( a, n, x )

!*****************************************************************************80
!
!! EVAL_POL evaluates a polynomial at X.
!
!  Discussion:
!
!    EVAL_POL = A(0) + A(1)*X + ... + A(N)*X**N
!
!  Modified:
!
!    15 December 1999
!
!  Parameters:
!
!    Input, real ( kind = 8 ) A(0:N), coefficients of the polynomial.
!
!    Input, integer ( kind = 4 ) N, length of A.
!
!    Input, real ( kind = 8 ) X, the point at which the polynomial
!    is to be evaluated.
!
!    Output, real ( kind = 8 ) EVAL_POL, the value of the polynomial at X.
!

  implicit none
!
!  integer, parameter :: rdp = selected_real_kind(15)
!  integer, parameter :: rint = selected_int_kind(6)

  integer ( kind = rint ) n

  real ( kind = rdp ) a(0:n)
  real ( kind = rdp ) eval_pol
  integer ( kind = rint ) i
  real ( kind = rdp ) term
  real ( kind = rdp ) x

  term = a(n)
  do i = n - 1, 0, -1
    term = term * x + a(i)
  end do

  eval_pol = term

  return
end function eval_pol


function cumnor ( arg )

!*****************************************************************************
! The original code was modified
!
!
!! CUMNOR computes the cumulative normal distribution.
!
!  Discussion:
!
!    This function evaluates the normal distribution function:
!
!                              / x
!                     1       |       -t*t/2
!          P(x) = ----------- |      e       dt
!                 sqrt(2 pi)  |
!                             /-oo
!
!    This transportable program uses rational functions that
!    theoretically approximate the normal distribution function to
!    at least 18 significant decimal digits.  The accuracy achieved
!    depends on the arithmetic system, the compiler, the intrinsic
!    functions, and proper selection of the machine dependent
!    constants.
!
!  Author:
!
!    William Cody
!    Mathematics and Computer Science Division
!    Argonne National Laboratory
!    Argonne, IL 60439
!
!  Reference:
!
!    William Cody,
!    Rational Chebyshev approximations for the error function,
!    Mathematics of Computation,
!    1969, pages 631-637.
!
!    William Cody,
!    Algorithm 715:
!    SPECFUN - A Portable FORTRAN Package of Special Function Routines
!    and Test Drivers,
!    ACM Transactions on Mathematical Software,
!    Volume 19, Number 1, 1993, pages 22-32.
!
!  Parameters:
!
!    Input, real ( kind = 8 ) ARG, the upper limit of integration.
!
!    Output, real ( kind = 8 ) the Normal density CDF.
!
!  Local Parameters:
!
!    Local, real ( kind = 8 ) EPS, the argument below which anorm(x)
!    may be represented by 0.5 and above which  x*x  will not underflow.
!    A conservative value is the largest machine number X
!    such that   1.0D+00 + X = 1.0D+00   to machine precision.
!

  implicit none
!
!  integer, parameter :: rdp = selected_real_kind(15)
!  integer, parameter :: rint = selected_int_kind(6)

  real ( kind = rdp ), parameter, dimension ( 5 ) :: a = (/ &
    2.2352520354606839287D+00, &
    1.6102823106855587881D+02, &
    1.0676894854603709582D+03, &
    1.8154981253343561249D+04, &
    6.5682337918207449113D-02 /)
  real ( kind = rdp ) arg
  real ( kind = rdp ), parameter, dimension ( 4 ) :: b = (/ &
    4.7202581904688241870D+01, &
    9.7609855173777669322D+02, &
    1.0260932208618978205D+04, &
    4.5507789335026729956D+04 /)
  real ( kind = rdp ), parameter, dimension ( 9 ) :: c = (/ &
    3.9894151208813466764D-01, &
    8.8831497943883759412D+00, &
    9.3506656132177855979D+01, &
    5.9727027639480026226D+02, &
    2.4945375852903726711D+03, &
    6.8481904505362823326D+03, &
    1.1602651437647350124D+04, &
    9.8427148383839780218D+03, &
    1.0765576773720192317D-08 /)
  real ( kind = rdp ) cumnor
  real ( kind = rdp ), parameter, dimension ( 8 ) :: d = (/ &
    2.2266688044328115691D+01, &
    2.3538790178262499861D+02, &
    1.5193775994075548050D+03, &
    6.4855582982667607550D+03, &
    1.8615571640885098091D+04, &
    3.4900952721145977266D+04, &
    3.8912003286093271411D+04, &
    1.9685429676859990727D+04 /)
  real ( kind = rdp ) del
  real ( kind = rdp ) eps
  integer ( kind = rint ) i
  real ( kind = rdp ), parameter, dimension ( 6 ) :: p = (/ &
    2.1589853405795699D-01, &
    1.274011611602473639D-01, &
    2.2235277870649807D-02, &
    1.421619193227893466D-03, &
    2.9112874951168792D-05, &
    2.307344176494017303D-02 /)
  real ( kind = rdp ), parameter, dimension ( 5 ) :: q = (/ &
    1.28426009614491121D+00, &
    4.68238212480865118D-01, &
    6.59881378689285515D-02, &
    3.78239633202758244D-03, &
    7.29751555083966205D-05 /)
  real ( kind = rdp ), parameter :: root32 = 5.656854248D+00
  real ( kind = rdp ), parameter :: sixten = 16.0D+00
  real ( kind = rdp ), parameter :: sqrpi = 3.9894228040143267794D-01
  real ( kind = rdp ), parameter :: thrsh = 0.66291D+00
  real ( kind = rdp ) x
  real ( kind = rdp ) xden
  real ( kind = rdp ) xnum
  real ( kind = rdp ) y
  real ( kind = rdp ) xsq
!
!  Machine dependent constants
!
  eps = epsilon ( 1.0D+00 ) * 0.5D+00

  x = arg
  y = abs ( x )

  if ( y <= thrsh ) then
!
!  Evaluate  anorm  for  |X| <= 0.66291
!
    if ( eps < y ) then
      xsq = x * x
    else
      xsq = 0.0D+00
    end if

    xnum = a(5) * xsq
    xden = xsq
    do i = 1, 3
      xnum = ( xnum + a(i) ) * xsq
      xden = ( xden + b(i) ) * xsq
    end do
    cumnor = x * ( xnum + a(4) ) / ( xden + b(4) )
    cumnor = 0.5D+00 + cumnor
!
!  Evaluate ANORM for 0.66291 <= |X| <= sqrt(32)
!
  else if ( y <= root32 ) then

    xnum = c(9) * y
    xden = y
    do i = 1, 7
      xnum = ( xnum + c(i) ) * y
      xden = ( xden + d(i) ) * y
    end do
    cumnor = ( xnum + c(8) ) / ( xden + d(8) )
    xsq = aint ( y * sixten ) / sixten
    del = ( y - xsq ) * ( y + xsq )
    cumnor = exp ( - xsq * xsq * 0.5D+00 ) * exp ( -del * 0.5D+00 ) * cumnor

    if ( 0.0D+00 < x ) then
       cumnor = 1D+00 - cumnor
    end if
!
!  Evaluate ANORM for sqrt(32) < |X|.
!
  else

    cumnor = 0.0D+00
    xsq = 1.0D+00 / ( x * x )
    xnum = p(6) * xsq
    xden = xsq
    do i = 1, 4
      xnum = ( xnum + p(i) ) * xsq
      xden = ( xden + q(i) ) * xsq
    end do

    cumnor = xsq * ( xnum + p(5) ) / ( xden + q(5) )
    cumnor = ( sqrpi - cumnor ) / y
    xsq = aint ( x * sixten ) / sixten
    del = ( x - xsq ) * ( x + xsq )
    cumnor = exp ( - xsq * xsq * 0.5D+00 ) &
      * exp ( - del * 0.5D+00 ) * cumnor

    if ( 0.0D+00 < x ) then
        cumnor = 1D+00 - cumnor
    end if

  end if

  if ( cumnor < tiny ( cumnor ) ) then
    cumnor = 0.0D+00
  end if

  return
 end function cumnor



! Subroutine to find the inverse of a square matrix
SUBROUTINE FINDinv(matrix, inverse, n, errorflag)

    implicit none

    !Declarations
    INTEGER(rint), INTENT(IN) :: n
    REAL(rdp), INTENT(IN) :: matrix(n,n)  !Input matrix
    INTEGER(rint), INTENT(OUT) :: errorflag  !Return error status. -1 for error, 0 for normal
    REAL(rdp), INTENT(OUT) :: inverse(n,n) !Inverted matrix

    integer :: ipiv(n), info, lwork
    real(rdp) :: work(n)

    external :: dgetrf, dgetri

    errorflag = 0

    inverse = matrix
    call dgetrf(n,n,inverse,n,ipiv,info)
    if (info > 0) then
        inverse = 0
        errorflag = -1
        return
    end if

    lwork = n
    call dgetri(n,inverse,n,ipiv,work,lwork,info)
    if (info > 0) then
        inverse = 0
        errorflag = -1
        return
    end if

END SUBROUTINE FINDinv


end subroutine estimate_bct_ordinal

fit <- glm(sent ~ ztrust + zfWHR + zAfro + glasses + attract + maturity +
             tattoos, family = binomial(), data = wilson)
BF_glm <- BF(fit)

ct <- lmtest::coeftest(fit)
BFdefault1 <- BF(ct[,1], Sigma = diag(ct[,2]^2), n = attr(ct,"nobs"))
BFcoeftest1 <- BF(ct)

test_that("BF.coeftest gives same exploratory results as BF.default", {
  expect_equivalent(
    BFdefault1$BFtu_exploratory,BFcoeftest1$BFtu_exploratory
  )})

test_that("for mlm object multiple hypotheses with equality and order constraints can be computed", {
  skip_on_cran()
  fmri.lm <- lm(cbind(Superficial,Middle,Deep) ~ Face + Vehicle, data=fmri)
  constraints.fmri <- "Face_on_Deep = Face_on_Superficial = Face_on_Middle < 0 <
    Vehicle_on_Deep = Vehicle_on_Superficial = Vehicle_on_Middle;
    Face_on_Deep < Face_on_Superficial = Face_on_Middle < 0 < Vehicle_on_Deep =
    Vehicle_on_Superficial = Vehicle_on_Middle"
  set.seed(123)
  BF_fmri1 <- BF(fmri.lm, hypothesis = constraints.fmri, BF.type = "AFBF")
  #check results
  matrixmeasures <- matrix(c(0.2741948, 0.19630283, 0.7186756, 0.9998622,  2.621041,  5.093468,  13.35019, 0.023006434,
                             0.2967301, 0.02525591, 5.1451176, 0.8243133, 17.339388, 32.638428, 565.93036, 0.975270261,
                             1.0000000, 1.00000000, 1.0000000, 1.0000000,  1.000000,  1.000000, 1.00000, 0.001723304),byrow=T,nrow=3)
  expect_equivalent(
    BF_fmri1$BFtable_confirmatory, matrixmeasures, tolerance = .01
  )
})

#no seed required
fmri.lm <- lm(cbind(Superficial,Middle,Deep) ~ Face + Vehicle, data=fmri)
constraints.fmri <- "Face_on_Deep = Face_on_Superficial = Face_on_Middle < 0;
    Face_on_Deep < Face_on_Superficial = Face_on_Middle = 0"
BF_fmri2 <- BF(x = fmri.lm, hypothesis = constraints.fmri, BF.type = "AFBF")
test_that("for mlm object multiple hypotheses for same IV", {
  expect_equivalent(
    BF_fmri2$PHP_confirmatory,c(0.18941726,0.72354824,0.08703451)
  )})
BF_fmri2b <- BF(fmri.lm, hypothesis = constraints.fmri, log=TRUE, BF.type = "AFBF")
test_that("for mlm object multiple hypotheses for same IV with log", {
  expect_equivalent(
    exp(BF_fmri2b$BFtu_confirmatory),BF_fmri2$BFtu_confirmatory
)})

#no seed required
constraints.fmri <- "Face_on_Deep = Vehicle_on_Deep < 0;
    Face_on_Middle = Vehicle_on_Middle < 0"
BF_fmri3 <- BF(x = fmri.lm, hypothesis = constraints.fmri, BF.type = "AFBF")
test_that("for mlm object multiple hypotheses on same DV", {
  expect_equivalent(
    BF_fmri3$PHP_confirmatory,c(0.05595577,0.38093401,0.56311022 )
  )})

# test multivariate t test
# via lm object
intercept <- rep(1,nrow(fmri))
lm1 <- lm(cbind(Superficial,Middle,Deep) ~ -1 + intercept, data=fmri)
BF1 <- BF(lm1,hypothesis="intercept_on_Superficial=1 & intercept_on_Middle=1.1 & intercept_on_Deep=1.36",
          BF.type = "AFBF")
BF1b <- BF(lm1,hypothesis="intercept_on_Superficial<intercept_on_Middle<intercept_on_Deep;
           intercept_on_Superficial=intercept_on_Middle=intercept_on_Deep", BF.type = "AFBF")
# via mvt_test object
mvt_fmri <- mvt_test(fmri[,3:5],null = c(1,1.1,1.36))
BF2 <- BF(mvt_fmri,hypothesis="Superficial<Middle<Deep;Superficial=Middle=Deep", BF.type = "AFBF")
test_that("test multivariate Student t", {
  expect_equivalent(
    BF1$BFtu_confirmatory,BF2$BFtu_exploratory
  )
  expect_equivalent(
    BF1b$BFtu_confirmatory,BF2$BFtu_confirmatory
  )
})

set.seed(123)
X <- as.data.frame(rmvnorm(10,mean=c(0,0)))
Y <- as.data.frame(rmvnorm(20,mean=c(1,1),sigma=.5*diag(2)+.5))
mvt_test1 <- mvt_test(X,Y,null=c(-1,-1))
BF3 <- BF(mvt_test1,hypothesis="difference_V1>difference_V2;
          difference_V1=difference_V2;
          0>difference_V1>difference_V2", BF.type = "AFBF")
test_that("test multivariate Student t, independent samples", {
  expect_equivalent(
    round(BF3$BFtu_exploratory[1],2),8.87
  )
  expect_equivalent(
    round(BF3$PHP_confirmatory[1:2],2),c(0.17,0.29)
  )
})

set.seed(123)
X <- as.data.frame(rmvnorm(10,mean=c(0,0)))
Y <- as.data.frame(rmvnorm(10,mean=c(1,1),sigma=.5*diag(2)+.5))
mvt_test1 <- mvt_test(X,Y,paired=TRUE)
BF3 <- BF(mvt_test1,hypothesis="(difference_V1,difference_V2)<0;difference_V1=difference_V2=0",log=TRUE,
          BF.type = "AFBF")
Diff1 <- X[,1] - Y[,1]
Diff2 <- X[,2] - Y[,2]
ones1 <- rep(1,length(Diff1))
mlm1 <- lm(cbind(Diff1,Diff2) ~ -1 + ones1)
BF3a <- BF(mlm1,hypothesis="(ones1_on_Diff1,ones1_on_Diff2)<0;ones1_on_Diff1=ones1_on_Diff2=0",BF.type = "AFBF")
test_that("test multivariate Student t, paired samples", {
  expect_equivalent(
    round(c(BF3$PHP_exploratory),3),c(0.519,0.481)
  )
  expect_equivalent(
    round(c(BF3$BFmatrix_confirmatory[1:2,3]),1),c(2.8,1.8)
  )
  expect_equivalent(
    round(BF3a$BFtu_confirmatory,3),round(exp(BF3$BFtu_confirmatory),3)
  )
})

set.seed(123)
X <- as.data.frame(rmvnorm(10,mean=c(0)))
mvt_test1 <- mvt_test(X)
BF3a <- BF(mvt_test1,log=TRUE, BF.type = "AFBF")
t_test1 <- t_test(X)
BF3b <- BF(t_test1,log=TRUE, BF.type = "AFBF")
test_that("test multivariate Student t, paired samples", {
  expect_equivalent(
    round(c(BF3a$BFtu_exploratory),3),round(c(BF3b$BFtu_exploratory),3)
  )
})

#' @importFrom pracma rref Rank
#' @importFrom mvtnorm dmvnorm pmvnorm rmvnorm dmvt pmvt rmvt
#' @importFrom stats rWishart qt
#' @importFrom MASS ginv
#' @describeIn BF S3 method for an object of class 'lm'
#' @method BF lm
#' @export
BF.lm <- function(x,
                  hypothesis = NULL,
                  prior.hyp.explo = NULL,
                  prior.hyp.conf = NULL,
                  prior.hyp = NULL,
                  complement = TRUE,
                  log = FALSE,
                  cov.prob = .95,
                  BF.type = NULL,
                  iter = 1e4,
                  ...){

  if(!(cov.prob>0 & cov.prob<1)){
    stop("The argument 'cov.prob' is a coverage probability for the interval estimates that should lie between 0 and 1. The default is 0.95.")
  }
  CrI_LB <- (1 - cov.prob)/2
  CrI_UB <- 1 - (1 - cov.prob)/2

  if(is.null(BF.type)){
    BF.type <- "FBF"
  }
  if(!is.null(BF.type)){
    if(is.na(BF.type) | (BF.type!="FBF" & BF.type!="AFBF"))
      stop("The argument 'BF.type' must be 'FBF' (for the fractional BF) or 'AFBF' (for the adjusted fractional BF).")
  }
  if(BF.type=="AFBF"){
    bayesfactor <- "generalized adjusted fractional Bayes factors"
  }else{
    bayesfactor <- "generalized fractional Bayes factors"
  }
  testedparameter <- "regression coefficients"

  logIN <- log

  # check proper usage of argument 'prior.hyp.conf' and 'prior.hyp.explo'
  if(!is.null(prior.hyp.conf)){
    prior.hyp <- prior.hyp.conf
  }
  prior.hyp.explo <- process.prior.hyp.explo(prior_hyp_explo = prior.hyp.explo, model=x)

  # default BF on location parameters in a univarite normal linear model
  # Note that it is recommended that the fitten model is based on standardized covariates.
  x$coefficients <- as.matrix(x$coefficients)
  x$residuals <- as.matrix(x$residuals)

  P <- ncol(x$residuals)
  N <- nrow(x$residuals)
  K <- length(x$coefficients)/P # dimension of predictors per dependent variable
  dummyX <- rep(F,K)
  names(dummyX) <- row.names(x$coefficients)

  Xmat <- model.matrix(x)
  Ymat <- model.matrix(x)%*%x$coefficients + x$residuals

  # Exploratory testing of regression coefficients
  if(length(x$xlevels)==0){ #no grouping covariates: all observations receive equal fractions
    J <- 1
    dummyX <- rep(F,K)
    Nj <- nrow(Xmat)
    dvec <- rep(1,Nj)
    #set minimal fractions for the group
    bj <- ((P+K)/J)/Nj
    #no dummy covariates for factors
    dummy01TRUE <- FALSE
  }else{
    # factors are present, which are used to identify groups, which are used to determine the fractions
    # fractions are chosen such get minimal information is used per unique group (which is defined by the
    # unique combination of group indicators)

    numlevels <- unlist(lapply(x$xlevels,length))
    main_names <- names(x$xlevels)
    names_coef <- colnames(model.matrix(x))

    dummyX <- apply(matrix(unlist(lapply(1:length(main_names),function(faclev){
      unlist(lapply(1:length(names_coef),function(cf){
        grepl(main_names[faclev],names_coef[cf],fixed=TRUE)
      }))
    })),nrow=length(names_coef)),1,max)==1
    #number of groups on variations of dummy combinations
    groupcode <- as.matrix(unique(Xmat[,dummyX]))

    N <- nrow(Xmat)

    rownames(groupcode) <- unlist(lapply(1:nrow(groupcode),function(r){
      paste0("groupcode",r)
    }))
    J <- nrow(groupcode)
    if(J==nrow(Xmat)){
      stop("Not enough observations for every group. Try fitting the model without factors.")
    }

    # group membership of each observation
    dvec <- unlist(lapply(1:N,function(i){
      which(rowSums(abs(t(matrix(rep(Xmat[i,dummyX],J),ncol=J)) - groupcode))==0)
    }))
    names(dvec) <- NULL
    Nj <- c(table(dvec))
    #set minimal fractions for each group
    bj <- ((P+K)/J)/Nj
    if(max(bj)>1){#then too few observations in certain groups, then use one single minimal fraction
      bj <- rep((P+K)/sum(Nj),length=J)
      if(bj[1]>1){
        stop("Not enough observations to compute a fractional Bayes factor.")
      }
    }
  }

  #Compute sufficient statistics for all groups
  tXXj <- lapply(1:J,function(j){
    if(Nj[j]==1){
      Xmat[dvec==j,]%*%t(Xmat[dvec==j,])
    }else t(Xmat[dvec==j,])%*%Xmat[dvec==j,]
  })
  tXXj_b <- lapply(1:J,function(j){
    tXXj[[j]]*bj[j]
  })
  tXYj <- lapply(1:J,function(j){
    if(Nj[j]==1){
      as.matrix(Xmat[dvec==j,])%*%t(Ymat[dvec==j,])
    } else {t(Xmat[dvec==j,])%*%Ymat[dvec==j,]}
  })
  tXYj_b <- lapply(1:J,function(j){
    tXYj[[j]]*bj[j]
  })
  tYYj <- lapply(1:J,function(j){
    if(Nj[j]==1){
      Ymat[dvec==j,]%*%t(Ymat[dvec==j,])
    }else t(Ymat[dvec==j,])%*%Ymat[dvec==j,]

  })
  tYYj_b <- lapply(1:J,function(j){
    tYYj[[j]]*bj[j]
  })
  tXX <- Reduce("+",tXXj)
  if(min(eigen(tXX)$values)<0){
    stop("Model matrix does not seem to be of full row rank.")
  }
  tXXi <- solve(tXX)
  tXY <- Reduce("+",tXYj)
  tYY <- Reduce("+",tYYj)
  tXX_b <- Reduce("+",tXXj_b)
  tXXi_b <- solve(tXX_b)
  tXY_b <- Reduce("+",tXYj_b)
  tYY_b <- Reduce("+",tYYj_b)
  BetaHat <- solve(tXX)%*%tXY          # same as x$coefficients
  S <- tYY - t(tXY)%*%solve(tXX)%*%tXY # same as sum((x$residuals)**2)
  # sufficient statistics based on fraction of the data
  BetaHat_b <- solve(tXX_b)%*%tXY_b
  S_b <- tYY_b - t(tXY_b)%*%solve(tXX_b)%*%tXY_b

  # BF computation for exploratory analysis of separate parameters
  if(P==1){
    if(sum(row.names(x$coefficients) == "(Intercept)") > 0){
      row.names(x$coefficients)[which(row.names(x$coefficients) == "(Intercept)")] <- "Intercept"
    }
    names_coef <- row.names(x$coefficients)
  }else{
    names_coef1 <- row.names(x$coefficients)
    if(sum(names_coef1 == "(Intercept)")>0){
      names_coef1[which(names_coef1 == "(Intercept)")] <- "Intercept"
    }
    names_coef2 <- colnames(x$coefficients)
    names_coef <- unlist(lapply(1:P,function(p){
      lapply(1:K,function(k){
        paste0(names_coef1[k],"_on_",names_coef2[p])
      })
    }))
  }

  # prior hyperparameters
  df0 <- 1 # should be the same as sum(rep(bj,times=Nj))-K-P+1
  Scale0 <- kronecker(S_b,tXXi_b)
  if(BF.type=="AFBF"){
    mean0 <- as.matrix(rep(0,K*P))
  }else{
    mean0 <- as.matrix(c(BetaHat_b))
  }
  # posterior hyperparameters
  dfN <- N-K-P+1
  ScaleN <- kronecker(S,tXXi)/(N-K-P+1)
  meanN <- as.matrix(c(BetaHat))
  row.names(meanN) <- row.names(mean0) <- names_coef

  # Hypotheses for exploratory test
  # H0: beta = 0
  # H1: beta < 0
  # H2: beta > 0
  relfit <- t(matrix(unlist(lapply(1:(K*P),function(k){
    c(dt((0-meanN[k,1])/sqrt(ScaleN[k,k]),df=dfN,log=TRUE) - log(sqrt(ScaleN[k,k])),
      pt((0-meanN[k,1])/sqrt(ScaleN[k,k]),df=dfN,lower.tail = TRUE,log.p=TRUE),
      pt((0-meanN[k,1])/sqrt(ScaleN[k,k]),df=dfN,lower.tail = FALSE,log.p=TRUE))
  })),nrow=3))
  relcomp <- t(matrix(unlist(lapply(1:(K*P),function(k){
    c(dt((0-mean0[k,1])/sqrt(Scale0[k,k]),df=df0,log=TRUE) - log(sqrt(Scale0[k,k])),
      pt((0-mean0[k,1])/sqrt(Scale0[k,k]),df=df0,lower.tail = TRUE,log.p=TRUE),
      pt((0-mean0[k,1])/sqrt(Scale0[k,k]),df=df0,lower.tail = FALSE,log.p=TRUE))
  })),nrow=3))
  colnames(relfit) <- colnames(relcomp) <- c("p(=0)","Pr(<0)","Pr(>0)")
  row.names(relcomp) <- row.names(relfit) <- names_coef

  BFtu_exploratory <- relfit - relcomp
  colnames(BFtu_exploratory) <- c("BF0u","BF1u","BF2u")
  maxrows <- apply(BFtu_exploratory,1,max)
  norm_BF_explo <- exp(BFtu_exploratory - maxrows %*% t(rep(1,3))) *
    (rep(1,nrow(relcomp)) %*% t(prior.hyp.explo[[1]]))
  PHP_exploratory <- norm_BF_explo / apply(norm_BF_explo,1,sum)
  colnames(PHP_exploratory) <- c("p(=0)","Pr(<0)","Pr(>0)")

  #compute estimates (test)
  postestimates <- cbind(meanN,meanN,
                         t(matrix(unlist(lapply(1:length(meanN),function(coef){
                           ub <- qt(p=CrI_UB,df=dfN)*sqrt(ScaleN[coef,coef])+meanN[coef,1]
                           lb <- qt(p=CrI_LB,df=dfN)*sqrt(ScaleN[coef,coef])+meanN[coef,1]
                           return(c(lb,ub))
                         })),nrow=2)
                         ),
                         1-pt((0-meanN[,1])/sqrt(diag(ScaleN)),df=dfN)
  )
  row.names(postestimates) <- names_coef
  colnames(postestimates) <- c("mean","median",paste0(as.character(round(CrI_LB*100,7)),"%"),
                               paste0(as.character(round(CrI_UB*100,7)),"%"),"Pr(>0)")

  # Additional exploratory tests of main effects and interaction effects
  # in the case of an aov type object
  if(sum(class(x)=="aov")==1 & J > 1){
    testedparameter <- "group means"

    mains <- unlist(lapply(1:length(x$xlevels),function(fac){
      unlist(lapply(1:length(x$xlevels[[fac]]),function(lev){
        paste0(names(x$xlevels)[fac],x$xlevels[[fac]][lev])
      }))
    }))

    # check main effects
    BFmain <- unlist(lapply(1:length(numlevels),function(fac){
      # if( number of levels of factor is less than the number of columns of main effect in model.matrix
      # then to test the main effect all effects of these columns should be zero, and if the number of
      # levels is equal to the number of columns then the effects of the columns should be equal.)
      name1 <- names(numlevels[fac])
      mains1 <- mains[sum(numlevels[1:fac])-numlevels[fac]+1:numlevels[fac]]
      which0 <- unlist(lapply(1:length(colnames(Xmat)),function(col){
        sum(colnames(Xmat)[col]==mains1)==1
      }))
      if(sum(which0) > 0){
        if(numlevels[fac]==sum(which0)){
          #there is no reference group
          #test whether the effects of all dummys are equal
          if(P > 1){
            RrE_f <- matrix(0,nrow=sum(which0)-1,ncol=length(colnames(Xmat)))
            for(r1 in 1:(sum(which0)-1)){
              RrE_f[r1,which(which0)[r1]] <- 1
              RrE_f[r1,which(which0)[r1+1]] <- -1
            }
            RrE_f <- cbind(kronecker(diag(P),RrE_f),0)
            relcomp_f <- MatrixStudent_measures(Mean1=matrix(mean0,ncol=P),Scale1=S_b,tXXi1=tXXi_b,
                                                df1=df0,RrE1=RrE_f,RrO1=NULL,Names1=NULL,constraints1=NULL,
                                                MCdraws=iter)
            relfit_f <- MatrixStudent_measures(Mean1=BetaHat,Scale1=S,tXXi1=tXXi,df1=dfN,RrE1=RrE_f,
                                               RrO1=NULL,Names1=NULL,constraints1=NULL,MCdraws=iter)
          }else{
            RrE_f <- matrix(0,nrow=sum(which0)-1,ncol=length(colnames(Xmat))+1)
            for(r1 in 1:(sum(which0)-1)){
              RrE_f[r1,which(which0)[r1]] <- 1
              RrE_f[r1,which(which0)[r1+1]] <- -1
            }
            relcomp_f <- Student_measures(mean1=mean0,Scale1=Scale0,df1=df0,RrE1=RrE_f,RrO1=NULL)
            relfit_f <- Student_measures(mean1=meanN,Scale1=ScaleN,df1=dfN,RrE1=RrE_f,RrO1=NULL)
          }

        }else{
          #there is a reference group
          #test whether the effects of the remaining dummys are zero of these indicators equals zero
          if(P > 1){
            RrE_f <- matrix(0,nrow=sum(which0),ncol=length(colnames(Xmat)))
            for(r1 in 1:sum(which0)){
              RrE_f[r1,which(which0)[r1]] <- 1
            }
            RrE_f <- cbind(kronecker(diag(P),RrE_f),rep(0,sum(which0)*P))
            relcomp_f <- MatrixStudent_measures(Mean1=matrix(mean0,ncol=P),Scale1=S_b,tXXi1=tXXi_b,
                                                df1=df0,RrE1=RrE_f,RrO1=NULL,Names1=NULL,constraints1=NULL,
                                                MCdraws=iter)
            relfit_f <- MatrixStudent_measures(Mean1=BetaHat,Scale1=S,tXXi1=tXXi,df1=dfN,RrE1=RrE_f,
                                               RrO1=NULL,Names1=NULL,constraints1=NULL,MCdraws=iter)
          }else{
            RrE_f <- matrix(0,nrow=sum(which0),ncol=length(colnames(Xmat))+1)
            for(r1 in 1:sum(which0)){
              RrE_f[r1,which(which0)[r1]] <- 1
            }
            relcomp_f <- Student_measures(mean1=mean0,Scale1=Scale0,df1=df0,RrE1=RrE_f,RrO1=NULL)
            relfit_f <- Student_measures(mean1=meanN,Scale1=ScaleN,df1=dfN,RrE1=RrE_f,RrO1=NULL)
          }
        }
        BFtu <- relfit_f[1] - relcomp_f[1]
        names(BFtu) <- name1
        return(c(BFtu,relfit_f[1],relcomp_f[1]))
      }
    }))
    #compute Bayes factors for testing main effects if present
    if(length(BFmain)>0){ # then there are main effects
      names_main <- names(BFmain[(0:(length(BFmain)/3-1))*3+1])
      BFtu_main <- matrix(c(BFmain[(0:(length(BFmain)/3-1))*3+1],rep(log(1),length(BFmain)/3)),
                          nrow=length(BFmain)/3)
      row.names(BFtu_main) <- names_main
      colnames(BFtu_main) <- c("BF0u","BFuu")
      maxBFtu <- apply(BFtu_main,1,max)
      norm_BF_main_explo <- exp(BFtu_main - maxBFtu %*% t(rep(1,ncol(BFtu_main)))) *
        (rep(1,nrow(BFtu_main)) %*% t(prior.hyp.explo[[2]]))
      PHP_main <- norm_BF_main_explo / apply(norm_BF_main_explo,1,sum)
      # PHP_main <- exp(BFtu_main - maxBFtu %*% t(rep(1,ncol(BFtu_main)))) /
      #   apply(exp(BFtu_main - maxBFtu %*% t(rep(1,ncol(BFtu_main)))),1,sum)
      colnames(PHP_main) <- c("Pr(no effect)","Pr(full model)")
      if(logIN == FALSE){
        BFtu_main <- exp(BFtu_main)
      }
    }else{ PHP_main <- BFtu_main <- NULL}
    #check whether interaction effects are present
    prednames <- names(attr(x$term,"dataClasses"))
    matcov <- cbind(matrix(unlist(lapply(1:K,function(col){
      colx <- colnames(Xmat)[col]
      unlist(lapply(1:length(prednames),function(pred){
        grepl(prednames[pred],colx)
      }))
    })),nrow=length(prednames)),rep(F,length(prednames)))
    row.names(matcov) <- prednames
    colnames(matcov) <- c(colnames(Xmat),"dummy")
    BFtu_interaction0 <- list()
    count_interaction <- 0

    for(c1 in 1:ncol(matcov)){
      if(c1 < ncol(matcov)){
        numeffects_c <- sum(matcov[,c1])
        if(numeffects_c>1){
          count_interaction <- count_interaction + 1
          interactionset <- names(which(matcov[,c1]))
          whichx <- apply(matcov[which(matcov[,c1]),],2,sum)==length(interactionset)
          if(P > 1){
            RrE_ia <- matrix(0,nrow=sum(whichx),ncol=K)
            for(r1 in 1:sum(whichx)){RrE_ia[r1,which(whichx)[r1]]<-1}
            RrE_ia <- cbind(kronecker(diag(P),RrE_ia),rep(0,sum(whichx)*P))
            relcomp_ia <- MatrixStudent_measures(Mean1=matrix(mean0,ncol=P),Scale1=S_b,tXXi1=tXXi_b,
                                                 df1=df0,RrE1=RrE_ia,RrO1=NULL,Names1=NULL,
                                                 constraints1=NULL,MCdraws=iter)
            relfit_ia <- MatrixStudent_measures(Mean1=BetaHat,Scale1=S,tXXi1=tXXi,df1=dfN,RrE1=RrE_ia,
                                                RrO1=NULL,Names1=NULL,constraints1=NULL,MCdraws=iter)
          }else{
            RrE_ia <- matrix(0,nrow=sum(whichx),ncol=K+1)
            for(r1 in 1:sum(whichx)){RrE_ia[r1,which(whichx)[r1]]<-1}
            relcomp_ia <- Student_measures(mean1=mean0,Scale1=Scale0,df1=df0,RrE1=RrE_ia,RrO1=NULL)
            relfit_ia <- Student_measures(mean1=meanN,Scale1=ScaleN,df1=dfN,RrE1=RrE_ia,RrO1=NULL)
          }
          names(relcomp_ia) <- c("c=","c>")
          names(relfit_ia) <- c("f=","f>")
          BFtu_ia <- relfit_ia[1] - relcomp_ia[1]
          names(BFtu_ia) <- paste(interactionset,collapse=":")
          BFtu_interaction0[[count_interaction]] <- c(BFtu_ia,relcomp_ia[1],relfit_ia[1])
          #exclude other columns from Xmat that have been used to avoid double results
          matcov <- matcov[,-(which(whichx[-c1])+1)]
        }
      }
    }
    #compute Bayes factors for testing interaction effects if present
    if(count_interaction>0){ # then there are interaction effects
      BFtu_interaction0 <- unlist(BFtu_interaction0)
      names_interaction <- names(BFtu_interaction0[(0:(length(BFtu_interaction0)/3-1))*3+1])
      BFtu_interaction <- matrix(c(BFtu_interaction0[(0:(length(BFtu_interaction0)/3-1))*3+1],
                                   rep(log(1),length(BFtu_interaction0)/3)),nrow=length(BFtu_interaction0)/3)
      row.names(BFtu_interaction) <- names_interaction
      colnames(BFtu_interaction) <- c("BF0u","BFuu")
      maxrows <- apply(BFtu_interaction,1,max)
      norm_BF_inter_explo <- exp(BFtu_interaction - maxrows %*% t(rep(1,ncol(BFtu_interaction)))) *
        (rep(1,nrow(BFtu_interaction)) %*% t(prior.hyp.explo[[3]]))
      PHP_interaction <- norm_BF_inter_explo / apply(norm_BF_inter_explo,1,sum)
      if(logIN == FALSE){
        BFtu_interaction <- exp(BFtu_interaction)
      }
      # PHP_interaction <- norm_BF_inter_explo / apply(norm_BF_inter_explo,1,sum)
      colnames(PHP_interaction) <- c("Pr(no effect)","Pr(full model)")
    }else{ PHP_interaction <- BFtu_interaction <- NULL}
    #BFtu_exploratory <- rbind(BFtu_main,BFtu_interaction)
    #PHP_exploratory <- rbind(PHP_main,PHP_interaction)
  }

  if(logIN == FALSE){
    BFtu_exploratory <- exp(BFtu_exploratory)
  }

  # confirmatory BF test
  if(!is.null(hypothesis)){

    #then constraints on regression coefficients
    matrixnames <- matrix(names_coef,nrow=K)

    # translate named constraints to matrices with coefficients for constraints
    parse_hyp <- parse_hypothesis(names_coef,hypothesis)
    parse_hyp$hyp_mat <- do.call(rbind, parse_hyp$hyp_mat)
    RrList <- make_RrList2(parse_hyp)
    RrE <- RrList[[1]]
    RrO <- RrList[[2]]

    if(length(RrE)==1){
      RrStack <- rbind(do.call(rbind,RrE),do.call(rbind,RrO))
      RrStack <- interval_RrStack(RrStack)
    }else{
      RrStack_list <- lapply(1:length(RrE),function(h){
        interval_RrStack(rbind(RrE[[h]],RrO[[h]]))
      })
      RrStack <- do.call(rbind,RrStack_list)
      RrStack <- interval_RrStack(RrStack)
      # RrStack_list <- do.call(rbind,lapply(1:length(RrE),function(h){
      #   rbind(RrE[[h]],RrO[[h]])
      # }))
      # RrStack <- interval_RrStack(RrStack_list)
    }
    if(nrow(RrStack)>1){
      RStack <- RrStack[,-(K*P+1)]
      rStack <- RrStack[,(K*P+1)]
    }else{
      RStack <- matrix(RrStack[,-(K*P+1)],nrow=1)
      rStack <- RrStack[,(K*P+1)]
    }

    # check if a common boundary exists for prior location under all constrained hypotheses
    if(nrow(RrStack) > 1 & BF.type == "AFBF"){
      rref_ei <- rref(RrStack)
      nonzero <- rref_ei[,P*K+1]!=0
      if(max(nonzero)>0){
        row1 <- max(which(nonzero))
        if(sum(abs(rref_ei[row1,1:(P*K)]))==0){
          stop("No common boundary point for prior location. Conflicting constraints.")
        }
      }
    }

    #number of hypotheses that are specified
    numhyp <- length(RrO)

    if(P > 1){

      #default prior location
      if(BF.type=="AFBF"){
        Mean0 <- matrix(c(ginv(RStack)%*%rStack),nrow=K,ncol=P)
      }else{
        Mean0 <- BetaHat #mean0
      }
      #Mean0 <- matrix(c(ginv(RStack)%*%rStack),nrow=K,ncol=P)

      relmeasunlist <- unlist(lapply(1:numhyp,function(h){
        # Check whether the constraints are on a single row or column, if so
        # use the analytic expression, else using a Monte Carlo estimate.
        RrStack <- rbind(RrE[[h]],RrO[[h]])
        Rcheck <- Reduce("+",lapply(1:nrow(RrStack),function(row1){
          abs(matrix(RrStack[row1,-(K*P+1)],nrow=K))
        }))
        RcheckRow <- apply(Rcheck,1,sum)
        RcheckCol <- apply(Rcheck,2,sum)
        if(sum(RcheckRow!=0)==1){ # use multivariate Student distributions
          K1 <- which(RcheckRow!=0)
          # posterior hyperparameters
          dfN <- N-K-P+1
          ScaleN <- S*tXXi[K1,K1]/(N-K-P+1) # off-diagonal elements have no meaning
          meanN <- as.matrix(c(BetaHat[K1,]))
          # exclude inactive rows
          if(is.null(RrE[[h]])){
            RrE_h=NULL
          }else{
            if(nrow(RrE[[h]])==1){
              RrE_h <- t(as.matrix(RrE[[h]][,c((0:(P-1))*K+K1,P*K+1)]))
            }else{
              RrE_h <- RrE[[h]][,c((0:(P-1))*K+K1,P*K+1)]
            }
          }
          if(is.null(RrO[[h]])){
            RrO_h=NULL
          }else{
            if(nrow(RrO[[h]])==1){
              RrO_h <- t(as.matrix(RrO[[h]][,c((0:(P-1))*K+K1,P*K+1)]))
            }else{
              RrO_h <- RrO[[h]][,c((0:(P-1))*K+K1,P*K+1)]
            }
          }
          # prior hyperparameters
          df0 <- 1 # should be the same as sum(rep(bj,times=Nj))-K-P+1
          Scale0 <- S_b*tXXi_b[K1,K1]
          mean0 <- Mean0[K1,]
          # compute relative measures of fit and complexity
          relcomp_h <- Student_measures(mean1=mean0,Scale1=Scale0,df1=df0,RrE1=RrE_h,
                                        RrO1=RrO_h,names1=matrixnames[K1,],
                                        constraints1=parse_hyp$original_hypothesis[h])
          relfit_h <- Student_measures(mean1=meanN,Scale1=ScaleN,df1=dfN,RrE1=RrE_h,
                                       RrO1=RrO_h,names1=matrixnames[K1,],
                                       constraints1=parse_hyp$original_hypothesis[h])

        }else if(sum(RcheckCol!=0)==1){ # use multivariate Student distributions
          P1 <- which(RcheckCol!=0)
          # prior hyperparameters
          df0 <- 1 # should be the same as sum(rep(bj,times=Nj))-K-P+1
          Scale0 <- S_b[P1,P1]*tXXi_b
          mean0 <- Mean0[,P1]
          # posterior hyperparameters
          dfN <- N-K-P+1
          ScaleN <- S[P1,P1]*tXXi/(N-K-P+1) # off-diagonal elements have no meaning
          meanN <- as.matrix(c(BetaHat[,P1]))
          # exclude inactive rows
          if(is.null(RrE[[h]])){RrE_h=NULL
          }else{
            if(nrow(RrE[[h]])==1){
              RrE_h <- t(as.matrix(RrE[[h]][,c((P1-1)*K+1:K,P*K+1)]))
            }else{
              RrE_h <- RrE[[h]][,c((P1-1)*K+1:K,P*K+1)]
            }
          }
          if(is.null(RrO[[h]])){RrO_h=NULL
          }else{
            if(nrow(RrO[[h]])==1){
              RrO_h <- t(as.matrix(RrO[[h]][,c((P1-1)*K+1:K,P*K+1)]))
            }else{
              RrO_h <- RrO[[h]][,c((P1-1)*K+1:K,P*K+1)]
            }
          }
          # compute relative measures of fit and complexity
          relcomp_h <- Student_measures(mean0,Scale0,df0,RrE_h,RrO_h,names1=matrixnames[,P1],
                                        constraints1=parse_hyp$original_hypothesis[h])
          relfit_h <- Student_measures(meanN,ScaleN,dfN,RrE_h,RrO_h,names1=matrixnames[,P1],
                                       constraints1=parse_hyp$original_hypothesis[h])

        }else{ #use Matrix-Student distributions with Monte Carlo estimate
          df0 <- 1
          dfN <- N-K-P+1
          relfit_h <- MatrixStudent_measures(Mean1=BetaHat,Scale1=S,tXXi1=tXXi,df1=dfN,RrE1=RrE[[h]],RrO1=RrO[[h]],
                                             Names1=matrix(names_coef,ncol=P),constraints1=parse_hyp$original_hypothesis[h],
                                             MCdraws=iter)
          relcomp_h <- MatrixStudent_measures(Mean1=Mean0,Scale1=S_b,tXXi1=tXXi_b,df1=df0,RrE1=RrE[[h]],RrO1=RrO[[h]],
                                              Names1=matrix(names_coef,ncol=P),constraints1=parse_hyp$original_hypothesis[h],
                                              MCdraws=iter)
        }
        return(list(relfit_h,relcomp_h))
      }))

      relfit <- t(matrix(unlist(relmeasunlist)[rep((0:(numhyp-1))*4,each=2)+rep(1:2,numhyp)],nrow=2))
      row.names(relfit) <- parse_hyp$original_hypothesis
      colnames(relfit) <- c("f_E","f_O")
      relcomp <- t(matrix(unlist(relmeasunlist)[rep((0:(numhyp-1))*4,each=2)+rep(3:4,numhyp)],nrow=2))
      row.names(relcomp) <- parse_hyp$original_hypothesis
      colnames(relcomp) <- c("c_E","c_O")

      # Compute relative fit/complexity for the complement hypothesis
      active.complement <- TRUE
      if(complement==TRUE){
        relfit <- MatrixStudent_prob_Hc(Mean1=BetaHat,Scale1=S,tXXi1=tXXi,df1=N-K-P+1,
                                        relmeas=as.matrix(relfit),RrO1=RrO)
        relcomp <- MatrixStudent_prob_Hc(Mean0,S_b,tXXi_b,1,as.matrix(relcomp),RrO)
        hypothesisshort <- unlist(lapply(1:nrow(relfit),function(h) paste0("H",as.character(h))))
        row.names(relcomp) <- row.names(relfit) <- hypothesisshort
        if(nrow(relfit) == numhyp){active.complement <- FALSE}
      }

    }else{
      # one dependent variable and posterior/prior have Student t distributions

      # prior hyperparameters
      df0 <- 1 # should be the same as sum(rep(bj,times=Nj))-K-P+1
      Scale0 <- kronecker(S_b,tXXi_b)
      #default prior location
      if(BF.type=="AFBF"){
        mean0 <- ginv(RStack)%*%rStack
      }else{
        mean0 <- BetaHat #mean0
      }
      # posterior hyperparameters
      dfN <- N-K-P+1
      ScaleN <- kronecker(S,tXXi)/(N-K-P+1)
      meanN <- as.matrix(c(BetaHat))

      relcomp <- t(matrix(unlist(lapply(1:numhyp,function(h){
        Student_measures(mean1=mean0,Scale1=Scale0,df1=df0,RrE1=RrE[[h]],RrO1=RrO[[h]],
                         names1=names_coef,constraints1=parse_hyp$original_hypothesis[h])
      })),nrow=2))

      relfit <- t(matrix(unlist(lapply(1:numhyp,function(h){
        Student_measures(mean1=meanN,Scale1=ScaleN,df1=dfN,RrE[[h]],RrO[[h]],
                         names1=names_coef,constraints1=parse_hyp$original_hypothesis[h])
      })),nrow=2))
      colnames(relcomp) <- c("c_E","c_O")
      colnames(relfit) <- c("f_E","f_O")
      row.names(relcomp)[1:numhyp] <- parse_hyp$original_hypothesis
      row.names(relfit)[1:numhyp] <- parse_hyp$original_hypothesis

      active.complement <- TRUE
      if(complement == TRUE){
        # Compute relative fit/complexity for the complement hypothesis
        relfit <- Student_prob_Hc(mean1=meanN,scale1=ScaleN,df1=dfN,relmeas1=relfit,
                                  constraints=hypothesis,RrO1=RrO)
        relcomp <- Student_prob_Hc(mean1=mean0,scale1=Scale0,df1=df0,relmeas1=relcomp,
                                   constraints=hypothesis,RrO1=RrO)
        row.names(relcomp)[1:numhyp] <- parse_hyp$original_hypothesis
        row.names(relfit)[1:numhyp] <- parse_hyp$original_hypothesis
        if(nrow(relfit) == numhyp){active.complement <- FALSE}
      }
    }

    # the BF for the complement hypothesis vs Hu needs to be computed.
    BFtu_conf <- relfit - relcomp
    BFtu_confirmatory <- c(apply(BFtu_conf, 1, sum))
    # Check input of prior probabilies
    if(is.null(prior.hyp)){
      priorprobs <- rep(1/length(BFtu_confirmatory),length(BFtu_confirmatory))
    }else{
      if(!is.numeric(prior.hyp) || length(prior.hyp)!=length(BFtu_confirmatory)){
        if(complement == TRUE && active.complement == FALSE && length(prior.hyp)==length(BFtu_confirmatory)+1){
          # the prior prob for the complement hypothesis is given but it is not used as the constrained hypotheses
          # cover the entire parameter space
          priorprobs <- prior.hyp[1:length(BFtu_confirmatory)]
          priorprobs <- priorprobs / sum(priorprobs)
        }else{
          warning(paste0("Argument 'prior.hyp' should be numeric and of length ",as.character(length(BFtu_confirmatory)),
                         ". Equal prior probabilities are used."))
          priorprobs <- rep(1/length(BFtu_confirmatory),length(BFtu_confirmatory))
        }
      }else{
        priorprobs <- prior.hyp
        priorprobs <- priorprobs / sum(priorprobs)
      }
    }

    names(priorprobs) <- names(BFtu_confirmatory)
    BFtu_scaled <- BFtu_confirmatory - max(BFtu_confirmatory)
    PHP_confirmatory <- exp(BFtu_scaled)*priorprobs / sum(exp(BFtu_scaled)*priorprobs)
    BFtable <- cbind(relcomp,relfit,relfit[,1]-relcomp[,1],relfit[,2]-relcomp[,2],
                     apply(relfit,1,sum)-apply(relcomp,1,sum),PHP_confirmatory)
    BFtable[,1:7] <- exp(BFtable[,1:7])
    row.names(BFtable) <- names(PHP_confirmatory)
    colnames(BFtable) <- c("comp_E","comp_O","fit_E","fit_O","BF_E","BF_O","BF","PHP")
    BFmatrix_confirmatory <- matrix(rep(BFtu_confirmatory,length(BFtu_confirmatory)),ncol=length(BFtu_confirmatory)) -
      t(matrix(rep(BFtu_confirmatory,length(BFtu_confirmatory)),ncol=length(BFtu_confirmatory)))
    diag(BFmatrix_confirmatory) <- log(1)
    row.names(BFmatrix_confirmatory) <- colnames(BFmatrix_confirmatory) <- names(BFtu_confirmatory)
    if(nrow(relfit)==length(parse_hyp$original_hypothesis)){
      hypotheses <- parse_hyp$original_hypothesis
    }else{
      hypotheses <- c(parse_hyp$original_hypothesis,"complement")
    }
    if(logIN == FALSE){
      BFtu_confirmatory <- exp(BFtu_confirmatory)
      BFmatrix_confirmatory <- exp(BFmatrix_confirmatory)
    }

  }else{
    BFtu_confirmatory <- PHP_confirmatory <- BFmatrix_confirmatory <- relfit <-
      relcomp <- hypotheses <- BFtable <- priorprobs <- NULL
  }

  if(sum(class(x)=="aov")==1 & J > 1){
    BFlm_out <- list(
      BFtu_exploratory=BFtu_exploratory,
      BFtu_main=BFtu_main,
      BFtu_interaction=BFtu_interaction,
      PHP_exploratory=PHP_exploratory,
      PHP_main=PHP_main,
      PHP_interaction=PHP_interaction,
      BFtu_confirmatory=BFtu_confirmatory,
      PHP_confirmatory=PHP_confirmatory,
      BFmatrix_confirmatory=BFmatrix_confirmatory,
      BFtable_confirmatory=BFtable,
      prior.hyp.conf=priorprobs,
      prior.hyp.explo=prior.hyp.explo,
      hypotheses=hypotheses,
      estimates=postestimates,
      model=x,
      bayesfactor=bayesfactor,
      parameter=testedparameter,
      log = logIN,
      fraction_number_groupIDs = J,
      fraction_groupID_observations = dvec,
      fraction_groupID = bj,
      call=match.call())
  }else{
    BFlm_out <- list(
      BFtu_exploratory=BFtu_exploratory,
      PHP_exploratory=PHP_exploratory,
      BFtu_confirmatory=BFtu_confirmatory,
      PHP_confirmatory=PHP_confirmatory,
      BFmatrix_confirmatory=BFmatrix_confirmatory,
      BFtable_confirmatory=BFtable,
      prior.hyp.explo=prior.hyp.explo,
      prior.hyp.conf=priorprobs,
      hypotheses=hypotheses,
      estimates=postestimates,
      model=x,
      bayesfactor=bayesfactor,
      parameter=testedparameter,
      log = logIN,
      fraction_number_groupIDs = J,
      fraction_groupID_observations = dvec,
      fraction_groupID = bj,
      call=match.call())
  }

  names(BFlm_out$fraction_group_identifier) <- NULL

  class(BFlm_out) <- "BF"

  return(BFlm_out)
}

# compute relative meausures (fit or complexity) under a multivariate Student t distribution
MatrixStudent_measures <- function(Mean1,Scale1,tXXi1,df1,RrE1,RrO1,Names1=NULL,
                                   constraints1=NULL,MCdraws=1e4,...){
  # constraints1 = parse_hyp$original_hypothesis
  # RrE1 <- matrix(0,nrow=1,ncol=ncol(RrO1))
  # RrE1[1,1:2] <- c(-1,1)
  K <- nrow(Mean1)
  P <- ncol(Mean1)
  # vectorize the mean
  mean1 <- c(Mean1)
  relE <- relO <- log(1)
  if(!is.null(RrE1) && is.null(RrO1)){ #only equality constraints
    RE1 <- RrE1[,-(K*P+1)]
    if(!is.matrix(RE1)){
      RE1 <- matrix(RE1,ncol=K*P)
    }
    rE1 <- RrE1[,(K*P+1)]
    qE1 <- nrow(RE1)

    temp1 <- rWishart(MCdraws,df1+P-1,solve(Scale1))
    temp2 <- lapply(seq(dim(temp1)[3]), function(x) temp1[,,x])
    SigmaList <- lapply(temp2,solve)
    covm1_E <- lapply(SigmaList,function(temp) RE1%*%(kronecker(temp,tXXi1))%*%t(RE1) )
    mean1_E <- c(RE1 %*% mean1)
    relE <- log(mean(unlist(lapply(covm1_E,function(temp) dmvnorm(rE1,mean=mean1_E,sigma=temp)))))

  }else{
    if(is.null(RrE1) && !is.null(RrO1)){ #only order constraints
      RO1 <- RrO1[,-(K*P+1)]
      if(!is.matrix(RO1)){
        RO1 <- matrix(RO1,ncol=K*P)
      }
      qO1 <- nrow(RO1)
      rO1 <- RrO1[,(K*P+1)]

      if(Rank(RO1)==nrow(RO1)){ #RO1 is of full row rank. So use transformation.

        Scale1inv <- solve(Scale1)
        relO <- unlist(lapply(1:1e3,function(s){
          Sigma1 <- solve(rWishart(1,df=df1+P-1,Sigma=Scale1inv)[,,1])
          meanO <- c(RO1%*%mean1)
          covmO <- RO1%*%kronecker(Sigma1,tXXi1)%*%t(RO1)
          pmvnorm(lower=rO1,upper=Inf,mean=meanO,sigma=covmO,...)[1]
        }))

        relO <- mean(relO[relO!="NaN"])
        if(relO>1){relO <- 1}
        if(relO<0){relO <- 0}
        relO <- log(relO)

      }else{ #no linear transformation can be used; pmvt cannot be used. Use bain with a multivariate normal approximation
        #compute covariance matrix for multivariate normal distribution
        mean1 <- c(Mean1)
        names(mean1) <- c(Names1)
        if(df1>2){ #posterior measures
          covm1 <- kronecker(Scale1,tXXi1)*df1/(df1-2) #estimate of covariance matrix
        }else if(df1==2){ #posterior measures
          covm1 <- kronecker(Scale1,tXXi1)*4
        }else{
          covm1 <- kronecker(Scale1,tXXi1)*5
        }
        bain_res <- bain(x=mean1,hypothesis=constraints1,Sigma=covm1,n=999) #n not used in computation
        relO <- log(bain_res$fit[1,3])

        # bain1 <- bain::bain(mean1,Sigma1=covm1,RrE1,RrO1,n=10) # choice of n does not matter
        # extract posterior probability (Fit_eq) from bain-object)
        # warning("Check if this works now")
      }
    }else{ #hypothesis with equality and order constraints

      RE1 <- RrE1[,-(K*P+1)]
      if(!is.matrix(RE1)){
        RE1 <- matrix(RE1,ncol=K*P)
      }
      rE1 <- RrE1[,(K*P+1)]
      qE1 <- nrow(RE1)
      RO1 <- RrO1[,-(K*P+1)]
      if(!is.matrix(RO1)){
        RO1 <- matrix(RO1,ncol=K*P)
      }
      qO1 <- nrow(RO1)
      rO1 <- RrO1[,(K*P+1)]
      Rr1 <- rbind(RrE1,RrO1)
      R1 <- rbind(RE1,RO1)
      r1 <- c(rE1,rO1)
      qC1 <- length(r1)

      temp1 <- rWishart(MCdraws,df1+P-1,solve(Scale1))
      temp2 <- lapply(seq(dim(temp1)[3]), function(x) temp1[,,x])
      SigmaList <- lapply(temp2,solve)
      covm1_E <- lapply(SigmaList,function(temp) RE1%*%(kronecker(temp,tXXi1))%*%t(RE1) )
      mean1_E <- RE1 %*% mean1
      relE <- log(mean(unlist(lapply(covm1_E,function(temp) dmvnorm(rE1,mean=mean1_E,sigma=temp)))))

      if(Rank(Rr1) == nrow(Rr1)){
        covm1_O <- lapply(SigmaList,function(temp) R1%*%(kronecker(temp,tXXi1))%*%t(R1) )
        mean1_O <- c(R1%*%mean1)

        mean1_OE <- lapply(covm1_O,function(temp){
          as.vector(mean1_O[(qE1+1):qC1] +
                      c(matrix(temp[(qE1+1):qC1,1:qE1],ncol=qE1)%*%solve(temp[1:qE1,1:qE1])%*%(rE1-mean1_E)))
        })
        covm1_OE <- lapply(covm1_O,function(temp) temp[(qE1+1):qC1,(qE1+1):qC1] -
                             matrix(temp[(qE1+1):qC1,1:qE1],ncol=qE1)%*%solve(temp[1:qE1,1:qE1])%*%
                             matrix(temp[1:qE1,(qE1+1):qC1],nrow=qE1))
        #check covariance because some can be nonsymmetric due to a generation error
        welk1 <- which(unlist(lapply(covm1_OE,function(temp) isSymmetric(temp,
                                                                         tol = sqrt(.Machine$double.eps),check.attributes = FALSE) &&
                                       min(eigen(temp)$values)>sqrt(.Machine$double.eps) )))
        covm1_OE <- covm1_OE[welk1]
        mean1_OE <- mean1_OE[welk1]
        relO <- mean(mapply(function(mu_temp,Sigma_temp)
          pmvnorm(lower=rO1,upper=rep(Inf,qO1),mean=mu_temp,sigma=Sigma_temp,...)[1],mean1_OE,covm1_OE))

        if(relO > 1){relO <- 1}
        if(relO < 0){relO <- 0}
        relO <- log(relO)
      }else{ #use bain for the computation of the probability

        mean1 <- c(Mean1)
        names(mean1) <- c(Names1)
        if(df1>2){ #posterior measures
          covm1 <- kronecker(Scale1,tXXi1)*df1/(df1-2)
        }else if(df1==2){ #posterior measures
          covm1 <- kronecker(Scale1,tXXi1)*4
        }else{
          covm1 <- kronecker(Scale1,tXXi1)*5
        }
        bain_res <- bain(x=mean1,hypothesis=constraints1,Sigma=covm1,n=999) #n not used in computation
        relO <- log(bain_res$fit[1,3])
      }
    }
  }
  return(c(relE,relO))
}

# compute relative meausures (fit or complexity) under a multivariate Student t distribution
Student_measures <- function(mean1,Scale1,df1,RrE1,RrO1,names1=NULL,constraints1=NULL,...){ # Volgens mij moet je hier ook N meegeven
  K <- length(mean1)
  relE <- relO <- log(1)
  if(!is.null(RrE1) && is.null(RrO1)){ #only equality constraints
    RE1 <- RrE1[,-(K+1)]
    if(!is.matrix(RE1)){
      RE1 <- matrix(RE1,ncol=K)
    }
    rE1 <- RrE1[,(K+1)]
    qE1 <- nrow(RE1)
    meanE <- RE1%*%mean1
    scaleE <- RE1%*%Scale1%*%t(RE1)
    relE <- dmvt(rE1,delta=c(meanE),sigma=scaleE,df=df1,log=TRUE)
  }
  if(is.null(RrE1) && !is.null(RrO1)){ #only order constraints
    RO1 <- RrO1[,-(K+1)]
    if(!is.matrix(RO1)){
      RO1 <- matrix(RO1,ncol=K)
    }
    qO1 <- nrow(RO1)
    rO1 <- RrO1[,(K+1)]

    if(Rank(RO1)==nrow(RO1)){ #RO1 is of full row rank. So use transformation.
      meanO <- c(RO1%*%mean1)
      scaleO <- RO1%*%Scale1%*%t(RO1)
      if(nrow(scaleO)==1){
        relO <- pt((rO1-meanO)/sqrt(scaleO[1,1]),df=df1,lower.tail=FALSE,log.p = TRUE)
      }else{
        relO <- pmvt(lower=rO1,upper=Inf,delta=meanO,sigma=scaleO,df=round(df1),
                     type="shifted",...)[1]
        if(relO<0){relO <- 0}
        if(relO>1){relO <- 1}
        relO <- log(relO)
      }
    }else{ # no linear transformation is used; Use proportion of draws satisfying the constraints
           # (bain was used before but is problematic with range constraints)

      relO <- log(mean(apply((rmvt(n=1e6,sigma=Scale1,df=df1) + rep(1,1e6) %*% t(mean1)) %*%
                               t(RO1) - rep(1,1e6) %*% t(rO1) > 0,1,prod)))
      # names(mean1) <- names1
      # if(df1>2){ # we need posterior measures
      #   covm1 <- Scale1*df1/(df1-2)
      # }else if(df1==2){
      #   covm1 <- Scale1*4
      # }else{
      #   covm1 <- Scale1*5
      # }
      # mean1vec <- c(mean1)
      # names(mean1vec) <- row.names(covm1) <-colnames(covm1) <- names(mean1)
      # bain_res <- bain(x=mean1vec,hypothesis=constraints1,Sigma=covm1,n=999) #n not used in computation
      # relO <- log(bain_res$fit[1,3])
    }
  }
  if(!is.null(RrE1) && !is.null(RrO1)){ #hypothesis with equality and order constraints

    RE1 <- RrE1[,-(K+1)]
    if(!is.matrix(RE1)){
      RE1 <- matrix(RE1,ncol=K)
    }
    rE1 <- RrE1[,(K+1)]
    qE1 <- nrow(RE1)
    RO1 <- RrO1[,-(K+1)]
    if(!is.matrix(RO1)){
      RO1 <- matrix(RO1,ncol=K)
    }
    qO1 <- nrow(RO1)
    rO1 <- RrO1[,(K+1)]

    #a)Transformation matrix
    D <- diag(K) - t(RE1) %*% solve(RE1 %*% t(RE1)) %*% RE1
    D2 <- unique(round(D, 5))
    if(length(as.logical(rowSums(D2 != 0)))==1){
      D2 <- matrix(D2[as.logical(rowSums(D2 != 0)),],nrow=1)
    }else{
      D2 <- D2[as.logical(rowSums(D2 != 0)),]
    }
    if(!is.matrix(D2)){
      D2 <- t(D2)
    }
    Tm <- rbind(RE1, D2)

    #b)
    Tmean1 <- Tm %*% mean1
    Tscale1 <- Tm %*% Scale1 %*% t(Tm)

    # relative meausure for equalities
    relE <- dmvt(x = t(rE1),
                 delta = Tmean1[1:qE1],
                 sigma = matrix(Tscale1[1:qE1, 1:qE1],ncol = qE1), df = df1, log = TRUE)

    # transform order constraints
    RO1tilde <- RO1 %*% ginv(D2)
    rO1tilde <- rO1 - RO1 %*% ginv(RE1) %*% rE1

    # Partitioning equality part and order part
    Tmean1E <- Tmean1[1:qE1]
    Tmean1O <- Tmean1[(qE1 + 1):K]

    Tscale1EE <- as.matrix(Tscale1[1:qE1, 1:qE1],nrow=qE1)
    Tscale1OE <- matrix(Tscale1[(qE1 + 1):K, 1:qE1],ncol=qE1)
    Tscale1OO <- as.matrix(Tscale1[(qE1 + 1):K, (qE1 + 1):K],ncol=K-qE1)

    #conditional location and scale matrix
    Tmean1OgE <- Tmean1O + Tscale1OE %*% solve(Tscale1EE) %*% matrix(rE1 - Tmean1E)
    Tscale1OgE <- as.vector((df1 + (t(matrix(rE1 - Tmean1E)) %*% solve(Tscale1EE) %*% matrix(rE1 - Tmean1E))) /
                              (df1 + qE1)) * (Tscale1OO - Tscale1OE %*% solve(Tscale1EE) %*% t(Tscale1OE))

    if(Rank(RO1tilde) == nrow(RO1tilde)){
      rO1tilde <- as.vector(rO1tilde)

      delta_trans <- as.vector(RO1tilde %*% Tmean1OgE)
      scale1_trans <- RO1tilde %*% Tscale1OgE %*% t(RO1tilde)

      if(nrow(scale1_trans)==1){ # univariate
        relO <- pt((rO1tilde - delta_trans) / sqrt(scale1_trans), df = df1+qE1, lower.tail = FALSE,
                   log.p = TRUE)[1]
      } else { # multivariate
        relO <- pmvt(lower = rO1tilde, upper = Inf, delta = delta_trans, sigma = scale1_trans,
                     df = round(df1+qE1), type = "shifted",...)[1]
        if(relO<0){relO <- 0}
        if(relO>1){relO <- 1}
        relO <- log(relO)
      }

    }else{ #use bain for the computation of the probability
      #compute covariance matrix for multivariate normal distribution
      names(mean1) <- names1
      if(df1>2){ # we need posterior measures
        covm1 <- Scale1*df1/(df1-2) # variance estimate used for Gaussian approximation
      }else if(df1==2){ # we need posterior measures (there is very little information)
        covm1 <- Scale1*4
      }else{ #then df=1, so we need prior measures
        covm1 <- Scale1*5 #for prior with df1==1, probability independent of common factor of scale1
      }
      mean1vec <- c(mean1)
      names(mean1vec) <- names(mean1)
      bain_res <- bain(x=mean1vec,hypothesis=constraints1,Sigma=covm1,n=999) #n not used
      relO <- log(bain_res$fit[1,3])
    }
  }

  return(c(relE,relO))
}

# The function computes the probability of an unconstrained draw falling in the complement subspace.
MatrixStudent_prob_Hc <- function(Mean1,Scale1,tXXi1,df1,relmeas,RrO1){

  P <- ncol(Mean1)
  K <- nrow(Mean1)
  numpara <- P*K
  numhyp <- nrow(relmeas)
  #  relmeas <- relmeas[1:numhyp,]
  which_eq <- relmeas[,1] != log(1)
  if(sum(which_eq)==numhyp){ # Then the complement is equivalent to the unconstrained hypothesis.
    relmeas <- rbind(relmeas,log(rep(1,2)))
    rownames(relmeas)[numhyp+1] <- "complement"
  }else{ # So there is at least one hypothesis with only order constraints
    welk <- which(!which_eq)
    if(length(welk)==1){ # There is one hypothesis with only order constraints. Hc is complement of this hypothesis.
      relmeas <- rbind(relmeas,rep(1,2))
      relmeas[numhyp+1,2] <- 1 - exp(relmeas[welk,2])
      relmeas[numhyp+1,] <- log(relmeas[numhyp+1,])
      rownames(relmeas)[numhyp+1] <- "complement"
    }else{ # So more than one hypothesis with only order constraints
      # First we check whether there is an overlap between the order constrained spaces.

      draws2 <- 1e4
      randomDraws <- rmvnorm(draws2,mean=rep(0,numpara),sigma=diag(numpara))
      #get draws that satisfy the constraints of the separate order constrained hypotheses
      checksOC <- lapply(welk,function(h){
        #Rorder <- as.matrix(RrO1[[h]][,-(1+numpara)])
        Rorder <- matrix(RrO1[[h]][,-(1+numpara)],nrow=nrow(RrO1[[h]]))
        #if(ncol(Rorder)==1){
        #  Rorder <- t(Rorder)
        #}
        rorder <- as.matrix(RrO1[[h]][,1+numpara])
        apply(randomDraws%*%t(Rorder) > rep(1,draws2)%*%t(rorder),1,prod)
      })
      checkOCplus <- Reduce("+",checksOC)

      if(sum(checkOCplus > 0) < draws2){ #then the joint order constrained hypotheses do not completely cover the parameter space.
        if(sum(checkOCplus>1)==0){ # then order constrained spaces are nonoverlapping
          relmeas <- rbind(relmeas,rep(1,2))
          relmeas[numhyp+1,2] <- 1 - sum(exp(relmeas[welk,2]))
          relmeas[numhyp+1,] <- log(relmeas[numhyp+1,])
          rownames(relmeas)[numhyp+1] <- "complement"
        }else{ #the order constrained subspaces at least partly overlap

          # function below gives a rough estimate of the posterior probability under Hc
          # a bain type of algorithm would be better of course. but for now this is ok.

          temp1 <- rWishart(draws2,df1+P-1,solve(Scale1))
          temp2 <- lapply(seq(dim(temp1)[3]), function(x) temp1[,,x])
          SigmaList <- lapply(temp2,solve)
          randomDraws <- matrix(unlist(lapply(SigmaList,function(temp){
            rmvnorm(1,mean=c(Mean1),sigma=kronecker(temp,tXXi1))
          })),nrow=numpara)

          checksOC <- lapply(welk,function(h){
            #Rorder <- as.matrix(RrO1[[h]][,-(1+numpara)])
            Rorder <- matrix(RrO1[[h]][,-(1+numpara)],nrow=nrow(RrO1[[h]]))
            #if(ncol(Rorder)==1){
            #  Rorder <- t(Rorder)
            #}
            rorder <- as.matrix(RrO1[[h]][,1+numpara])
            apply(Rorder%*%randomDraws > rorder%*%t(rep(1,draws2)),2,prod)
          })
          relmeas <- rbind(relmeas,rep(1,2))
          relmeas[numhyp+1,] <- log(c(1,sum(Reduce("+",checksOC)==0)/draws2))
          rownames(relmeas)[numhyp+1] <- "complement"
        }
      }
    }
  }

  return(relmeas)
}

# The function computes the probability of an unconstrained draw falling in the complement subspace.
Student_prob_Hc <- function(mean1,scale1,df1,relmeas1,constraints,RrO1=NULL){

  numpara <- length(mean1)
  numhyp <- nrow(relmeas1)
  if(numhyp==1){
    relmeas <- t(relmeas1[1:numhyp,])
  }else{ relmeas <- relmeas1[1:numhyp,]}
  which_eq <- relmeas[,1] != log(1)
  if(sum(which_eq)==numhyp){ # Then the complement is equivalent to the unconstrained hypothesis.
    relmeas <- rbind(relmeas,log(rep(1,2)))
    rownames(relmeas)[numhyp+1] <- "complement"
  }else{ # So there is at least one hypothesis with only order constraints
    welk <- which(!which_eq)
    if(length(welk)==1){ # There is one hypothesis with only order constraints. Hc is complement of this hypothesis.
      relmeas <- rbind(relmeas,rep(1,2))
      relmeas[numhyp+1,2] <- 1 - exp(relmeas[welk,2])
      relmeas[numhyp+1,] <- log(relmeas[numhyp+1,])
      rownames(relmeas)[numhyp+1] <- "complement"
    }else{ # So more than one hypothesis with only order constraints
      # First we check whether ther is an overlap between the order constrained spaces.

      draws2 <- 1e5
      randomDraws <- rmvnorm(draws2,mean=rep(0,numpara),sigma=diag(numpara))
      #get draws that satisfy the constraints of the separate order constrained hypotheses
      checksOC <- lapply(welk,function(h){
        #Rorder <- as.matrix(RrO1[[h]][,-(1+numpara)])
        Rorder <- matrix(RrO1[[h]][,-(1+numpara)],nrow=nrow(RrO1[[h]]))
        #if(ncol(Rorder)==1){
        #  Rorder <- t(Rorder)
        #}
        rorder <- as.matrix(RrO1[[h]][,1+numpara])
        apply(randomDraws%*%t(Rorder) > rep(1,draws2)%*%t(rorder),1,prod)
      })
      checkOCplus <- Reduce("+",checksOC)

      if(sum(checkOCplus > 0) < draws2){ #then the joint order constrained hypotheses do not completely cover the parameter space.
        if(sum(checkOCplus>1)==0){ # then order constrained spaces are nonoverlapping
          relmeas <- rbind(relmeas,rep(log(1),2))
          relmeas[numhyp+1,2] <- log(1 - sum(exp(relmeas[welk,2])))
          rownames(relmeas)[numhyp+1] <- "complement"
        }else{ #the order constrained subspaces at least partly overlap

          # the function below gives a rough estimate of the posterior probability under Hc
          # a bain type of algorithm but then using the Student distribution would be better.

          randomDraws <- rmvt(draws2,delta=mean1,sigma=scale1,df=df1)
          checksOC <- lapply(welk,function(h){
            #Rorder <- as.matrix(RrO1[[h]][,-(1+numpara)])
            Rorder <- matrix(RrO1[[h]][,-(1+numpara)],nrow=nrow(RrO1[[h]]))
            #if(ncol(Rorder)==1){
            #  Rorder <- t(Rorder)
            #}
            rorder <- as.matrix(RrO1[[h]][,1+numpara])
            apply(randomDraws%*%t(Rorder) > rep(1,draws2)%*%t(rorder),1,prod)
          })
          relmeas <- rbind(relmeas,rep(1,2))
          relmeas[numhyp+1,2] <- sum(Reduce("+",checksOC) == 0) / draws2
          relmeas[numhyp+1,] <- log(relmeas[numhyp+1,])
          rownames(relmeas)[numhyp+1] <- "complement"
        }
      }
    }
  }

  return(relmeas)
}

#' @method BF hetcor
#' @export
BF.hetcor <- function(x,
                       hypothesis = NULL,
                       prior.hyp.explo = NULL,
                       prior.hyp.conf = NULL,
                       prior.hyp = NULL,
                       complement = TRUE,
                       log = FALSE,
                       cov.prob = .95,
                       ...){

  if(!(cov.prob>0 & cov.prob<1)){
    stop("The argument 'cov.prob' is a coverage probability for the interval estimates that should lie between 0 and 1. The default is 0.95.")
  }
  CrI_LB <- (1 - cov.prob)/2
  CrI_UB <- 1 - (1 - cov.prob)/2

  logIN <- log

  # check proper usage of argument 'prior.hyp.conf' and 'prior.hyp.explo'
  if(!is.null(prior.hyp.conf)){
    prior.hyp <- prior.hyp.conf
  }
  prior.hyp.explo <- process.prior.hyp.explo(prior_hyp_explo = prior.hyp.explo, model=x)

  get_est <- get_estimates(x)
  P <- nrow(x$std.errors)
  numcorr <- P*(P-1)/2
  estimates <- get_est$estimate
  errcov <- get_est$Sigma[[1]]
  # use Fisher transformed for both exploratory and confirmatory to get consistent results.
  # skewness in the likelihood is ignored.
  est.var.F <- do.call(cbind,lapply(1:numcorr,function(c){
    draws.norm <- rnorm(1e5,mean=estimates[c],sd=sqrt(errcov[c,c]))
    draws.norm.F <- FisherZ(draws.norm[draws.norm < 1 & draws.norm > -1])
    return(c(median(draws.norm.F),var(draws.norm.F)))
  }))
  estimates.F <- est.var.F[1,]
  if(numcorr > 1){
    errcov.F <- diag(est.var.F[2,])
  }else{
    errcov.F <- as.matrix(est.var.F[2,])
  }
  corr_names <- names(get_est$estimate)
  #matrix_names <- matrix(corr_names,nrow=P)
  names(estimates.F) <- colnames(errcov.F) <- row.names(errcov.F) <- corr_names

  #exploratory BF testing
  relfit <- matrix(c(dnorm(0,mean=estimates.F,sd=sqrt(diag(errcov.F)),log=TRUE),
                     pnorm(0,mean=estimates.F,sd=sqrt(diag(errcov.F)),log.p=TRUE),
                     pnorm(0,mean=estimates.F,sd=sqrt(diag(errcov.F)),log.p=TRUE,
                           lower.tail=FALSE)),ncol=3)
  # get draws from joint uniform prior to compute relative measures
  if(sum(P==Fcor$P)==0){
    #number of draws to get 1e7 draws for the marginal of 1 Fisher transformation correlation
    numdraws <- round(1e7/(P*(P-1)/2))
    drawsJU <- draw_ju_r(P,samsize=numdraws,Fisher=1)
    approx_studt <- QRM::fit.st(c(drawsJU))$par.ests[c(1,3)]
  }else{
    # use the estimate of the scale from the Fcor object
    # for the df the estimates show some numerical error for P>20, so use fitted line
    approx_studt <- unlist(c(Fcor[which(P==Fcor$P),1:2]))
    if(P > 20){
      # use fitted linear line rather than rough estimate of df
      slpe1 <- 2.944494
      intcept1 <- 1.864901
      approx_studt[1] <- P * slpe1 + intcept1
    }
  }
  relcomp0 <- dt(0,df=approx_studt[1],log=TRUE)-log(approx_studt[2]) # all marginal priors are the same
  relcomp <- matrix(c(rep(relcomp0,numcorr),rep(log(.5),numcorr*2)),ncol=3)
  row.names(relfit) <- row.names(relcomp) <- names(estimates.F)

  BFtu_exploratory <- relfit - relcomp
  colnames(BFtu_exploratory) <- colnames(BFtu_exploratory) <-  c("Pr(=0)","Pr(<0)","Pr(>0)")
  maxrow <- apply(BFtu_exploratory,1,max)
  BFtu_explo_norm <- exp(BFtu_exploratory - maxrow %*% t(rep(1,ncol(BFtu_exploratory)))) *
    rep(1,nrow(BFtu_exploratory)) %*% t(prior.hyp.explo[[1]])
  PHP_exploratory <- BFtu_explo_norm / apply(BFtu_explo_norm,1,sum)

  if(logIN == FALSE){
    BFtu_exploratory <- exp(BFtu_exploratory)
  }

  #confirmatory BF testing
  if(!is.null(hypothesis)){
    numG <- 1
    numcorrgroup <- numcorr
    varnames <- list(row.names(x$correlations))
    # get all names combinations for correlations (similar as BF.cor_test)
    corrnames <- lapply(1:numG,function(g){
      matrix(unlist(lapply(1:P,function(p2){
        unlist(lapply(1:P,function(p1){
          if(numG==1){
            paste0(varnames[[g]][p1],"_with_",varnames[[g]][p2])
          }else{
            paste0(varnames[[g]][p1],"_with_",varnames[[g]][p2],"_in_g",as.character(g))
          }
        }))
      })),nrow=P)
    })
    x$corrnames <- corrnames

    params_in_hyp1 <- params_in_hyp(hypothesis)

    corr_names <- unlist(lapply(1:length(x$corrnames),function(g){
      c(x$corrnames[[g]][lower.tri(x$corrnames[[g]])],
        t(x$corrnames[[g]])[lower.tri(x$corrnames[[g]])])
    })) #which includes Y1_with_Y2 and Y2_with_Y1

    parse_hyp <- parse_hypothesis(corr_names,hypothesis)
    parse_hyp$hyp_mat <- do.call(rbind, parse_hyp$hyp_mat)
    if(nrow(parse_hyp$hyp_mat)==1){
      select1 <- rep(1:numcorrgroup,numG) + rep((0:(numG-1))*2*numcorrgroup,each=numcorrgroup)
      select2 <- rep(numcorrgroup+1:numcorrgroup,numG) + rep((0:(numG-1))*2*numcorrgroup,each=numcorrgroup)
      parse_hyp$hyp_mat <-
        t(as.matrix(c(parse_hyp$hyp_mat[,select1] + parse_hyp$hyp_mat[,select2],parse_hyp$hyp_mat[,numcorrgroup*2*numG+1])))
    }else{
      #combine equivalent correlations, e.g., cor(Y1,Y2)=corr(Y2,Y1).
      select1 <- rep(1:numcorrgroup,numG) + rep((0:(numG-1))*2*numcorrgroup,each=numcorrgroup)
      select2 <- rep(numcorrgroup+1:numcorrgroup,numG) + rep((0:(numG-1))*2*numcorrgroup,each=numcorrgroup)
      parse_hyp$hyp_mat <-
        cbind(parse_hyp$hyp_mat[,select1] + parse_hyp$hyp_mat[,select2],parse_hyp$hyp_mat[,numcorrgroup*2*numG+1])
    }
    #create coefficient with equality and order constraints
    RrList <- make_RrList2(parse_hyp)
    RrE <- RrList[[1]]
    RrO <- RrList[[2]]

    numhyp <- length(RrE)
    relfit <- t(matrix(unlist(lapply(1:numhyp,function(h){
      Gaussian_measures(estimates,errcov,RrE1=RrE[[h]],RrO1=RrO[[h]],names1=names(estimates),
                        constraints1=parse_hyp$original_hypothesis[h])
    })),nrow=2))
    # approximate unconstrained Fisher transformed correlations with a multivariate Student t
    mean0 <- rep(0,numcorr)
    if(numcorr==1){
      Scale0 <- as.matrix(approx_studt[2]**2)
      df0 <- round(approx_studt[1])
    }else{
      Scale0 <- diag(rep(approx_studt[2]**2,numcorr))
      df0 <- round(approx_studt[1])
    }
    mean0 <- rep(0,numcorr)
    relcomp <- t(matrix(unlist(lapply(1:numhyp,function(h){
      relcomp_h <- Student_measures(mean1=mean0,
                                    Scale1=Scale0,
                                    df1=df0,
                                    RrE1=RrE[[h]],
                                    RrO1=RrO[[h]])
      return(relcomp_h)

    })),nrow=2))

    row.names(relcomp) <- parse_hyp$original_hypothesis
    row.names(relfit) <- parse_hyp$original_hypothesis
    # evaluation of complement hypothesis
    active.complement <- TRUE
    if(complement == TRUE){
      relfit <- Gaussian_prob_Hc(estimates.F,errcov.F,relfit,RrO)
      relcomp <- Student_prob_Hc(mean1=mean0,scale1=Scale0,df1=df0,relmeas1=relcomp,
                                 constraints=NULL,RrO1=RrO)
      if(nrow(relfit) == numhyp){active.complement <- FALSE}
    }
    hypothesisshort <- unlist(lapply(1:nrow(relfit),function(h) paste0("H",as.character(h))))
    row.names(relfit) <- row.names(relfit) <- hypothesisshort

    colnames(relcomp) <- c("c_E","c_O")
    colnames(relfit) <- c("f_E","f_O")
    # computation of exploratory BFs and PHPs
    # the BF for the complement hypothesis vs Hu needs to be computed.
    BFtu_confirmatory <- c(apply(relfit - relcomp, 1, sum))
    # Check input of prior probabilies
    if(is.null(prior.hyp)){
      priorprobs <- rep(1/length(BFtu_confirmatory),length(BFtu_confirmatory))
    }else{
      if(!is.numeric(prior.hyp) || length(prior.hyp)!=length(BFtu_confirmatory)){
        if(complement == TRUE && active.complement == FALSE && length(prior.hyp)==length(BFtu_confirmatory)+1){
          # the prior prob for the complement hypothesis is given but it is not used as the constrained hypotheses
          # cover the entire parameter space
          priorprobs <- prior.hyp[1:length(BFtu_confirmatory)]
          priorprobs <- priorprobs / sum(priorprobs)
        }else{
          warning(paste0("Argument 'prior.hyp' should be numeric and of length ",as.character(length(BFtu_confirmatory)),
                         ". Equal prior probabilities are used."))
          priorprobs <- rep(1/length(BFtu_confirmatory),length(BFtu_confirmatory))
        }
      }else{
        priorprobs <- prior.hyp
        priorprobs <- priorprobs / sum(priorprobs)
      }
    }
    names(priorprobs) <- names(BFtu_confirmatory)
    maxBFtu <- max(BFtu_confirmatory)
    PHP_confirmatory <- exp(BFtu_confirmatory-maxBFtu)*priorprobs /
      sum(exp(BFtu_confirmatory-maxBFtu)*priorprobs)
    BFtable <- cbind(relcomp,relfit,relfit[,1]-relcomp[,1],relfit[,2]-relcomp[,2],
                     apply(relfit,1,sum)-apply(relcomp,1,sum),PHP_confirmatory)
    BFtable[,1:7] <- exp(BFtable[,1:7])
    row.names(BFtable) <- names(BFtu_confirmatory)
    colnames(BFtable) <- c("complex=","complex>","fit=","fit>","BF=","BF>","BF","PHP")
    BFmatrix_confirmatory <- matrix(rep(BFtu_confirmatory,length(BFtu_confirmatory)),ncol=length(BFtu_confirmatory)) -
      t(matrix(rep(BFtu_confirmatory,length(BFtu_confirmatory)),ncol=length(BFtu_confirmatory)))
    diag(BFmatrix_confirmatory) <- 0
    row.names(BFmatrix_confirmatory) <- colnames(BFmatrix_confirmatory) <- names(BFtu_confirmatory)
    hypotheses <- row.names(relcomp)

    if(logIN == FALSE){
      BFtu_confirmatory <- exp(BFtu_confirmatory)
      BFmatrix_confirmatory <- exp(BFmatrix_confirmatory)
    }

  }else{
    BFtu_confirmatory <- PHP_confirmatory <- BFmatrix_confirmatory <- relfit <-
      relcomp <- hypotheses <- BFtable <- priorprobs <- NULL
  }

  # Store in output
  BF_out <- list(
    BFtu_exploratory=BFtu_exploratory,
    PHP_exploratory=PHP_exploratory,
    BFtu_confirmatory=BFtu_confirmatory,
    PHP_confirmatory=PHP_confirmatory,
    BFmatrix_confirmatory=BFmatrix_confirmatory,
    BFtable_confirmatory=BFtable,
    prior.hyp=priorprobs,
    hypotheses=hypotheses,
    estimates=estimates,
    model=x,
    bayesfactor="Bayes factors based on joint uniform priors",
    parameter="measures of association",
    log=logIN,
    call=match.call())

  class(BF_out) <- "BF"

  return(BF_out)

}

#' fMRI data
#'
#' fMRI data assessing relation between individual differences in the ability to recognize
#' faces and cars and thickness of the superficial, middle, and deep layers of the
#' fusiform face area, as assessed by high-resolution fMRI recognition (Williams et al, 2019, under review)
#'
#'   \tabular{lll}{
#'   \strong{Subject}\tab\code{numeric}\tab Particicpant ID number\cr
#'    \strong{Face} \tab \code{numeric} \tab Standardized score on face recognition battery\cr
#'    \strong{Vehicle} \tab \code{numeric} \tab Standardized score on vehicle recognition battery\cr
#'    \strong{Superficial} \tab \code{numeric} \tab Depth in mm of superficial layer of FFA\cr
#'    \strong{Middle} \tab \code{numeric} \tab Depth in mm of middle layer of FFA\cr
#'    \strong{Bform} \tab \code{numeric} \tab Depth in mm of deep layer of FFA\cr
#' }
#' @docType data
#' @keywords datasets
#' @name fmri
#' @usage data(fmri)
#' @references McGuigin, R.W., Newton, A.T., Tamber-Rosenau, B., Tomarken, A.J, & Gauthier, I. (under review). Thickness of deep layers in the fusiform face area predicts face recognition.
#' @format A data.frame with 13 rows and 6 columns.
NULL

# old BF test
BF_rma.uni <- function(x,
                    hypothesis = NULL,
                    prior.hyp.explo = NULL,
                    prior.hyp.conf = NULL,
                    prior.hyp = NULL,
                    complement = TRUE,
                    log = FALSE,
                    cov.prob = .95,
                    BF.type,
                    iter = 2e4,
                    ...){

  # x should be of class rma.uni
  if(class(x)[1]!="rma.uni"){
    stop("Only objects of class 'rma.uni' are currently supported.")
  }
  if(!is.null(x$formula.mods)){
    stop("Only an object without moderators is currently supported.")
  }
  if(!is.null(hypothesis)){
    message("Note that confirmatory testing via the 'hypothesis' argument is currently not supported for object of class 'rma.uni'.")
  }
  if(!(cov.prob>0 & cov.prob<1)){
    stop("The argument 'cov.prob' is a coverage probability for the interval estimates that should lie between 0 and 1. The default is 0.95.")
  }
  CrI_LB <- (1 - cov.prob)/2
  CrI_UB <- 1 - (1 - cov.prob)/2

  if(!exists("BF.type")){
    stop("The argument 'BF.type' is missing. See documentation. See ?BF")
  }

  logIN <- log

  # check proper usage of argument 'prior.hyp.conf' and 'prior.hyp.explo'
  if(!is.null(prior.hyp.conf)){
    prior.hyp <- prior.hyp.conf
  }
  prior.hyp.explo <- process.prior.hyp.explo(prior_hyp_explo = prior.hyp.explo, model = x)

  ### Extract effect sizes and sampling variances from the metafor object
  yi <- c(x$yi)
  vi <- x$vi
  ni <- x$ni #only used in case of unit-information prior: BF.type = "unit.info"
  N <- sum(ni)
  K <- length(vi)

  wi <- 1/vi # Weights equal-effects model
  typ_vi <- sum(wi*(length(wi)-1))/(sum(wi)^2 - sum(wi^2)) # Typical within-study sampling variance
  est.mu <- sum(yi*wi) / sum(wi)
  se.mu <- x$se

  est.tau2 <- x$tau2
  se.tau2 <- x$se.tau2 # check if we can use it as step sd for posterior sampling of tau
  if(is.na(se.tau2)){se.tau2 <- .5}
  tau2.min <- -min(vi)

  # prior tau2
  prior.tau2.Jeffreys1 <- function(tau2.arg,log=FALSE){
    ptau2 <- -1/K*sum(log(vi+tau2.arg))
    #ptau2 <- .5*log(sum(1/(vi+tau2.arg)^2)) # Tibshirani (1989)
    #ptau2 <- 0 #uniform prior
    #ptau2 <- -.5*log(tau2.arg) #right-Haar prior (e.g., Daniels, 1999)
    ifelse(log,ptau2,exp(ptau2))
    } # Berger & Deely (1988)


  if(is(BF.type,"character")){
    if(BF.type == "stand.effect"){
      # conjugate normal prior assuming average effects of about .5
      prior.mu <- function(x,tau2,log=FALSE){dnorm(x, mean = 0, sd = 1, log = log)}
      prior.muGR0 <- .5
      bayesfactor.name <- "Bayes factor based on a normal prior (mu ~ norm(mean=0, sd=1))"
      parameter.name <- "standardized effect"
    }else if(BF.type == "log.odds"){
      # Student t prior which approximates the implied distribution of the log odds ratio based on uniform success probabilities
      prior.mu <- function(x,tau2,log=FALSE){
        dt1 <- dt(x/2.36,df=13.1,log=TRUE)-log(2.36)
        ifelse(log,dt1,exp(dt1))
        }
      prior.muGR0 <- .5
      bayesfactor.name <- "Bayes factor based on uniform priors for success probabilities (log.odds ~ t(0,2.36,13.1))."
      parameter.name <- "log odds"
    }else if(BF.type == "correlation"){
      # logistic prior for the Fisher transformed correlation corresponding to a uniform prior for the correlation in (-1,1)
      prior.mu <- function(x,tau2,log=FALSE){dlogis(x, scale = .5, log = log)}
      prior.muGR0 <- .5
      bayesfactor.name <- "Bayes factor based on a uniform prior for the correlation in (-1,1) (Fisher(cor)~logis(0.5))."
      parameter.name <- "correlation"
    }else if(BF.type == "unit.info"){
      if(N==0){
        stop("In order to use a unit-information prior, the sample sizes need to be part of the element 'ni'
              of the 'rma.uni' object (e.g., using the 'ni' argument of the 'rma' function).")
      }
      prior.mu <- function(x,tau2,log=FALSE){dnorm(x, mean = 0, sd = sqrt(N/sum(1/(vi+tau2))), log = log)}
      prior.muGR0 <- .5
      bayesfactor.name <- "Bayes factor based on unit information prior."
      parameter.name <- "general parameter"
    }else{
      stop("The argument 'BF.type' is not correctly specified for an object of type 'rma.uni'. See documentation. ?BF")
    }
  }else if(is(BF.type,"prior")){
    prior.mu <- function(x,tau2,log=FALSE){BF.type(x,log)}
    seq1 <- seq(0,1e3,length=1e5)
    seq1 <- seq1+(seq1[2]-seq1[1])/2
    prior.muGR0 <- sum(prior.mu(seq1,0))*(seq1[2]-seq1[1]) #riemann estimate of prior prob that mu > 0.
    bayesfactor.name <- paste0("Bayes factor based on manually chosen prior")
    parameter.name <- "general parameter"
    message("Be sure that the specified prior is not truncated in a specific interval.")
  }else{
    stop("The argument 'BF.type' is not correctly specified for an object of type 'rma.uni'. See documentation. ?BF")
  }
  prior.mu.estimation <- function(x,tau2,log=FALSE){
    logdens <- 0
    ifelse(log,logdens,exp(logdens))
  }

  #exploratory testing
  if (x$method != "EE" & x$method != "FE") ### random effects / marema model
  {

    #### MAREMA MODEL

    # initial values for MCMC run
    start_mu <- est.mu
    sdstep.mu <- x$se
    start_tau2 <- ifelse(x$tau2>0, x$tau2,.05)
    sdstep.tau2 <- x$se.tau2

    # marginal likelihood full unconstrained model. marema.unc.post.draws
    marg.like.marema.full <- log_marg_like_full(yi, vi, tau2_min=tau2.min,
                                                prior.mu=prior.mu,
                                                prior.tau2=prior.tau2.Jeffreys1,
                                                start_mu=start_mu, start_tau2=start_tau2,
                                                sdstep.mu=sdstep.mu, sdstep.tau2=sdstep.tau2,
                                                burnin = round(iter/2), iters1=iter, iters2=iter)

    # marginal likelihood under mu = 0
    marg.like.marema.muEQ0 <- log_marg_like_cond.mu(yi, vi, tau2_min=tau2.min,
                                                    muIN=0, prior.tau2=prior.tau2.Jeffreys1,
                                                    start_tau2=start_tau2+start_mu^2, sdstep.tau2=sdstep.tau2,
                                                    burnin=iter, iters1=iter, iters2=iter)

    # marginal likelihoods when constraining mu
    # mu = 0
    marg.like.marema.muEQ0 <- marg.like.marema.muEQ0[[1]]
    # mu > 0
    marg.like.marema.muGR0 <- marg.like.marema.full[[1]] + log(mean(marg.like.marema.full$post.draws[,1]>0)/prior.muGR0)
    # mu < 0
    marg.like.marema.muSM0 <- marg.like.marema.full[[1]] + log(mean(marg.like.marema.full$post.draws[,1]<0)/(1-prior.muGR0))

    #### random effects MODEL

    # marginal likelihood full unconstrained model
    marg.like.ranef.full <- log_marg_like_full(yi, vi, tau2_min=0,
                                               prior.mu = prior.mu,
                                               prior.tau2 = prior.tau2.Jeffreys1,
                                               start_mu = start_mu, start_tau2 = ifelse(start_tau2<0,sdstep.tau2,start_tau2),
                                               sdstep.mu = sdstep.mu, sdstep.tau2 = sdstep.tau2,
                                               burnin=round(iter/2), iters1=iter, iters2=iter)

    # marginal likelihood under mu = 0
    marg.like.ranef.muEQ0 <- log_marg_like_cond.mu(yi, vi, tau2_min=0,
                                                   muIN=0, prior.tau2=prior.tau2.Jeffreys1,
                                                   start_tau2=ifelse(start_tau2<0,sdstep.tau2,start_tau2)+start_mu^2, sdstep.tau2=sdstep.tau2,
                                                   burnin = round(iter/2), iters1=iter, iters2=iter)

    marg.like.ranef.muEQ0 <- marg.like.ranef.muEQ0[[1]]
    # mu > 0
    marg.like.ranef.muGR0 <- marg.like.ranef.full[[1]] + log(mean(marg.like.ranef.full$post.draws[,1]>0)/prior.muGR0)
    # mu < 0
    marg.like.ranef.muSM0 <- marg.like.ranef.full[[1]] + log(mean(marg.like.ranef.full$post.draws[,1]<0)/(1-prior.muGR0))

    ### Compute Bayes factors model vs. unconstrained mu and I^2 (marema)
    BF0u_ranef <- marg.like.ranef.muEQ0 - marg.like.ranef.full[[1]]
    BF1u_ranef <- marg.like.ranef.muSM0 - marg.like.ranef.full[[1]]
    BF2u_ranef <- marg.like.ranef.muGR0 - marg.like.ranef.full[[1]]
    BF0u_marema <- marg.like.marema.muEQ0 - marg.like.marema.full[[1]]
    BF1u_marema <- marg.like.marema.muSM0 - marg.like.marema.full[[1]]
    BF2u_marema <- marg.like.marema.muGR0 - marg.like.marema.full[[1]]

    BFtu_exploratory_marema <- matrix(c(BF0u_marema,BF1u_marema,BF2u_marema),nrow=1)
    rowmax <- apply(BFtu_exploratory_marema,1,max)
    norm_BF_explo <- exp(BFtu_exploratory_marema - rowmax %*% t(rep(1,ncol(BFtu_exploratory_marema)))) *
      (rep(1,nrow(BFtu_exploratory_marema)) %*% t(prior.hyp.explo[[1]]))
    PHP_exploratory_marema <- norm_BF_explo / apply(norm_BF_explo,1,sum)

    BFtu_exploratory_ranef <- matrix(c(BF0u_ranef,BF1u_ranef,BF2u_ranef),nrow=1)
    rowmax <- apply(BFtu_exploratory_ranef,1,max)
    norm_BF_explo <- exp(BFtu_exploratory_ranef - rowmax %*% t(rep(1,ncol(BFtu_exploratory_ranef)))) *
      (rep(1,nrow(BFtu_exploratory_ranef)) %*% t(prior.hyp.explo[[1]]))
    PHP_exploratory_ranef <- norm_BF_explo / apply(norm_BF_explo,1,sum)

    BFtu_exploratory <- rbind(BFtu_exploratory_marema,BFtu_exploratory_ranef)
    PHP_exploratory <- rbind(PHP_exploratory_marema,PHP_exploratory_ranef)

    colnames(BFtu_exploratory) <- c("=0","<0",">0")
    colnames(PHP_exploratory) <- c("Pr(=0)","Pr(<0)","Pr(>0)")
    row.names(BFtu_exploratory) <- row.names(PHP_exploratory) <- c("mu (marema)", "mu (ranef)")

    if(logIN == FALSE){
      BFtu_exploratory <- exp(BFtu_exploratory)
    }

    #store descriptives of estimates
    marema_estimation <- log_marg_like_full(yi, vi, tau2_min=tau2.min,
                                                prior.mu=prior.mu.estimation,
                                                prior.tau2=prior.tau2.Jeffreys1,
                                                start_mu=start_mu, start_tau2=start_tau2,
                                                sdstep.mu=sdstep.mu, sdstep.tau2=sdstep.tau2,
                                                burnin = round(iter/2), iters1=iter, iters2=1e3)
    ranef_estimation <- log_marg_like_full(yi, vi, tau2_min=0,
                                               prior.mu = prior.mu,
                                               prior.tau2 = prior.tau2.Jeffreys1,
                                               start_mu = start_mu, start_tau2 = ifelse(start_tau2<0,sdstep.tau2,start_tau2),
                                               sdstep.mu = sdstep.mu, sdstep.tau2 = sdstep.tau2,
                                               burnin=round(iter/2), iters1=iter, iters2=iter)
    estimates_marema <- cbind(apply(marema_estimation$post.draws,2,mean),
                              apply(marema_estimation$post.draws,2,median),
                              apply(marema_estimation$post.draws,2,quantile,CrI_LB),
                              apply(marema_estimation$post.draws,2,quantile,CrI_UB),
                              apply(marema_estimation$post.draws,2,function(x){mean(x>0)}))
    row.names(estimates_marema) <- c("mu      (marema)", "tau2    (marema)")
    estimates_ranef <- cbind(apply(ranef_estimation$post.draws,2,mean),
                             apply(ranef_estimation$post.draws,2,median),
                             apply(ranef_estimation$post.draws,2,quantile,CrI_LB),
                             apply(ranef_estimation$post.draws,2,quantile,CrI_UB),
                             apply(ranef_estimation$post.draws,2,function(x){mean(x>0)}))
    row.names(estimates_ranef) <- c("mu      (ranef)", "tau2    (ranef)")
    colnames(estimates_marema) <- colnames(estimates_ranef) <-
      c("mean","median",paste0(as.character(round(CrI_LB*100,7)),"%"),
        paste0(as.character(round(CrI_UB*100,7)),"%"),"Pr(>0)")

    uncestimates <- estimates_marema
    # estimates of separate group means
    #under marema model
    marema_post.draws_mu <- marema_estimation$post.draws[,1]
    marema_post.draws_tau2 <- marema_estimation$post.draws[,2]
    tau2draws_trunc <- marema_estimation$post.draws[marema_post.draws_tau2>0,2]
    mudraws_trunc <- marema_estimation$post.draws[marema_post.draws_tau2>0,1]
    numzerotau2 <- sum(marema_post.draws_tau2 <= 0)
    mudraws_studies <- do.call(cbind,lapply(1:length(vi),function(s){
      var_s <- 1/(1/vi[s] + 1/tau2draws_trunc)
      mean_s <- (yi[s]/vi[s] + mudraws_trunc/tau2draws_trunc) * var_s
      draws_s <- c(rnorm(length(mudraws_trunc),mean=mean_s,sd=sqrt(var_s)),
                   marema_post.draws_mu[marema_post.draws_tau2<0])
      c(draws_s,mean(draws_s),median(draws_s),quantile(draws_s,.025),quantile(draws_s,.975),mean(draws_s>0))
    }))
    uncestimates <- rbind(uncestimates,t(mudraws_studies[length(marema_post.draws_mu)+1:5,]))
    row.names(uncestimates)[-(1:2)] <- paste0("theta_",1:ncol(mudraws_studies)," (marema)")
    #under random effects model
    uncestimates <- rbind(uncestimates,estimates_ranef)
    ranef_post.draws_mu <- ranef_estimation$post.draws[,1]
    ranef_post.draws_tau2 <- ranef_estimation$post.draws[,2]
    mudraws_studies <- do.call(cbind,lapply(1:length(vi),function(s){
      var_s <- 1/(1/vi[s] + 1/ranef_post.draws_tau2)
      mean_s <- (yi[s]/vi[s] + ranef_post.draws_mu/ranef_post.draws_tau2) * var_s
      draws_s <- rnorm(length(ranef_post.draws_mu),mean=mean_s,sd=sqrt(var_s))
      c(draws_s,mean(draws_s),median(draws_s),quantile(draws_s,.025),quantile(draws_s,.975),mean(draws_s>0))
    }))
    uncestimates <- rbind(uncestimates,t(mudraws_studies[length(ranef_post.draws_mu)+1:5,]))
    row.names(uncestimates)[-(1:(4+length(vi)))] <- paste0("theta_",1:ncol(mudraws_studies)," (ranef)")

    #store posterior draws based on noninformative improper prior
    post.draws <- list()
    post.draws$marema <- marema_estimation$post.draws
    post.draws$ranef <- ranef_estimation$post.draws

  } else if (x$method == "EE" | x$method == "FE") { ### common effect model

    # testing
    marg.like.comef.muEQ0 <- log_marg_like_cond.tau2(yi, vi, tau2IN=0,
                                                     prior.mu,
                                                     start_mu=est.mu, sdstep.mu=se.mu,
                                                     burnin = round(iter/2), iters1 = iter, iters2 = iter)

    ### log marginal likelihoods
    # mu = 0
    marg.like.muEQ0 <- sum(dnorm(yi,mean=0,sd=sqrt(vi+0),log=TRUE))
    # mu > 0
    marg.like.muGR0 <- marg.like.comef.muEQ0[[1]] + log(mean(marg.like.comef.muEQ0[[2]]>0)/prior.muGR0)
    # mu < 0
    marg.like.muSM0 <- marg.like.comef.muEQ0[[1]] + log(mean(marg.like.comef.muEQ0[[2]]<0)/(1-prior.muGR0))

    ### Compute Bayes factors model vs. unconstrained model
    BF0u_comef <- marg.like.muEQ0 - marg.like.comef.muEQ0[[1]]
    BF1u_comef <- marg.like.muSM0 - marg.like.comef.muEQ0[[1]]
    BF2u_comef <- marg.like.muGR0 - marg.like.comef.muEQ0[[1]]

    BFtu_exploratory <- matrix(c(BF0u_comef,BF1u_comef,BF2u_comef),nrow=1)
    rowmax <- apply(BFtu_exploratory,1,max)
    norm_BF_explo <- exp(BFtu_exploratory - rowmax %*% t(rep(1,ncol(BFtu_exploratory)))) *
      (rep(1,nrow(BFtu_exploratory)) %*% t(prior.hyp.explo[[1]]))
    PHP_exploratory <- norm_BF_explo / apply(norm_BF_explo,1,sum)
    colnames(BFtu_exploratory) <- c("=0","<0",">0")
    colnames(PHP_exploratory) <- c("Pr(=0)","Pr(<0)","Pr(>0)")
    row.names(BFtu_exploratory) <- row.names(PHP_exploratory) <- c("mu (comef)")

    #descriptives of estimates
    comef_estimation <- log_marg_like_cond.tau2(yi, vi, tau2IN=0,
                                                     prior.mu=prior.mu.estimation,
                                                     start_mu=est.mu, sdstep.mu=se.mu,
                                                     burnin = round(iter/2), iters1 = iter, iters2 = 1e3)
    uncestimates <- cbind(mean(comef_estimation$post.draws), median(comef_estimation$post.draws),
                          quantile(comef_estimation$post.draws,CrI_LB),quantile(comef_estimation$post.draws,CrI_UB),
                          mean(comef_estimation$post.draws>0))
    colnames(uncestimates) <- c("mean","median",paste0(as.character(round(CrI_LB*100,2)),"%"),
                                paste0(as.character(round(CrI_UB*100,2)),"%"),"Pr(>0)")
    row.names(uncestimates) <- "mu"

    #store posterior draws based on noninformative improper prior
    post.draws <- comef_estimation$post.draws
    colnames(post.draws) <- row.names(uncestimates)

    if(logIN == FALSE){
      BFtu_exploratory <- exp(BFtu_exploratory)
    }

  }

  if(!is.null(hypothesis)){
    message("The 'hypothesis' argument for this model class is not yet supported. Please inform us about the types of hypotheses you wish to test :)")
  }

  BFtu_confirmatory <- PHP_confirmatory <- BFmatrix_confirmatory <- BFtable <-
    priorprobs <- hypotheses <- NULL

  BF_out <- list(
    BFtu_exploratory=BFtu_exploratory,
    PHP_exploratory=PHP_exploratory,
    BFtu_confirmatory=BFtu_confirmatory,
    PHP_confirmatory=PHP_confirmatory,
    BFmatrix_confirmatory=BFmatrix_confirmatory,
    BFtable_confirmatory=BFtable,
    prior.hyp.explo=prior.hyp.explo,
    prior.hyp.conf=priorprobs,
    hypotheses=hypotheses,
    estimates=uncestimates,
    model=x,
    bayesfactor=bayesfactor.name,
    parameter=parameter.name,
    log=logIN,
    call=match.call(),
    post.draws=post.draws
  )

  class(BF_out) <- "BF"

  BF_out

}

#1. unconstrained sampling under marema using 'prior.mu' as prior for mu and a noninformative prior for tau2 in (prior.tau2,Inf)
gibbs_unc_prior <- function(yi, vi, tau2_min,
                            prior.mu, prior.tau2,
                            start_mu, start_tau2,
                            sdstep.mu, sdstep.tau2,
                            burnin, iters){

  #check sd for random walk every 100 iterations
  check1 <- 100
  upper1 <- .5 # define region where the sdstep does not have to be changed
  lower1 <- .15
  sdsteptel <- 1
  optimal1 <- .35
  gamma1 <- function(prem) 1 / (prem + check1)^0.6
  #stretched beta prior for rho is a uniform: alpha1 = 1 & beta1 = 1

  #initialization
  mu <- start_mu
  if(start_tau2>tau2_min){
    tau2 <- start_tau2
  }else{
    tau2 <- sdstep.tau2
  }
  cur_vars <- vi+tau2
  acceptMat <- matrix(0, nrow = iters + burnin, ncol = 2)
  sdstepseq <- matrix(0, nrow = (iters + burnin) / check1, 2)
  sdstep <- c(sdstep.mu, sdstep.tau2)
  store <- matrix(NA, ncol=2, nrow = iters + burnin)
  colnames(store) <- colnames(acceptMat) <- colnames(sdstepseq) <- c("mu","tau2")

  for(i in 1:(burnin+iters)){
    #draw candidate tau2 from truncated normal
    tau2_star <- rtnorm_local(1, mean = tau2, sd = sdstep[2], a = tau2_min)

    #evaluate Metropolis-Hastings acceptance probability
    can_vars <- vi+tau2_star
    R_MH <- exp( sum(dnorm(yi,mean=mu,sd=sqrt(can_vars),log=TRUE)) -
                   sum(dnorm(yi,mean=mu,sd=sqrt(cur_vars),log=TRUE)) +
                   prior.mu(mu,tau2_star,log=TRUE) - prior.mu(mu,tau2,log=TRUE) +
                   prior.tau2(tau2_star,log=TRUE) - prior.tau2(tau2,log=TRUE)) *
      (1 - pnorm(tau2_min,mean=tau2,sd=sdstep[2])) / (1 - pnorm(tau2_min,mean=tau2_star,sd=sdstep[2]))
    tau2 <- ifelse(runif(1) < R_MH, tau2_star, tau2)
    acceptMat[i,2] <- tau2_star == tau2
    cur_vars <- vi+tau2

    #draw mu
    mu_star <- rnorm(1, mean = mu, sd = sdstep[1])
    R_MH <- exp( sum(dnorm(yi,mean=mu_star,sd=sqrt(cur_vars),log=TRUE)) -
                   sum(dnorm(yi,mean=mu,sd=sqrt(cur_vars),log=TRUE)) +
      prior.mu(mu_star,tau2,log=TRUE) - prior.mu(mu,tau2,log=TRUE)
      )
    mu <- ifelse(runif(1) < R_MH, mu_star, mu)
    acceptMat[i,1] <- mu_star == mu

    #if needed update random walk sd depending on acceptance rate
    if(i %% check1 == 0){
      probs <- apply(as.matrix(acceptMat[(i-check1+1):i,]),2,mean)
      # sdstep <- exp(log(sdstep) + gamma1(i) * (probs - optimal1))
      sdstep[probs>upper1] <- sdstep[probs>upper1] * ( (probs[probs>upper1]-upper1)/(1-upper1) + 1)
      sdstep[probs<lower1] <- sdstep[probs<lower1] * 1/( 2 - (probs[probs<lower1])/lower1 )
      # store sdstep
      sdstepseq[sdsteptel,] <- sdstep
      sdsteptel <- sdsteptel + 1
    }

    store[i,] <- c(mu,tau2)
    if(i > burnin){
      check1 <- min(1e3,burnin)
    }
  }

  return(list(postdraws = as.matrix(store[-(1:burnin),]), sdstepseq=sdstepseq))

}
gibbs_unc_prior1 <- function(yi, vi, tau2_min,
                            prior.mu, prior.tau2,
                            start_mu, start_tau2,
                            sdstep.mu, sdstep.tau2,
                            burnin, iters){

  #check sd for random walk every 100 iterations
  check1 <- 100
  upper1 <- .5 # define region where the sdstep does not have to be changed
  lower1 <- .2
  sdsteptel <- 1
  #stretched beta prior for rho is a uniform: alpha1 = 1 & beta1 = 1

  # transform logtau2 to tau2
  fun_logtau2TOtau2 <- function(logtau2.arg){
    exp(logtau2.arg) + tau2_min
  }

  # implied prior for transformed parameter logtau2
  prior.logtau2.func <- function(logtau2.arg,log=FALSE){

    plogtau2 <- prior.tau2(fun_logtau2TOtau2(logtau2.arg),log=TRUE) + logtau2.arg

    ifelse(log,plogtau2,exp(plogtau2))
  }

  #initialization
  mu <- start_mu
  draws.for.start.logtau2 <- log(rtnorm_local(1e4,mean=start_tau2,sd=sdstep.tau2,a=tau2_min) - tau2_min)
  sdstep.logtau2 <- sd(draws.for.start.logtau2)
  logtau2 <- mean(draws.for.start.logtau2)
  # if(start_tau2 > tau2_min){
  #   logtau2 <- log(start_tau2 - tau2_min)
  # }else{
  #   logtau2 <- log(sdstep.tau2 - tau2_min)
  # }

  cur_vars <- vi+fun_logtau2TOtau2(logtau2)
  acceptMat <- matrix(0, nrow = iters + burnin, ncol = 2)
  sdstepseq <- matrix(0, nrow = (iters + burnin) / check1, 2)
  sdstep <- c(sdstep.mu, sdstep.logtau2)
  store <- matrix(NA, ncol=2, nrow = iters + burnin)
  colnames(store) <- colnames(acceptMat) <- colnames(sdstepseq) <- c("mu","tau2")

  for(i in 1:(burnin+iters)){
    #draw candidate logtau2 from a normal (tau2)
    logtau2_star <- rnorm(1, mean = logtau2, sd = sdstep[2])

    #evaluate Metropolis-Hastings acceptance probability
    can_vars <- vi+fun_logtau2TOtau2(logtau2_star)
    R_MH <- exp( sum(dnorm(yi,mean=mu,sd=sqrt(can_vars),log=TRUE)) -
                   sum(dnorm(yi,mean=mu,sd=sqrt(cur_vars),log=TRUE)) +
                   prior.mu(mu,fun_logtau2TOtau2(logtau2_star),log=TRUE) -
                   prior.mu(mu,fun_logtau2TOtau2(logtau2),log=TRUE) +
                   prior.logtau2.func(logtau2_star,log=TRUE) -
                   prior.logtau2.func(logtau2,log=TRUE) )
    logtau2 <- ifelse(runif(1) < R_MH, logtau2_star, logtau2)
    acceptMat[i,2] <- logtau2_star == logtau2
    cur_vars <- vi+fun_logtau2TOtau2(logtau2)

    #draw mu
    mu_star <- rnorm(1, mean = mu, sd = sdstep[1])
    R_MH <- exp( sum(dnorm(yi,mean=mu_star,sd=sqrt(cur_vars),log=TRUE)) -
                   sum(dnorm(yi,mean=mu,sd=sqrt(cur_vars),log=TRUE)) +
                   prior.mu(mu_star,fun_logtau2TOtau2(logtau2),log=TRUE) -
                   prior.mu(mu,fun_logtau2TOtau2(logtau2),log=TRUE)
    )
    mu <- ifelse(runif(1) < R_MH, mu_star, mu)
    acceptMat[i,1] <- mu_star == mu

    #if needed update random walk sd depending on acceptance rate
    if((i <= burnin) && (i %% check1 == 0)){
      probs <- apply(as.matrix(acceptMat[(i-check1+1):i,]),2,mean)
      # sdstep <- exp(log(sdstep) + gamma1(i) * (probs - optimal1))
      sdstep[probs>upper1] <- sdstep[probs>upper1] * ( (probs[probs>upper1]-upper1)/(1-upper1) + 1)
      sdstep[probs<lower1] <- sdstep[probs<lower1] * 1/( 2 - (probs[probs<lower1])/lower1 )
      # store sdstep
      sdstepseq[sdsteptel,] <- sdstep
      sdsteptel <- sdsteptel + 1
    }

    store[i,] <- c(mu,fun_logtau2TOtau2(logtau2))
    # if(i > burnin){
    #   check1 <- min(1e3,burnin)
    # }
  }

  return(list(postdraws = as.matrix(store[-(1:burnin),]), sdstepseq=sdstepseq))

}

#2. unconstrained sampling under marema while fixing 'mu' and a noninformative prior for tau in (prior.tau2,Inf)
gibbs_cond.mu_prior <- function(yi, vi, tau2_min,
                                muIN, prior.tau2,
                                start_tau2, sdstep.tau2, burnin, iters){

  #check sd for random walk every 100 iterations
  check1 <- 100
  upper1 <- .5 # define region where the sdstep does not have to be changed
  lower1 <- .15
  sdsteptel <- 1
  #stretched beta prior for rho is a uniform: alpha1 = 1 & beta1 = 1

  #initialization
  mu <- muIN
  if(start_tau2>tau2_min){
    tau2 <- start_tau2
  }else{
    tau2 <- sdstep.tau2
  }
  cur_vars <- vi+tau2
  acceptMat <- matrix(0, nrow = iters + burnin, ncol = 1)
  sdstepseq <- matrix(0, nrow = (iters + burnin) / check1, 1)
  sdstep <- sdstep.tau2
  store <- matrix(NA, ncol=1, nrow = iters + burnin)
  colnames(store) <- colnames(acceptMat) <- colnames(sdstepseq) <- c("tau2")

  for(i in 1:(burnin+iters)){
    #draw rho
    #draw candidate from truncated normal
    tau2_star <- rtnorm_local(1, mean = tau2, sd = sdstep, a = tau2_min)

    #evaluate Metropolis-Hastings acceptance probability
    can_vars <- vi+tau2_star
    R_MH <- exp( sum(dnorm(yi,mean=mu,sd=sqrt(can_vars),log=TRUE)) -
                   sum(dnorm(yi,mean=mu,sd=sqrt(cur_vars),log=TRUE)) ) *
      (1 - pnorm(tau2_min,mean=tau2,sd=sdstep)) / (1 - pnorm(tau2_min,mean=tau2_star,sd=sdstep)) *
      prior.tau2(tau2_star) / prior.tau2(tau2)
    tau2 <- ifelse(runif(1) < R_MH, tau2_star, tau2)
    acceptMat[i,1] <- tau2_star == tau2
    cur_vars <- vi+tau2

    #if needed update random walk sd depending on acceptance rate
    if(ceiling(i/check1)==i/check1){
      probs <- mean(as.matrix(acceptMat[(i-check1+1):i,]))
      sdstep[probs>upper1] <- sdstep[probs>upper1] * ( (probs[probs>upper1]-upper1)/(1-upper1) + 1)
      sdstep[probs<lower1] <- sdstep[probs<lower1] * 1/( 2 - (probs[probs<lower1])/lower1 )
      # store sdstep
      sdstepseq[sdsteptel,1] <- sdstep
      sdsteptel <- sdsteptel + 1
    }

    store[i,] <- c(tau2)
  }

  return(list(postdraws = as.matrix(store[-(1:burnin),]), sdstepseq=sdstepseq))

}
gibbs_cond.mu_prior1 <- function(yi, vi, tau2_min,
                                muIN, prior.tau2,
                                start_tau2, sdstep.tau2, burnin, iters){

  #check sd for random walk every 100 iterations
  check1 <- 100
  upper1 <- .5 # define region where the sdstep does not have to be changed
  lower1 <- .2
  sdsteptel <- 1
  # optimal1 <- .35
  # gamma1 <- function(prem) 1 / (prem + check1)^0.6
  #stretched beta prior for rho is a uniform: alpha1 = 1 & beta1 = 1

  # transform logtau2 to tau2
  fun_logtau2TOtau2 <- function(logtau2.arg){
    exp(logtau2.arg) + tau2_min
  }

  # implied prior for transformed parameter logtau2
  prior.logtau2.func <- function(logtau2.arg,log=FALSE){

    plogtau2 <- prior.tau2(fun_logtau2TOtau2(logtau2.arg),log=TRUE) + logtau2.arg

    ifelse(log,plogtau2,exp(plogtau2))
  }

  #initialization
  draws.for.start.logtau2 <- log(rtnorm_local(1e4,mean=start_tau2,sd=sdstep.tau2,a=tau2_min) - tau2_min)
  sdstep.logtau2 <- sd(draws.for.start.logtau2)
  logtau2 <- mean(draws.for.start.logtau2)
  # if(start_tau2 > tau2_min){
  #   logtau2 <- log(start_tau2 - tau2_min)
  # }else{
  #   logtau2 <- log(sdstep.tau2 - tau2_min)
  # }
  mu <- muIN
  # if(start_tau2>tau2_min){
  #   tau2 <- start_tau2
  # }else{
  #   tau2 <- sdstep.tau2
  # }
  cur_vars <- vi+fun_logtau2TOtau2(logtau2)
  acceptMat <- matrix(0, nrow = iters + burnin, ncol = 1)
  sdstepseq <- matrix(0, nrow = (iters + burnin) / check1, 1)
  sdstep <- sdstep.logtau2
  store <- matrix(NA, ncol=1, nrow = iters + burnin)
  store_check <- matrix(NA, ncol=2, nrow = iters + burnin)
  colnames(store) <- colnames(acceptMat) <- colnames(sdstepseq) <- c("tau2")

  for(i in 1:(burnin+iters)){
    #draw rho
    #draw candidate from truncated normal
    logtau2_star <- rnorm(1, mean = logtau2, sd = sdstep)
    store_check[i,] <- c(logtau2,logtau2_star)

    #evaluate Metropolis-Hastings acceptance probability
    can_vars <- vi+fun_logtau2TOtau2(logtau2_star)
    R_MH <- exp( sum(dnorm(yi,mean=mu,sd=sqrt(can_vars),log=TRUE)) -
                   sum(dnorm(yi,mean=mu,sd=sqrt(cur_vars),log=TRUE)) +
                   prior.logtau2.func(logtau2_star,log=TRUE) -
                   prior.logtau2.func(logtau2,log=TRUE)
                 )
    logtau2 <- ifelse(runif(1) < R_MH, logtau2_star, logtau2)
    acceptMat[i,1] <- logtau2_star == logtau2
    cur_vars <- vi+fun_logtau2TOtau2(logtau2)

    #if needed update random walk sd depending on acceptance rate
    if((i <= burnin) && (i %% check1 == 0)){
      probs <- mean(as.matrix(acceptMat[(i-check1+1):i,]))
#      sdstep <- exp(log(sdstep) + gamma1(i) * (probs - optimal1))
      sdstep[probs>upper1] <- sdstep[probs>upper1] * ( (probs[probs>upper1]-upper1)/(1-upper1) + 1)
      sdstep[probs<lower1] <- sdstep[probs<lower1] * 1/( 2 - (probs[probs<lower1])/lower1 )
      # store sdstep
      sdstepseq[sdsteptel,1] <- sdstep
      sdsteptel <- sdsteptel + 1
    }

    store[i,] <- c(fun_logtau2TOtau2(logtau2))
  }

  return(list(postdraws = as.matrix(store[-(1:burnin),]), sdstepseq=sdstepseq))

}

#3. unconstrained sampling under marema using 'prior.mu' as prior while fixing 'tau2'
gibbs_cond.tau_prior <- function(yi, vi, tau2IN,
                                 prior.mu,
                                 start_mu, sdstep.mu, burnin, iters){

  #check sd for random walk every 100 iterations
  check1 <- 100
  upper1 <- .5 # define region where the sdstep does not have to be changed
  lower1 <- .15
  sdsteptel <- 1
  #stretched beta prior for rho is a uniform: alpha1 = 1 & beta1 = 1

  #initialization
  mu <- start_mu
  tau2 <- tau2IN
  cur_vars <- vi+tau2
  acceptMat <- matrix(0, nrow = iters + burnin, ncol = 1)
  sdstepseq <- matrix(0, nrow = (iters + burnin) / check1, 1)
  sdstep <- c(sdstep.mu)
  store <- matrix(NA, ncol=1, nrow = iters + burnin)
  colnames(store) <- colnames(acceptMat) <- colnames(sdstepseq) <- c("mu")

  for(i in 1:(burnin+iters)){

    #draw mu
    mu_star <- rnorm(1, mean = mu, sd = sdstep[1])

    R_MH <- exp( sum(dnorm(yi,mean=mu_star,sd=sqrt(cur_vars),log=TRUE)) -
                   sum(dnorm(yi,mean=mu,sd=sqrt(cur_vars),log=TRUE)) ) *
      prior.mu(mu_star,tau2) / prior.mu(mu,tau2)
    mu <- ifelse(runif(1) < R_MH, mu_star, mu)
    acceptMat[i,1] <- mu_star == mu

    #if needed update random walk sd depending on acceptance rate
    if(ceiling(i/check1)==i/check1){
      probs <- mean(as.matrix(acceptMat[(i-check1+1):i,]))
      sdstep[probs>upper1] <- sdstep[probs>upper1] * ( (probs[probs>upper1]-upper1)/(1-upper1) + 1)
      sdstep[probs<lower1] <- sdstep[probs<lower1] * 1 / ( 2 - (probs[probs<lower1])/lower1 )
      # store sdstep
      sdstepseq[sdsteptel,1] <- sdstep
      sdsteptel <- sdsteptel + 1
    }

    store[i,] <- c(mu)
    if(i > burnin){
      check1 <- min(1e3,burnin)
    }
    #print(c(mu,tau2,sdstep))

  }

  return(list(postdraws = as.matrix(store[-(1:burnin),]), sdstepseq=sdstepseq))

}

# marginal likelihood computation for #1
log_marg_like_full <- function(yi, vi, tau2_min,
                               prior.mu,
                               prior.tau2,
                               start_mu, start_tau2, sdstep.mu, sdstep.tau2,
                               burnin, iters1, iters2){

  #estimate unconstrained marema model with proper prior for mu
  unc.draws.marglike <- gibbs_unc_prior1(yi, vi, tau2_min,
                                        prior.mu,
                                        prior.tau2,
                                        start_mu, start_tau2, sdstep.mu, sdstep.tau2,
                                        burnin, iters=iters1)

  #approximate posterior tau2 with a log normal for constructing proposal distribution
  para.tau2 <- fitdistr(unc.draws.marglike[[1]][,2]-tau2_min, "log-normal", lower = 0.0000001)$estimate
  para.tau2.IS <- c(para.tau2[1],para.tau2[2]*2)
  draws.IS.tau2 <- rlnorm(iters2, meanlog = para.tau2.IS[1], sdlog = para.tau2.IS[2]) + tau2_min

  #approximate posterior mu with a normal
  para.mu <- c(mean(unc.draws.marglike[[1]][,1]),sd(unc.draws.marglike[[1]][,1]))
  para.mu.IS <- para.mu
  para.mu.IS[2] <- para.mu.IS[2] * 2
  draws.IS.mu <- rnorm(iters2, mean = para.mu.IS[1], sd = para.mu.IS[2])

  #evaluate integrand of marginal likelihood for proposal draws
  logintegrands <- unlist(lapply(1:iters2,function(s){
    sum(dnorm(yi,mean=draws.IS.mu[s],sd=sqrt(vi+draws.IS.tau2[s]),log=TRUE)) +
      prior.mu(draws.IS.mu[s],draws.IS.tau2[s],log=TRUE) + prior.tau2(draws.IS.tau2[s],log=TRUE) -
      dnorm(draws.IS.mu[s], mean = para.mu.IS[1], sd = para.mu.IS[2],log=TRUE) -
      dlnorm(draws.IS.tau2[s]-tau2_min, meanlog = para.tau2.IS[1], sdlog = para.tau2.IS[2], log=TRUE)
  }))
  logintegrands <- logintegrands[which(!is.na(logintegrands) & (abs(logintegrands)!=Inf) )]

  logmuu <- log(mean(exp(logintegrands-max(logintegrands)))) + max(logintegrands)

  return(list(log.marg.like = logmuu, post.draws = unc.draws.marglike[[1]], sdstepseq = unc.draws.marglike$sdstepseq))

}

# marginal likelihood computation for #2
log_marg_like_cond.mu <- function(yi, vi, tau2_min,
                                  muIN, prior.tau2,
                                  start_tau2, sdstep.tau2, burnin, iters1, iters2){

  draws.marglike.cond.mu <- gibbs_cond.mu_prior1(yi, vi, tau2_min,
                                                muIN, prior.tau2=prior.tau2,
                                                start_tau2, sdstep.tau2, burnin, iters=iters1)

  #approximate posterior tau2 with a shifted log-normal as proposal distribution
  para.tau2 <- fitdistr(draws.marglike.cond.mu[[1]][,1]-tau2_min, "log-normal", lower = 0.000001)$estimate
  para.tau2.IS <- c(para.tau2[1],para.tau2[2]*2)
  draws.IS.tau2 <- rlnorm(iters2, meanlog = para.tau2.IS[1], sdlog = para.tau2.IS[2]) + tau2_min

  #evaluate integrand of marginal likelihood for proposal draws
  logintegrands <- unlist(lapply(1:iters2,function(s){
    sum(dnorm(yi,mean=muIN,sd=sqrt(vi+draws.IS.tau2[s]),log=TRUE)) +
      prior.tau2(draws.IS.tau2[s],log=TRUE) -
      dlnorm(draws.IS.tau2[s]-tau2_min, meanlog = para.tau2.IS[1], sdlog = para.tau2.IS[2], log=TRUE)
  }))
  logintegrands <- logintegrands[which(!is.na(logintegrands) & (abs(logintegrands)!=Inf) )]
  logmuu_cond.mu <- log(mean(exp(logintegrands-max(logintegrands)))) + max(logintegrands)

  return(list(log.marg.like = logmuu_cond.mu, post.draws = draws.marglike.cond.mu[[1]], sdstepseq = draws.marglike.cond.mu[[2]]))

}

# marginal likelihood computation for #3
log_marg_like_cond.tau2 <- function(yi, vi, tau2IN,
                                    prior.mu,
                                    start_mu, sdstep.mu, burnin, iters1, iters2){

  draws.marglike.cond.tau2 <- gibbs_cond.tau_prior(yi, vi, tau2IN,
                                                   prior.mu, start_mu, sdstep.mu,
                                                   burnin, iters1)

  #approximate posterior mu with a normal
  para.mu <- c(mean(draws.marglike.cond.tau2[[1]][,1]),sd(draws.marglike.cond.tau2[[1]][,1]))
  para.mu.IS <- para.mu
  para.mu.IS[2] <- para.mu.IS[2] * 3
  draws.IS.mu <- rnorm(iters2, mean = para.mu.IS[1], sd = para.mu.IS[2])

  #evaluate integrand of marginal likelihood for proposal draws
  logintegrands <- unlist(lapply(1:iters2,function(s){
    sum(dnorm(yi,mean=draws.IS.mu[s],sd=sqrt(vi+tau2IN),log=TRUE)) +
      prior.mu(draws.IS.mu[s],tau2IN,log=TRUE) -
      dnorm(draws.IS.mu[s], mean = para.mu.IS[1], sd = para.mu.IS[2],log=TRUE)
  }))
  logintegrands <- logintegrands[!is.na(logintegrands)]

  logmu0 <- log(mean(exp(logintegrands-max(logintegrands)))) + max(logintegrands)

  return(list(log.marg.like = logmu0, post.draws = draws.marglike.cond.tau2[[1]], sdstepseq = draws.marglike.cond.tau2[[2]]))

}

# function to draw from truncated normal using local function
rtnorm_local <- function(n, mean = 0, sd = 1, a = -Inf, b = Inf) {
  la <- (a - mean) / sd
  lb <- (b - mean) / sd

  Fa <- stats::pnorm(la)
  Fb <- stats::pnorm(lb)

  u <- stats::runif(n, min = Fa, max = Fb)
  mean + sd * stats::qnorm(u)
}


#' @importFrom metaBMA prior
#' @importFrom stats dlogis rlnorm dlnorm
#' @importFrom MASS fitdistr
#' @importFrom methods is
#' @describeIn BF BF S3 method for an object of class 'rma.uni'
#' @method BF rma.uni
#' @export
BF.rma.uni <- function(x,
                       hypothesis = NULL,
                       prior.hyp.explo = NULL,
                       prior.hyp.conf = NULL,
                       prior.hyp = NULL,
                       complement = TRUE,
                       log = FALSE,
                       cov.prob = .95,
                       BF.type,
                       #prior.tau2 = 1,
                       iter = 2e4,
                       ...){

  # x should be of class rma.uni
  if(class(x)[1]!="rma.uni"){
    stop("Only objects of class 'rma.uni' are currently supported.")
  }
  if(!is.null(x$formula.mods)){
    stop("Only an object without moderators is currently supported.")
  }
  if(!is.null(hypothesis)){
    message("Note that confirmatory testing via the 'hypothesis' argument is currently not supported for object of class 'rma.uni'.")
  }
  if(!(cov.prob>0 & cov.prob<1)){
    stop("The argument 'cov.prob' is a coverage probability for the interval estimates that
         should lie between 0 and 1. The default is 0.95.")
  }
  CrI_LB <- (1 - cov.prob)/2
  CrI_UB <- 1 - (1 - cov.prob)/2

  if(!exists("BF.type")){
    stop("The argument 'BF.type' is missing. See documentation. See ?BF")
  }

  logIN <- log

  # check proper usage of argument 'prior.hyp.conf' and 'prior.hyp.explo'
  if(!is.null(prior.hyp.conf)){
    prior.hyp <- prior.hyp.conf
  }
  prior.hyp.explo <- process.prior.hyp.explo(prior_hyp_explo = prior.hyp.explo, model = x)

  ### Extract effect sizes and sampling variances from the metafor object
  yi <- c(x$yi)
  vi <- x$vi
  ni <- x$ni #only used in case of unit-information prior: BF.type = "unit.info"
  N <- sum(ni)
  K <- length(vi)

  wi <- 1/vi # Weights equal-effects model
  typ_vi <- sum(wi*(length(wi)-1))/(sum(wi)^2 - sum(wi^2)) # Typical within-study sampling variance
  est.mu <- sum(yi*wi) / sum(wi)
  se.mu <- x$se

  est.tau2 <- x$tau2
  se.tau2 <- x$se.tau2 # check if we can use it as step sd for posterior sampling of tau
  if(is.na(se.tau2)){se.tau2 <- .5}
  nugget_boundary <- 1e-6
  tau2.min <- -min(vi) + nugget_boundary

  if(length(BF.type)==1){

    prior.tau2.func <- prior.tau2.func.RE <- function(tau2.arg,log=FALSE){
      ptau2 <- -1/K*sum(log(vi+tau2.arg)) #Berger-Deely prior
      ifelse(log,ptau2,exp(ptau2))
    }

    if(is(BF.type,"character")){
      if(BF.type == "stand.effect"){
        # conjugate normal prior with sd of 1
        prior.mu <- function(x,tau2,log=FALSE){dnorm(x, mean = 0, sd = 1, log = log)}
        prior.muGR0 <- .5
        bayesfactor.name <- "Bayes factor based on a normal prior (mu ~ norm(mean=0, sd=1)) and noninformative Berger-Deely prior for tau^2"
        parameter.name <- "standardized effect"
      }else if(BF.type == "log.odds"){
        # Student t prior which approximates the implied distribution of the log odds ratio based on uniform success probabilities
        prior.mu <- function(x,tau2,log=FALSE){
          dt1 <- dt(x/2.36,df=13.1,log=TRUE)-log(2.36)
          ifelse(log,dt1,exp(dt1))
        }
        prior.muGR0 <- .5
        bayesfactor.name <- "Bayes factor based on uniform priors for success probabilities (log.odds ~ t(0,2.36,13.1)) and noninformative Berger-Deely prior for tau^2."
        parameter.name <- "log odds"
      }else if(BF.type == "correlation"){
        # logistic prior for the Fisher transformed correlation corresponding to a uniform prior for the correlation in (-1,1)
        prior.mu <- function(x,tau2,log=FALSE){dlogis(x, scale = .5, log = log)}
        prior.muGR0 <- .5
        bayesfactor.name <- "Bayes factor based on a uniform prior for the correlation in (-1,1) (Fisher(cor)~logis(0.5)) and noninformative Berger-Deely prior for tau^2."
        parameter.name <- "Fisher transformed correlation"
      }else if(BF.type == "unit.info"){
        if(N==0){
          stop("In order to use a unit-information prior, the sample sizes need to be part of the element 'ni'
              of the 'rma.uni' object (e.g., using the 'ni' argument of the 'rma' function).")
        }
        prior.mu <- function(x,tau2,log=FALSE){dnorm(x, mean = 0, sd = sqrt(N/sum(1/(vi+tau2))), log = log)}
        prior.muGR0 <- .5
        bayesfactor.name <- "Bayes factor based on unit information prior and noninformative Berger-Deely prior for tau^2."
        parameter.name <- "mean parameter"
      }else{
        stop("The argument 'BF.type' is not correctly specified for an object of type 'rma.uni'. See documentation. ?BF")
      }
    }else if(is(BF.type,"prior")){
      prior.mu <- function(x,tau2,log=FALSE){BF.type(x,log)}
      seq1 <- seq(0,1e3,length=1e5)
      seq1 <- seq1+(seq1[2]-seq1[1])/2
      prior.muGR0 <- sum(prior.mu(seq1,0))*(seq1[2]-seq1[1]) #riemann estimate of prior prob that mu > 0.
      bayesfactor.name <- paste0("Bayes factor based on manually chosen prior")
      parameter.name <- "general parameter"
      message("Be sure that the specified prior is not truncated in a specific interval.")
    }else{
      stop("The argument 'BF.type' is not correctly specified for an object of type 'rma.uni'. See documentation. ?BF")
    }
  }else if(length(BF.type)==2){
    # both elements of BF.type should be of class 'prior' (from metaBMA package)
    # first element of BF.type is the prior for the mean
    # section element of BF.type is the prior for tau^2
    if( is(BF.type[[1]],"prior") & is(BF.type[[2]],"prior") ){

      prior.mu <- function(x,tau2,log=FALSE){BF.type[[1]](x,log)}
      seq1 <- seq(0,1e3,length=1e5)
      seq1 <- seq1+(seq1[2]-seq1[1])/2
      prior.muGR0 <- sum(prior.mu(seq1,0))*(seq1[2]-seq1[1]) #riemann estimate of prior prob that mu > 0.
      bayesfactor.name <- paste0("Bayes factor based on manually chosen prior")
      parameter.name <- "general parameter"
      message("The prior for the mean should not be truncated in a specific interval.")
      prior.tau2.func <- function(tau2.arg,log=FALSE){
        #ptau2 <- -1/K*sum(log(vi+tau2.arg))
        #ptau2 <- .5*log(sum(1/(vi+tau2.arg)^2)) # Tibshirani (1989)
        #ptau2 <- 0 #uniform prior
        ptau2 <- BF.type[[2]](tau2.arg,log=TRUE) # the right Haar prior for tau^2 is used by default
        ifelse(log,ptau2,exp(ptau2))
      }
    }else{
      stop("The argument 'BF.type' is not correctly specified for an object of type 'rma.uni'. See documentation. ?BF")
    }
  }else{
    stop("The argument 'BF.type' is not correctly specified for an object of type 'rma.uni'. See documentation. ?BF")
  }

  # # try different improper priors for tau2
  # if(prior.tau2 == 1){
  #   prior.tau2.func <- prior.tau2.func.RE <- function(tau2.arg,log=FALSE){
  #     ptau2 <- -1/K*sum(log(vi+tau2.arg))
  #     #ptau2 <- .5*log(sum(1/(vi+tau2.arg)^2)) # Tibshirani (1989)
  #     #ptau2 <- 0 #uniform prior
  #     ifelse(log,ptau2,exp(ptau2))
  #   }
  # }else if(prior.tau2 == 2){
  #   prior.tau2.func <- prior.tau2.func.RE <- function(tau2.arg,log=FALSE){
  #     #ptau2 <- -1/K*sum(log(vi+tau2.arg))
  #     #ptau2 <- .5*log(sum(1/(vi+tau2.arg)^2)) # Tibshirani (1989)
  #     ptau2 <- 0 #uniform prior
  #     ifelse(log,ptau2,exp(ptau2))
  #   }
  # }else if(prior.tau2 == 3){
  #   prior.tau2.func <- prior.tau2.func.RE <- function(tau2.arg,log=FALSE){
  #     #ptau2 <- -1/K*sum(log(vi+tau2.arg))
  #     #ptau2 <- .5*log(sum(1/(vi+tau2.arg)^2)) # Tibshirani (1989)
  #     #ptau2 <- 0 #uniform prior
  #     ptau2 <- -.5 * log(abs(tau2.arg))
  #     ifelse(log,ptau2,exp(ptau2))
  #   }
  # }else if(prior.tau2 == 4){
  #   prior.tau2.func <- function(tau2.arg,log=FALSE){
  #     #ptau2 <- -1/K*sum(log(vi+tau2.arg))
  #     #ptau2 <- .5*log(sum(1/(vi+tau2.arg)^2)) # Tibshirani (1989)
  #     #ptau2 <- 0 #uniform prior
  #     #ptau2 <- -.5 * log(abs(tau2.arg))
  #     ptau2 <- dinvgamma(sqrt(tau2.arg-tau2.min),alpha=1,beta=.15,log=TRUE) + log(.5) -.5 * log(tau2.arg-tau2.min)
  #     ifelse(log,ptau2,exp(ptau2))
  #   }
  #   prior.tau2.func.RE <- function(tau2.arg,log=FALSE){
  #     #ptau2 <- -1/K*sum(log(vi+tau2.arg))
  #     #ptau2 <- .5*log(sum(1/(vi+tau2.arg)^2)) # Tibshirani (1989)
  #     #ptau2 <- 0 #uniform prior
  #     #ptau2 <- -.5 * log(abs(tau2.arg))
  #     ptau2 <- dinvgamma(sqrt(tau2.arg),alpha=1,beta=.15,log=TRUE) + log(.5) -.5 * log(tau2.arg)
  #     ifelse(log,ptau2,exp(ptau2))
  #   }
  # }

  #exploratory testing
  if (x$method != "EE" & x$method != "FE") ### random effects / marema model
  {

    #### MAREMA MODEL

    # initial values for MCMC run
    start_mu <- mean(yi)
    sdstep.mu <- x$se
    start_tau2 <- ifelse(x$tau2>0, x$tau2,.05)
    sdstep.tau2 <- x$se.tau2
    if(is.na(sdstep.tau2)){
      sdstep.tau2 <- 1
    }

    # marginal likelihood full unconstrained model
    marg.like.marema.full <- log_marg_like_full(yi, vi, tau2_min=tau2.min,
                                                prior.mu=prior.mu,
                                                prior.tau2=prior.tau2.func,
                                                start_mu=start_mu, start_tau2=start_tau2,
                                                sdstep.mu=sdstep.mu, sdstep.tau2=sdstep.tau2,
                                                burnin = round(iter/2), iters1=iter, iters2=iter)

    # marginal likelihood under mu = 0
    marg.like.marema.muEQ0 <- log_marg_like_cond.mu(yi, vi, tau2_min=tau2.min,
                                                    muIN=0, prior.tau2=prior.tau2.func,
                                                    start_tau2=start_tau2+start_mu^2, sdstep.tau2=sdstep.tau2,
                                                    burnin=round(iter/2), iters1=iter, iters2=iter)

    # marginal likelihoods when constraining mu
    # mu = 0
    marg.like.marema.muEQ0 <- marg.like.marema.muEQ0[[1]]
    # mu > 0
    marg.like.marema.muGR0 <- marg.like.marema.full[[1]] + log(mean(marg.like.marema.full$post.draws[,1]>0)/prior.muGR0)
    # mu < 0
    marg.like.marema.muSM0 <- marg.like.marema.full[[1]] + log(mean(marg.like.marema.full$post.draws[,1]<0)/(1-prior.muGR0))

    #### random effects MODEL

    # marginal likelihood full unconstrained model
    marg.like.ranef.full <- log_marg_like_full(yi, vi, tau2_min=0+nugget_boundary,
                                               prior.mu = prior.mu,
                                               prior.tau2 = prior.tau2.func.RE,
                                               start_mu = start_mu, start_tau2 = start_tau2,
                                               sdstep.mu = sdstep.mu, sdstep.tau2 = sdstep.tau2,
                                               burnin=round(iter/2), iters1=iter, iters2=iter)

    # marginal likelihood under mu = 0
    marg.like.ranef.muEQ0 <- log_marg_like_cond.mu(yi, vi, tau2_min=0+nugget_boundary,
                                                   muIN=0, prior.tau2=prior.tau2.func.RE,
                                                   start_tau2=start_tau2+start_mu^2, sdstep.tau2=sdstep.tau2,
                                                   burnin = round(iter/2), iters1=iter, iters2=iter)

    marg.like.ranef.muEQ0 <- marg.like.ranef.muEQ0[[1]]
    # mu > 0
    marg.like.ranef.muGR0 <- marg.like.ranef.full[[1]] + log(mean(marg.like.ranef.full$post.draws[,1]>0)/prior.muGR0)
    # mu < 0
    marg.like.ranef.muSM0 <- marg.like.ranef.full[[1]] + log(mean(marg.like.ranef.full$post.draws[,1]<0)/(1-prior.muGR0))

    ### Compute Bayes factors model vs. unconstrained mu and I^2 (marema)
    BF0u_ranef <- marg.like.ranef.muEQ0 - marg.like.ranef.full[[1]]
    BF1u_ranef <- marg.like.ranef.muSM0 - marg.like.ranef.full[[1]]
    BF2u_ranef <- marg.like.ranef.muGR0 - marg.like.ranef.full[[1]]
    BF0u_marema <- marg.like.marema.muEQ0 - marg.like.marema.full[[1]]
    BF1u_marema <- marg.like.marema.muSM0 - marg.like.marema.full[[1]]
    BF2u_marema <- marg.like.marema.muGR0 - marg.like.marema.full[[1]]

    BFtu_exploratory_marema <- matrix(c(BF0u_marema,BF1u_marema,BF2u_marema),nrow=1)
    rowmax <- apply(BFtu_exploratory_marema,1,max)
    norm_BF_explo <- exp(BFtu_exploratory_marema - rowmax %*% t(rep(1,ncol(BFtu_exploratory_marema)))) *
      (rep(1,nrow(BFtu_exploratory_marema)) %*% t(prior.hyp.explo[[1]]))
    PHP_exploratory_marema <- norm_BF_explo / apply(norm_BF_explo,1,sum)

    BFtu_exploratory_ranef <- matrix(c(BF0u_ranef,BF1u_ranef,BF2u_ranef),nrow=1)
    rowmax <- apply(BFtu_exploratory_ranef,1,max)
    norm_BF_explo <- exp(BFtu_exploratory_ranef - rowmax %*% t(rep(1,ncol(BFtu_exploratory_ranef)))) *
      (rep(1,nrow(BFtu_exploratory_ranef)) %*% t(prior.hyp.explo[[1]]))
    PHP_exploratory_ranef <- norm_BF_explo / apply(norm_BF_explo,1,sum)

    BFtu_exploratory <- rbind(BFtu_exploratory_marema,BFtu_exploratory_ranef)
    PHP_exploratory <- rbind(PHP_exploratory_marema,PHP_exploratory_ranef)

    colnames(BFtu_exploratory) <- c("=0","<0",">0")
    colnames(PHP_exploratory) <- c("Pr(=0)","Pr(<0)","Pr(>0)")
    row.names(BFtu_exploratory) <- row.names(PHP_exploratory) <- c("mu (marema)", "mu (ranef)")

    if(logIN == FALSE){
      BFtu_exploratory <- exp(BFtu_exploratory)
    }

    #store descriptives of estimates
    estimates_marema <- cbind(apply(marg.like.marema.full$post.draws,2,mean),
                              apply(marg.like.marema.full$post.draws,2,median),
                              apply(marg.like.marema.full$post.draws,2,quantile,.025),
                              apply(marg.like.marema.full$post.draws,2,quantile,.975),
                              apply(marg.like.marema.full$post.draws,2,function(x){mean(x>0)}))
    row.names(estimates_marema) <- c("mu      (marema)", "tau2    (marema)")
    estimates_ranef <- cbind(apply(marg.like.ranef.full$post.draws,2,mean),
                             apply(marg.like.ranef.full$post.draws,2,median),
                             apply(marg.like.ranef.full$post.draws,2,quantile,.025),
                             apply(marg.like.ranef.full$post.draws,2,quantile,.975),
                             apply(marg.like.ranef.full$post.draws,2,function(x){mean(x>0)}))
    row.names(estimates_ranef) <- c("mu      (ranef)", "tau2    (ranef)")
    colnames(estimates_marema) <- colnames(estimates_ranef) <- #c("mean","median","2.5%","97.5%","Pr(>0)")
      c("mean","median",paste0(as.character(round(CrI_LB*100,7)),"%"),
        paste0(as.character(round(CrI_UB*100,7)),"%"),"Pr(>0)")
    uncestimates <- estimates_marema
    # estimates of separate group means
    #under marema model
    marema_post.draws_mu <- marg.like.marema.full$post.draws[,1]
    marema_post.draws_tau2 <- marg.like.marema.full$post.draws[,2]
    tau2draws_trunc <- marg.like.marema.full$post.draws[marema_post.draws_tau2>0,2]
    mudraws_trunc <- marg.like.marema.full$post.draws[marema_post.draws_tau2>0,1]
    numzerotau2 <- sum(marema_post.draws_tau2 <= 0)
    mudraws_studies <- do.call(cbind,lapply(1:length(vi),function(s){
      var_s <- 1/(1/vi[s] + 1/tau2draws_trunc)
      mean_s <- (yi[s]/vi[s] + mudraws_trunc/tau2draws_trunc) * var_s
      draws_s <- c(rnorm(length(mudraws_trunc),mean=mean_s,sd=sqrt(var_s)),
                   marema_post.draws_mu[marema_post.draws_tau2<0])
      c(draws_s,mean(draws_s),median(draws_s),quantile(draws_s,CrI_LB),quantile(draws_s,CrI_UB),mean(draws_s>0))
    }))
    uncestimates <- rbind(uncestimates,t(mudraws_studies[length(marema_post.draws_mu)+1:5,]))
    row.names(uncestimates)[-(1:2)] <- paste0("theta_",1:ncol(mudraws_studies)," (marema)")
    #under random effects model
    uncestimates <- rbind(uncestimates,estimates_ranef)
    ranef_post.draws_mu <- marg.like.ranef.full$post.draws[,1]
    ranef_post.draws_tau2 <- marg.like.ranef.full$post.draws[,2]
    mudraws_studies <- do.call(cbind,lapply(1:length(vi),function(s){
      var_s <- 1/(1/vi[s] + 1/ranef_post.draws_tau2)
      mean_s <- (yi[s]/vi[s] + ranef_post.draws_mu/ranef_post.draws_tau2) * var_s
      draws_s <- rnorm(length(ranef_post.draws_mu),mean=mean_s,sd=sqrt(var_s))
      c(draws_s,mean(draws_s),median(draws_s),quantile(draws_s,CrI_LB),quantile(draws_s,CrI_UB),mean(draws_s>0))
    }))
    uncestimates <- rbind(uncestimates,t(mudraws_studies[length(ranef_post.draws_mu)+1:5,]))
    row.names(uncestimates)[-(1:(4+length(vi)))] <- paste0("theta_",1:ncol(mudraws_studies)," (ranef)")

    #store posterior draws based on noninformative improper prior
    post.draws <- list()
    post.draws$marema <- marg.like.marema.full$post.draws
    post.draws$ranef <- marg.like.ranef.full$post.draws

  } else if (x$method == "EE" | x$method == "FE") { ### common effect model

    # testing
    marg.like.comef.muEQ0 <- log_marg_like_cond.tau2(yi, vi, tau2IN=0,
                                                     prior.mu,
                                                     start_mu=est.mu, sdstep.mu=se.mu,
                                                     burnin = round(iter/2), iters1 = iter, iters2 = iter)

    ### log marginal likelihoods
    # mu = 0
    marg.like.muEQ0 <- sum(dnorm(yi,mean=0,sd=sqrt(vi+0),log=TRUE))
    # mu > 0
    marg.like.muGR0 <- marg.like.comef.muEQ0[[1]] + log(mean(marg.like.comef.muEQ0[[2]]>0)/prior.muGR0)
    # mu < 0
    marg.like.muSM0 <- marg.like.comef.muEQ0[[1]] + log(mean(marg.like.comef.muEQ0[[2]]<0)/(1-prior.muGR0))

    ### Compute Bayes factors model vs. unconstrained model
    BF0u_comef <- marg.like.muEQ0 - marg.like.comef.muEQ0[[1]]
    BF1u_comef <- marg.like.muSM0 - marg.like.comef.muEQ0[[1]]
    BF2u_comef <- marg.like.muGR0 - marg.like.comef.muEQ0[[1]]

    BFtu_exploratory <- matrix(c(BF0u_comef,BF1u_comef,BF2u_comef),nrow=1)
    rowmax <- apply(BFtu_exploratory,1,max)
    norm_BF_explo <- exp(BFtu_exploratory - rowmax %*% t(rep(1,ncol(BFtu_exploratory)))) *
      (rep(1,nrow(BFtu_exploratory)) %*% t(prior.hyp.explo[[1]]))
    PHP_exploratory <- norm_BF_explo / apply(norm_BF_explo,1,sum)
    colnames(BFtu_exploratory) <- c("=0","<0",">0")
    colnames(PHP_exploratory) <- c("Pr(=0)","Pr(<0)","Pr(>0)")
    row.names(BFtu_exploratory) <- row.names(PHP_exploratory) <- c("mu (comef)")

    #descriptives of estimates
    uncestimates <- cbind(mean(marg.like.comef.muEQ0$post.draws), median(marg.like.comef.muEQ0$post.draws),
                          quantile(marg.like.comef.muEQ0$post.draws,CrI_LB),quantile(marg.like.comef.muEQ0$post.draws,CrI_UB),
                          mean(marg.like.comef.muEQ0$post.draws>0))
    colnames(uncestimates) <- c("mean","median",paste0(as.character(round(CrI_LB*100,7)),"%"),
                                paste0(as.character(round(CrI_UB*100,7)),"%"),"Pr(>0)")
    row.names(uncestimates) <- "mu"

    #store posterior draws based on noninformative improper prior
    post.draws <- marg.like.comef.muEQ0$post.draws
    colnames(post.draws) <- row.names(uncestimates)

    if(logIN == FALSE){
      BFtu_exploratory <- exp(BFtu_exploratory)
    }

  }

  if(!is.null(hypothesis)){
    message("The 'hypothesis' argument for this model class is not yet supported. Please inform us about the types of hypotheses you wish to test :)")
  }

  BFtu_confirmatory <- PHP_confirmatory <- BFmatrix_confirmatory <- BFtable <-
    priorprobs <- hypotheses <- NULL

  BF_out <- list(
    BFtu_exploratory=BFtu_exploratory,
    PHP_exploratory=PHP_exploratory,
    BFtu_confirmatory=BFtu_confirmatory,
    PHP_confirmatory=PHP_confirmatory,
    BFmatrix_confirmatory=BFmatrix_confirmatory,
    BFtable_confirmatory=BFtable,
    prior.hyp.explo=prior.hyp.explo,
    prior.hyp.conf=priorprobs,
    hypotheses=hypotheses,
    estimates=uncestimates,
    model=x,
    bayesfactor=bayesfactor.name,
    parameter=parameter.name,
    log=logIN,
    call=match.call(),
    post.draws=post.draws
  )

  class(BF_out) <- "BF"

  BF_out

}

#' Precision of the Anchor Influences the Amount of Adjustment
#'
#' Data from an experimental study where participants have to guess the price
#' of a plasma tv. There were two experimental conditions.
#' These data were simulated using the sufficient statistics from Janiszewski &
#' Uy (2008).
#'
#' \tabular{lll}{
#'    \strong{price} \tab \code{numeric} \tab Participant z-scores of price\cr
#'    \strong{anchor} \tab \code{factor} \tab Participant anchor\cr
#'    \strong{motivation} \tab \code{factor} \tab motivation to change\cr
#' }
#' @docType data
#' @keywords datasets
#' @name tvprices
#' @usage data(tvprices)
#' @references Janiszewski, C., & Uy, D. (2008). Precision of the anchor influences the amount of adjustment. Psychological Science, 19(2), 121–127. doi:10.1111/j.1467-9280.2008.02057.x
#' @format A data.frame with 59 rows and 3 columns.
NULL

#BF method for glm classes


#' @method BF glm
#' @export
BF.glm <- function(x,
                   hypothesis = NULL,
                   prior.hyp.explo = NULL,
                   prior.hyp.conf = NULL,
                   prior.hyp = NULL,
                   complement = TRUE,
                   log = FALSE,
                   cov.prob = .95,
                   ...){

  if(as.character(x$family)[1]=="gaussian"){

    # then use BF.lm
    class(x) <- "lm"
    Args <- as.list(match.call()[-1])
    Args$x <- x
    Args$hypothesis <- hypothesis
    Args$prior.hyp <- prior.hyp
    Args$prior.hyp.conf <- prior.hyp.conf
    Args$prior.hyp.explo <- prior.hyp.explo
    Args$complement <- complement
    Args$log <- log
    Args$cov.prob <- cov.prob
    out <- do.call(BF, Args)
    out$model <- x
    out$call <- match.call()

  }else{

    Args <- as.list(match.call()[-1])
    get_est <- get_estimates(x)
    Args$x <- get_est$estimate
    Args$Sigma <- get_est$Sigma[[1]]
    Args$n <- nobs(x)
    Args$hypothesis <- hypothesis
    Args$prior.hyp <- prior.hyp
    Args$prior.hyp.conf <- prior.hyp.conf
    Args$prior.hyp.explo <- prior.hyp.explo
    Args$complement <- complement
    Args$log <- log
    Args$cov.prob <- cov.prob
    out <- do.call(BF, Args)
    out$model <- x
    out$call <- match.call()

  }

  out
}

.onAttach <- function(libname, pkgname) {
  packageStartupMessage(
    sprintf("BFpack (%s) loaded", utils::packageVersion("BFpack"))
  )
  packageStartupMessage(
    "Citation: https://doi.org/10.18637/jss.v100.i18"
  )
}

# .onAttach <- function(libname, pkgname) {
#   version <- read.dcf(
#     file = system.file("DESCRIPTION", package = pkgname),
#     fields = "Version"
#   )
#   packageStartupMessage(
#     "\n",
#     "This is ", paste(pkgname, version),".", "\n",
#     "BFpack is free software. Please report any bugs."
#   )
# }

# .onAttach <- function(libname, pkgname) {
#   packageStartupMessage("test")
# }

#BF method for ergm class

#' @importFrom ergm ergmMPLE
#' @importFrom Bergm bergm
#' @method BF ergm
#' @export
BF.ergm <- function(x,
                    hypothesis = NULL,
                    prior.hyp.explo = NULL,
                    prior.hyp.conf = NULL,
                    prior.hyp = NULL,
                    complement = TRUE,
                    log = FALSE,
                    cov.prob = .95,
                    ...){

  if(!(cov.prob>0 & cov.prob<1)){
    stop("The argument 'cov.prob' is a coverage probability for the interval estimates that should lie between 0 and 1. The default is 0.95.")
  }

  logIN <- log

  # extract coefficients
  estimate <- coef(x)
  K1 <- length(estimate)
  # get design matrix of pseudo likelihood to construct prior covariance matrix
  nw <- x$network
  form.char <- paste(format(x$formula), collapse = '')
  location_tilde <- regexpr("~",form.char)[1]
  form.new <- as.formula(paste0("nw ~",substr(form.char,start=location_tilde+1,stop=nchar(form.char))))
  x_MPLE <- ergmMPLE(form.new,output="dyadlist")
  design.X <- x_MPLE$predictor[,2+1:K1]
  which.edges <- which(colnames(design.X)=="edges")
  if(length(which.edges)==0){ #no intercept 'edges'
    Xdelta <- as.matrix(design.X)
    priorcov <- solve(t(Xdelta)%*%Xdelta) * nrow(Xdelta)
  }else{
    Xdelta <- as.matrix(design.X[,-which.edges])
    priorcov.Xdelta <- solve(t(Xdelta)%*%Xdelta) * nrow(Xdelta)
    priorcov <- matrix(0,ncol=K1,nrow=K1)
    priorcov[which.edges,which.edges] <- 100000 #flat prior for the 'edges' parameter
    if(which.edges==1){
      priorcov[2:K1,2:K1] <- priorcov.Xdelta
    }else{
      if(which.edges==K1){
        priorcov[1:(K1-1),1:(K1-1)] <- priorcov.Xdelta
      }else{
        priorcov[1:(which.edges-1),1:(which.edges-1)] <- priorcov.Xdelta[1:(which.edges-1),1:(which.edges-1)]
        priorcov[(which.edges+1):K1,(which.edges+1):K1] <- priorcov.Xdelta[which.edges:(K1-1),which.edges:(K1-1)]
        priorcov[1:(which.edges-1),(which.edges+1):K1] <- priorcov.Xdelta[1:(which.edges-1),which.edges:(K1-1)]
        priorcov[(which.edges+1):K1,1:(which.edges-1)] <- t(priorcov[1:(which.edges-1),(which.edges+1):K1])
      }
    }
  }
  Bergm.out <- bergm(form.new,prior.mean=rep(0,K1),prior.sigma=priorcov,...)
  #get robust estimates for the Gaussian mean and covariance matrix
  post.mean <- apply(Bergm.out$Theta,2,median)
  names(post.mean) <- names(estimate)
  #get robust estimate of posterior covariance matrix
  mlm1 <- lm(Bergm.out$Theta ~ 1)
  post.Sigma <- sandwich(mlm1) * nrow(Bergm.out$Theta)

  # use Savage-Dickey approximation of the BF
  if(length(which.edges)==0){
    prior.mean = rep(0,K1)
    names(prior.mean) <- names(estimate)
    row.names(priorcov) <- colnames(priorcov) <- names(estimate)
    BFergm_out <- Savage.Dickey.Gaussian(prior.mean = prior.mean,
                     prior.sigma = priorcov,
                     post.mean = post.mean,
                     post.sigma = post.Sigma,
                     hypothesis = hypothesis,
                     prior.hyp = prior.hyp,
                     prior.hyp.explo = prior.hyp.explo,
                     prior.hyp.conf = prior.hyp.conf,
                     complement = complement,
                     log = logIN)
  }else{
    prior.mean = rep(0,K1)
    names(prior.mean) <- names(estimate)
    row.names(priorcov) <- colnames(priorcov) <- names(estimate)
    BFergm_out <- Savage.Dickey.Gaussian(prior.mean = prior.mean[-which.edges],
                                         prior.sigma = priorcov[-which.edges,-which.edges],
                                         post.mean = post.mean[-which.edges],
                                         post.sigma = post.Sigma[-which.edges,-which.edges],
                                         hypothesis = hypothesis,
                                         prior.hyp = prior.hyp,
                                         prior.hyp.explo = prior.hyp.explo,
                                         prior.hyp.conf = prior.hyp.conf,
                                         complement = complement,
                                         log = logIN)
  }

  CrI_LB <- (1 - cov.prob)/2
  CrI_UB <- 1 - (1 - cov.prob)/2
  postestimates <- cbind(apply(Bergm.out$Theta,2,mean),
                          apply(Bergm.out$Theta,2,median),
                          apply(Bergm.out$Theta,2,quantile,CrI_LB),
                          apply(Bergm.out$Theta,2,quantile,CrI_UB))
  rownames(postestimates) <- names(estimate)
  colnames(postestimates) <- c("mean","median","2.5%","97.5%")

  BFergm_out$estimates <- postestimates
  BFergm_out$model <- x
  BFergm_out$call <- match.call()
  BFergm_out$bayesfactor <- "Bayes factors based on unit-information priors and Gaussian approximations"
  BFergm_out$parameter <- "ERGM coefficients"
  BFergm_out$model_update <- Bergm.out
  BFergm_out$prior.parameters <- list(prior.mean=prior.mean,prior.cov=priorcov)

  return(BFergm_out)
}

#' @method get_estimates ergm
#' @export
get_estimates.ergm <- function(x, ...){

  nw <- x$network
  form.char <- paste(format(x$formula), collapse = '')
  location_tilde <- regexpr("~",form.char)[1]
  form.new <- as.formula(paste0("nw ~",substr(form.char,start=location_tilde+1,stop=nchar(form.char))))

  estimate <- coef(x)
  K1 <- length(estimate)
  x_MPLE <- ergmMPLE(form.new,output="dyadlist")
  design.X <- x_MPLE$predictor[,2+1:K1]
  which.edges <- which(colnames(design.X)=="edges")
  out <- list()
  if(length(which.edges)==0){ #no intercept 'edges'
    out$estimate <- coef(x)
    out$Sigma <- list(vcov(x))
  }else{
    out$estimate <- coef(x)[-which.edges]
    out$Sigma <- list(as.matrix(vcov(x)[-which.edges,-which.edges]))
  }
  class(out) <- "model_estimates"
  attr(out, "analysisType") <- "ergm"
  out
}


#' @method BF bergm
#' @export
BF.bergm <- function(x,
                    hypothesis = NULL,
                    prior.hyp.explo = NULL,
                    prior.hyp.conf = NULL,
                    prior.hyp = NULL,
                    complement = TRUE,
                    log = FALSE,
                    ...){

  logIN <- log

  form.char <- paste(format(x$formula), collapse = '')
  location_tilde <- regexpr("~",form.char)[1]
  name.nw <- substr(form.char,start=1,stop=location_tilde-2)
  if(!exists(name.nw)){
    stop(paste0("For an object of class 'bergm', the function 'BF()' only runs if the network data object '",name.nw,
                   "' is also present in the environment."))
  }

  # first check if effect names in hypothesis argument correspond with names in x
  coef_names_hyp <- names(get_estimates(x)$estimate)
  if(!is.null(hypothesis)){
    test0 <- parse_hypothesis(coef_names_hyp,hypothesis)
  }

  # extract coefficients
  estimate <- apply(x$Theta,2,median)
  K1 <- length(estimate)
  coef_names <- paste0("theta",1:K1)
  # get design matrix of pseudo likelihood to construct prior covariance matrix
  x_MPLE <- ergmMPLE(formula=x$formula,output="dyadlist")
  design.X <- x_MPLE$predictor[,2+1:K1]
  which.edges <- which(colnames(design.X)=="edges")
  if(length(which.edges)==0){ #no intercept 'edges'
    Xdelta <- as.matrix(design.X)
    priorcov <- solve(t(Xdelta)%*%Xdelta) * nrow(Xdelta)
  }else{
    Xdelta <- as.matrix(design.X[,-which.edges])
    priorcov.Xdelta <- solve(t(Xdelta)%*%Xdelta) * nrow(Xdelta)
    priorcov <- matrix(0,ncol=K1,nrow=K1)
    priorcov[which.edges,which.edges] <- 100000 #flat prior for the 'edges' parameter
    if(which.edges==1){
      priorcov[2:K1,2:K1] <- priorcov.Xdelta
    }else{
      if(which.edges==K1){
        priorcov[1:(K1-1),1:(K1-1)] <- priorcov.Xdelta
      }else{
        priorcov[1:(which.edges-1),1:(which.edges-1)] <- priorcov.Xdelta[1:(which.edges-1),1:(which.edges-1)]
        priorcov[(which.edges+1):K1,(which.edges+1):K1] <- priorcov.Xdelta[which.edges:(K1-1),which.edges:(K1-1)]
        priorcov[1:(which.edges-1),(which.edges+1):K1] <- priorcov.Xdelta[1:(which.edges-1),which.edges:(K1-1)]
        priorcov[(which.edges+1):K1,1:(which.edges-1)] <- t(priorcov[1:(which.edges-1),(which.edges+1):K1])
      }
    }
  }
  Bergm.out <- bergm(x$formula,prior.mean=rep(0,K1),prior.sigma=priorcov,...)
  #get robust estimates for the Gaussian mean and covariance matrix
  post.mean <- apply(Bergm.out$Theta,2,median)
  names(post.mean) <- paste0("theta",1:K1)
  #get robust estimate of posterior covariance matrix
  mlm1 <- lm(Bergm.out$Theta ~ 1)
  post.Sigma <- sandwich(mlm1) * nrow(Bergm.out$Theta)

  # use Savage-Dickey approximation of the BF
  if(length(which.edges)==0){
    prior.mean = rep(0,K1)
    names(prior.mean) <- coef_names
    row.names(priorcov) <- colnames(priorcov) <- coef_names
    BFergm_out <- Savage.Dickey.Gaussian(prior.mean = prior.mean,
                                         prior.sigma = priorcov,
                                         post.mean = post.mean,
                                         post.sigma = post.Sigma,
                                         hypothesis = hypothesis,
                                         prior.hyp = prior.hyp,
                                         prior.hyp.explo = prior.hyp.explo,
                                         prior.hyp.conf = prior.hyp.conf,
                                         complement = complement,
                                         log = logIN)
  }else{
    prior.mean = rep(0,K1)
    names(prior.mean) <- coef_names
    row.names(priorcov) <- colnames(priorcov) <- coef_names
    BFergm_out <- Savage.Dickey.Gaussian(prior.mean = prior.mean[-which.edges],
                                         prior.sigma = priorcov[-which.edges,-which.edges],
                                         post.mean = post.mean[-which.edges],
                                         post.sigma = post.Sigma[-which.edges,-which.edges],
                                         hypothesis = hypothesis,
                                         prior.hyp = prior.hyp,
                                         prior.hyp.explo = prior.hyp.explo,
                                         prior.hyp.conf = prior.hyp.conf,
                                         complement = complement,
                                         log = logIN)
  }

  postestimates <- cbind(apply(Bergm.out$Theta,2,mean),
                         apply(Bergm.out$Theta,2,median),
                         apply(Bergm.out$Theta,2,quantile,.025),
                         apply(Bergm.out$Theta,2,quantile,.975))
  rownames(postestimates) <- names(estimate)
  colnames(postestimates) <- c("mean","median","2.5%","97.5%")

  BFergm_out$estimates <- postestimates
  BFergm_out$model <- x
  BFergm_out$call <- match.call()
  BFergm_out$bayesfactor <- "Bayes factors based on unit-information priors and Gaussian approximations"
  BFergm_out$parameter <- "ERGM coefficients"
  BFergm_out$model_update <- Bergm.out
  BFergm_out$prior.parameters <- list(prior.mean=prior.mean,prior.cov=priorcov)

  return(BFergm_out)
}

#' @method get_estimates bergm
#' @export
get_estimates.bergm <- function(x, ...){

  form.char <- paste(format(x$formula), collapse = '')
  location_tilde <- regexpr("~",form.char)[1]
  name.nw <- substr(form.char,start=1,stop=location_tilde-2)
  if(!exists(name.nw)){
    stop(paste0("For an object of class 'bergm', the function 'BF()' only runs if the network data object '",name.nw,
                "' is also present in the environment."))
  }

  K1 <- length(apply(x$Theta,2,median))
  names.bergm.coef <- paste0("theta",1:K1)
  x_MPLE <- ergmMPLE(formula=x$formula,output="dyadlist")
  design.X <- x_MPLE$predictor[,2+1:K1]
  which.edges <- which(colnames(design.X)=="edges")
  out <- list()
  if(length(which.edges)==0){ #no intercept 'edges'
    out$estimate <- apply(x$Theta,2,median)
    names(out$estimate) <- names.bergm.coef
    mlm1 <- lm(x$Theta ~ 1)
    out$Sigma <- list(sandwich(mlm1) * nrow(x$Theta))
    colnames(out$Sigma[[1]]) <- row.names(out$Sigma[[1]]) <- names(out$estimate)
  }else{
    out$estimate <- apply(x$Theta,2,median)
    names(out$estimate) <- names.bergm.coef
    out$estimate <- out$estimate[-which.edges]
    mlm1 <- lm(x$Theta ~ 1)
    out$Sigma <- list(sandwich(mlm1)[-which.edges,-which.edges] * nrow(x$Theta))
    colnames(out$Sigma[[1]]) <- row.names(out$Sigma[[1]]) <- names(out$estimate)
  }
  class(out) <- "model_estimates"
  attr(out, "analysisType") <- "ergm"
  out
}

#


#' @importFrom stats qnorm dnorm pnorm
#' @describeIn BF S3 method for a named vector 'x'
#' @method BF default
#' @export
BF.default <- function(x,
                       hypothesis = NULL,
                       prior.hyp.explo = NULL,
                       prior.hyp.conf = NULL,
                       prior.hyp = NULL,
                       complement = TRUE,
                       log = FALSE,
                       cov.prob = .95,
                       Sigma,
                       n,
                       ...){

  #Input is a named mean vector x, covariance matrix and number of observations
  #These are extracted by the relevant method functions from a model object and
  #passed together with the hypothesis and prior to the Gaussian_estimator

  # use Savage-Dickey approximation of the BF
  BF_out <- Savage.Dickey.Gaussian(prior.mean = rep(0, length(x)),
                                        prior.sigma = Sigma * n,
                                        post.mean = x,
                                        post.sigma = Sigma,
                                        hypothesis = hypothesis,
                                        prior.hyp.explo = prior.hyp.explo,
                                        prior.hyp.conf = prior.hyp.conf,
                                        prior.hyp = prior.hyp,
                                        complement = complement,
                                        log = log,
                                   cov.prob=cov.prob)

  BF_out$model <- x
  BF_out$call <- match.call()
  BF_out$bayesfactor <- "adjusted fractional Bayes factors using Gaussian approximations"
  BF_out$parameter <- "general parameters"

  BF_out

}

# extended Savage-Dickey density ratio for multivariate normal prior and posterior
Savage.Dickey.Gaussian <- function(prior.mean,
                                   prior.sigma,
                                   post.mean,
                                   post.sigma,
                                   hypothesis,
                                   prior.hyp.explo,
                                   prior.hyp.conf,
                                   prior.hyp,
                                   complement,
                                   log = FALSE,
                                   cov.prob = .95){

  if(!(cov.prob>0 & cov.prob<1)){
    stop("The argument 'cov.prob' is a coverage probability for the interval estimates that should lie between 0 and 1. The default is 0.95.")
  }
  CrI_LB <- (1 - cov.prob)/2
  CrI_UB <- 1 - (1 - cov.prob)/2

  #prior.mean is a normal prior mean of key parameters
  #prior.sigma is a normal prior covariance matrix of key parameters
  #post.mean is a normal posterior mean of key parameters
  #post.sigma is a normal posterior covariance matrix of key parameters
  #These are passed together with the hypothesis and prior to the Gaussian_estimator
  #the prior of the nuisance parameters under the constrained models is a conditional version of the full prior

  meanN <- post.mean     #the posterior mean is approximated with the estimate
  covmN <- post.sigma    #the posterior covariance matrix is approximated with the error covariance matrix

  names_coef <- names(meanN)
  covm0 <- prior.sigma
  mean0 <- prior.mean # for constrained testing prior mean is relocated to 'boundary of constrained space'

  logIN <- log

  # check proper usage of argument 'prior.hyp.conf' and 'prior.hyp.explo'
  if(!is.null(prior.hyp.conf)){
    prior.hyp <- prior.hyp.conf
  }
  dummy <- 123
  class(dummy) <- "dummy"
  prior.hyp.explo <- process.prior.hyp.explo(prior_hyp_explo = prior.hyp.explo, model=dummy)

  # compute exploratory BFs for each parameter
  relfit <- matrix(c(dnorm(0,mean=meanN,sd=sqrt(diag(covmN)),log=TRUE),
                     pnorm(0,mean=meanN,sd=sqrt(diag(covmN)),log.p=TRUE),
                     pnorm(0,mean=meanN,sd=sqrt(diag(covmN)),log.p=TRUE,lower.tail = FALSE)),ncol=3)

  relcomp <- matrix(c(dnorm(0,mean=mean0,sd=sqrt(diag(covm0)),log=TRUE),
                      pnorm(0,mean=mean0,sd=sqrt(diag(covm0)),log.p=TRUE),
                      pnorm(0,mean=mean0,sd=sqrt(diag(covm0)),log.p=TRUE,lower.tail = FALSE)),ncol=3)
  BFtu_exploratory <- relfit - relcomp
  maxrow <- apply(BFtu_exploratory,1,max)
  norm_BF_explo <- exp(BFtu_exploratory - maxrow %*% t(rep(1,3))) * (rep(1,nrow(relcomp)) %*% t(prior.hyp.explo[[1]]))
  PHP_exploratory <- norm_BF_explo / apply(norm_BF_explo,1,sum)
  # PHP_exploratory <- round(exp(BFtu_exploratory - maxrow %*% t(rep(1,3))) /
  #                            apply(exp(BFtu_exploratory - maxrow %*% t(rep(1,3))),1,sum),3)
  colnames(PHP_exploratory) <- c("Pr(=0)","Pr(<0)","Pr(>0)")
  row.names(PHP_exploratory) <- row.names(BFtu_exploratory) <- names_coef
  colnames(BFtu_exploratory) <- c("H(=0)","H(<0)","H(>0)")

  # compute posterior estimates
  postestimates <- cbind(meanN,meanN,
                         t(matrix(unlist(lapply(1:length(meanN),function(coef){
                           ub <- qnorm(p=.975)*sqrt(covmN[coef,coef])+meanN[coef]
                           lb <- qnorm(p=.025)*sqrt(covmN[coef,coef])+meanN[coef]
                           return(c(lb,ub))
                         })),nrow=2)),
                         1-pnorm(0,mean=meanN,sd=sqrt(diag(covmN)))
  )
  row.names(postestimates) <- names_coef
  colnames(postestimates) <- c("mean","median",paste0(as.character(round(CrI_LB*100,7)),"%"),
                               paste0(as.character(round(CrI_UB*100,7)),"%"),"Pr(>0)")

  if(logIN == FALSE){
    BFtu_exploratory <- exp(BFtu_exploratory)
  }

  if(is.null(hypothesis)){
    BFtu_confirmatory <- PHP_confirmatory <- BFmatrix_confirmatory <- relfit <-
      relcomp <- BFtable <- hypotheses <- priorprobs <- NULL
  }else{
    # confirmatory tests based on input constraints
    parse_hyp <- parse_hypothesis(names_coef,hypothesis)
    parse_hyp$hyp_mat <- do.call(rbind, parse_hyp$hyp_mat)
    #create coefficient with equality and order constraints
    RrList <- make_RrList2(parse_hyp)
    RrE <- RrList[[1]]
    RrO <- RrList[[2]]

    # RrStack is used to check conflicting constraints, and for the default prior location
    if(length(RrE)==1){
      RrStack <- rbind(do.call(rbind,RrE),do.call(rbind,RrO))
      RrStack <- interval_RrStack(RrStack)
    }else{
      RrStack_list <- lapply(1:length(RrE),function(h){
        interval_RrStack(rbind(RrE[[h]],RrO[[h]]))
      })
      RrStack <- do.call(rbind,RrStack_list)
    }
    K <- length(meanN)
    if(nrow(RrStack)>1){
      RStack <- RrStack[,-(K+1)]
      rStack <- RrStack[,(K+1)]
    }else{
      RStack <- matrix(RrStack[,-(K+1)],nrow=1)
      rStack <- RrStack[,(K+1)]
    }

    # check if a common boundary exists for prior location under all constrained hypotheses
    if(nrow(RrStack) > 1){
      rref_ei <- rref(RrStack)
      nonzero <- rref_ei[,K+1]!=0
      if(max(nonzero)>0){
        row1 <- max(which(nonzero))
        if(sum(abs(rref_ei[row1,1:K]))==0){
          stop("No common boundary point for prior location. Conflicting constraints.")
        }
      }
      #determine fraction via number of independent rows (constraints)
      if(is.matrix(rref_ei[,-(K+1)])){
        numindep <- sum(apply(abs(rref_ei[,-(K+1)]),1,sum)!=0)
      }else{
        numindep <- sum(apply(abs(as.matrix(rref_ei[,-(K+1)])),1,sum)!=0)
      }
    } else {
      numindep <- 1
    }

    #get relative fit and complexity of hypotheses
    numhyp <- length(RrE)

    relcomp <- t(matrix(unlist(lapply(1:numhyp,function(h){
      Gaussian_measures(mean1 = mean0, Sigma1 = covm0, RrE1 = RrE[[h]], RrO1 = RrO[[h]],
                        names1=names_coef,constraints1=parse_hyp$original_hypothesis[h])
    })),nrow=2))
    relfit <- t(matrix(unlist(lapply(1:numhyp,function(h){
      Gaussian_measures(mean1 = meanN, Sigma1 = covmN, RrE1 = RrE[[h]], RrO1 = RrO[[h]],
                        names1=names_coef,constraints1=parse_hyp$original_hypothesis[h])
    })),nrow=2))
    row.names(relfit) <- row.names(relcomp) <- parse_hyp$original_hypothesis

    active.complement <- TRUE
    if(complement == TRUE){
      # get relative fit and complexity of complement hypothesis
      #Note that input is a bit strange here, Gaussian_prob_Hc needs fixing
      relcomp <- Gaussian_prob_Hc(mean1 = mean0, Sigma1 = covm0, relmeas = relcomp, RrO = RrO)
      relfit <- Gaussian_prob_Hc(mean1 = meanN, Sigma1 = covmN, relmeas = relfit, RrO = RrO)
      if(nrow(relfit) == numhyp){active.complement <- FALSE}
    }
    hypothesisshort <- unlist(lapply(1:nrow(relfit),function(h) paste0("H",as.character(h))))
    row.names(relfit) <- row.names(relfit) <- hypothesisshort
    colnames(relcomp) <- c("c_E", "c_0")
    colnames(relfit) <- c("f_E", "f_0")

    # the BF for the complement hypothesis vs Hu needs to be computed.
    BFtu_confirmatory <- c(apply(relfit - relcomp, 1, sum))
    # Check input of prior probabilies
    if(is.null(prior.hyp)){
      priorprobs <- rep(1/length(BFtu_confirmatory),length(BFtu_confirmatory))
    }else{
      if(!is.numeric(prior.hyp) || length(prior.hyp)!=length(BFtu_confirmatory)){
        if(complement == TRUE && active.complement == FALSE && length(prior.hyp)==length(BFtu_confirmatory)+1){
          # the prior prob for the complement hypothesis is given but it is not used as the constrained hypotheses
          # cover the entire parameter space
          priorprobs <- prior.hyp[1:length(BFtu_confirmatory)]
          priorprobs <- priorprobs / sum(priorprobs)
        }else{
          warning(paste0("Argument 'prior.hyp' should be numeric and of length ",as.character(length(BFtu_confirmatory)),
                         ". Equal prior probabilities are used."))
          priorprobs <- rep(1/length(BFtu_confirmatory),length(BFtu_confirmatory))
        }
      }else{
        priorprobs <- prior.hyp
        priorprobs <- priorprobs / sum(priorprobs)
      }
    }
    names(priorprobs) <- hypothesisshort

    maxBFtu <- max(BFtu_confirmatory)
    PHP_confirmatory <- round(exp(BFtu_confirmatory-maxBFtu)*priorprobs /
                                sum(exp(BFtu_confirmatory-maxBFtu)*priorprobs),3)
    BFtable <- cbind(relcomp,relfit,relfit[,1]-relcomp[,1],relfit[,2]-relcomp[,2],
                     apply(relfit,1,sum)-apply(relcomp,1,sum),PHP_confirmatory)
    BFtable[,1:7] <- exp(BFtable[,1:7])
    row.names(BFtable) <- names(PHP_confirmatory)
    colnames(BFtable) <- c("complex=","complex>","fit=","fit>","BF=","BF>","BF","PHP")
    BFmatrix_confirmatory <- matrix(rep(BFtu_confirmatory,length(BFtu_confirmatory)),ncol=length(BFtu_confirmatory)) -
      t(matrix(rep(BFtu_confirmatory,length(BFtu_confirmatory)),ncol=length(BFtu_confirmatory)))
    diag(BFmatrix_confirmatory) <- log(1)
    # row.names(BFmatrix_confirmatory) <- Hnames
    # colnames(BFmatrix_confirmatory) <- Hnames
    if(nrow(relfit)==length(parse_hyp$original_hypothesis)){
      hypotheses <- parse_hyp$original_hypothesis
    }else{
      hypotheses <- c(parse_hyp$original_hypothesis,"complement")
    }

    if(logIN == FALSE){
      BFtu_confirmatory <- exp(BFtu_confirmatory)
      BFmatrix_confirmatory <- exp(BFmatrix_confirmatory)
    }
  }

  BF_out <- list(
    BFtu_exploratory=BFtu_exploratory,
    PHP_exploratory=PHP_exploratory,
    BFtu_confirmatory=BFtu_confirmatory,
    PHP_confirmatory=PHP_confirmatory,
    BFmatrix_confirmatory=BFmatrix_confirmatory,
    BFtable_confirmatory=BFtable,
    prior.hyp.explo=prior.hyp.explo,
    prior.hyp.conf=priorprobs,
    hypotheses=hypotheses,
    estimates=postestimates,
    model=NULL,
    bayesfactor="Bayes factors using Gaussian approximations",
    parameter="general parameters",
    log = logIN,
    call=NULL)

  class(BF_out) <- "BF"

  BF_out

}

# compute relative meausures (fit or complexity) under a multivariate Gaussian distribution
#' @importFrom mvtnorm dmvnorm pmvnorm
#' @importFrom berryFunctions is.error
Gaussian_measures <- function(mean1,Sigma1,n1=0,RrE1,RrO1,names1=NULL,constraints1=NULL,...){
  K <- length(mean1)
  relE <- relO <- log(1)
  if(!is.null(RrE1) && is.null(RrO1)){ #only equality constraints
    RE1 <- RrE1[,-(K+1)]
    if(!is.matrix(RE1)){
      RE1 <- matrix(RE1,ncol=K)
    }
    rE1 <- RrE1[,(K+1)]
    qE1 <- nrow(RE1)
    meanE <- RE1%*%mean1
    SigmaE <- RE1%*%Sigma1%*%t(RE1)
    relE <- dmvnorm(rE1,mean=c(meanE),sigma=SigmaE,log=TRUE)
  }
  if(is.null(RrE1) && !is.null(RrO1)){ #only order constraints
    RO1 <- RrO1[,-(K+1)]
    if(!is.matrix(RO1)){
      RO1 <- matrix(RO1,ncol=K)
    }
    qO1 <- nrow(RO1)
    rO1 <- RrO1[,(K+1)]

    if(Rank(RO1)==nrow(RO1)){ #RO1 is of full row rank. So use transformation.
      meanO <- c(RO1%*%mean1)
      SigmaO <- RO1%*%Sigma1%*%t(RO1)
      #check_vcov(SigmaO)
      relO <- log(pmvnorm(lower=rO1,upper=Inf,mean=meanO,sigma=SigmaO,...)[1])
    }else{ #no linear transformation can be used; pmvt cannot be used. Use bain with a multivariate normal approximation
      names(mean1) <- names1
      mean1vec <- c(mean1)
      names(mean1vec) <- row.names(Sigma1) <- colnames(Sigma1) <- names1
      if(n1>0){ # we need prior measures
        bain_res <- bain(x=mean1vec,hypothesis=constraints1,Sigma=Sigma1,n=n1)
        relO <- log(bain_res$fit[1,4])
      }else { # we need posterior measures (there is very little information)
        if(!is.error(bain(x=mean1vec,hypothesis=constraints1,Sigma=Sigma1,n=999))){
          bain_res <- bain(x=mean1vec,hypothesis=constraints1,Sigma=Sigma1,n=999) #n not used in computation
          relO <- log(bain_res$fit[1,3])
        }else{
          # simply compute the proportion based on 1e6 draws
          relO <- log(mean(apply(rmvnorm(1e6,mean=mean1vec,sigma=Sigma1) %*% t(RO1) - rep(1,1e6) %*% t(rO1) > 0,1,prod)))
        }
      }

    }
  }
  if(!is.null(RrE1) && !is.null(RrO1)){ #hypothesis with equality and order constraints

    RE1 <- RrE1[,-(K+1)]
    if(!is.matrix(RE1)){
      RE1 <- matrix(RE1,ncol=K)
    }
    rE1 <- RrE1[,(K+1)]
    qE1 <- nrow(RE1)
    RO1 <- RrO1[,-(K+1)]
    if(!is.matrix(RO1)){
      RO1 <- matrix(RO1,ncol=K)
    }
    qO1 <- nrow(RO1)
    rO1 <- RrO1[,(K+1)]
    Rr1 <- rbind(RrE1,RrO1)

    if(Rank(Rr1) == nrow(Rr1)){

      R1 <- rbind(RE1,RO1)

      #b)
      Tmean1 <- R1 %*% mean1
      TSigma1 <- R1 %*% Sigma1 %*% t(R1)

      # relative meausure for equalities
      relE <- dmvnorm(x=rE1,mean=Tmean1[1:qE1],sigma=matrix(TSigma1[1:qE1,1:qE1],ncol=qE1),log=TRUE)

      # Partitioning equality part and order part
      Tmean1E <- Tmean1[1:qE1]
      Tmean1O <- Tmean1[qE1+1:qO1]

      TSigma1EE <- TSigma1[1:qE1,1:qE1]
      TSigma1OE <- matrix(c(TSigma1[qE1+1:qO1,1:qE1]),nrow=qO1)
      TSigma1OO <- TSigma1[qE1+1:qO1,qE1+1:qO1]

      #conditional location and covariance matrix
      Tmean1OgE <- Tmean1O + TSigma1OE %*% solve(TSigma1EE) %*% (rE1-Tmean1E)
      TSigma1OgE <- TSigma1OO - TSigma1OE %*% solve(TSigma1EE) %*% t(TSigma1OE)

      relO <- log(pmvnorm(lower=rO1,upper=Inf,mean=c(Tmean1OgE),sigma=TSigma1OgE,...)[1])

    }else{ #use bain for the computation of the probability
      names(mean1) <- names1
      if(n1>0){ # we need prior measures
        bain_res <- bain(x=c(mean1),hypothesis=constraints1,Sigma=Sigma1,n=n1)
        relO <- log(bain_res$fit[1,4])
        relE <- log(bain_res$fit[1,2])
      }else { # we need posterior measures (there is very little information)
        bain_res <- bain(x=c(mean1),hypothesis=constraints1,Sigma=Sigma1,n=999) #n not used in computation
        relO <- log(bain_res$fit[1,3])
        relE <- log(bain_res$fit[1,1])
      }
    }
  }

  return(c(relE,relO))
}

# The function computes the probability of an unconstrained draw falling in the complement subspace under a multivariate Gaussian distribution.
#' @importFrom mvtnorm rmvnorm
Gaussian_prob_Hc <- function(mean1,Sigma1,relmeas,RrO){

  numpara <- length(mean1)
  numhyp <- nrow(relmeas)
  which_eq <- relmeas[,1] != log(1)
  if(sum(which_eq)==numhyp){ # Then the complement is equivalent to the unconstrained hypothesis.
    relmeas <- rbind(relmeas,rep(log(1),2))
    rownames(relmeas)[numhyp+1] <- "complement"
  }else{ # So there is at least one hypothesis with only order constraints
    welk <- which(!which_eq)
    if(length(welk)==1){ # There is one hypothesis with only order constraints. Hc is complement of this hypothesis.
      relmeas <- rbind(relmeas,rep(log(1),2))
      relmeas[numhyp+1,2] <- log(1 - exp(relmeas[welk,2]))
      rownames(relmeas)[numhyp+1] <- "complement"
    }else{ # So more than one hypothesis with only order constraints
      # First we check whether ther is an overlap between the order constrained spaces.
      draws2 <- 1e5
      randomDraws <- rmvnorm(draws2,mean=rep(0,numpara),sigma=diag(numpara))
      #get draws that satisfy the constraints of the separate order constrained hypotheses
      checksOC <- lapply(welk,function(h){
        Rorder <- matrix(RrO[[h]][,-(1+numpara)],nrow=nrow(RrO[[h]]))
        # if(ncol(Rorder)==1){
        #   Rorder <- t(Rorder)
        # }
        rorder <- as.matrix(RrO[[h]][,1+numpara])
        apply(randomDraws%*%t(Rorder) > rep(1,draws2)%*%t(rorder),1,prod)
      })
      checkOCplus <- Reduce("+",checksOC)

      if(sum(checkOCplus > 0) < draws2){ #then the joint order constrained hypotheses do not completely cover the parameter space.
        if(sum(checkOCplus>1)==0){ # then order constrained spaces are nonoverlapping
          relmeas <- rbind(relmeas,rep(log(1),2))
          relmeas[numhyp+1,2] <- log(1 - sum(exp(relmeas[welk,2])))
          rownames(relmeas)[numhyp+1] <- "complement"
        }else{ #the order constrained subspaces at least partly overlap

          # funtion below gives a rough estimate of the posterior probability under Hc
          # a bain type of algorithm would be better of course. but for now this is ok.

          randomDraws <- rmvnorm(draws2,mean=mean1,sigma=Sigma1)
          checksOCpost <- lapply(welk,function(h){
            Rorder <- matrix(RrO[[h]][,-(1+numpara)],nrow=nrow(RrO[[h]]))
            # if(ncol(Rorder)==1){
            #   Rorder <- t(Rorder)
            # }
            rorder <- as.matrix(RrO[[h]][,1+numpara])
            apply(randomDraws%*%t(Rorder) > rep(1,draws2)%*%t(rorder),1,prod)
          })
          relmeas <- rbind(relmeas,rep(log(1),2))
          relmeas[numhyp+1,2] <- log(sum(Reduce("+",checksOCpost) == 0) / draws2)
          rownames(relmeas)[numhyp+1] <- "complement"
        }
      }
    }
  }

  return(relmeas)
}

#' @title Bayes factors for Bayesian exploratory and confirmatory hypothesis
#' testing
#' @description The \code{BF} function can be used for hypothesis testing and
#'  model selection using the Bayes factor. By default exploratory hypothesis tests are
#' performed of whether each model parameter equals zero, is negative, or is
#'  positive. Confirmatory hypothesis tests can be executed by specifying hypotheses with
#' equality and/or order constraints on the parameters of interest. Depending on the
#' class of the fitted model different Bayes factors are used as described in the output.
#'
#' @param x An R object containing the outcome of a statistical analysis.
#' An R object containing the outcome of a statistical analysis. Currently, the
#' following objects can be processed: t_test(), bartlett_test(), lm(), aov(),
#' manova(), cor_test(), lmer() (only for testing random intercep variances),
#' glm(), coxph(), survreg(), polr(), zeroinfl(), rma(), ergm(), bergm(), or named
#' vector objects. In the case \code{x} is a named vector, the arguments \code{Sigma}
#' and \code{n} are also needed. See vignettes for elaborations.
#' @param hypothesis A character string containing the constrained (informative) hypotheses to
#' evaluate in a confirmatory test. The default is NULL, which will result in standard exploratory testing
#' under the model \code{x}.
#' @param prior.hyp.explo The prior probabilities of the hypotheses in the exploratory tests. Except for
#' objects of class \code{aov} (for (M)ANOVA, etc.), this argument should be a vector with three
#' elements reflecting the prior probability of a zero effect, a negative effect, and a positive effect, respectively.
#' For objects of class \code{aov}, the argument
#' should be a list where the first element should be a vector of length 3 specifying the prior probabilities of each parameter
#' being zero, negative, or positive, the second element should be a vector of length 2 specifying the prior probabilities
#' of a model where is no main effect for a factor and the full model, and the third element should be a vector of length 2
#' specifying the prior probabilities of a model where is no interaction effect (if present) for two factors and the full model.
#' The default (\code{NULL}) specifies equal prior probabilities for each hypothesis per exploratory test.
#' @param prior.hyp.conf The prior probabilities of the constrained hypotheses in the confirmatory test.
#' @param prior.hyp Deprecated. Please use the argument \code{prior.hyp.conf}.
#' @param complement a logical specifying whether the complement should be added
#' to the tested hypothesis under \code{hypothesis}.
#' @param log a logical specifying whether the Bayes factors should be computed on a log scale.
#' Default is \code{FALSE}.
#' @param cov.prob coverage probability of the Bayesian credibility interval in the \code{estimates} element of
#' the output object. The argument is only used for objects for which a Bayesion estimation algorithm is used,
#' such as a \code{cor_test} or \code{rma.uni} object. The default coverage probability is 0.95.
#' @param BF.type For certain object classes of \code{x}, different types of Bayes factor tests are supported.
#' This can be specified using this argument. For models of class 'lm' and 't_test', setting this argument to
#' \code{'FBF'} the fractional Bayes factor (O'Hagan, 1995) is used and setting this argument to
#' \code{'AFBF'} the adjusted fractional Bayes factor (Mulder, 2014) is used. The default for these model classes
#' is the fractional Bayes factor. For models of class 'rma.uni' (meta-analyses), 'BF.type' controls the prior
#' that is used for the test. Certain defaults are provided for a standardized effect (\code{BF.type="stand.effect"}),
#' a normal prior with mean 0 and sd 0.5 is used, for log odds (\code{BF.type="log.odds"}), a Student prior with mean 0,
#' scale 2.36, and 13.1 degrees of freedom is used (corresponding to uniform priors on the success probabilities), for
#' a correlation (\code{BF.type="correlation"}), a logistic prior is used for the Fisher transformed correlation having
#' scale 0.5 (corresponding to a uniform prior for the correlation in (-1,1)), for a unit-information prior
#' (\code{BF.type="unit.info"}), the total sample size \code{sum(ni)} is used to scale a normal unit information prior,
#' and for a manually specified prior \code{BF.type} needs to be an object of class
#' \code{prior} from the \code{metaBMA} package.
#' @param iter Number of iterations that are used to compute the Monte Carlo estimates.
#' Only used for certain hypothesis tests of class \code{mlm} (multivariate regression) where constraints are formulated
#' across different dependent variables as well as different predictors.
#' @param Sigma An approximate posterior covariance matrix (e.g,. error covariance
#' matrix) of the parameters of interest. This argument is only required when \code{x}
#' is a named vector.
#' @param n The (effective) sample size that was used to acquire the estimates in the named vector
#' \code{x} and the error covariance matrix \code{Sigma}. This argument is only required when \code{x}
#' is a named vector.
#' @param ... Parameters passed to and from other functions.
#' @usage NULL
#' @return The output is an object of class \code{BF}. The object has elements:
#' \itemize{
#' \item \code{BFtu_exploratory}: The Bayes factors of the constrained hypotheses against
#' the unconstrained hypothesis in the exploratory test.
#' \item \code{BFtu_main} (only for \code{aov} objects with predictors of class \code{factor}):
#' The Bayes factors of a constrained model where all levels of a \code{factor} are assumed
#' to have the same effect on the outcome variable versus an unconstrained (full) model with
#' no constraints.
#' \item \code{BFtu_interaction} (only for \code{aov} objects with interaction effects with
#' predictors of class \code{factor}): The Bayes factors of a constrained model where the effect
#' of the dummy variables corresponding to an interaction effects are assumed to be zero versus
#' an unconstrained (full) model with no constraints.
#' \item \code{PHP_exploratory:} The posterior probabilities of the constrained hypotheses
#' in the exploratory test.
#' \item \code{PHP_main} (only for \code{aov} objects with predictors of class \code{factor}):
#' The posterior probabilities a constrained model where all levels of a \code{factor} are assumed
#' to have the same effect on the outcome variable versus an unconstrained (full) model with
#' no constraints.
#' \item \code{PHP_interaction} (only for \code{aov} objects with interaction effects with
#' predictors of class \code{factor}): The posterior probabilities of a constrained model where the
#' effect of the dummy variables corresponding to an interaction effects are assumed to be zero versus
#' an unconstrained (full) model with no constraints.
#' \item \code{BFtu_confirmatory}: The Bayes factors of the constrained hypotheses against
#' the unconstrained hypothesis in the confirmatory test using the \code{hypothesis}
#' argument.
#' \item \code{PHP_confirmatory}: The posterior probabilities of the constrained hypotheses
#' in the confirmatory test using the \code{hypothesis} argument.
#' \item \code{BFmatrix_confirmatory}: The evidence matrix which contains the Bayes factors
#' between all possible pairs of hypotheses in the confirmatory test.
#' \item \code{BFtable_confirmatory}: The \code{Specification table} (output when printing the
#' \code{summary} of a \code{BF} for a confirmatory test) which contains the different
#' elements of the extended Savage Dickey density ratio where
#' \itemize{
#' \item The first column `\code{complex=}' quantifies the relative complexity of the
#' equality constraints of a hypothesis (the prior density at the equality constraints in the
#' extended Savage Dickey density ratio).
#' \item The second column `\code{complex>}' quantifies the relative complexity of the
#' order constraints of a hypothesis (the prior probability of the order constraints in the extended
#' Savage Dickey density ratio).
#' \item The third column `\code{fit=}' quantifies the relative fit of the equality
#' constraints of a hypothesis (the posterior density at the equality constraints in the extended
#' Savage Dickey density ratio).
#' \item The fourth column `\code{fit>}' quantifies the relative fit of the order
#' constraints of a hypothesis (the posterior probability of the order constraints in the extended
#' Savage Dickey density ratio)
#' \item The fifth column `\code{BF=}' contains the Bayes factor of the equality constraints
#' against the unconstrained hypothesis.
#' \item The sixth column `\code{BF>}' contains the Bayes factor of the order constraints
#' against the unconstrained hypothesis.
#' \item The seventh column `\code{BF}' contains the Bayes factor of the constrained hypothesis
#' against the unconstrained hypothesis.
#' \item The eighth column `\code{PHP}' contains the posterior probabilities of the hypotheses.
#' }
#' \item \code{prior.hyp.explo}: The prior probabilities of the constrained hypotheses in the exploratory tests.
#' \item \code{prior.hyp.conf}: The prior probabilities of the constrained hypotheses in the confirmatory test.
#' \item \code{hypotheses}: The tested constrained hypotheses in a confirmatory test.
#' \item \code{estimates}: Descriptives of unconstrained estimates based on flat priors (also for \code{rma.uni} objects for
#' Bayesian meta-analyses).
#' \item \code{model}: The input model \code{x}.
#' \item \code{bayesfactor}: The type of Bayes factor that is used for this model.
#' \item \code{parameter}: The type of parameter that is tested.
#' \item \code{log}: \code{logical} whether the Bayes factors were reported on a log scale.
#' \item \code{fraction_number_groupIDs} (only for objects of class \code{lm}): The number of
#' 'group identifiers' that were identified based on the number of unique combinations of \code{level}s
#' of predictor variables of class \code{factor} in the data. These group identifiers are used to automatically
#' specify the minimal fractions that are used to compute (adjusted) fractional Bayes factors.
#' \item \code{fraction_groupID_observations} (only for objects of class \code{lm}): A vector that
#' specifies to which 'group identifier' an observation belongs. The group identifiers are constructed
#' based on the unique combination of the \code{levels} based on the predictor variables of class \code{factor}
#' of the observations.
#' \item \code{call}: The call of the \code{BF} function.
#' }
#' @details The function requires a fitted modeling object. Current analyses
#' that are supported: \code{\link[bain]{t_test}},
#' \code{\link[BFpack]{bartlett_test}},
#' \code{\link[stats]{aov}}, \code{\link[stats]{manova}},
#' \code{\link[stats]{lm}}, \code{mlm},
#' \code{\link[stats]{glm}}, \code{\link[polycor]{hetcor}},
#' \code{\link[lme4]{lmer}}, \code{\link[survival]{coxph}},
#' \code{\link[survival]{survreg}}, \code{\link[ergm]{ergm}},
#' \code{\link[Bergm]{bergm}},
#' \code{\link[pscl]{zeroinfl}}, \code{\link[metafor]{rma}} and \code{\link[MASS]{polr}}.
#'
#' For testing parameters from the results of t_test(), lm(), aov(),
#' manova(), and bartlett_test(), hypothesis testing is done using
#' adjusted fractional Bayes factors are computed (using minimal fractions).
#' For testing measures of association (e.g., correlations) via \code{cor_test()},
#' Bayes factors are computed using joint uniform priors under the correlation
#' matrices. For testing intraclass correlations (random intercept variances) via
#' \code{lmer()}, Bayes factors are computed using uniform priors for the intraclass
#' correlations. For all other tests,  approximate adjusted fractional Bayes factors
#' (with minimal fractions) are computed using Gaussian approximations, similar as
#' a classical Wald test.
#'
#' @references Mulder, Williams, Gu, Tomarken, Böing-Messing, Olsson-Collentine, Meyerink,
#' Menke, Fox, Rosseel, Wagenmakers, Hoijtink, and van Lissa (2021). BFpack: Flexible Bayes
#' Factor Testing of Scientific Theories in R. Journal of Statistical Software, 100.
#' <https://doi.org/10.18637/jss.v100.i18>
#' @references Mulder and Xin (2022). Bayesian Testing of Scientific Expectations under
#' Multivariate Normal Linear Models. <https://doi.org/10.1080/00273171.2021.1904809>
#' @references Mulder and Gelissen (2021). Bayes factor testing of equality and order
#' constraints on measures of association in social research. Journal of Applied Statistics,
#' 50. <https://doi.org/10.1080/02664763.2021.1992360>
#' @references Mulder and Fox (2019). Bayes Factor Testing of Multiple Intraclass Correlations.
#' Bayesian Analysis, 14. <http://doi.org/10.1214/18-BA1115>
#' @references Hoijtink, Mulder, van Lissa, and Gu (2018). A tutorial on testing hypotheses
#' using the Bayes factor. Psychological Methods, 24(5), 539–556.
#' <http://doi.org/10.1037/met0000201>
#' @references Boeing-Messing, van Assen, Hofman, Hoijtink, and Mulder (2017).
#' Bayesian evaluation of constrained hypotheses on variances of multiple independent groups.
#' Psycholological Methods, 22(2), 262-287. <https://doi.org/10.1037/met0000116>
#' @references van Aert and Mulder (2021). Bayesian hypothesis testing and estimation under
#' the marginalized random-effects meta-analysis model. Psychonomic Bulletin and Review, 29,
#' 55–69. <https://doi.org/10.3758/s13423-021-01918-9>
#'
#' @examples
#' \donttest{
#' # EXAMPLE 1. One-sample t test
#' ttest1 <- t_test(therapeutic, mu = 5)
#' print(ttest1)
#' # confirmatory Bayesian one sample t test
#' BF1 <- BF(ttest1, hypothesis = "mu = 5")
#' summary(BF1)
#' # exploratory Bayesian one sample t test
#' BF(ttest1)
#'
#' # EXAMPLE 2. ANOVA
#' aov1 <- aov(price ~ anchor * motivation,data = tvprices)
#' BF1 <- BF(aov1, hypothesis = "anchorrounded = motivationlow;
#'                               anchorrounded < motivationlow")
#' summary(BF1)
#'
#' # EXAMPLE 3. linear regression
#' lm1 <- lm(mpg ~ cyl + hp + wt, data = mtcars)
#' BF(lm1, hypothesis = "wt < cyl < hp = 0")
#'
#' # EXAMPLE 4. Logistic regression
#' fit <- glm(sent ~ ztrust + zfWHR + zAfro + glasses + attract + maturity +
#'    tattoos, family = binomial(), data = wilson)
#' BF1 <- BF(fit, hypothesis = "ztrust > zfWHR > 0;
#'                              ztrust > 0 & zfWHR = 0")
#' summary(BF1)
#'
#' # EXAMPLE 5. Correlation analysis
#' set.seed(123)
#' cor1 <- cor_test(memory[1:20,c(1,2,6)])
#' BF1 <- BF(cor1)
#' summary(BF1)
#' BF2 <- BF(cor1, hypothesis = "Rat_with_Im > Rat_with_Del > 0;
#'                              Rat_with_Im = Rat_with_Del = 0")
#' summary(BF2)
#' # correlations can also be computed between continuous/ordinal variables
#' memory_test <- memory[1:20,c(1,2,6)]
#' memory_test[,3] <- as.ordered(memory_test[,3])
#' cor2 <- cor_test(memory_test)
#' BF(cor2)
#'
#' # EXAMPLE 6. Bayes factor testing on a named vector
#' # A Poisson regression model is used to illustrate the computation
#' # of Bayes factors with a named vector as input
#' poisson1 <- glm(formula = breaks ~ wool + tension,
#'   data = datasets::warpbreaks, family = poisson)
#' # extract estimates, error covariance matrix, and sample size:
#' estimates <- poisson1$coefficients
#' covmatrix <- vcov(poisson1)
#' samplesize <- nobs(poisson1)
#' # compute Bayes factors on equal/order constrained hypotheses on coefficients
#' BF1 <- BF(estimates, Sigma = covmatrix, n = samplesize, hypothesis =
#' "woolB > tensionM > tensionH; woolB = tensionM = tensionH")
#' summary(BF1)
#' }
#' @rdname BF
#' @export
#' @useDynLib BFpack, .registration = TRUE
#'
BF <- function(x, hypothesis, prior.hyp.explo, prior.hyp.conf, prior.hyp, complement, log, cov.prob, ...) {
  UseMethod("BF", x)
}

### Joris Mulder 2019. Bayes factor testing of multiple random intercept models
### via multiple lmer-objects based on Mulder & Fox (2013, 2019).
### with extension to unbalanced data

#' @importFrom stats rgamma rnorm rbeta dbeta runif
#' @importFrom lme4 getME VarCorr
#' @method BF lmerMod
#' @export
BF.lmerMod <- function(x,
                       hypothesis = NULL,
                       prior.hyp.explo = NULL,
                       prior.hyp.conf = NULL,
                       prior.hyp = NULL,
                       complement = TRUE,
                       log = FALSE,
                       cov.prob = .95,
                       ...){

  if(!(cov.prob>0 & cov.prob<1)){
    stop("The argument 'cov.prob' is a coverage probability for the interval estimates that should lie between 0 and 1. The default is 0.95.")
  }

  logIN <- log

  # check proper usage of argument 'prior.hyp.conf' and 'prior.hyp.explo'
  if(!is.null(prior.hyp.conf)){
    prior.hyp <- prior.hyp.conf
  }
  prior.hyp.explo <- process.prior.hyp.explo(prior_hyp_explo = prior.hyp.explo, model=x)

  numcat <- length(x@cnms)
  namescat <- unlist(lapply(1:numcat,function(ca){
    x@cnms[[ca]]
  }))
  if(numcat>1){
    iccnames <- unlist(lapply(1:numcat,function(nc){namescat[nc]}))
  }else{ iccnames <- "icc" }

  #check if names of effects in hypothesis argument match
  if(!is.null(hypothesis)){ # confirmatory test with constrained hypotheses on icc's.
    test1 <- parse_hypothesis(iccnames,hypothesis)
  }

  # check if the lmer-model only has a random intercept or category specific random intercepts
  Zstack <- Reduce(cbind,getME(x,"mmList"))
  if(numcat>1){ #check if the random effects are category specific random intercepts
    for(ca in 1:numcat){
      freq_ca <- table(Zstack[,ca])
      if(sum(abs(sort(as.integer(names(freq_ca))) - c(0,1))) !=0){
        stop("only models with a single random intercept or category specific random intercepts are currently supported when testing icc's.")
      }
    }
  }else{
    freq_ca <- table(getME(x,"mmList")[[1]])
    if(as.integer(names(freq_ca))!=1 || length(as.integer(names(freq_ca)))!=1){
      stop("only models with a single random intercept or category specific random intercepts are currently supported when testing icc's.")
    }
  }

  # sort data per cluster
  clusterindex <- x@flist[[1]]
  if(length(table(table(clusterindex)))>1){
    balanced <- FALSE
    pvec <- table(clusterindex)
  }else{
    balanced <- TRUE
    pvec <- rep(table(clusterindex)[1],length(table(clusterindex)))
  }

  levels(clusterindex) <- 1:length(levels(clusterindex))
  reorder1 <- order(as.integer(clusterindex))
  nclusters <- length(levels(clusterindex)) #total number of groups/clusters
  ystack <- getME(x,"y")[reorder1]
  Xstack <- as.matrix(getME(x,"X")[reorder1,])
  Zstack <- Zstack[reorder1,]
  # next sort data per category
  if(numcat>1){
    # sort rows per category
    firstofcluster <- cumsum(c(0,pvec[1:(nclusters-1)])) + 1
    catassign <- unlist(lapply(firstofcluster,function(cluster){
      ifelse(sum(Zstack[cluster,1:numcat])==1,
             which(Zstack[cluster,1:numcat]==1),
             0)
    }))
    if(sum(names(table(catassign))=="0")==0){ #all groups belong to a category
      ngroups <- table(catassign)
      numgroups <- sum(ngroups)
      names(ngroups) <- NULL
      #reorder data matrices according to categories
      reorder2 <- unlist(lapply(1:numcat,function(ca){
        welk_ca <- which(catassign==ca)
        names(welk_ca) <- NULL
        unlist(lapply(welk_ca,function(cluster){
          firstofcluster[cluster]+0:(pvec[cluster]-1)
        }))
      }))
      #update order of group sizes
      pvec <- pvec[unlist(lapply(1:numcat,function(ca){
        welk_ca <- which(catassign==ca)
      }))]
      Xstack <- Xstack[reorder2,]
      if(is.null(dim(Xstack))){ #then create column matrix from vector
        Xstack <- t(t(Xstack))
      }
      ystack <- ystack[reorder2]
      Zstack <- Zstack[reorder2,]
    }else{ #only include groups that belong to a category
      stop("Some groups don't belong to a group category. Exclude these groups from the data.")
    }
  }else{
    ngroups <- numgroups <- nclusters
  }

  #transform data matrices with Helmert matrix
  firstofcluster0 <- c(0,cumsum(pvec[1:(numgroups-1)]))
  zWstack <- do.call(rbind,
                     lapply(1:length(pvec),function(j){
                       H_j <- Helmert(pvec[j])
                       H_j%*%cbind(ystack[firstofcluster0[j]+1:pvec[j]],
                                   as.matrix(Xstack[firstofcluster0[j]+1:pvec[j],]) )
                     })
  )
  #extract ML estimates for rho
  tau2ML <- unlist(VarCorr(x))
  sigma2ML <- attr(VarCorr(x),"sc")**2
  rhoML <- tau2ML/(tau2ML+sigma2ML)

  shape0 <- c(1,1) # set uniform priors for icc's
  cat("First, unconstrained icc analysis...")
  cat("\n")
  cat("\n")
  numuncdraws <- 5e4
  marglike_Hu <- MargLikeICC_Hq(rhoML,zW=zWstack,ngroups,pvec,samsize1=numuncdraws,
                                samsize2=4e4,unique1=1:numcat,cov.prob=cov.prob)

  postestimates <- marglike_Hu[[4]]
  colnames(postestimates) <- iccnames
  cat("Second, exploratory testing of icc's...")
  cat("\n")
  BFtu_exploratory_icc <- t(matrix(unlist(lapply(1:numcat,function(nc){

    cat(paste0(iccnames[nc],"; "))

    marglike_explo <- rep(0,3)

    if(numcat>1){
      unique_c <- rep(1,numcat)
      unique_c[nc] <- 0
      unique_c[-nc] <- 1:(numcat-1)
    }else {
      unique_c <- 0
    }

    # zero icc
    marglike_explo[1] <- MargLikeICC_Hq(rhoML,zWstack,ngroups,pvec,unique1=unique_c)[[1]]

    # positive icc
    marglike_positive <- marglike_Hu[[1]] + log(marglike_Hu$postprobpositive[nc]) -
      log(marglike_Hu$priorprobpositive[nc])
    marglike_explo[3] <- marglike_positive
    # negative icc
    marglike_negative <- marglike_Hu[[1]] + log(1-marglike_Hu$postprobpositive[nc]) -
      log(1-marglike_Hu$priorprobpositive[nc])
    marglike_explo[2] <- marglike_negative

    return(marglike_explo - marglike_Hu[[1]])
  })),nrow=3))

  colnames(BFtu_exploratory_icc) <- c("icc=0","icc<0","icc>0")
  row.names(BFtu_exploratory_icc) <- iccnames
  rowmax <- apply(BFtu_exploratory_icc,1,max)
  norm_BF_explo <- exp(BFtu_exploratory_icc - rowmax %*% t(rep(1,ncol(BFtu_exploratory_icc)))) *
    (rep(1,nrow(BFtu_exploratory_icc)) %*% t(prior.hyp.explo[[1]]))
  PHP_exploratory_icc <- norm_BF_explo / apply(norm_BF_explo,1,sum)
  if(logIN == FALSE){
    BFtu_exploratory_icc <- exp(BFtu_exploratory_icc)
  }
  priorprobs <- rep(1,3)/3 #prior probs for exploratory tests
  cat("\n")
  cat("\n")
  if(!is.null(hypothesis)){ # confirmatory test with constrained hypotheses on icc's.
    cat("Third, confirmatory testing of icc's...")
    cat("\n")
    parse_hyp <- parse_hypothesis(iccnames,hypothesis)
    parse_hyp$hyp_mat <- do.call(rbind, parse_hyp$hyp_mat)
    RrList <- make_RrList2(parse_hyp)
    RrE <- RrList[[1]]
    RrO <- RrList[[2]]
    # check if icc's are only tested against each other or against zero
    numhyp <- length(RrE)
    for(h in 1:numhyp){
      if(!is.null(RrE[[h]])){
        for(r in 1:nrow(RrE[[h]])){
          row1 <- RrE[[h]][r,]
          if( !(sum(abs(row1))==1 || sum(row1)==0) ){
            stop("icc's can only be compared with each other or to zero.")
          }
        }
      }
      if(!is.null(RrO[[h]])){
        for(r in 1:nrow(RrO[[h]])){
          freq1 <- table(sort(RrO[[h]][r,]))
          row1 <- RrO[[h]][r,]
          if( !(sum(abs(row1))==1 || sum(row1)==0) ){
            stop("icc's can only be compared with each other or to zero.")
          }
        }
      }
    }

    output_marglike_icc <- t(matrix(unlist(lapply(1:numhyp, function(h){

      cat(paste0(parse_hyp$original_hypothesis[h],"; "))
      cat("\n")

      # get unconstrained prior draws, if needed for computing prior probabilities
      # in case of only order constraints
      pcat <- rep(1:length(ngroups),times=ngroups)
      unique1 <- 1:numcat
      LB <- unlist(unlist(lapply(1:numcat,function(c){
        -1/(max(pvec[pcat==c])-1)
      })))
      priordraws <- matrix(unlist(lapply(1:numcat,function(c){
        rbeta(numuncdraws,1,1) * (1 - LB[c]) + LB[c]
      })),ncol=numcat)

      # code equal icc's with same integer for marglike2_Hq function
      unique_h <- 1:numcat
      if(!is.null(RrE[[h]])){
        unique_h <- rep(NA,numcat)
        zeroMat <- RrE[[h]][which(apply(RrE[[h]],1,sum)==1),]
        if(!is.matrix(zeroMat)){
          zeroMat <- matrix(zeroMat,nrow=1)
        }
        unique_h[which(apply(zeroMat,2,sum)==1)] <- 0
        teller <- 0
        for(row1 in which(apply(RrE[[h]],1,sum)==0)){
          welk1 <- which(RrE[[h]][row1,]!=0)
          isna_h <- is.na(unique_h[welk1])
          if(sum(isna_h)==2){
            teller <- teller + 1
            unique_h[welk1] <- teller
          }else{ #one is already assigned a unique code
            unique_h[welk1] <- unique_h[welk1[!isna_h]]
            #
          }

        }
        if(sum(is.na(unique_h))>0){ #unconstrained icc's receive unique code
          unique_h[is.na(unique_h)] <- teller + 1:sum(is.na(unique_h))
          teller <- teller + sum(is.na(unique_h))
        }
      }
      if(!is.null(RrO[[h]])){
        unicum <- unique(unique_h[unique_h!=0])
        inequalities_h <- matrix(0,nrow=nrow(RrO[[h]]),ncol=max(unicum)+1)
        for(u in sort(unicum)){

          welk <- which(unique_h == u)
          if(length(welk) > 1){
            if(nrow(RrO[[h]]) > 1){
              inequalities_h[,u] <- apply(RrO[[h]][,which(unique_h == u)],1,sum)
            }else{
              inequalities_h[,u] <- apply(t(RrO[[h]][,which(unique_h == u)]),1,sum)
            }
          }else{ #length is 1
            inequalities_h[,u] <- RrO[[h]][,which(unique_h == u)]
          }
        }
      } else inequalities_h = 0
      if(is.null(RrE[[h]])){ #only order constraints; use output from unconstrained analysis
        priorprob_h <- mean(apply(cbind(priordraws,rep(-1,length(numuncdraws)))%*%t(inequalities_h)>0,1,prod))
        postprob_h <- mean(apply(cbind(marglike_Hu$postdraws,rep(-1,length(numuncdraws)))%*%
                                   t(inequalities_h)>0,1,prod))
        marglike2_h <- list(marglike_Hu[[1]],postprob_h,priorprob_h)
      }else{
        marglike2_h <- MargLikeICC_Hq(rhoML,zW=zWstack,ngroups,pvec,unique1=unique_h,
                                      inequalities=inequalities_h)[1:3]
      }
      return(c(unlist(marglike2_h),ifelse(is.null(RrE[[h]]),1,0)))
    })),nrow=4))
    if(complement == TRUE){
      #compute BF for complement hypothesis
      if(sum(output_marglike_icc[,4])==0){ #the complement is equivalent to the unconstrained model
        output_marglike_icc <- rbind(output_marglike_icc,c(unlist(marglike_Hu)[1:3],1))
      } else { #the complement is the complement of the joint of the order hypotheses
        which_order <- which(output_marglike_icc[,4]==1)
        if(length(which_order)==1){
          probs <- output_marglike_icc[which_order,2:3]
          marglike_Hc <- marglike_Hu[[1]] #+ log(1-probs[1]) - log(1-probs[2])
          output_marglike_icc <- rbind(output_marglike_icc,c(marglike_Hc,1-probs[1],1-probs[2],1))
        }else {
          probs <- apply(output_marglike_icc[which_order,2:3],2,sum)
          marglike_Hc <- marglike_Hu[[1]] #+ log(1-probs[1]) - log(1-probs[2])
          output_marglike_icc <- rbind(output_marglike_icc,c(marglike_Hc,1-probs[1],1-probs[2],1))
        }
      }
      row.names(output_marglike_icc) <- c(parse_hyp$original_hypothesis,"complement")
    }else{
      row.names(output_marglike_icc) <- c(parse_hyp$original_hypothesis)
    }
    relcomp <- matrix(c(rep(NA,nrow(output_marglike_icc)),output_marglike_icc[,3]),ncol=2)
    relfit <- matrix(c(rep(NA,nrow(output_marglike_icc)),output_marglike_icc[,2]),ncol=2)
    #compute log marginal likelihood for H* without order constraints
    BF_E <- exp(output_marglike_icc[,1] - marglike_Hu[[1]])
    output_marglike_icc <- cbind(output_marglike_icc,output_marglike_icc[,1] + log(output_marglike_icc[,2]) -
                                   log(output_marglike_icc[,3]))
    BFtu_confirmatory_icc <- output_marglike_icc[,5]  - marglike_Hu[[1]]
    #compute BFmatrix and PHPs
    logBFmatrix <- matrix(rep(output_marglike_icc[,5],numhyp+complement),nrow=numhyp+complement) -
      matrix(rep(output_marglike_icc[,5],each=numhyp+complement),nrow=numhyp+complement)
    diag(logBFmatrix) <- 0
    if(complement == TRUE){
      row.names(logBFmatrix) <- colnames(logBFmatrix) <- c(parse_hyp$original_hypothesis,"complement")
    }else{
      row.names(logBFmatrix) <- colnames(logBFmatrix) <- c(parse_hyp$original_hypothesis)
    }
    BFmatrix_confirmatory_icc <- logBFmatrix
    BFta_confirmatory_icc <- exp(output_marglike_icc[,5] - max(output_marglike_icc[,5]))
    # Change prior probs in case of default setting
    if(is.null(prior.hyp)){
      priorprobs <- rep(1/length(BFtu_confirmatory_icc),length(BFtu_confirmatory_icc))
    }else{
      if(!is.numeric(prior.hyp) || length(prior.hyp)!=length(BFtu_confirmatory_icc)){
        warning(paste0("Argument 'prior.hyp' should be numeric and of length ",as.character(length(BFtu_confirmatory_icc)),". Equal prior probabilities are used."))
        priorprobs <- rep(1/length(BFtu_confirmatory_icc),length(BFtu_confirmatory_icc))
      }else{
        priorprobs <- prior.hyp
      }
    }
    PHP_confirmatory_icc <- priorprobs*BFta_confirmatory_icc / sum(priorprobs*BFta_confirmatory_icc)
    BFtable <- cbind(relcomp,relfit,BF_E,relfit[,2]/relcomp[,2],
                     BF_E*relfit[,2]/relcomp[,2],PHP_confirmatory_icc)
    row.names(BFtable) <- names(PHP_confirmatory_icc)
    colnames(BFtable) <- c("complex=","complex>","fit=","fit>","BF=","BF>","BF","PHP")
    hypotheses <- names(BFta_confirmatory_icc)

    if(logIN == FALSE){
      BFtu_confirmatory_icc <- exp(BFtu_confirmatory_icc)
      BFmatrix_confirmatory_icc <- exp(BFmatrix_confirmatory_icc)
    }

  }else{
    BFmatrix_confirmatory_icc <- PHP_confirmatory_icc <- BFtu_confirmatory_icc <- relfit <-
      relcomp <- hypotheses <- BFtable <- priorprobs <- NULL
  }
  BFlm_out <- list(
    BFtu_exploratory=BFtu_exploratory_icc,
    PHP_exploratory=PHP_exploratory_icc,
    BFtu_confirmatory=BFtu_confirmatory_icc,
    PHP_confirmatory=PHP_confirmatory_icc,
    BFmatrix_confirmatory=BFmatrix_confirmatory_icc,
    BFtable_confirmatory=BFtable,
    prior.hyp=priorprobs,
    hypotheses=hypotheses,
    estimates=postestimates,
    model=x,
    bayesfactor="Bayes factors based on uniform priors",
    parameter="intraclass correlations",
    log=logIN,
    call=match.call())

  class(BFlm_out) <- "BF"

  return(BFlm_out)

}

int_lhood <- function(rhoS,ngroups,pvec,N,K,Wmat1,zvec1,tWW2,tWz2,tzz2){
  #the log integrated likelihood

  diagDi <- c( (1-rep(rhoS,times=ngroups))/(1+(pvec-1)*rep(rhoS,times=ngroups)) )
  tWDiW1 <- t(Wmat1)%*%(diagDi*Wmat1)
  tWDiz1 <- t(Wmat1)%*%(diagDi*zvec1)
  tzDiz1 <- sum(zvec1**2*diagDi)
  if(K > 0){
    s2 <- tzDiz1 + tzz2 - c(t(tWDiz1+tWz2)%*%solve(tWDiW1+tWW2)%*%(tWDiz1+tWz2) )
    return(
      .5*sum(log(diagDi)) - .5*log(det(tWDiW1+tWW2)) - (N-K)/2*log(s2)
    )
  }else{
    s2 <- tzDiz1 + tzz2
    return(
      .5*sum(log(diagDi)) - (N-K)/2*log(s2)
    )
  }
}

#' @importFrom stats rnorm rbeta dbeta
MargLikeICC_Hq <- function(rhoML,zW,ngroups,pvec,samsize1=5e4,samsize2=5e4,
                           unique1,inequalities=0,complement=FALSE,cov.prob=.95){

  #E.g., for categories=5, unique1=c(0,2,0,1,1),inequalities=[1 -1 0],
  #the hypothesis equals, Hq:rho1=rho3=0, rho4=rho5>rho2
  #samsize1 sets the sample size for the number of draws from the proposal distribution for the importance
  #sample estimate of the marginal likelihood excluding the inequality constraints.
  #samsize2 sets the sample size for computing the probability that the inequality constraints hold,
  #and for sampling from the proposal distribution. to get the IS estimate.

  CrI_LB <- (1 - cov.prob)/2
  CrI_UB <- 1 - (1 - cov.prob)/2

  numcat <- length(ngroups)
  N <- sum(pvec)

  zvec <- zW[,1]
  Wmat <- as.matrix(zW[,-1])

  K <- ncol(Wmat)
  numgroups <- sum(ngroups)
  firstofcluster <- c(0,cumsum(pvec[1:(numgroups-1)]))+1
  restcluster <- (1:N)[-firstofcluster]
  zvec1 <- zvec[firstofcluster]
  Wmat1 <- as.matrix(Wmat[firstofcluster,])
  zvec2 <- zvec[restcluster]
  Wmat2 <- as.matrix(Wmat[restcluster,])
  tWW2 <- t(Wmat2)%*%Wmat2
  tWz2 <- t(Wmat2)%*%zvec2
  tzz2 <- sum(zvec2**2)

  if(sum(unique1)==0){ #null model with no random effects
    marglikeHstar <- int_lhood(rhoS=rep(0,length=numcat),ngroups,pvec,N,K,Wmat1,zvec1,tWW2,tWz2,tzz2) +
      lgamma((N-K)/2) - (N-K)/2*log(pi)
    postprobpositive <- priorprobpositive <- postestimates <- postdraws <- drawsMat <- NULL
    priorprob <- postprob <- 1

  }else{
    numHrho <- max(unique1)

    transMatrix <- matrix(0,nrow=numcat,ncol=numHrho)
    for(cc in 1:numHrho){
      transMatrix[which(unique1==cc),cc] <- 1
    }
    #minimal value for rho under H (without order constraints)
    pcat <- rep(1:length(ngroups),times=ngroups)
    LB <- unlist(unlist(lapply(1:numHrho,function(c){
      welk <- which(unique1==c)
      -1/(max(pvec[unlist(lapply(1:length(welk),function(j){
        which(pcat==welk[j])
      }))])-1)
    })))

    #initial value rho
    rhoH0ML <- unlist(lapply(1:numHrho,function(c){
      mean(rhoML[which(unique1==c)])
    }))
    rhoH <- rhoH0ML
    rhoS <- c(transMatrix%*%rhoH)
    RWsd <- rep(.1,numHrho)
    #initial unstandardized posterior (with uniform prior on rhoH)
    unpost <- int_lhood(rhoS,ngroups,pvec,N,K,Wmat1,zvec1,tWW2,tWz2,tzz2)

    #start sampling rho
    check1 <- 300 #check every 'check1' draws whether random walk sd needs to be increased/decreased
    acceptMat <- drawsMat <- matrix(0,nrow=samsize1,ncol=numHrho)
    #wer <- Sys.time()
    for(s in 1:samsize1){
      #sampling unique rho's under H
      for(j in 1:numHrho){
        #random walks
        rhoH_can <- rhoH
        rhoH_can[j] <- rnorm(1,mean=rhoH[j],sd=RWsd[j])
        if(rhoH_can[j]>LB[j] && rhoH_can[j]<1){
          #MH acceptance
          welk <- which(unique1==j)
          rhoS_can <- c(transMatrix%*%rhoH_can)
          unpost_can <- int_lhood(rhoS_can,ngroups,pvec,N,K,Wmat1,zvec1,tWW2,tWz2,tzz2)
          MHprob <- exp(unpost_can - unpost)
          if(runif(1) < MHprob){
            rhoH <- rhoH_can
            rhoS <- rhoS_can
            unpost <- unpost_can
            acceptMat[s,j] <- 1
          }
        }
      }
      drawsMat[s,] <- rhoH

      if(ceiling(s/check1)==s/check1){ #adjust sd of random walk based on acceptance proportions of last 'check1' draws
        probs <- apply(as.matrix(acceptMat[(s-check1+1):s,]),2,mean)
        upper1 <- .5
        lower1 <- .15
        RWsd[probs>upper1] <- RWsd[probs>upper1] * ( (probs[probs>upper1]-upper1)/(1-upper1) + 1)
        RWsd[probs<lower1] <- RWsd[probs<lower1] * 1/( 2 - (probs[probs<lower1])/lower1 )
      }
    }
    #Sys.time() - wer
    discard <- .2
    drawsMat <- as.matrix(drawsMat[(discard*samsize1+1):samsize1,]) #discard with 20% for burn-in
    samsize1 <- nrow(drawsMat)
    # get a tailored importance sampling estimate
    meanICC <- apply(drawsMat,2,mean)
    varICC <- apply(drawsMat,2,var)
    shape1IM <- - ((LB-meanICC)*((meanICC-1)*meanICC-meanICC*LB+LB+varICC)/((LB-1)*varICC))
    shape2IM <- - shape1IM*(meanICC-1)/(meanICC-LB)

    postestimates <- rbind(t(apply(drawsMat,2,mean)),
                           t(apply(drawsMat,2,median)),
                           apply(drawsMat,2,quantile,probs=c(CrI_LB,CrI_UB)))
    row.names(postestimates)[c(1,2)] <- c("mean","median")
    iccnames <- unlist(lapply(1:ncol(drawsMat),function(nc){paste0("icc",as.character(nc))}))
    colnames(postestimates) <- iccnames
    #compute posterior probs of positive icc of the free parameters
    postprobpositive <- apply(drawsMat>0,2,mean)
    priorprobpositive <- 1/(1 - LB)

    if(is.matrix(inequalities)){
      #compute prior probability of order constraints
      priordraws <- matrix(unlist(lapply(1:numHrho,function(cc){
        rbeta(samsize1,shape1=1,shape2=1)*(1-LB[cc])+LB[cc]
      })),ncol=numHrho)
      priorprob <- mean(apply(cbind(priordraws,rep(-1,samsize1))%*%t(inequalities) > 0,1,prod) )
      priorprob <- priorprob * (1-complement) + (1 - priorprob) * complement
      remove(priordraws)
      #compute posterior probability of order constraints
      postprob <- mean(apply(cbind(drawsMat,rep(-1,samsize1))%*%t(inequalities) > 0,1,prod) )
      postprob <- postprob * (1-complement) + (1 - postprob) * complement
    }else{
      priorprob <- postprob <- 1
    }
    #wer <- Sys.time()
    factor1 <- .6
    shape1IM <- shape1IM * factor1
    shape2IM <- shape2IM * factor1

    ISdraws <- matrix(unlist(lapply(1:numHrho,function(c){
      (rbeta(samsize2,shape1IM[c],shape2IM[c]) * (1 - LB[c]) + LB[c] ) * .99999
    })),ncol=numHrho)

    logintegrands <- unlist(lapply(1:samsize2,function(s){
      int_lhood(rhoS=c(transMatrix%*%ISdraws[s,]),ngroups,pvec,N,K,Wmat1,zvec1,tWW2,tWz2,tzz2) +
        sum(log(1/(1-LB))) -
        sum(dbeta((ISdraws[s,]-LB)/(1-LB),shape1=shape1IM,shape2=shape2IM,log=TRUE) +
              log(1/(1-LB)) )
    }))
    #marginal likelihood for H*, which excludes the order constraints probabilities
    marglikeHstar <- log(mean(exp(logintegrands - max(logintegrands)))) +
      max(logintegrands) + lgamma((N-K)/2) - (N-K)/2*log(pi)
  }
  return(list(marglike=marglikeHstar,postprob=postprob,priorprob=priorprob,
              postestimates=postestimates,postprobpositive=postprobpositive,
              priorprobpositive=priorprobpositive,postdraws=drawsMat))
}

Helmert = function(p){
  Helm <- diag(p)
  Helm[1,] <- 1/sqrt(p)
  for(pp in 2:p){
    Helm[pp,1:(pp-1)] <- 1/sqrt(pp*(pp-1))
    Helm[pp,pp] <- -(pp-1)/sqrt(pp*(pp-1))
  }
  return(Helm)
}

#' @method get_estimates hetcor
#' @import bain
#' @export
get_estimates.hetcor <- function(x, ...){
  cl <- match.call()
  cl[[1]] <- as.name("get_estimates")
  cl[["x"]] <- x$correlations
  P <- nrow(x$std.errors)
  out <- eval.parent(cl)
  retain <- matrix(1:length(out$estimate), nrow = nrow(x$std.errors))
  out$estimate <- out$estimate[retain[lower.tri(retain)]]
  errcov <- x$std.errors**2
  errcov <- errcov[lower.tri(retain)]
  if(length(errcov) == 1){
    out$Sigma <- list(matrix(errcov))
  } else {
    out$Sigma <- list(diag(errcov))
  }
  class(out) <- "model_estimates"
  attr(out, "analysisType") <- "hetcor"
  out
}


#' @method get_estimates coxph
#' @export
get_estimates.coxph <- function(x, ...){
  out <- list()
  out$estimate <- coef(x)
  out$Sigma <- list(vcov(x))
  class(out) <- "model_estimates"
  attr(out, "analysisType") <- "cophx"
  out
}


#' @method get_estimates glm
#' @export
get_estimates.glm <- function(x, ...){
  out <- list()
  out$estimate <- coef(x)
  out$Sigma <- list(vcov(x))
  class(out) <- "model_estimates"
  attr(out, "analysisType") <- "glm"
  out
}

#' @method get_estimates polr
#' @export
get_estimates.polr <- function(x, ...){
  out <- list()
  out$estimate <- c(coef(x),x$zeta)
  out$Sigma <- list(vcov(x))
  class(out) <- "model_estimates"
  attr(out, "analysisType") <- "polr"
  out
}


#' @method get_estimates bartlett_htest
#' @export
get_estimates.bartlett_htest <- function(x, ...){
  out <- list()
  out$estimate <- x$vars
  out$Sigma <- NULL
  class(out) <- "model_estimates"
  attr(out, "analysisType") <- "bartlett_htest"
  out
}


#' @method get_estimates survreg
#' @export
get_estimates.survreg <- function(x, ...){
  out <- list()
  out$estimate <- x$coefficients
  out$Sigma <- list(x$var)
  class(out) <- "model_estimates"
  attr(out, "analysisType") <- "survreg"
  out
}


#' @method get_estimates zeroinfl
#' @export
get_estimates.zeroinfl <- function(x, ...){
  out <- list()
  out$estimate <- c(coef(x),x$zeta)
  out$Sigma <- list(vcov(x))
  class(out) <- "model_estimates"
  attr(out, "analysisType") <- "zeroinfl"
  out
}


#' @method get_estimates lm
#' @export
get_estimates.lm <- function(x, ...){
  out <- list()
  P <- ncol(x$coefficients)
  K <- nrow(x$coefficients)
  N <- nrow(x$residuals)
  if(!is.matrix(x$coefficients)){
    if(sum(names(x$coefficients) == "(Intercept)")>0){
      names(x$coefficients)[which(names(x$coefficients) == "(Intercept)")] <- "Intercept"
    }
    out$estimate <- coef(x)
    out$Sigma <- list(vcov(x))
    class(out) <- "model_estimates"
    attr(out, "analysisType") <- "lm"
    out
  }else{
    names_coef1 <- row.names(x$coefficients)
    if(sum(names_coef1 == "(Intercept)")>0){
      names_coef1[which(names_coef1 == "(Intercept)")] <- "Intercept"
    }
    names_coef2 <- colnames(x$coefficients)
    names_coef <- unlist(lapply(1:P,function(p){
      lapply(1:K,function(k){
        paste0(names_coef1[k],"_on_",names_coef2[p])
      })
    }))
    # estimates of regression coefficients
    estimatesBeta <- c(x$coefficients)
    names(estimatesBeta) <- names_coef
    Xmat <- model.matrix(x)
    Ymat <- model.matrix(x)%*%x$coefficients + x$residuals
    SigmaEst <- t(x$residuals)%*%x$residuals/N
    covmBeta <- kronecker(SigmaEst,solve(t(Xmat)%*%Xmat))
    row.names(covmBeta) <- colnames(covmBeta) <- names_coef

    out$estimate <- estimatesBeta
    out$Sigma <- list(covmBeta)
    class(out) <- "model_estimates"
    attr(out, "analysisType") <- "mlm"
    out
  }
}


#' @method get_estimates cor_test
#' @export
get_estimates.cor_test <- function(x, ...){
  out <- list()
  out$estimate <- x$meanF
  out$Sigma <- list(x$covmF)
  class(out) <- "model_estimates"
  attr(out, "analysisType") <- "corr_htest"
  out
}


#' @method get_estimates t_test
#' @export
get_estimates.t_test <- function(x, ...){
  out <- list()
  if(length(x$estimate)>1){
    difference <- x$estimate[1] - x$estimate[2]
    names(difference) <- "difference"
    out$estimate <- difference
    out$Sigma <- list((x$stderr)**2)
  }else if(names(x$estimate) == "mean difference"){
    difference <- x$estimate
    names(difference) <- "difference"
    out$estimate <- difference
    out$Sigma <- list((x$stderr)**2)
  }else{
    populationmean <- x$estimate
    names(populationmean) <- "mu"
    out$estimate <- populationmean
    out$Sigma <- list((x$stderr)**2)
  }
  class(out) <- "model_estimates"
  attr(out, "analysisType") <- "t_test"
  out
}

### Joris Mulder 2019. Bayes factor for a one sample Student t test
### via adjusted FBFs (Mulder, 2014) using (m)lm-objects using a t.test object.

#' @method BF htest
#' @export
BF.htest <-
  function(x,
           hypothesis = NULL,
           prior.hyp.explo = NULL,
           prior.hyp.conf = NULL,
           prior.hyp = NULL,
           ...) {
    stop("Please use the function t_test() from the 'bain' package for a t-test or bartlett_test() from the 'BFpack' package for a test on group variances.")
  }



#' @importFrom stats approxfun
#' @describeIn BF BF S3 method for an object of class 't_test'
#' @method BF t_test
#' @export
BF.t_test <- function(x,
                      hypothesis = NULL,
                      prior.hyp.explo = NULL,
                      prior.hyp.conf = NULL,
                      prior.hyp = NULL,
                      complement = TRUE,
                      log = FALSE,
                      cov.prob = .95,
                      BF.type = NULL,
                      iter = 1e6,
                      ...){

  if(!(cov.prob>0 & cov.prob<1)){
    stop("The argument 'cov.prob' is a coverage probability for the interval estimates that should lie between 0 and 1. The default is 0.95.")
  }
  CrI_LB <- (1 - cov.prob)/2
  CrI_UB <- 1 - (1 - cov.prob)/2

  numpop <- length(x$estimate)

  if(is.null(BF.type)){
    BF.type <- "FBF"
  }
  if(!is.null(BF.type)){
    if(is.na(BF.type) | (BF.type!="FBF" & BF.type!="AFBF"))
      stop("The argument 'BF.type' must be 'FBF' (for the fractional BF) or 'AFBF' (for the adjusted fractional BF).")
  }
  if(BF.type=='AFBF'){
    bayesfactor <- "generalized adjusted fractional Bayes factors"
  }else{
    bayesfactor <- "generalized fractional Bayes factors"
  }

  logIN <- log

  if(numpop==1){ #one sample t test

    tvalue <- x$statistic
    mu0 <- x$null.value
    xbar <- x$estimate
    df <- x$parameter
    n <- df + 1
    stderr <- (xbar - mu0) / tvalue #standard error
    sigmaML <- stderr*sqrt(n-1)

    sampledata <- rnorm(n)
    sampledata <- (sampledata - mean(sampledata)) / sd(sampledata)
    sampledata <- sampledata * sqrt(x$v[1]) + xbar
    sampledata_explo <- sampledata - mu0
    if(x$method == "One Sample t-test"){
      mu <- rep(1,length(sampledata))
      df.draws <- data.frame(obs=sampledata,obs_explo=sampledata_explo,mu=mu)
      lm_conf <- lm(obs~mu-1,data=df.draws)
      sampledata_explo <- sampledata - x$null.value
      lm_explo <- lm(obs_explo~mu-1,data=df.draws)
    }else{
      difference <- rep(1,length(sampledata))
      df.draws <- data.frame(obs=sampledata,obs_explo=sampledata_explo,difference=difference)
      lm_conf <- lm(obs~difference-1,data=df.draws)
      sampledata_explo <- sampledata - x$null.value
      lm_explo <- lm(obs_explo~difference-1,data=df.draws)
    }
    BF_conf <- BF(lm_conf,
                  hypothesis=hypothesis,
                  prior.hyp.explo = prior.hyp.explo,
                  prior.hyp.conf = prior.hyp.conf,
                  prior.hyp=prior.hyp,
                  complement=complement,
                  BF.type=BF.type,
                  log=logIN,
                  cov.prob=cov.prob)
    BF_explo <- BF(lm_explo,
                   hypothesis=NULL,
                   prior.hyp.explo = prior.hyp.explo,
                   prior.hyp.conf = prior.hyp.conf,
                   prior.hyp=prior.hyp,
                   complement=complement,
                   BF.type=BF.type,
                   log=logIN,
                   cov.prob=cov.prob)
    BF_out <- BF_conf
    BF_out$BFtu_exploratory <- BF_explo$BFtu_exploratory
    BF_out$PHP_exploratory <- BF_explo$PHP_exploratory
    colnames(BF_out$PHP_exploratory) <-
      c(paste0("Pr(=",as.character(mu0),")"),paste0("Pr(<",as.character(mu0),")"),
        paste0("Pr(>",as.character(mu0),")"))
    colnames(BF_out$BFtu_exploratory) <- c("BF0u","BF1u","BF2u")

    estimates_ttest <- BF_explo$estimates

  }else{ # two samples t test

    if(!grepl("Welch",x$method)){ # equal variances assumed
      # compute via a lm-analysis
      x1 <- c(rep(1,x$n[1]),rep(0,x$n[2]))
      y1 <- c(rep(0,x$n[1]),rep(1,x$n[2]))
      matx1y1 <- cbind(x1,y1)
      draw1 <- rnorm(x$n[1])
      out1 <- (draw1 - mean(draw1))/sd(draw1)*sqrt(x$v[1])+x$estimate[1]
      draw2 <- rnorm(x$n[2])
      out2 <- (draw2 - mean(draw2))/sd(draw2)*sqrt(x$v[2])+x$estimate[2]
      out <- c(out2,out1)
      # perform the test via a lm object using a factor for the group indicator
      # such that the name of the key variable (the difference between the means)
      # is called 'difference'
      df1 <- data.frame(out=out,differenc=factor(c(rep("a",x$n[2]),rep("e",x$n[1]))))
      lm1 <- lm(out ~ differenc,df1)
      BF_out <- BF(lm1,
                   hypothesis=hypothesis,
                   prior.hyp.explo=prior.hyp.explo,
                   prior.hyp.conf=prior.hyp.conf,
                   prior.hyp=prior.hyp,
                   complement=complement,
                   BF.type=BF.type,
                   log=logIN,
                   cov.prob=cov.prob)
      BF_out$BFtu_exploratory <- t(as.matrix(BF_out$BFtu_exploratory[2,]))
      BF_out$PHP_exploratory <- t(as.matrix(BF_out$PHP_exploratory[2,]))
      row.names(BF_out$BFtu_exploratory) <- row.names(BF_out$PHP_exploratory) <- "difference"

      BF_out$estimates <- t(BF_out$estimates[2,])
      row.names(BF_out$estimates) <- "difference"

    }else{ #equal variances not assumed. BF.lm cannot be used

      # check proper usage of argument 'prior.hyp.conf' and 'prior.hyp.explo'
      if(!is.null(prior.hyp.conf)){
        prior.hyp <- prior.hyp.conf
      }
      prior.hyp.explo <- process.prior.hyp.explo(prior_hyp_explo = prior.hyp.explo, model=x)

      diff.obs <- x$estimate[1] - x$estimate[2]
      nvec <- x$n
      bvec <- 2/nvec
      s2vec <- (x$v)*(x$n-1)
      #draws for Monte Carlo estimates
      sigma2_1.draws <- rgamma(iter,shape=(nvec[1]-1)/2,rate=s2vec[1]/2)
      sigma2b_1.draws <- rgamma(iter,shape=(nvec[1]*bvec[1]-1)/2,rate=s2vec[1]*bvec[1]/2)
      sigma2_2.draws <- rgamma(iter,shape=(nvec[2]-1)/2,rate=s2vec[2]/2)
      sigma2b_2.draws <- rgamma(iter,shape=(nvec[2]*bvec[2]-1)/2,rate=s2vec[2]*bvec[2]/2)
      relfit0 <- log(mean(dnorm(x$null.value,
                                mean=diff.obs,
                                sd=sqrt(1/sigma2_1.draws*1/nvec[1] + 1/sigma2_2.draws*1/nvec[2]))))
      relfit1 <- log(mean(pnorm(x$null.value,
                                mean=diff.obs,
                                sd=sqrt(1/sigma2_1.draws*1/nvec[1] + 1/sigma2_2.draws*1/nvec[2]))))

      diff.draws <- rnorm(iter,mean=diff.obs,sd=sqrt(1/sigma2_1.draws*1/nvec[1] + 1/sigma2_2.draws*1/nvec[2]))
      estimates_ttest <- t(c(mean(diff.draws),median(diff.draws),quantile(diff.draws,CrI_LB),
                         quantile(diff.draws,CrI_UB)))
      row.names(estimates_ttest) <- "difference"
      colnames(estimates_ttest) <- c("mean","median",paste0(as.character(round(CrI_LB*100,7)),"%"),
                                    paste0(as.character(round(CrI_UB*100,7)),"%"))

      relfit2 <- log(1 - exp(relfit1))
      if(BF.type == 'AFBF'){
        prior.mean <- x$null.value
      }else{
        prior.mean <- diff.obs
      }
      relcomp0 <- log(mean(dnorm(x$null.value,
                                 mean=prior.mean,
                                 sd=sqrt(1/sigma2b_1.draws*1/(nvec[1]*bvec[1]) + 1/sigma2b_2.draws*1/(nvec[2]*bvec[2])))))
      relcomp1 <- log(mean(pnorm(x$null.value,
                                 mean=prior.mean,
                                 sd=sqrt(1/sigma2b_1.draws*1/(nvec[1]*bvec[1]) + 1/sigma2b_2.draws*1/(nvec[2]*bvec[2])))))
      relcomp2 <- log(1 - exp(relcomp1))

      #exploratory Bayes factor test
      hypotheses_exploratory <- c("difference=0","difference<0","difference>0")
      logBFtu_exploratory <- c(relfit0-relcomp0,relfit1-relcomp1,relfit2-relcomp2)
      names(logBFtu_exploratory) <- hypotheses_exploratory
      BFtu_exploratory <- logBFtu_exploratory
      maxBFtu <- max(BFtu_exploratory)
      BFtu_explo_norm <- exp(BFtu_exploratory - maxBFtu) * prior.hyp.explo[[1]]
      PHP_exploratory <- matrix(BFtu_explo_norm/sum(BFtu_explo_norm),nrow=1)
      colnames(PHP_exploratory) <- c("Pr(=0)","Pr(<0)","Pr(>0)")
      row.names(PHP_exploratory) <- "difference"

      if(!is.null(hypothesis)){
        name1 <- "difference"
        parse_hyp <- parse_hypothesis(name1,hypothesis)
        parse_hyp$hyp_mat <- do.call(rbind, parse_hyp$hyp_mat)
        RrList <- make_RrList2(parse_hyp)
        RrE <- RrList[[1]]
        RrO <- RrList[[2]]

        if(length(RrE)==1){
          RrStack <- rbind(RrE[[1]],RrO[[1]])
          RrStack <- interval_RrStack(RrStack)
        }else{
          RrStack_list <- lapply(1:length(RrE),function(h){
            interval_RrStack(rbind(RrE[[h]],RrO[[h]]))
          })
          RrStack <- do.call(rbind,RrStack_list)
        }
        if(nrow(RrStack)>1){
          RStack <- RrStack[,-2]
          rStack <- RrStack[,2]
        }else{
          RStack <- matrix(RrStack[,-2],nrow=1)
          rStack <- RrStack[,2]
        }

        if(BF.type=='AFBF'){
          # check if a common boundary exists for prior location under all constrained hypotheses
          # necessary for the adjusted FBF
          if(nrow(RrStack) > 1){
            rref_ei <- rref(RrStack)
            nonzero <- rref_ei[,2]!=0
            if(max(nonzero)>0){
              row1 <- max(which(nonzero))
              if(sum(abs(rref_ei[row1,1]))==0){
                stop("No common boundary point for prior location. Conflicting constraints.")
              }
            }
          }
          prior.mean <- c(ginv(RStack)%*%rStack)
        }else{
          prior.mean <- diff.obs
        }

        relfit <- t(matrix(unlist(lapply(1:length(RrE),function(h){
          if(!is.null(RrE[[h]]) & is.null(RrO[[h]])){ #only an equality constraint
            nullvalue <- RrE[[h]][1,2]/RrE[[h]][1,1]
            relfit_h <- c(log(mean(dnorm(nullvalue,mean=diff.obs,
                                      sd=sqrt(1/sigma2_1.draws*1/nvec[1] + 1/sigma2_2.draws*1/nvec[2])))),
                          log(1))
            # relfit_h <- c(log(densN(nullvalue)),0)
          }else if(is.null(RrE[[h]]) & !is.null(RrO[[h]])){
            which.LB <- which(RrO[[h]][,1]>0)
            if(length(which.LB) == 0){
              LB <- -Inf
            }else{
              LB <- max(RrO[[h]][which.LB,2] / RrO[[h]][which.LB,1])
            }
            which.UB <- which(RrO[[h]][,1]<0)
            if(length(which.UB) == 0){
              UB <- Inf
            }else{
              UB <- min(RrO[[h]][which.UB,2] / RrO[[h]][which.UB,1])
            }
            relfit_h <- c(log(1),
                          log(mean(pnorm(UB,mean=diff.obs,
                                         sd=sqrt(1/sigma2_1.draws*1/nvec[1] + 1/sigma2_2.draws*1/nvec[2])) -
                                     pnorm(LB,mean=diff.obs,
                                           sd=sqrt(1/sigma2_1.draws*1/nvec[1] + 1/sigma2_2.draws*1/nvec[2]))
                                   )
                              )
                          )
          }else stop("hypothesis should either contain one equality constraint or inequality constraints on 'difference'.")
          return(relfit_h)
        })),nrow=2))
        #relfit <- exp(relfit)
        relcomp <- t(matrix(unlist(lapply(1:length(RrE),function(h){
          if(!is.null(RrE[[h]]) & is.null(RrO[[h]])){ #only an equality constraint
            nullvalue <- RrE[[h]][1,2]/RrE[[h]][1,1]
            relcomp_h <- c(log(mean(dnorm(nullvalue,mean=prior.mean,
                                          sd=sqrt(1/sigma2b_1.draws*1/(nvec[1]*bvec[1]) + 1/sigma2b_2.draws*1/(nvec[1]*bvec[1]))))),
                           log(1))
          }else if(is.null(RrE[[h]]) & !is.null(RrO[[h]])){ #order constraint(s)
            which.LB <- which(RrO[[h]][,1]>0)
            if(length(which.LB) == 0){
              LB <- -Inf
            }else{
              LB <- max(RrO[[h]][which.LB,2] / RrO[[h]][which.LB,1])
            }
            which.UB <- which(RrO[[h]][,1]<0)
            if(length(which.UB) == 0){
              UB <- Inf
            }else{
              UB <- min(RrO[[h]][which.UB,2] / RrO[[h]][which.UB,1])
            }
            relcomp_h <- c(log(1),
                          log(mean(pnorm(UB,mean=prior.mean,
                                         sd=sqrt(1/sigma2b_1.draws*1/(nvec[1]*bvec[1]) + 1/sigma2b_2.draws*1/(nvec[1]*bvec[1]))) -
                                     pnorm(LB,mean=prior.mean,
                                           sd=sqrt(1/sigma2b_1.draws*1/(nvec[1]*bvec[1]) + 1/sigma2b_2.draws*1/(nvec[1]*bvec[1])))
                                   )
                              )
                          )
          }else stop("hypothesis should either contain one equality constraint or inequality constraints on 'difference'.")
          return(relcomp_h)
        })),nrow=2))
        #relcomp <- exp(relcomp)
        row.names(relfit) <- row.names(relcomp) <- parse_hyp$original_hypothesis
        colnames(relfit) <- c("f=","f>")
        colnames(relcomp) <- c("c=","c>")
        active.complement <- TRUE
        if(complement == TRUE){
          #add complement to analysis
          if(sum(relcomp[,1]==0)>0){ #then there are order hypotheses
            # check if one-sided hypotheses are (partly) overlapping
            check.draws <- c(rnorm(iter,sd=1),rnorm(iter,sd=100))
            which.O <- which(unlist(lapply(RrO,function(x) {!is.null(x)})))
            checks <- Reduce("+",lapply(which.O,function(h){
              apply(as.matrix(RrO[[h]][,1]) %*% t(check.draws) > RrO[[h]][,2] %*%
                      t(rep(1,length(check.draws))),2,prod)
            }))
            if(max(checks)>1){ #(partly) overlapping
              warning("Complement hypothesis omitted because of (partly) overlapping hypotheses")
            }else{
              if(sum(exp(relcomp[which.O,2]))<1){#complement nonempty
                relcomp_c <- log(1-sum(exp(relcomp[which.O,2])))
                relcomp <- rbind(relcomp,c(0,relcomp_c))
                relfit_c <- log(1-sum(exp(relfit[which.O,2])))
                relfit <- rbind(relfit,c(0,relfit_c))
                row.names(relfit) <- row.names(relcomp) <- c(parse_hyp$original_hypothesis,"complement")
              }else{
                active.complement <- FALSE
              }
            }
          }else{ #no order constraints
            relcomp <- rbind(relcomp,c(0,0))
            relfit <- rbind(relfit,c(0,0))
            row.names(relcomp) <- row.names(relfit) <- c(parse_hyp$original_hypothesis,"complement")
          }
        }
        # Check input of prior probabilies
        if(is.null(prior.hyp)){
          priorprobs <- rep(1/nrow(relcomp),nrow(relcomp))
        }else{
          if(!is.numeric(prior.hyp) || length(prior.hyp)!=nrow(relcomp)){
            if(complement == TRUE && active.complement == FALSE && length(prior.hyp)==nrow(relcomp)+1){
              # the prior prob for the complement hypothesis is given but it is not used as the constrained hypotheses
              # cover the entire parameter space
              priorprobs <- prior.hyp[1:nrow(relcomp)]
              priorprobs <- priorprobs / sum(priorprobs)
            }else{
              warning(paste0("Argument 'prior.hyp' should be numeric and of length ",as.character(nrow(relcomp)),". Equal prior probabilities are used."))
              priorprobs <- rep(1/nrow(relcomp),nrow(relcomp))
            }
          }else{
            priorprobs <- prior.hyp
          }
        }

        #compute Bayes factors and posterior probabilities for confirmatory test
        BFtu_confirmatory <- c(apply(relfit - relcomp, 1, sum))
        maxBFtu <- max(BFtu_confirmatory)
        PHP_confirmatory <- exp(BFtu_confirmatory-maxBFtu)*priorprobs / sum(exp(BFtu_confirmatory-maxBFtu)*priorprobs)
        BFmatrix_confirmatory <- matrix(rep(BFtu_confirmatory,length(BFtu_confirmatory)),ncol=length(BFtu_confirmatory)) -
          t(matrix(rep(BFtu_confirmatory,length(BFtu_confirmatory)),ncol=length(BFtu_confirmatory)))
        diag(BFmatrix_confirmatory) <- log(1)
        row.names(BFmatrix_confirmatory) <- colnames(BFmatrix_confirmatory) <- names(BFtu_confirmatory)
        relative_fit <- relfit
        relative_complexity <- relcomp

        BFtable <- cbind(relative_complexity,
                         relative_fit,
                         relative_fit[,1]-relative_complexity[,1],
                         relative_fit[,2]-relative_complexity[,2],
                         apply(relfit - relcomp, 1, sum),
                         PHP_confirmatory)
        BFtable[,1:7] <- exp(BFtable[,1:7])
        row.names(BFtable) <- names(BFtu_confirmatory)
        colnames(BFtable) <- c("complex=","complex>","fit=","fit>","BF=","BF>","BF","PHP")
        hypotheses <- row.names(relative_complexity)

      }

      if(logIN == FALSE){
        BFtu_exploratory <- exp(BFtu_exploratory)
        if(!is.null(hypothesis)){
          BFtu_confirmatory <- exp(BFtu_confirmatory)
          BFmatrix_confirmatory <- exp(BFmatrix_confirmatory)
        }
      }

      if(is.null(hypothesis)){
        BFtu_confirmatory <- PHP_confirmatory <- BFmatrix_confirmatory <- relative_fit <-
          relative_complexity <- BFtable <- hypotheses <- priorprobs <- NULL
      }

      rm(sigma2_1.draws);rm(sigma2_2.draws);rm(sigma2b_1.draws);rm(sigma2b_2.draws)
      parameter <- "mean difference"

      BF_out <- list(
        BFtu_exploratory=BFtu_exploratory,
        PHP_exploratory=PHP_exploratory,
        BFtu_confirmatory=BFtu_confirmatory,
        PHP_confirmatory=PHP_confirmatory,
        BFmatrix_confirmatory=BFmatrix_confirmatory,
        BFtable_confirmatory=BFtable,
        prior.hyp.explo=prior.hyp.explo,
        prior.hyp.conf=priorprobs,
        hypotheses=hypotheses,
        estimates=estimates_ttest,
        model=x,
        bayesfactor=bayesfactor,
        parameter=parameter,
        log = logIN,
        fraction_number_groupIDs=2,
        fraction_groupID_observation=rep(1:2,times=nvec),
        call=match.call())
      class(BF_out) <- "BF"
    }
  }

  return(BF_out)
}

#' BFpack: Flexible Bayes factor testing of scientific expectations
#'
#'
#' @description The \code{R} package \strong{BFpack} provides tools for exploratory and
#' confirmatory Bayesian hypothesis testing using Bayes factors and posterior probabilities
#' under common statistical models. The main function `BF` needs a fitted model `x` as input
#' argument. Depending on the class of the fitted model, a standard hypothesis test is
#' executed by default. For example, if `x` is a
#' fitted regression model of class `lm` then posterior probabilities are computed of whether
#' each separate coefficient is zero, negative, or positive (assuming equal prior probabilities).
#' If one has specific hypotheses with
#' equality and/or order constraints on the parameters under the fitted model `x` then these
#' can be formulated using the `hypothesis` argument (a character string), possibly together
#' prior probabilities for the hypotheses via the `prior` argument (default all hypotheses are
#' equally likely a priori), and the `complement` argument which is a logical stating whether
#' the complement hypotheses should be included in the case (`TRUE` by default).
#'
#' Use compilation for Fortran functions
#'
#' @references
#' Mulder, J., D.R. Williams, Gu, X., A. Tomarken,
#' F. Böing-Messing, J.A.O.C. Olsson-Collentine, Marlyne Meyerink, J. Menke,
#' J.-P. Fox, Y. Rosseel, E.J. Wagenmakers, H. Hoijtink., and van Lissa, C.
#' (submitted). BFpack: Flexible Bayes Factor Testing of Scientific Theories
#' in R. \url{https://arxiv.org/abs/1911.07728}
#'
#' Mulder, J., van Lissa, C., Gu, X., Olsson-Collentine, A., Boeing-Messing, F., Williams,
#' D. R., Fox, J.-P., Menke, J., et al. (2019). BFpack: Flexible Bayes Factor Testing of
#' Scientific Expectations. (Version 0.2.1) \url{https://CRAN.R-project.org/package=BFpack}
#'
#'
#' @examples
#' \dontrun{
#' # EXAMPLE 1. One-sample t test
#' ttest1 <- t_test(therapeutic, mu = 5)
#' print(ttest1)
#' # confirmatory Bayesian one sample t test
#' BF1 <- BF(ttest1, hypothesis = "mu = 5")
#' summary(BF1)
#' # exploratory Bayesian one sample t test
#' BF(ttest1)
#'
#' # EXAMPLE 2. ANOVA
#' aov1 <- aov(price ~ anchor * motivation,data = tvprices)
#' BF1 <- BF(aov1, hypothesis = "anchorrounded = motivationlow;
#'                               anchorrounded < motivationlow")
#' summary(BF1)
#'
#' # EXAMPLE 3. Logistic regression
#' fit <- glm(sent ~ ztrust + zfWHR + zAfro + glasses + attract + maturity +
#'    tattoos, family = binomial(), data = wilson)
#' BF1 <- BF(fit, hypothesis = "ztrust > zfWHR > 0;
#'                              ztrust > 0 & zfWHR = 0")
#' summary(BF1)
#'
#' # EXAMPLE 4. Correlation analysis
#' set.seed(123)
#' cor1 <- cor_test(memory[1:20,1:3])
#' BF1 <- BF(cor1)
#' summary(BF1)
#' BF2 <- BF(cor1, hypothesis = "Wmn_with_Im > Wmn_with_Del > 0;
#'                               Wmn_with_Im = Wmn_with_Del = 0")
#' summary(BF2)
#' }
#'
#' @docType package
#'
#'
#' @name BFpack-package
#'
"_PACKAGE"

#BF method for survreg classes


#' @method BF survreg
#' @export
BF.survreg <- function(x,
                   hypothesis = NULL,
                   prior.hyp.explo = NULL,
                   prior.hyp.conf = NULL,
                   prior.hyp = NULL,
                   complement = TRUE,
                   log = FALSE,
                   cov.prob = .95,
                   ...){

  #Extract summary statistics
  get_est <- get_estimates(x)
  Args <- as.list(match.call()[-1])
  get_est <- get_estimates(x)
  Args$x <- get_est$estimate
  Args$Sigma <- get_est$Sigma[[1]]
  Args$n <- length(x$y)
  Args$hypothesis <- hypothesis
  Args$prior.hyp <- prior.hyp
  Args$prior.hyp.explo <- prior.hyp.explo
  Args$prior.hyp.conf <- prior.hyp.conf
  Args$complement <- complement
  Args$log <- log
  Args$cov.prob <- cov.prob
  out <- do.call(BF, Args)
  out$model <- x
  out$call <- match.call()
  out
}

#' @method print cor_test
#' @export
print.cor_test <- function(x,
                           digits = 3, ...){

  estimates <- x$correstimates
  names <- x$corrnames
  groups <- length(names)
  P <- nrow(names[[1]])
  numcorr <- P*(P-1)/2
  countg = 0
  corrlist <- lapply(1:groups,function(g){
    lapply(1:3,function(b){
      matje <- matrix(NA,P,P)
      row.names(matje) <- colnames(matje) <- x$variables[[1]]
      matje[lower.tri(diag(P))] <- estimates[numcorr*(g-1)+1:numcorr,1+b]
      matje
    })
  })

  cat("\n")
  if(x$prior.cor=="joint.unif"){
    cat("Unconstrained Bayesian estimation (joint uniform prior)","\n", sep = "")
  }else{
    cat("Unconstrained Bayesian estimation (marginally uniform prior)","\n", sep = "")
  }
  cat("\n")

  if(groups > 1){
    for(g in 1:groups){
      cat(paste0("Group g",as.character(g),":"),"\n", sep = "")
      cat("\n")
      cat("correlation types","\n")
      print(x$cor.type[[g]],na.print="",quote=FALSE)
      cat("\n")
      cat("Posterior 2.5% lower bounds:","\n", sep = "")
      print(round(corrlist[[g]][[2]],digits), na.print = "")
      cat("\n")
      cat("Posterior median:","\n", sep = "")
      print(round(corrlist[[g]][[1]],digits), na.print = "")
      cat("\n")
      cat("Posterior 97.5% upper bounds:","\n", sep = "")
      print(round(corrlist[[g]][[3]],digits), na.print = "")
      cat("\n")
    }
  }else{
    cat("correlation types","\n")
    print(x$cor.type[[1]],na.print="",quote=FALSE)
    cat("\n")
    cat("Posterior 2.5% lower bounds:","\n", sep = "")
    print(round(corrlist[[1]][[2]],digits), na.print = "")
    cat("\n")
    cat("Posterior median:","\n", sep = "")
    print(round(corrlist[[1]][[1]],digits), na.print = "")
    cat("\n")
    cat("Posterior 97.5% upper bounds:","\n", sep = "")
    print(round(corrlist[[1]][[3]],digits), na.print = "")
    cat("\n")
  }

}


#' @method summary cor_test
#' @export
summary.cor_test <- function(object, digits = 3, ...){

  cor.df <- round(as.data.frame(object$correstimates),digits)
  cor.df$cor.type <- unlist(lapply(1:length(object$cor.type),function(g){
    object$cor.type[[g]][lower.tri(object$cor.type[[g]])]
  }))

  cor.df

}


#' @importFrom coda mcmc
#' @method plot cor_test
#' @export
plot.cor_test <- function(x, ...){

  numgroups <- length(x$corrdraws)
  P <- dim(x$corrdraws[[1]])[2]
  numcor <- P*(P-1)/2
  numcor.total <- numcor * numgroups

  cor.draws.matrix <- mcmc(data=do.call(cbind,lapply(1:numgroups,function(g){
    do.call(cbind,lapply(1:(P-1),function(p){
      draws_g_p <- as.matrix(x$corrdraws[[g]][,(p+1):P,p])
      colnames(draws_g_p) <- x$corrnames[[g]][(p+1):P,p]
      draws_g_p
    }))
  })),start=1,end=length(x$corrdraws[[1]][,1,1]),thin=1)

  plot(cor.draws.matrix, ...)

}

#' A sequence of innovation-related e-mail messages
#'
#' A time-ordered sequence of 247 communication messages between 25 actors.
#'
#' The related data files 'actors', 'same_location', 'same_culture' contain information
#' on the actors and three event statistics respectively.
#'
#'
#' @name relevents
#' @docType data
#' @usage data(relevents)
#' @format dataframe (247 rows, 3 columns)
#'
#' \tabular{lll}{
#'    \strong{relevents$time} \tab \code{numeric} \tab Time of the e-mail message,
#'    in seconds since onset of the observation \cr
#'    \strong{relevents$sender} \tab \code{integer} \tab ID of the sender, corresponding to
#'    the employee IDs in the actors dataframe \cr
#'    \strong{relevents$receiver} \tab \code{integer} \tab ID of the receiver \cr
#' }
#'
#' @keywords datasets
NULL

#' Same culture event statistic
#'
#' A matrix coding whether senders of events (in the rows) and receivers of events
#' (in the column) have the background culture. Related to the 'events' data object,
#' that contains a relational event sequence, and the 'actors' object, that contains
#' information on the 25 actors involved in the relational event sequence.
#'
#'
#' @name same_culture
#' @docType data
#' @usage data(same_culture)
#' @format dataframe (25 rows, 4 columns)
#'
#' \tabular{lll}{
#'    \strong{same_culture} \tab \code{integer} \tab Event statistic. Matrix with senders in the
#'    rows and receivers in the columns. The event statistic is 1 if sender and receiver have
#'    the same culture and 0 otherwise. \cr
#' }
#'
#' @keywords datasets
NULL

#' Multiple Sources of Attentional Dysfunction in Adults With Tourette's Syndrome
#'
#' Data from a psychological study comparing attentional performances of
#' Tourette's syndrome (TS) patients, ADHD patients, and controls.
#' These data were simulated using the sufficient statistics from Silverstein,
#' Como, Palumbo, West, and Osborn (1995).
#'
#' \tabular{lll}{
#'    \strong{accuracy} \tab \code{numeric} \tab Participant's accuracy in the attentional task\cr
#'    \strong{group} \tab \code{factor} \tab Participant's group membership (TS patient, ADHD patient, or control)\cr
#' }
#' @docType data
#' @keywords datasets
#' @name attention
#' @usage data(attention)
#' @references Silverstein, S. M., Como, P. G., Palumbo, D. R., West, L. L., & Osborn, L. M. (1995). Multiple sources of attentional dysfunction in adults with Tourette's syndrome: Comparison with attention deficit-hyperactivity disorder. Neuropsychology, 9(2), 157-164. doi:10.1037/0894-4105.9.2.157
#' @format A data.frame with 51 rows and 2 columns.
NULL

! rngfuncs.f90
module rngfuncs1

implicit none

public

! See the R header `R_ext/Random.h` for details
interface
    ! double unif_rand(void);
    function unif_rand() bind(c,name="unif_rand")
        use, intrinsic :: iso_c_binding, only: c_double
        real(c_double) :: unif_rand
    end function

    ! void GetRNGstate(void);
    subroutine getrngstate() bind(c,name="GetRNGstate")
    end subroutine

    ! void PutRNGstate(void);
    subroutine putrngstate() bind(c,name="PutRNGstate")
    end subroutine

  end interface

end module

module rkinds1
   use, intrinsic :: iso_c_binding
   use, intrinsic :: iso_fortran_env
   private
   integer, parameter, public :: rint = int32   ! Using int32 from iso_fortran_env
   integer, parameter, public :: rdp = real64   ! Using real64 from iso_fortran_env
   ! Using real64 from iso_fortran_env
end module


subroutine draw_ju(P,drawscorr,samsize,numcorrgroup,Fisher)
    ! Fortran implementation of the algorithm proposed by Joe (2006)

    use rkinds1, only: rint, rdp
    use rngfuncs1

    implicit none

    integer(rint), intent(in) :: P, samsize, numcorrgroup, Fisher
    integer(rint)             :: s1,r1, r2, i1, i2, k1, corrIndex(P,P), teldummy,&
                               t1, t2, error1
    real (rdp)              :: corrMat(P,P),draw1(1),&
                               R2inv(P,P), vec1(P,1), vec2(P,1),&
                               dummy11(1,1), dummy12(1,1), dummy22(1,1),&
                               Di1i2, preinv(P,P)
    real(rdp), intent (out) :: drawscorr(samsize,numcorrgroup)
    real(rdp)               :: alpha

!========================================================================================!

    ! create corrIndex matrix
    teldummy = 1

    do r1=2,P
        do r2=1,r1-1
            corrIndex(r1,r2) = teldummy
            corrIndex(r2,r1) = teldummy
            teldummy = teldummy + 1
        end do
    end do

    do s1=1,samsize

        ! create identity matrix
        do t1=1,P
            do t2=1,P
                if (t1==t2) then
                    corrMat(t1,t2)=1.0
                else
                    corrMat(t1,t2)=0.0
                ENDIF
            end do
        end do
        do r1 = 1,P-1
            alpha=P/2.0
            draw1 = random_beta(alpha, alpha, .true.)
            draw1 = draw1*2.0-1.0
            corrMat(r1,r1+1) = draw1(1)
            corrMat(r1+1,r1) = corrMat(r1,r1+1)
            drawscorr(s1,corrIndex(r1+1,r1)) = corrMat(r1,r1+1)
        end do
        R2inv(:,:) = 0
        preinv(:,:)= 0
        do r1 = 3,P
            do r2 = 1,P-r1+1
                i1 = r2
                i2 = r2+r1-1
                k1 = i2 - i1
                !draw partial correlations
                alpha = .5*(P+1-k1)
                draw1 = random_beta(alpha, alpha, .true.)
                draw1=draw1*2-1.0
                !rbeta(1,.5*(dim+1-k),.5*(dim+1-k))*2-1
                vec1(1:(k1-1),1) = corrMat(i1,(i1+1):(i1+k1-1))
                vec2(1:(k1-1),1) = corrMat(i2,(i1+1):(i1+k1-1))
                preinv(1:(i2-i1-1),1:(i2-i1-1)) = ((corrMat((i1+1):(i2-1),(i1+1):(i2-1))))
        !        R2inv(1:(i2-i1-1),1:(i2-i1-1)) = inverse(preinv(1:(i2-i1-1),1:(i2-i1-1)),(i2-i1-1))
                call FINDinv(preinv(1:(i2-i1-1),1:(i2-i1-1)),R2inv(1:(i2-i1-1),1:(i2-i1-1)),(i2-i1-1),error1)

                dummy11 = matmul(matmul(transpose(vec1(1:(k1-1),:)),R2inv(1:(i2-i1-1),1:(i2-i1-1))),vec1(1:(k1-1),:))
                dummy22 = matmul(matmul(transpose(vec2(1:(k1-1),:)),R2inv(1:(i2-i1-1),1:(i2-i1-1))),vec2(1:(k1-1),:))
                dummy12 = matmul(matmul(transpose(vec1(1:(k1-1),:)),R2inv(1:(i2-i1-1),1:(i2-i1-1))),vec2(1:(k1-1),:))
                Di1i2 = sqrt((1-dummy11(1,1))*(1-dummy22(1,1)))

                corrMat(i1,i2) = dummy12(1,1) + Di1i2*draw1(1)
                corrMat(i2,i1) = corrMat(i1,i2)

                drawscorr(s1,corrIndex(i1,i2)) = corrMat(i1,i2)
                end do
            end do
        end do
    if(Fisher==1) then
        drawscorr(1:samsize,1:numcorrgroup) = .5*log((1.0+drawscorr(1:samsize,1:numcorrgroup)) &
                                        /(1.0-drawscorr(1:samsize,1:numcorrgroup)))
    end if


contains


! Subroutine to find the inverse of a square matrix
SUBROUTINE FINDinv(matrix, inverse, n, errorflag)

    implicit none

    !Declarations
    INTEGER(rint), INTENT(IN) :: n
    REAL(rdp), INTENT(IN) :: matrix(n,n)  !Input matrix
    INTEGER(rint), INTENT(OUT) :: errorflag  !Return error status. -1 for error, 0 for normal
    REAL(rdp), INTENT(OUT) :: inverse(n,n) !Inverted matrix

    integer :: ipiv(n), info, lwork
    real(rdp) :: work(n)

    external :: dgetrf, dgetri

    errorflag = 0

    inverse = matrix
    call dgetrf(n,n,inverse,n,ipiv,info)
    if (info > 0) then
        inverse = 0
        errorflag = -1
        return
    end if

    lwork = n
    call dgetri(n,inverse,n,ipiv,work,lwork,info)
    if (info > 0) then
        inverse = 0
        errorflag = -1
        return
    end if

END SUBROUTINE FINDinv



FUNCTION random_beta(aa, bb, first) RESULT(fn_val)

! Adapted from Fortran 77 code from the book:
!     Dagpunar, J. 'Principles of random variate generation'
!     Clarendon Press, Oxford, 1988.   ISBN 0-19-852202-9

!     Author: Alan Miller
!             CSIRO Division of Mathematical & Information Sciences
!             Private Bag 10, Clayton South MDC
!             Clayton 3169, Victoria, Australia
!     Phone: (+61) 3 9545-8016      Fax: (+61) 3 9545-8080
!     e-mail: amiller @ bigpond.net.au

! FUNCTION GENERATES A RANDOM VARIATE IN [0,1]
! FROM A BETA DISTRIBUTION WITH DENSITY
! PROPORTIONAL TO BETA**(AA-1) * (1-BETA)**(BB-1).
! USING CHENG'S LOG LOGISTIC METHOD.

!     AA = SHAPE PARAMETER FROM DISTRIBUTION (0 < REAL)
!     BB = SHAPE PARAMETER FROM DISTRIBUTION (0 < REAL)

    implicit none

    REAL(rdp), INTENT(IN)    :: aa, bb
    LOGICAL, INTENT(IN) :: first
    !INTEGER(rint), INTENT(IN) :: iseed
    REAL ( kind = rdp )   :: fn_val

    !     Local variables
    REAL(rdp), PARAMETER  :: aln4 = 1.3862944, one=1.0, two=2.0, &
                             vlarge = HUGE(1.0), vsmall = TINY(1.0), zero = 0.0
    REAL ( kind = rdp ) :: a, b, g, r, s, x, y, z
    REAL ( kind = rdp ), SAVE        :: d, f, h, t, c
    LOGICAL, SAVE     :: swap



    IF (first) THEN                        ! Initialization, if necessary
    a = aa
    b = bb
    swap = b > a
    IF (swap) THEN
        g = b
        b = a
        a = g
    END IF
    d = a/b
    f = a+b
    IF (b > one) THEN
        h = SQRT((two*a*b - f)/(f - two))
        t = one
    ELSE
        h = b
        t = one/(one + (a/(vlarge*b))**b)
    END IF
    c = a+h
    END IF

    DO
    r = unif_rand()
    x = unif_rand()
    !print*, r,x

    s = r*r*x
    IF (r < vsmall .OR. s <= zero) CYCLE
    IF (r < t) THEN
        x = LOG(r/(one - r))/h
        y = d*EXP(x)
        z = c*x + f*LOG((one + d)/(one + y)) - aln4
        IF (s - one > z) THEN
        IF (s - s*z > one) CYCLE
        IF (LOG(s) > z) CYCLE
        END IF
        fn_val = y/(one + y)
    ELSE
        IF (4.0*s > (one + one/d)**f) CYCLE
        fn_val = one
    END IF
    EXIT
    END DO

    IF (swap) fn_val = one - fn_val
    RETURN
END FUNCTION random_beta



function eye(n)

    implicit none

    integer(rint):: i,n
    real(rdp):: eye(n,n)
    real(rdp):: check(n,n)

    check=0
    do i=1,n
        check(i,i)= 1
    enddo

    eye(:,:) = check(:,:)
    return

end function eye


end subroutine draw_ju

#' Wason task performance and morality
#'
#' Data from an experimental study, using the Wason selection task (Wason 1968)
#' to examine whether humans have cognitive adaptations for detecting violations
#' of rules in multiple moral domains. Moral domains are operationalized in
#' terms of the five domains of the Moral Foundations Questionnaire
#' (Graham et al. 2011).
#' These data were simulated using the
#' R-package \code{synthpop}, based on the characteristics of the original data.
#'
#' \tabular{lll}{
#'    \strong{sex} \tab \code{factor} \tab Participant sex\cr
#'    \strong{age} \tab \code{integer} \tab Participant age\cr
#'    \strong{nationality} \tab \code{factor} \tab Participant nationality\cr
#'    \strong{politics} \tab \code{integer} \tab How would you define your political opinions? Likert type scale, from 1 (Liberal) to 6 (Conservative)\cr
#'    \strong{WasonOrder} \tab \code{factor} \tab Was the Wason task presented before, or after the MFQ? \cr
#'    \strong{Harm} \tab \code{numeric} \tab MFQ harm domain.\cr
#'    \strong{Fairness} \tab \code{numeric} \tab MFQ fairness domain.\cr
#'    \strong{Loyalty} \tab \code{numeric} \tab MFQ loyalty domain.\cr
#'    \strong{Purity} \tab \code{numeric} \tab MFQ purity domain.\cr
#'    \strong{Tasktype} \tab \code{ordered} \tab How was the Wason task framed?\cr
#'    \strong{GotRight} \tab \code{factor} \tab Did the participant give the correct answer to the Wason task?
#' }
#' @docType data
#' @keywords datasets
#' @name sivan
#' @usage data(sivan)
#' @references Sivan, J., Curry, O. S., & Van Lissa, C. J. (2018). Excavating the Foundations: Cognitive Adaptations for Multiple Moral Domains. Evolutionary Psychological Science, 4(4), 408–419. doi:10.1007/s40806-018-0154-8
#' @format A data.frame with 887 rows and 12 columns.
NULL

#' @title Multivariate Student t test
#' @description First step to performs a Bayesian multivariate one sample Student t test using the
#' (adjusted) fractional Bayes factor using the \code{BF()} function.
#'
#'@details \code{X} must be a data matrix and \code{null}
#'must be a vector of the assumed null values of the variables.
#'
#'@param X a data matrix with the variables in the columns.
#'@param Y an optional data matrix with the variables in the columns.
#'@param null a vector of the null values of the variables.
#'@param paired a logical indicating whether you want a multivariate paired t-test.
#'@param ... further arguments to be passed to or from methods.
#'
#'@return An object that can be applied to the \code{BF()}.
#'
#'@references Mulder, J. and Gu, X. (2023). Bayesian Testing of Scientific
#'Expectations under Multivariate Normal Linear Models. Multivariate Behavioral
#'Research, 57, 767-783. DOI: 10.1080/00273171.2021.1904809.
#'
#'@examples
#'
#'mvt_fmri <- mvt_test(fmri[,1:2],null = c(0,0))
#'BF(mvt_fmri)
#'
#'# the same test can be executed via the lm() function
#'intercept <- rep(1,nrow(fmri))
#'lm1 <- lm(cbind(Face,Vehicle) ~ -1 + intercept, data=fmri)
#'BF(lm1,hypothesis="intercept_on_Face=intercept_on_Vehicle=0")
#'
#' @rdname mvt_test
#' @export
mvt_test <- function(X, Y, null = NULL, paired = FALSE, ...){

  if(missing(Y)){

    X <- as.data.frame(as.matrix(X))
    p <- ncol(X)
    n <- nrow(X)

    if(is.null(null)){
      null <- rep(0,p)
    }

    if(length(null) != p){
      stop("'null' must be a vector of length equal to the number of variables.")
    }

    if(p == 1){
      out <- t_test(x=X,mu=null)
    }else{
      intercept <- rep(1,n)
      varnames <- colnames(X)
      formu <- as.formula(paste0("cbind(",paste0(varnames,collapse = ","),") ~ -1 + intercept"))
      out <- lm(formu,data=X)
      out$null <- null
      out$numpop <- 1
      out$paired <- paired
      class(out) <- "mvt_test"
      #
    }

  }else{

    if(paired==TRUE){

      p <- ncol(X)
      n <- nrow(X)

      if(is.null(null)){
        null <- rep(0,p)
      }

      if(length(null) != p){
        stop("'null' must be a vector of length equal to the number of variables.")
      }

      if(dim(X)[1]!=dim(Y)[1] || dim(X)[2]!=dim(Y)[2]){
        stop("X and Y must have same dimension in case of paired test")
      }

      if(p == 1){
        out <- t_test(x=X-Y,mu=null,paired=paired)
      }else{
        intercept <- rep(1,n)
        varnames <- colnames(X)
        formu <- as.formula(paste0("cbind(",paste0(varnames,collapse = ","),") ~ -1 + intercept"))
        df.XY <- as.data.frame(X-Y)
        colnames(df.XY) <- varnames
        out <- lm(formu,data=df.XY)
        out$null <- null
        out$numpop <- 1
        out$paired <- paired
        class(out) <- "mvt_test"
        #
      }

    }else{ # independent samples multivariate t test
      # equal covariances assumed

      p <- ncol(X)
      n1 <- nrow(X)
      n2 <- nrow(Y)
      N <- n1 + n2

      if(is.null(null)){
        null <- rep(0,p)
      }

      if(length(null) != p){
        stop("'null' must be a vector of length equal to the number of variables.")
      }

      if(dim(X)[2]!=dim(Y)[2]){
        stop("X and Y must have same number of columns")
      }

      if(p == 1){
        out <- t_test(x=X,y=Y,mu=null,paired=paired)
      }else{
        intercept <- rep(1,N)
        difference <- as.factor(c(rep("e",n1),rep("0",n2)))
        varnames <- colnames(X)
        formu <- as.formula(paste0("cbind(",paste0(varnames,collapse = ","),") ~ 1 + differenc"))
        df.XY <- as.data.frame(rbind(X,Y))
        df.XY$differenc <- difference
        colnames(df.XY)[1:p] <- varnames
        out <- lm(formu,data=df.XY)
        out$null <- null
        out$numpop <- 2
        out$paired <- paired
        class(out) <- "mvt_test"
        #
      }

    }

  }

  return(out)
}

#' @method get_estimates mvt_test
#' @export
get_estimates.mvt_test <- function(x, ...){
  class(x) <- "lm"
  x_est <- get_estimates(x)
  if(x$numpop==1){
    if(x$paired==FALSE){
      names(x_est$estimate) <- colnames(x_est$Sigma[[1]]) <-
        row.names(x_est$Sigma[[1]]) <- gsub("intercept_on_","",names(x_est$estimate))
    }else{
      names(x_est$estimate) <- colnames(x_est$Sigma[[1]]) <-
        row.names(x_est$Sigma[[1]]) <- gsub("intercept_on_","difference_",names(x_est$estimate))
    }
  }else{
    P <- ncol(x$coefficients)
    select_diff <- 2*(1:P)
    names(x_est$estimate) <- colnames(x_est$Sigma[[1]]) <-
      row.names(x_est$Sigma[[1]]) <- gsub("difference_on_","difference_",names(x_est$estimate))
    x_est$estimate <- x_est$estimate[select_diff]
    x_est$Sigma[[1]] <- x_est$Sigma[[1]][select_diff,select_diff]
  }
  x_est
}

#' @importFrom stats as.formula
#' @method BF mvt_test
#' @export
BF.mvt_test <- function(x,
                        hypothesis = NULL,
                        prior.hyp.explo = NULL,
                        prior.hyp.conf = NULL,
                        prior.hyp = NULL,
                        complement = TRUE,
                        log = FALSE,
                        cov.prob = .95,
                        BF.type = NULL,
                        ...) {

  if(!(cov.prob>0 & cov.prob<1)){
    stop("The argument 'cov.prob' is a coverage probability for the interval estimates that should lie between 0 and 1. The default is 0.95.")
  }

  if(x$numpop==1 & x$paired==FALSE){
    parameters <- "means"
  }else{
    parameters <- "differences"
  }

  if(x$numpop == 1){
    P <- length(x$coefficients)
    names1 <- colnames(x$coefficients)
    names2 <- paste0("intercept_on_",names1)
  }else{
    P <- ncol(x$coefficients)
    names1 <- colnames(x$coefficients)
    names2 <- paste0("difference_on_",names1)
  }
  #exploratory test of joint equality to null
  hypothesis.explo <- paste0(unlist(lapply(1:P,function(p){
    paste0(names2[p],"=",x$null[p])
  })),collapse = " & ")
  x1 <- x
  class(x1) <- "lm"
  BF.explo <- BF(x1,
                 hypothesis=hypothesis.explo,
                 prior.hyp.conf=prior.hyp.explo,
                 log=log,
                 cov.prob=cov.prob,
                 BF.type=BF.type)
  BF.explo$BFtu_confirmatory <- t(as.matrix(BF.explo$BFtu_confirmatory))
  BF.explo$PHP_confirmatory <- t(as.matrix(BF.explo$PHP_confirmatory))
  row.names(BF.explo$BFtu_confirmatory) <- row.names(BF.explo$PHP_confirmatory) <- parameters
  colnames(BF.explo$BFtu_confirmatory) <- c("BF0u","BFuu")
  colnames(BF.explo$PHP_confirmatory) <- c("Pr(=null)","Pr(not null)")

  if(!is.null(hypothesis)){
    variable.names <- colnames(x$coefficients)
    if(x$numpop==1 | x$paired==TRUE){
      add.name <- "intercept_on_"
      rem.name <- ""
      if(x$numpop==1 & x$paired==TRUE){
        rem.name <- "difference_"
      }
    }else{
      add.name <- "difference_on_"
      rem.name <- "difference_"
    }
    hypothesis.updated <- hypothesis
    for(varname in variable.names){
      hypothesis.updated <- gsub(paste0(rem.name,varname),paste0(add.name,varname),hypothesis.updated)
    }
  }else{
    hypothesis.updated <- NULL
  }

  BF.conf <- BF(x1,
                hypothesis=hypothesis.updated,
                prior.hyp.conf=prior.hyp.conf,
                log=log,
                complement=complement,
                BF.type=BF.type)
  if(!is.null(hypothesis)){
    BF.conf$hypotheses <- gsub(add.name,rem.name,BF.conf$hypotheses)
  }

  BF.explo_estimates <- BF.explo$estimates
  if(x$numpop == 1){
    row.names(BF.explo_estimates) <- names(get_estimates(x)$estimate)
  }else{
    BF.explo_estimates <- BF.explo_estimates[(1:P)*2,]
    row.names(BF.explo_estimates) <- names(get_estimates(x)$estimate)
  }

  BFlm_out <- list(
    BFtu_exploratory=BF.explo$BFtu_confirmatory,
    PHP_exploratory=BF.explo$PHP_confirmatory,
    BFtu_confirmatory=BF.conf$BFtu_confirmatory,
    PHP_confirmatory=BF.conf$PHP_confirmatory,
    BFmatrix_confirmatory=BF.conf$BFmatrix_confirmatory,
    BFtable_confirmatory=BF.conf$BFtable_confirmatory,
    prior.hyp.explo=BF.explo$prior.hyp.conf,
    prior.hyp.conf=BF.conf$prior.hyp.conf,
    hypotheses=BF.conf$hypotheses,
    estimates=BF.explo_estimates,
    model=x1,
    bayesfactor=BF.conf$bayesfactor,
    parameter=parameters,
    log = BF.conf$log,
    fraction_number_groupIDs = BF.conf$fraction_number_groupIDs,
    fraction_groupID_observations = BF.conf$fraction_groupID_observations,
    fraction_groupID = BF.conf$fraction_groupID,
    call=match.call())

  class(BFlm_out) <- "BF"

  return(BFlm_out)

}

#' @method print BF
#' @export
print.BF <- function(x,
                     digits = 3,
                     na.print = "", ...){


  cat("Call:")
  cat("\n")
  print(x$call)

  cat("\n")

  digits <- 3

  if(is.null(x$BFtu_confirmatory)){

    cat("Bayesian hypothesis test","\n", sep = "")
    cat("Type: exploratory","\n", sep = "")
    cat("Object: ",class(x$model)[1],"\n", sep = "")
    cat("Parameter: ",x$parameter,"\n", sep = "")
    cat("Method: ",x$bayesfactor,"\n\n", sep = "")

    cat("Posterior probabilities of the hypotheses","\n", sep = "")
    cat("\n")
    cat("Model parameters:")
    cat("\n")
    print(round(x$PHP_exploratory,digits))

    cat("\n")

    if(sum(class(x$model)=="aov")>0){
      if(!is.null(x$PHP_main)){
        cat("main effects:")
        cat("\n")
        print(round(x$PHP_main,digits))

        cat("\n")
      }
      if(!is.null(x$PHP_interaction)){
        cat("interaction effects:")
        cat("\n")
        print(round(x$PHP_interaction,digits))

        cat("\n")
      }
    }
    if(class(x$model)[1]=="rma.uni"){
      cat("Posterior support for random effects","\n", sep = "")
      cat("\n")
      tau2out <- matrix(x$estimates[2,5])
      row.names(tau2out) <- "tau2 (marema)"
      colnames(tau2out) <- "Pr(>0)"
      print(round(tau2out,digits))
      cat("\n")
    }
    if(class(x$model)[1]=="cor_test"){
      cat("Bartlett's test of sphericity (all correlations are zero):","\n", sep = "")
      cat("\n")
      if(nrow(x$PHP_Bartlett)>1){
        print(round(x$PHP_Bartlett,digits))
      }else{
        print(round(x$PHP_Bartlett[1,],digits))
      }
      cat("\n")
    }

  }else{

    cat("Bayesian hypothesis test","\n", sep = "")
    cat("Type: confirmatory","\n", sep = "")
    cat("Object: ",class(x$model)[1],"\n", sep = "")
    cat("Parameter: ",x$parameter,"\n", sep = "")
    cat("Method: ",x$bayesfactor,"\n\n", sep = "")

    cat("Posterior probabilities of the hypotheses")
    cat("\n")

    PHPmatrix <- as.matrix(round(x$PHP_confirmatory,digits))
    colnames(PHPmatrix) <- "Pr(hypothesis|data)"
    hypnumbers <- unlist(lapply(1:nrow(PHPmatrix),function(r){
      paste0("H",as.character(r))
    }))
    row.names(PHPmatrix) <- hypnumbers
    print(PHPmatrix)

    cat("\n")
    if(x$log==FALSE){
      cat("Bayes factors (rows versus columns):")
    }else{
      cat("Log Bayes factors (rows versus columns):")
    }
    cat("\n")

    BFmat <- round(x$BFmatrix_confirmatory,digits)
    row.names(BFmat) <- colnames(BFmat) <- hypnumbers
    print(BFmat)

    cat("\n")
    cat("Hypotheses:")
    cat("\n")

    for(h in 1:length(x$hypotheses)){
      cat(paste0(hypnumbers[h],": ",x$hypotheses[h]))
      cat("\n")
    }

  }

}

#' @importFrom utils getFromNamespace
#' @importFrom stats approxfun coef complete.cases cov dbeta density dnorm dt lm median model.matrix
#' @importFrom stats nobs pchisq pnorm pt quantile rWishart rbeta rgamma rnorm rt sd setNames var vcov
parse_hypothesis <- getFromNamespace("parse_hypothesis", "bain")
constraint_to_equation <- getFromNamespace("constraint_to_equation", "bain")
constraint_to_row <- getFromNamespace("constraint_to_row", "bain")
expand_compound_constraints <- getFromNamespace("expand_compound_constraints", "bain")
expand_parentheses <- getFromNamespace("expand_parentheses", "bain")
flip_inequality <- getFromNamespace("flip_inequality", "bain")
order_terms <- getFromNamespace("order_terms", "bain")
params_in_hyp <- getFromNamespace("params_in_hyp", "bain")

#' @method summary BF
#' @export
summary.BF <- function(object, ...){

  cat("Call:")
  cat("\n")
  print(object$call)

  cat("\n")
  digits <- 3

  cat("Bayesian hypothesis test","\n", sep = "")
  cat("Type: exploratory","\n", sep = "")
  cat("Object: ",class(object$model)[1],"\n", sep = "")
  cat("Parameter: ",object$parameter,"\n", sep = "")
  cat("Method: ",object$bayesfactor,"\n\n", sep = "")

  cat("Posterior probabilities of the hypotheses","\n", sep = "")
  cat("\n")
  cat("Model parameters:")
  cat("\n")
  print(round(object$PHP_exploratory,digits))

  cat("\n")

  if(sum(class(object$model)=="aov")>0){
    if(!is.null(object$PHP_main)){
      cat("main effects:")
      cat("\n")
      print(round(object$PHP_main,digits))

      cat("\n")
    }
    if(!is.null(object$PHP_interaction)){
      cat("interaction effects:")
      cat("\n")
      print(round(object$PHP_interaction,digits))

      cat("\n")
    }
  }
  if(class(object$model)[1]=="rma.uni"){
    cat("Posterior support for random effects","\n", sep = "")
    cat("\n")
    tau2out <- matrix(object$estimates[2,5])
    row.names(tau2out) <- "tau2 (marema)"
    colnames(tau2out) <- "Pr(>0)"
    print(round(tau2out,digits))
    cat("\n")
  }
  if(class(object$model)[1]=="cor_test"){
    cat("Bartlett's test of sphericity (all correlations are zero):","\n", sep = "")
    cat("\n")
    if(nrow(object$PHP_Bartlett)>1){
      print(round(object$PHP_Bartlett,digits))
    }else{
      print(round(object$PHP_Bartlett[1,],digits))
    }
    cat("\n")
  }

  if(!is.null(object$BFtu_confirmatory)){

    cat("Bayesian hypothesis test","\n", sep = "")
    cat("Type: confirmatory","\n", sep = "")
    cat("Object: ",class(object$model)[1],"\n", sep = "")
    cat("Parameter: ",object$parameter,"\n", sep = "")
    cat("Method: ",object$bayesfactor,"\n\n", sep = "")

    cat("Posterior probabilities of the hypotheses")
    cat("\n")

    PHPmatrix <- as.matrix(round(object$PHP_confirmatory,digits))
    colnames(PHPmatrix) <- "Pr(hypothesis|data)"
    hypnumbers <- unlist(lapply(1:nrow(PHPmatrix),function(r){
      paste0("H",as.character(r))
    }))
    row.names(PHPmatrix) <- hypnumbers
    print(PHPmatrix)

    cat("\n")
    if(object$log==FALSE){
      cat("Bayes factors (rows versus columns):")
    }else{
      cat("Log Bayes factors (rows versus columns):")
    }
    cat("\n")

    BFmat <- round(object$BFmatrix_confirmatory,digits)
    row.names(BFmat) <- colnames(BFmat) <- hypnumbers
    print(BFmat)

    cat("\n")
    cat("Specification table:")
    cat("\n")

    BFtable <- round(object$BFtable_confirmatory,digits)
    row.names(BFtable) <- hypnumbers
    print(BFtable)

    cat("\n")
    cat("Hypotheses:")
    cat("\n")

    for(h in 1:length(object$hypotheses)){
      cat(paste0(hypnumbers[h],": ",object$hypotheses[h]))
      cat("\n")
    }
  }
}

#' Actors from a small hypothetical network
#'
#' The related data files 'events', 'same_location', 'same_culture' contain
#' information on the  event sequence and the two event statistics respectively.
#'
#'
#' @name actors
#' @docType data
#' @usage data(actors)
#' @keywords datasets
#' @format dataframe (25 rows, 4 columns)
#'
#' \tabular{lll}{
#'    \strong{actors$id} \tab \code{integer} \tab ID of the employee, corresponding to
#'    the sender and receiver IDs in the events dataframe \cr
#'    \strong{actors$location} \tab \code{numeric} \tab Location of the actor,
#'    ranging from 1-4 \cr
#'    \strong{actors$culture} \tab \code{character} \tab Categorical variable, indicating the
#'    culture of the employee \cr
#' }
#'
#' @keywords datasets
NULL

#' Trends in International Mathematics and Science Study (TIMSS) 2011-2015
#'
#' A stratified sample was drawn by country and school to obtain a balanced
#' sample of p = 15 grade-4 students
#' per school for each of four countries (The Netherlands (NL), Croatia (HR),
#' Germany
#' (DE), and Denmark (DK)) and two measurement occasions (2011, 2015).
#' Achievement scores
#' (first plausible value) of overall mathematics were considered. Performances
#' of fourth
#' and eight graders from more than 50 participating countries around the world
#' can be found at (https://www.iea.nl/timss)
#' The TIMSS achievement scale is centered at 500 and the standard deviation is
#' equal to 100 scale score points.
#' The TIMSS data set has a three-level structure, where students are nested
#' within classrooms/schools, and
#' the classrooms/schools are nested within countries. Only one classroom was
#' sampled per school.
#' Changes in the mathematics achievement can be investigated by examining the
#' grouping of
#' students in schools across countries. Changes in country-specific intraclass
#' correlation coefficient
#' from 2011 to 2015, representing heterogeneity in mathematic achievements
#' within and between schools across years,
#' can be tested. When detecting a decrease in average performance together
#' with an increase
#' of the intraclass correlation, a subset of schools performed worse. For a
#' constant
#' intraclass correlation across years the drop in performance applied to the
#' entire population
#' of schools. For different countries, changes in the intraclass correlation
#' across years
#' can be tested concurrently to examine also differences across countries.
#'
#' \tabular{lll}{
#'    \strong{math} \tab \code{numeric} \tab math score child\cr
#'    \strong{groupNL11} \tab \code{numeric} \tab
#'    Indicator for child from NL in 2011\cr
#'    \strong{groupNL15} \tab \code{numeric} \tab
#'    Indicator for child from NL in 2015\cr
#'    \strong{groupHR11} \tab \code{numeric} \tab
#'    Indicator for child from HR in 2011\cr
#'    \strong{groupHR15} \tab \code{numeric} \tab
#'    Indicator for child from HR in 2015\cr
#'    \strong{groupDE11} \tab \code{numeric} \tab
#'    Indicator for child from DE in 2011\cr
#'    \strong{groupDE15} \tab \code{numeric} \tab
#'    Indicator for child from DE in 2015\cr
#'    \strong{groupDR11} \tab \code{numeric} \tab
#'    Indicator for child from DK in 2011\cr
#'    \strong{groupDR15} \tab \code{numeric} \tab
#'    Indicator for child from DK in 2015\cr
#'    \strong{gender} \tab \code{numeric} \tab Female=0,Male=1 \cr
#'    \strong{weight} \tab \code{numeric} \tab Child sampling weight \cr
#'    \strong{yeargender} \tab \code{numeric} \tab
#'    Interaction for occassion and gender \cr
#'    \strong{lln} \tab \code{numeric} \tab
#'    total number of children in school-class \cr
#'    \strong{groupschool} \tab \code{factor} \tab
#'    Nested indicator for school in country\cr
#'    \strong{schoolID} \tab \code{factor} \tab
#'    Unique indicator for school
#' }
#' @docType data
#' @keywords datasets
#' @name timssICC
#' @usage data(timssICC)
#' @references Mulder, J. & Fox, J.-P. (2019). Bayes factor testing of multiple
#' intraclass correlations. Bayesian Analysis. 14, 2, p. 521-552.
#' @format A data.frame with 16770 rows and 15 columns.
NULL

#' @importFrom sandwich sandwich
#' @importFrom coda as.mcmc mcmc.list gelman.diag
#' @importFrom stats terms
remove_predictors_helper <- function(Y_groups, formula){

  # number of groups
  groups <- length(Y_groups)

  # model matrix terms
  mm_terms <- attr(terms(formula), "term.labels")

  if(length(mm_terms) == 0){

    Y_groups

  } else {
    lapply(seq_len(groups), function(x){

      # check for factors
      factor_pred <- which(paste0("as.factor(", colnames(Y_groups[[x]]), ")") %in% mm_terms)

      # check for non factors
      cont_pred <- which(colnames(Y_groups[[x]]) %in% mm_terms)

      # remove predictors
      Y_groups[[x]][,-c(factor_pred, cont_pred)]

    })
  }
}


FisherZ <- function(r){.5*log((1+r)/(1-r))}


globalVariables(c("Fcor"))

#' @importFrom mvtnorm dmvnorm pmvnorm rmvnorm
#' @importFrom utils globalVariables
#' @importFrom stats dnorm pnorm
#' @importFrom QRM fit.st
#' @method BF cor_test
#' @export
BF.cor_test <- function(x,
                        hypothesis = NULL,
                        prior.hyp.explo = NULL,
                        prior.hyp.conf = NULL,
                        prior.hyp = NULL,
                        complement = TRUE,
                        log = FALSE,
                        cov.prob = .95,
                        ...){

  if(x$prior.cor == "joint.unif"){
    bayesfactor <- "Bayes factors based on joint uniform priors"
  }
  if(x$prior.cor == "marg.unif"){
    bayesfactor <- "Bayes factors based on marginally uniform priors"
  }
  if( !(x$prior.cor == "marg.unif" | x$prior.cor == "joint.unif") ){
    stop("argument 'prior' of cor_test object should be 'joint.unif' or 'marg.unif'. See ?cor_test")
  }
  testedparameter <- "correlation coefficients"

  if(!(cov.prob>0 & cov.prob<1)){
    stop("The argument 'cov.prob' is a coverage probability for the interval estimates that should lie between 0 and 1. The default is 0.95.")
  }
  CrI_LB <- (1 - cov.prob)/2
  CrI_UB <- 1 - (1 - cov.prob)/2

  logIN <- log

  # check proper usage of argument 'prior.hyp.conf' and 'prior.hyp.explo'
  if(!is.null(prior.hyp.conf)){
    prior.hyp <- prior.hyp.conf
  }
  prior.hyp.explo <- process.prior.hyp.explo(prior_hyp_explo = prior.hyp.explo, model = x)

  P <- dim(x$corrdraws[[1]])[2]
  numG <- length(x$corrdraws)
  numcorrgroup <- P*(P-1)/2
  get_est <- get_estimates(x)
  corrmeanN <- get_est$estimate
  corrcovmN <- get_est$Sigma[[1]]

  # Exploratory testing of correlation coefficients
  # get height of prior density at 0 of Fisher transformed correlation
  # slope and intercept of df as function of P based on Fcor

  if(sum(P==Fcor$P)==0){
    #number of draws to get 1e7 draws for the marginal of 1 Fisher transformation correlation
    numdraws <- round(1e7/(P*(P-1)/2))
    drawsJU <- draw_ju_r(P,samsize=numdraws,Fisher=1)
    approx_studt <- fit.st(c(drawsJU))$par.ests[c(1,3)]
  }else{
    # use the estimate of the scale from the Fcor object
    # for the df the estimates show some numerical error for P>20, so use fitted line
    approx_studt <- unlist(c(Fcor[which(P==Fcor$P),1:2]))
    if(P > 20){
      # use fitted linear line rather than rough estimate of df
      slpe1 <- 2.944494
      intcept1 <- 1.864901
      approx_studt[1] <- P * slpe1 + intcept1
    }
  }
  if(x$prior.cor == "joint.unif"){
    relcomp0 <- dt(0,df=approx_studt[1],log=TRUE)-log(approx_studt[2]) # all marginal priors are the same
    relcomp0_bartlett <- dmvt(rep(0,numcorrgroup),df=approx_studt[1],sigma=approx_studt[1]^2*diag(numcorrgroup),
                              log=TRUE)
  }
  if(x$prior.cor == "marg.unif"){
    relcomp0 <- dt(0,df=Fcor[1,1],log=TRUE)-log(Fcor[1,2]) #Fisher transformed height for a uniform(-1,1) prior
  }

  # compute exploratory BFs
  corr_names <- rownames(x$correstimates)
  numcorr <- length(corrmeanN)
  relfit <- matrix(c(dnorm(0,mean=corrmeanN,sd=sqrt(diag(corrcovmN)),log=TRUE),
                     pnorm(0,mean=corrmeanN,sd=sqrt(diag(corrcovmN)),log.p=TRUE),
                     pnorm(0,mean=corrmeanN,sd=sqrt(diag(corrcovmN)),log.p=TRUE,
                             lower.tail=FALSE)),ncol=3)
  relfit0_bartlett <- unlist(lapply(1:numG,function(gr){
    dmvnorm(rep(0,numcorrgroup),mean=corrmeanN[(gr-1)*numcorrgroup+1:numcorrgroup],
            sigma=as.matrix(corrcovmN[(gr-1)*numcorrgroup+1:numcorrgroup,(gr-1)*numcorrgroup+1:numcorrgroup]),
            log=TRUE)
  }))
  relcomp <- matrix(c(rep(relcomp0,numcorr),rep(log(.5),numcorr*2)),ncol=3)
  colnames(relcomp) <- colnames(relfit) <- c("p(=0)","Pr(<0)","Pr(>0)")
  BFtu_exploratory <- relfit - relcomp
  row.names(BFtu_exploratory) <- rownames(x$correstimates)
  colnames(BFtu_exploratory) <- c("BF0u","BF1u","BF2u")
  rowmax <- apply(BFtu_exploratory,1,max)
  norm_BF_explo <- exp(BFtu_exploratory - rowmax %*% t(rep(1,ncol(BFtu_exploratory)))) *
    (rep(1,nrow(BFtu_exploratory)) %*% t(prior.hyp.explo[[1]]))
  PHP_exploratory <- norm_BF_explo / apply(norm_BF_explo,1,sum)
  colnames(PHP_exploratory) <- c("P(=0)","P(<0)","P(>0)")
  BF_Bartlett <- relfit0_bartlett - relcomp0_bartlett
  PHP_Bartlett <- exp(BF_Bartlett) / (exp(BF_Bartlett) + 1)
  PHP_Bartlett <- cbind(PHP_Bartlett,1-PHP_Bartlett)
  BF_Bartlett <- cbind(BF_Bartlett,1)
  colnames(BF_Bartlett) <- c("BF0u","BFuu")
  row.names(PHP_Bartlett) <- row.names(BF_Bartlett) <- paste0("group",1:numG)
  colnames(PHP_Bartlett) <- c("Pr(all zero)","Pr(not all zero)")

  # posterior estimates
  postestimates_correlations <- Reduce(rbind,
                                       lapply(1:numG,function(g){
                                         draws_stack_g <- as.matrix(do.call(cbind,lapply(1:P,function(p){x$corrdraws[[g]][,,p]}))[,which(c(lower.tri(diag(P))))])
                                         means <- apply(draws_stack_g,2,mean)
                                         medians <- apply(draws_stack_g,2,median)
                                         lb <- apply(draws_stack_g,2,quantile,CrI_LB)
                                         ub <- apply(draws_stack_g,2,quantile,CrI_UB)
                                         probGr0 <- apply(draws_stack_g>0,2,mean)
                                         rm(draws_stack_g)
                                         return(cbind(means,medians,lb,ub,probGr0))
                                       }))

  colnames(postestimates_correlations) <- c("mean","median",paste0(as.character(round(CrI_LB*100,7)),"%"),
                                            paste0(as.character(round(CrI_UB*100,7)),"%"),"Pr(>0)")
  postestimates <- postestimates_correlations
  rownames(postestimates) <- corr_names

  if(logIN == FALSE){
    BFtu_exploratory <- exp(BFtu_exploratory)
  }

  # confirmatory testing if hypothesis argument is used
  if(!is.null(hypothesis)){
    if(x$prior.cor != "joint.unif"){
      BFtu_confirmatory <- PHP_confirmatory <- BFmatrix_confirmatory <- relfit <-
        relcomp <- hypotheses <- BFtable <- priorprobs <- NULL
      warning(
        "The 'hypothesis' argument is not supported when using the marginally uniform prior. Only testing
single correlations via the standard tests are supported. Comparing correlations is not recommended when
using the marginally uniform prior (Mulder and Gelissen, 2023, Section 4.2.2).")
    }else{

      #check if constraints are formulated on correlations in different populations
      #if so, then the correlation names contains the string "_in_g" at the end
      params_in_hyp1 <- params_in_hyp(hypothesis)

      corr_names <- unlist(lapply(1:length(x$corrnames),function(g){
        c(x$corrnames[[g]][lower.tri(x$corrnames[[g]])],
          t(x$corrnames[[g]])[lower.tri(x$corrnames[[g]])])
      })) #which includes Y1_with_Y2 and Y2_with_Y1

      parse_hyp <- parse_hypothesis(corr_names,hypothesis)
      parse_hyp$hyp_mat <- do.call(rbind, parse_hyp$hyp_mat)
      if(nrow(parse_hyp$hyp_mat)==1){
        select1 <- rep(1:numcorrgroup,numG) + rep((0:(numG-1))*2*numcorrgroup,each=numcorrgroup)
        select2 <- rep(numcorrgroup+1:numcorrgroup,numG) + rep((0:(numG-1))*2*numcorrgroup,each=numcorrgroup)
        parse_hyp$hyp_mat <-
          t(as.matrix(c(parse_hyp$hyp_mat[,select1] + parse_hyp$hyp_mat[,select2],parse_hyp$hyp_mat[,numcorrgroup*2*numG+1])))
      }else{
        #combine equivalent correlations, e.g., cor(Y1,Y2)=corr(Y2,Y1).
        select1 <- rep(1:numcorrgroup,numG) + rep((0:(numG-1))*2*numcorrgroup,each=numcorrgroup)
        select2 <- rep(numcorrgroup+1:numcorrgroup,numG) + rep((0:(numG-1))*2*numcorrgroup,each=numcorrgroup)
        parse_hyp$hyp_mat <-
          cbind(parse_hyp$hyp_mat[,select1] + parse_hyp$hyp_mat[,select2],parse_hyp$hyp_mat[,numcorrgroup*2*numG+1])
      }
      #create coefficient with equality and order constraints
      RrList <- make_RrList2(parse_hyp)
      RrE <- RrList[[1]]
      RrO <- RrList[[2]]
      numhyp <- length(RrE)
      #Fisher transform constants in constraints
      for(h in 1:numhyp){
        if(!is.null(RrE[[h]])){
          RrE[[h]][,ncol(RrE[[h]])] <- FisherZ(RrE[[h]][,ncol(RrE[[h]])])
        }
        if(!is.null(RrO[[h]])){
          RrO[[h]][,ncol(RrO[[h]])] <- FisherZ(RrO[[h]][,ncol(RrO[[h]])])
        }
      }

      relfit <- t(matrix(unlist(lapply(1:numhyp,function(h){
        Gaussian_measures(mean1=corrmeanN,Sigma1=corrcovmN,RrE1=RrE[[h]],RrO1=RrO[[h]],names1=names(corrmeanN),
                          constraints1=parse_hyp$original_hypothesis[h])
      })),nrow=2))
      #names1 and constraints1 ... to fix ...
      # approximate unconstrained Fisher transformed correlations with a multivariate Student t
      if(numcorrgroup==1){
        if(numcorr==1){
          Scale0 <- as.matrix(approx_studt[2]**2)
        }else{
          Scale0 <- diag(rep(approx_studt[2]**2,numG))
        }
        mean0 <- rep(0,numG)
        df0 <- approx_studt[1]
      }else{
        mean0 <- rep(0,numcorrgroup*numG)
        Scale0 <- diag(rep(approx_studt[2]**2,numcorrgroup*numG))
        df0 <- approx_studt[1]
      }
      names(mean0) <- row.names(Scale0) <- colnames(Scale0) <- names(corrmeanN)
      relcomp <- t(matrix(unlist(lapply(1:numhyp,function(h){
        relcomp_h <- Student_measures(mean1=mean0,
                                      Scale1=Scale0,
                                      df1=df0,
                                      constraints1=parse_hyp$original_hypothesis[h],
                                      names1=names(corrmeanN),
                                      RrE1=RrE[[h]],
                                      RrO1=RrO[[h]])
        return(relcomp_h)

      })),nrow=2))

      row.names(relfit) <- row.names(relcomp) <- parse_hyp$original_hypothesis
      if(complement == TRUE){
        relfit <- Gaussian_prob_Hc(mean1=corrmeanN,Sigma1=corrcovmN,relmeas=relfit,RrO)
        relcomp <- Student_prob_Hc(mean1=mean0,scale1=Scale0,df1=df0,relmeas1=relcomp,
                                   constraints=NULL,RrO1=RrO)
      }
      hypothesisshort <- unlist(lapply(1:nrow(relfit),function(h) paste0("H",as.character(h))))
      row.names(relfit) <- row.names(relfit) <- hypothesisshort

      # the BF for the complement hypothesis vs Hu needs to be computed.
      BFtu_confirmatory <- c(apply(relfit - relcomp, 1, sum))
      # Check input of prior probabilies
      if(is.null(prior.hyp)){
        priorprobs <- rep(1/length(BFtu_confirmatory),length(BFtu_confirmatory))
      }else{
        if(!is.numeric(prior.hyp) || length(prior.hyp)!=length(BFtu_confirmatory)){
          warning(paste0("Argument 'prior.hyp' should be numeric and of length ",as.character(length(BFtu_confirmatory)),". Equal prior probabilities are used."))
          priorprobs <- rep(1/length(BFtu_confirmatory),length(BFtu_confirmatory))
        }else{
          priorprobs <- prior.hyp
        }
      }
      names(priorprobs) <- names(BFtu_confirmatory)
      maxBFtu <- max(BFtu_confirmatory)
      PHP_confirmatory <- exp(BFtu_confirmatory-maxBFtu)*priorprobs / sum(exp(BFtu_confirmatory-maxBFtu)*priorprobs)
      BFtable <- cbind(relcomp,relfit,relfit[,1]-relcomp[,1],relfit[,2]-relcomp[,2],
                       apply(relfit,1,sum)-apply(relcomp,1,sum),PHP_confirmatory)
      BFtable[,1:7] <- exp(BFtable[,1:7])
      row.names(BFtable) <- names(BFtu_confirmatory)
      colnames(BFtable) <- c("complex=","complex>","fit=","fit>","BF=","BF>","BF","PHP")
      BFmatrix_confirmatory <- matrix(rep(BFtu_confirmatory,length(BFtu_confirmatory)),ncol=length(BFtu_confirmatory))-
        t(matrix(rep(BFtu_confirmatory,length(BFtu_confirmatory)),ncol=length(BFtu_confirmatory)))
      diag(BFmatrix_confirmatory) <- 0
      row.names(BFmatrix_confirmatory) <- colnames(BFmatrix_confirmatory) <- names(BFtu_confirmatory)
      if(nrow(relfit)==length(parse_hyp$original_hypothesis)){
        hypotheses <- parse_hyp$original_hypothesis
      }else{
        hypotheses <- c(parse_hyp$original_hypothesis,"complement")
      }

      if(logIN == FALSE){
        BFtu_confirmatory <- exp(BFtu_confirmatory)
        BFmatrix_confirmatory <- exp(BFmatrix_confirmatory)
      }
    }
  }else{
    BFtu_confirmatory <- PHP_confirmatory <- BFmatrix_confirmatory <- relfit <-
      relcomp <- hypotheses <- BFtable <- priorprobs <- NULL
  }

  BFcorr_out <- list(
    BFtu_exploratory=BFtu_exploratory,
    BF_Bartlett=BF_Bartlett,
    PHP_exploratory=PHP_exploratory,
    PHP_Bartlett=PHP_Bartlett,
    BFtu_confirmatory=BFtu_confirmatory,
    PHP_confirmatory=PHP_confirmatory,
    BFmatrix_confirmatory=BFmatrix_confirmatory,
    BFtable_confirmatory=BFtable,
    prior.hyp.explo=prior.hyp.explo,
    prior.hyp.conf=priorprobs,
    hypotheses=hypotheses,
    estimates=postestimates,
    model=x,
    bayesfactor=bayesfactor,
    parameter=testedparameter,
    log=logIN,
    call=match.call()
  )

  class(BFcorr_out) <- "BF"

  return(BFcorr_out)

}


#get draws from joint uniform prior in Fisher transformed space
#Call Fortran subroutine in from bct_prior.f90
draw_ju_r <- function(P, samsize=50000, Fisher=1){
  testm <- matrix(0,ncol=.5*P*(P-1),nrow=samsize)
  #  random1 <- rnorm(1)
  #  random1 <- (random1 - floor(random1))*1e6
  res <-.Fortran("draw_ju",P = as.integer(P),
                 drawscorr=testm,
                 samsize=as.integer(samsize),
                 numcorrgroup=as.integer(.5*P*(P-1)),
                 Fisher=as.integer(Fisher),
                 #seed=as.integer( sample.int(1e6,1) ),
                 PACKAGE="BFpack")
  return(res$drawscorr)

}


#' @title Bayesian correlation analysis
#'
#' @name cor_test
#'
#' @description Estimate the unconstrained posterior for the correlations using a joint uniform prior (Mulder and Gelissen,
#' 2023) or a marginally uniform prior (Barnard et al., 2000, Mulder, 2016). Correlation matrices are sampled from the posterior
#' using the MCMC algorithm of Talhouk et al. (2012).
#'
#' @param ...  matrices (or data frames) of dimensions \emph{n} (observations) by  \emph{p} (variables)
#' for different groups (in case of multiple matrices or data frames).
#'
#' @param formula an object of class \code{\link[stats]{formula}}. This allows for including
#' control variables in the model (e.g., \code{~ education}).
#'
#' @param iter total number of iterations from posterior. The total is split across three chains. By default, the total number of iterations is 10000,
#' implying a rounded integer of 3333 iterations per chain.
#'
#' @param burnin number of iterations for burnin (default is 2000).
#'
#' @param method \code{"LKJ"} for onion method sampler of correlation matrix proposed by Lewandowski, Kurowicka,
#'  & Joe (2009), which is the default, or \code{"LD"} for 'PX-RPMH' correlation matrix sampler of Liu and Daniels (2009).
#'  \code{"onion"} method tends to mix better for larger correlation matrices.
#'
#' @param nugget.scale When using \code{method = "LD"}, a nugget.scale can be used to avoid computational issues due to posterior draws
#' for the correlations too close to 1 in absolute value. Posterior draws for the correlations are multiplied with this nugget.scale.
#' So \code{nugget.scale} should be close to 1 (the default is .999). If the traceplots show that draws are stuck
#' at 1 or -1 too long try a slightly smaller \code{nugget.scale}.
#'
#' @return list of class \code{cor_test}:
#' \itemize{
#' \item \code{meanF} posterior means of Fisher transform correlations
#' \item \code{covmF} posterior covariance matrix of Fisher transformed correlations
#' \item \code{correstimates} posterior estimates of correlation coefficients
#' \item \code{corrdraws} list of posterior draws of correlation matrices per group
#' \item \code{corrnames} names of all correlations
#' }
#'
#' @references Barnard, J., McCulloch, R., & Meng, X. L. (2000). Modeling covariance matrices in terms of standard deviations and
#' correlations, with application to shrinkage. Statistica Sinica, 1281-1311. <https://www.jstor.org/stable/24306780>
#'
#' @references Joe, H. (2006). Generating random correlation matrices based on partial correlations. Journal of Multivariate
#' Analysis, 97(10), 2177-2189. <https://doi.org/10.1016/j.jmva.2005.05.010>
#'
#' @references Mulder, J., & Gelissen, J. P. (2023). Bayes factor testing of equality and order constraints on measures of
#' association in social research. Journal of Applied Statistics, 50(2), 315-351. <https://doi.org/10.1080/02664763.2021.1992360>
#'
#' @references Mulder, J. (2016). Bayes factors for testing order-constrained hypotheses on correlations. Journal of Mathematical
#' Psychology, 72, 104-115. <https://doi.org/10.1016/j.jmp.2014.09.004>
#'
#' @references Lewandowski, D., Kurowicka, D., & Joe, H. (2009). Generating random correlation matrices based on vines and extended
#' onion method. Journal of Multivariate Analysis, 100(9), 1989–2001. <https://doi.org/10.1016/j.jmva.2009.04.008>
#'
#' @references Liu, X., & Daniels, M. J. (2006). A new algorithm for simulating a correlation matrix based on parameter expansion and
#' reparameterization. Journal of Computational and Graphical Statistics, 15(4), 897-914. <https://doi.org/10.1198/106186006X160681>
#'
#' @examples
#' \donttest{
#' # Bayesian correlation analysis of the 6 variables in 'memory' object
#' # we consider a correlation analysis of the first three variable of the memory data.
#' fit <- cor_test(BFpack::memory[,1:3])
#'
#' # Bayesian correlation of variables in memory object in BFpack while controlling
#' # for the Cat variable
#' fit <- cor_test(BFpack::memory[,c(1:4)],formula = ~ Cat)
#'
#' # Example of Bayesian estimation of polyserial correlations
#' memory_example <- memory[,c("Im","Rat")]
#' memory_example$Rat <- as.ordered(memory_example$Rat)
#' fit <- cor_test(memory_example)
#'
#' # Bayesian correlation analysis of first three variables in memory data
#' # for two different groups
#' HC <- subset(BFpack::memory[,c(1:3,7)], Group == "HC")[,-4]
#' SZ <- subset(BFpack::memory[,c(1:3,7)], Group == "SZ")[,-4]
#' fit <- cor_test(HC,SZ)
#'
#' }
#' @rdname cor_test
#' @export
cor_test <- function(..., formula = NULL, iter = 1e4, burnin = 2e3, method = "LKJ", nugget.scale = .999){

  # if(is.na(prior.cor)){stop("'prior.cor' argument needs to be either 'joint.unif' or 'marg.unif'. See ?cor_test.")}
  # if(is.null(prior.cor)){stop("'prior.cor' argument needs to be either 'joint.unif' or 'marg.unif'. See ?cor_test.")}
  # if(!(prior.cor == "joint.unif" | prior.cor == "marg.unif")){
  #   stop("'prior.cor' argument needs to be either 'joint.unif' or 'marg.unif'. See ?cor_test.")}
  # if(prior.cor == "joint.unif"){
  #   priorchoice <- 1
  # }else{
  #   priorchoice <- 2
  # }
  prior.cor <- "joint.unif"
  priorchoice <- 1

  if(method=="LKJ"){
    methodnum <- 2
  }else if(method=="LD"){
    methodnum <- 1
  }else{
    warning("argument 'method' ill-specified. Default 'LKJ' method used.")
  }

  if(!is.numeric(nugget.scale)){stop("'nugget.scale' should be a numerical scalar.")}
  if(nugget.scale > 1 | nugget.scale < 0){stop("'nugget.scale' should be very close 1. If should not exceed 1 nor fall below 0.")}
  nugget.scale <- nugget.scale[1]

  Y_input <- list(...)
  numG <- length(Y_input)

  if(is.null(formula)){
    formula <- ~ 1
  }
  Xnames <- attr(terms(formula), "term.labels")
  whichDV <- lapply(Y_input,function(y){
    unlist(lapply(colnames(y),function(x){sum(x==Xnames)==0}))
  })
  if(numG>1){ #check that the same number of DVs are present in each group (that's how dimensions are coded)
    numDV <- rep(NA,numG)
    for(gg in 1:numG){
      numDV[gg] <- sum(whichDV[[gg]])
    }
    if(sum(abs(diff(numDV)))!=0){
      stop("Each group should contain same number of dependent variables.")
    }
  }

  #check measurement level of dependent variables
  P <- sum(whichDV[[1]])
  ordi <- numcats <- matrix(0,nrow=numG,ncol=P)
  Ylevel <- matrix(0,nrow=numG,ncol=P)
  Y_groups <- Y_input
  for(gg in 1:numG){
    teller <- 1
    for(pp in which(whichDV[[gg]])){
      if(class(Y_groups[[gg]][,pp])[1] == "numeric" | class(Y_groups[[gg]][,pp])[1] == "integer"){
        teller <- teller + 1
        Ylevel[gg,pp] <- "numeric"
      }else{
        if(class(Y_groups[[gg]][,pp])[1] == "ordered"){
          #levels(Y_groups[[gg]][,pp]) <- 1:length(levels(Y_groups[[gg]][,pp]))
          old_levels <- sort(as.numeric(unique(Y_groups[[gg]][,pp])))
          for(index_levels_g_p in 1:length(old_levels)){
            Y_groups[[gg]][which(Y_groups[[gg]][,pp] == old_levels[index_levels_g_p]),pp] <- index_levels_g_p
          }
          Y_groups[[gg]][,pp] <- as.numeric(Y_groups[[gg]][,pp])
          ordi[gg,teller] <- 1
          numcats[gg,teller] <- max(Y_groups[[gg]][,pp])
          Ylevel[gg,pp] <- "ordinal"
          if(numcats[gg,teller]==2){Ylevel[gg,pp] <- "dichotomous"}
          teller <- teller + 1
          if(max(Y_groups[[gg]][,pp])>11){
            stop("Ordinal variables are not allowed to have more than 11 categories")
          }
        }else{
          if(class(Y_groups[[gg]][,pp])[1] == "factor"){
            if(length(levels(Y_groups[[gg]][,pp]))==2){
              Ylevel[gg,pp] <- "dichotomous"
              levels(Y_groups[[gg]][,pp]) <- 1:length(levels(Y_groups[[gg]][,pp]))
              Y_groups[[gg]][,pp] <- as.numeric(Y_groups[[gg]][,pp])
              ordi[gg,teller] <- 1
              numcats[gg,teller] <- 2
              teller <- teller + 1
            }else{
              stop("Outcome variables should be either of class 'numeric', 'ordered', or a 2-level 'factor'.")
            }
          }else{
            stop("Outcome variables should be either of class 'numeric', 'ordered', or a 2-level 'factor'.")
          }
        }
      }
    }
    ordi[gg,] <- as.double(ordi[gg,])
    numcats[gg,] <- as.double(numcats[gg,])
  }
  if(max(numcats) == 1){
    stop("One categorical variable is constant.")
  }
  #because ordinal variables are not yet supported we set these indicators to '0'
  #ordi <- numcats <- matrix(0,nrow=numG,ncol=P)
  cor.type <- lapply(1:numG,function(g){matrix(NA,ncol=P,nrow=P)})
  for(gg in 1:numG){
    for(p1 in 2:P){
      for(p2 in 1:(p1-1)){
        if(Ylevel[gg,p1]=="numeric" & Ylevel[gg,p2]=="numeric"){
          cor.type[[gg]][p1,p2]<-"product-moment"
        }else{
          if(Ylevel[gg,p1]=="dichotomous" & Ylevel[gg,p2]=="dichotomous"){
            cor.type[[gg]][p1,p2]<-"tetrachoric"
          }else{
            if(Ylevel[gg,p1]=="ordinal" & Ylevel[gg,p2]=="ordinal"){
              cor.type[[gg]][p1,p2]<-"polychoric"
            }else{
              if((Ylevel[gg,p1]=="ordinal" & Ylevel[gg,p2]=="numeric") |
                 (Ylevel[gg,p1]=="numeric" & Ylevel[gg,p2]=="ordinal")){
                cor.type[[gg]][p1,p2]<-"polyserial"
              }else{
                if((Ylevel[gg,p1]=="ordinal" & Ylevel[gg,p2]=="dichotomous")|
                   (Ylevel[gg,p1]=="dichotomous" & Ylevel[gg,p2]=="ordinal")){
                  cor.type[[gg]][p1,p2]<-"tetrachoric"
                }else{
                  if((Ylevel[gg,p1]=="numeric" & Ylevel[gg,p2]=="dichotomous")|
                     (Ylevel[gg,p1]=="dichotomous" & Ylevel[gg,p2]=="numeric")){
                    cor.type[[gg]][p1,p2]<-"biserial"
                  }
                }
              }
            }
          }
        }
      }
    }
  }

  model_matrices <- lapply(seq_len(numG) , function(x) {
    model.matrix(formula, Y_groups[[x]])
  })

  correlate <- remove_predictors_helper(Y_groups = Y_groups, formula)

  YXlist <- lapply(1:length(model_matrices),function(g){
    list(as.matrix(correlate[[g]]),as.matrix(model_matrices[[g]]))
  })

  K <- ncol(YXlist[[1]][[2]])
  numcorr <- numG*P*(P-1)/2
  ngroups <- unlist(lapply(1:numG,function(g){nrow(YXlist[[g]][[1]])}))
  if(min(ngroups)<2*P+1){
    stop("sample size in each group must at least be equal to 2*P+1, with P the number of DVs.")
  }
  Ntot <- max(ngroups)
  Ygroups <- array(0,dim=c(numG,Ntot,P))
  Xgroups <- array(0,dim=c(numG,Ntot,K))
  XtXi <- array(0,dim=c(numG,K,K))
  BHat <- array(0,dim=c(numG,K,P))
  sdHat <- matrix(0,nrow=numG,ncol=P)
  CHat <- array(0,dim=c(numG,P,P))
  SumSq <- array(0,dim=c(numG,P,P))
  SumSqInv <- array(0,dim=c(numG,P,P))
  sdsd <- matrix(0,nrow=numG,ncol=P)

  for(g in 1:numG){
    Y_g <- YXlist[[g]][[1]]
    for(p in 1:P){
      if(ordi[g,p]==0){
        Y_g[,p] <- c(scale(Y_g[,p]))
      }
    }
    X_g <- YXlist[[g]][[2]]
    Ygroups[g,1:ngroups[g],] <- as.double(Y_g)
    #standardize data to get a more stable sampler for the correlations.
    tableX <- apply(X_g,2,table)
    catX <- unlist(lapply(1:length(tableX),function(xcol){
      length(tableX[[xcol]])
    }))
    if(sum(catX>1)){
      X_g[1:ngroups[g],which(catX>1)] <- apply(as.matrix(X_g[1:ngroups[g],which(catX>1)]),2,scale)
    }
    Xgroups[g,1:ngroups[g],] <- as.double(X_g)
    XtXi[g,,] <- as.double(solve(t(X_g)%*%X_g))
    BHat[g,,] <- as.double(XtXi[g,,]%*%t(X_g)%*%Y_g)
    SumSq[g,,] <- t(Y_g - X_g%*%BHat[g,,])%*%(Y_g - X_g%*%BHat[g,,])
    SumSqInv[g,,] <- solve(SumSq[g,,])
    Sigma_g <- SumSq[g,,]/ngroups[g]
    sdHat[g,] <- as.double(sqrt(diag(Sigma_g)))
    CHat[g,,] <- as.double(diag(1/sdHat[g,])%*%Sigma_g%*%diag(1/sdHat[g,]))
    #get rough estimate of posterior sd of the standard deviations (used for random walk sd)
    drawsSigma_g <- rWishart(1e2,df=ngroups[g],Sigma=SumSqInv[g,,])
    sdsd[g,] <- unlist(lapply(1:P,function(p){
      as.double(sd(sqrt(drawsSigma_g[p,p,])))
    }))
  }
  samsize0 <- round(iter / 3)
  gLiuSab <- array(as.double(0),dim=c(samsize0,numG,P))
  Njs <- matrix(as.double(ngroups),nrow=numG,ncol=1)

  # call Fortran subroutine for Gibbs sampling using noninformative improper priors
  # for regression coefficients, Jeffreys priors for standard deviations, and a proper
  # joint uniform prior for the correlation matrices.

  res <- lapply(1:3,function(chain){
    .Fortran("estimate_bct_ordinal",
                    postZmean=matrix(as.double(0),nrow=numcorr,ncol=1),
                    postZcov=matrix(as.double(0),nrow=numcorr,ncol=numcorr),
                    P=as.integer(P),
                    numcorr=as.integer(numcorr),
                    K=as.integer(K),
                    numG=as.integer(numG),
                    BHat=BHat,
                    sdHat=sdHat,
                    CHat=CHat,
                    XtXi=XtXi,
                    samsize0=as.integer(samsize0),
                    burnin=as.integer(burnin),
                    Ntot=as.integer(Ntot),
                    Njs_in=Njs,
                    Xgroups=Xgroups,
                    Ygroups=Ygroups,
                    C_quantiles=array(as.double(0),dim=c(numG,P,P,3)),
                    sigma_quantiles=array(as.double(0),dim=c(numG,P,3)),
                    B_quantiles=array(as.double(0),dim=c(numG,K,P,3)),
                    BDrawsStore=array(as.double(0),dim=c(samsize0,numG,K,P)),
                    sigmaDrawsStore=array(as.double(0),dim=c(samsize0,numG,P)),
                    CDrawsStore=array(as.double(0),dim=c(samsize0,numG,P,P)),
                    sdMH=sdsd,
                    ordinal_in=ordi,
                    Cat_in=numcats,
                    maxCat=as.integer(max(numcats)),
                    gLiuSab=gLiuSab,
                    nuggetscale=as.double(nugget.scale),
                    corrAlg=as.integer(methodnum)
    )
  })
  CDrawsStore <- array(NA,dim=c(length(res)*samsize0,numG,P,P))
  #BDrawsStore <- array(NA,dim=c(length(res)*samsize0,numG,K,P))
  #sigmaDrawsStore <- array(NA,dim=c(length(res)*samsize0,numG,P))
  for(chain in 1:length(res)){
    CDrawsStore[((chain-1)*samsize0+1):(chain*samsize0),,,] <- res[[chain]]$CDrawsStore
    #BDrawsStore[((chain-1)*samsize0+1):(chain*samsize0),,,] <- res[[chain]]$BDrawsStore
    #sigmaDrawsStore[((chain-1)*samsize0+1):(chain*samsize0),,] <- res[[chain]]$sigmaDrawsStore
  }
  samsize0 <- samsize0 * length(res)

  # compute Gelman-Rubin Rhat
  chain1_mcmc <- as.mcmc(FisherZ(do.call(cbind,lapply(1:numG,function(gr){
    do.call(cbind,lapply(1:(P-1),function(p2){
      res[[1]]$CDrawsStore[,gr,(p2+1):P,p2]
    }))
  }))))
  chain2_mcmc <- as.mcmc(FisherZ(do.call(cbind,lapply(1:numG,function(gr){
    do.call(cbind,lapply(1:(P-1),function(p2){
      res[[2]]$CDrawsStore[,gr,(p2+1):P,p2]
    }))
  }))))
  chain3_mcmc <- as.mcmc(FisherZ(do.call(cbind,lapply(1:numG,function(gr){
    do.call(cbind,lapply(1:(P-1),function(p2){
      res[[3]]$CDrawsStore[,gr,(p2+1):P,p2]
    }))
  }))))
  chains_list <- mcmc.list(chain1_mcmc, chain2_mcmc, chain3_mcmc)
  gelmanrubin_check <- gelman.diag(chains_list, autoburnin = FALSE)
  if(P==2 & numG==1){
    if(gelmanrubin_check$psrf[1] > 1.05){
      warning(paste0("Gelman-Rubin's Rhat is ",round(gelmanrubin_check$psrf[1],2),", which is larger than 1.05. Check
    the traceplot for possible convergence issues. To resolve,
    possibly use a sligthly smaller 'nugget.scale'."))
    }
  }else{
    if(gelmanrubin_check$mpsrf[1] > 1.05){
      warning(paste0("Gelman-Rubin's Rhat is ",round(gelmanrubin_check$mpsrf[1],2),", which is larger than 1.05. Check
    the traceplot for possible convergence issues. To resolve,
    possibly use a sligthly smaller 'nugget.scale'."))
    }
  }

  varnames <- lapply(1:numG,function(g){
    names(correlate[[g]])
  })
  corrnames <- lapply(1:numG,function(g){
    matrix(unlist(lapply(1:P,function(p2){
      unlist(lapply(1:P,function(p1){
        if(numG==1){
          paste0(varnames[[g]][p1],"_with_",varnames[[g]][p2])
        }else{
          paste0(varnames[[g]][p1],"_with_",varnames[[g]][p2],"_in_g",as.character(g))
        }
      }))
    })),nrow=P)
  })

  FmeansCovCorr <- lapply(1:numG,function(g){
    Fdraws_g <- FisherZ(t(matrix(unlist(lapply(1:samsize0,function(s){
      CDrawsStore[s,g,,][lower.tri(diag(P))]
    })),ncol=samsize0)))
    mean_g <- apply(Fdraws_g,2,median)
    names(mean_g) <- corrnames[[g]][lower.tri(diag(P))]
    #covm_g <- cov(Fdraws_g)

    #get robust estimate of posterior covariance matrix
    mlm1 <- lm(Fdraws_g ~ 1)
    covm_g <- sandwich(mlm1) * nrow(Fdraws_g)

    return(list(mean_g,covm_g))
  })

  meansCovCorr <- lapply(1:numG,function(g){
    matcor_g <- unlist(lapply(1:(P-1),function(p2){
      unlist(lapply((p2+1):P,function(p1){
        #mean(res$CDrawsStore[,g,p1,p2])
        mean(CDrawsStore[,g,p1,p2])
      }))
    }))
    names(matcor_g) <- corrnames[[g]][lower.tri(diag(P))]
    return(matcor_g)
  })

  meanN <- unlist(lapply(1:numG,function(g){
    FmeansCovCorr[[g]][[1]]
  }))
  covmN <- matrix(0,nrow=numcorr,ncol=numcorr)
  numcorrg <- numcorr/numG

  corrdraws <- lapply(1:numG,function(g){
    #array_g <- res$CDrawsStore[,g,,]
    array_g <- CDrawsStore[,g,,]
    dimnames(array_g) <- list(NULL,varnames[[g]],varnames[[g]])
    return(array_g)
  })

  for(g in 1:numG){
    covmN[(g-1)*numcorrg+1:numcorrg,(g-1)*numcorrg+1:numcorrg] <- FmeansCovCorr[[g]][[2]]
    colnames(cor.type[[g]]) <- row.names(cor.type[[g]]) <- varnames[[g]]
  }

  # posterior estimates
  postestimates_correlations <- Reduce(rbind,
                                       lapply(1:numG,function(g){
                                         means <- meansCovCorr[[g]]
                                         medians <- lb <- ub <- rep(NA,length(sum(lower.tri(diag(P)))))
                                         tel <- 1
                                         for(p2 in 1:(P-1)){
                                           for(p1 in (p2+1):P){
                                             medians[tel] <- quantile(CDrawsStore[,g,p1,p2],.5)
                                             lb[tel] <- quantile(CDrawsStore[,g,p1,p2],.025)
                                             ub[tel] <- quantile(CDrawsStore[,g,p1,p2],.975)
                                             tel <- tel + 1
                                           }
                                         }
                                         return(cbind(means,medians,lb,ub))
                                       }))
  colnames(postestimates_correlations) <- c("mean","median","2.5%","97.5%")
  row.names(gelmanrubin_check$psrf) <- row.names(postestimates_correlations)

  cor_out <- list(meanF=meanN,covmF=covmN,correstimates=postestimates_correlations,
                  corrdraws=corrdraws,corrnames=corrnames,variables=varnames,
                  cor.type=cor.type,res=res,prior.cor=prior.cor,Rhat_gelmanrubin=gelmanrubin_check)
  class(cor_out) <- "cor_test"

  return(cor_out)
}

set.seed(123)
#check results
test_that("correlation test on cor_test object with two groups correctly evaluated", {
  skip_on_cran()
  #split data frame in two data frame for each group
  memoryHC <- subset(memory,Group=="HC")[,-(4:7)]
  memorySZ <- subset(memory,Group=="SZ")[,-(4:7)]
  cor1 <- cor_test(memoryHC,memorySZ)
  BF6_cor <- BF(cor1, hypothesis=
    "Del_with_Im_in_g1 > Del_with_Im_in_g2 &
    Del_with_Wmn_in_g1 > Del_with_Wmn_in_g2 &
    Im_with_Wmn_in_g1 > Im_with_Wmn_in_g2")
  expect_equivalent(
    log(BF6_cor$BFmatrix_confirmatory[1,2]), 4.8, tolerance = .5
)})

# exploratory testing correlations in multivariate normal model
set.seed(123)
cor1 <- cor_test(mtcars[,4:6],iter = 1e3,burnin = 0)
BF1 <- BF(cor1,prior.hyp.explo = c(1,1,1),cov.prob=.99)
test_that("BF.cor_test use of cov.prob argument", {
  expect_equivalent(
    colnames(BF1$estimates)[3],"0.5%"
  )
  expect_equivalent(
    BF1$estimates[1,3],-0.73,tol=.05
  )
})
BF1a <- BF(cor1,prior.hyp.explo = 3:5)
PHPexplo <- matrix(
  c(0.09,  0.9,  0.01,
    0.0,  0.0,  1.0,
    0.0,  1.0,  0.0),nrow=3,byrow=T)
test_that("BF.cor_test exploratory hypotheses on correlations correctly evaluated", {
  expect_equivalent(
    round(BF1$PHP_exploratory,2),PHPexplo, tolerance = .1
  )
  expect_equivalent(
    BF1a$PHP_exploratory[1,],
    BF1$BFtu_exploratory[1,]*(3:5)/sum(BF1$BFtu_exploratory[1,]*(3:5)),
    tolerance = .02
  )
})
# confirmatory hypothesis test on the correlations
BF2 <- BF(cor1,hypothesis="wt_with_drat<drat_with_hp<wt_with_hp;
   wt_with_drat=drat_with_hp<0<wt_with_hp")
BFtable <- matrix(c(   0, 1.2, 5.9,
                      -1.2, 0, 4.6,
                    -5.9, -4.6, 0),byrow=TRUE,nrow=3)
test_that("BF.cor_test confirmatory hypotheses on correlations correctly evaluated", {
  expect_equivalent(
    round(log(BF2$BFmatrix_confirmatory),1),BFtable, tolerance = .5
  )})

BF2.log <- BF(cor1,hypothesis="wt_with_drat<drat_with_hp<wt_with_hp;
   wt_with_drat=drat_with_hp<0<wt_with_hp",log=TRUE)
test_that("log BF.cor_test confirmatory hypotheses on correlations correctly evaluated", {
  expect_equivalent(
    round(BF2.log$BFmatrix_confirmatory,1),BFtable, tolerance = .5
  )})

# test a single correlation
set.seed(123)
cor2 <- cor_test(mtcars[,1:2],burnin=1e2)
#print(cor2)
BF2 <- BF(cor2,hypothesis="cyl_with_mpg= -.9")
logBFexplo <- matrix(
  c(-19.7,  .69,  -23.3),nrow=1,byrow=T)
# exploratory hypothesis test on the correlation
test_that("BF.cor_test exploratory hypotheses on correlations correctly evaluated", {
  expect_equivalent(
    log(BF2$BFtu_exploratory),logBFexplo, tolerance = .3
  )})
# confirmatory hypothesis test on the correlation
test_that("BF.cor_test confirmatory hypotheses on correlations correctly evaluated", {
  expect_equivalent(
    log(BF2$BFmatrix_confirmatory[1,2]),2.4, tolerance = .2
  )})

# estimate correlations for unequal groups
set.seed(123)
cor2b <- cor_test(mtcars[1:10,2:4],mtcars[11:32,2:4])
#print(cor2b)
estimates_check <- c(.75,.7,.68,.87,.77,.71)
test_that("check estimates of correlations of multiple groups", {
  expect_equivalent(
    round(cor2b$correstimates[,1],2),estimates_check, tolerance = .1
  )})

# test a single correlation in multiple groups
test_that("BF.cor_test exploratory hypotheses on correlations correctly evaluated", {
  skip_on_cran()
  set.seed(123)
  cor3 <- cor_test(mtcars[,3:4],mtcars[,5:6])
  BF3 <- BF(cor3,hypothesis="hp_with_disp_in_g1= -wt_with_drat_in_g2")
  expect_equivalent(
    round(BF3$PHP_confirmatory,2),c(.77,.23), tolerance = .1
  )})

# test a single correlation on categorical outcomes
set.seed(123)
mtcars_test <- mtcars[,8:9]
mtcars_test[,2] <- as.factor(mtcars_test[,2])
mtcars_test[,1] <- as.factor(mtcars_test[,1])
cor2 <- cor_test(mtcars_test,burnin = 5e2,iter=3000,nugget.scale=.995)
#print(cor2)
test_that("check estimate of polychoric correlation", {
  expect_equivalent(
    round(cor2$correstimates[1,1],2),.23, tolerance = .1
  )})
BF2 <- BF(cor2,hypothesis="am_with_vs= .1",prior.hyp.explo = c(1,1,1))
PHPexplo <- matrix(
  c(.496,  .091,  .413),nrow=1,byrow=T)
# exploratory hypothesis test on the correlation
test_that("BF.cor_test exploratory hypotheses on correlations correctly evaluated", {
  expect_equivalent(
    round(BF2$PHP_exploratory,3),PHPexplo, tolerance = .1
  )})
# confirmatory hypothesis test on the correlation
test_that("BF.cor_test confirmatory hypotheses on correlations correctly evaluated", {
  expect_equivalent(
    round(log(BF2$BFmatrix_confirmatory[1,2]),2),.95, tolerance = .1
  )})

# test with combinations between continuous and ordinal (categorical) outcome variables
set.seed(187)
mtcars_test <- rbind(mtcars[,c(1,2,9,10)],mtcars[,c(1,2,9,10)])
mtcars_test[,2] <- as.ordered(mtcars_test[,2])
mtcars_test[,3] <- as.factor(mtcars_test[,3])
mtcars_test[,4] <- as.integer(mtcars_test[,4])
cor4 <- cor_test(mtcars_test,iter = 1e3,burnin = 3e2,nugget.scale = .98,method = "LD")
#print(cor4)
BF4 <- BF(cor4,log = TRUE)
test_that("BF.cor_test exploratory hypotheses on correlations mixed measurement levels", {
  expect_equivalent(
    round(BF4$BFtu_exploratory[,2],1),c(0.7,-13.4,-10.3,.7,.7,-18.9), tolerance = .1
  )})
cor4 <- cor_test(mtcars_test,iter = 1e3,burnin = 3e2,nugget.scale = .98,method = "LKJ")
#print(cor4)
BF4 <- BF(cor4,log = TRUE)
test_that("BF.cor_test exploratory hypotheses on correlations mixed measurement levels", {
  expect_equivalent(
    round(BF4$BFtu_exploratory[,2],1),c(0.7,-9.7,-10.2,.7,.7,-40.7), tolerance = 1
  )})
#

#test ordinal correlations multiple groups
set.seed(123)
group1 <- data.frame(cbind(round(runif(20)*2+1),rnorm(20)))
class(group1$X1) <- "ordered"
group2 <- data.frame(cbind(round(runif(14)*2+1),rnorm(14)))
class(group2$X1) <- "ordered"
cor4 <- cor_test(group1,group2,iter = 1e3,burnin = 3e2)
test_that("test ordinal correlations multiple groups", {
  expect_equivalent(
    round(cor4$meanF,2),c(-0.35,0.21), tolerance = .1
  )
})

# test marginally uniform prior
# set.seed(13)
# cor5 <- cor_test(group1,group2,iter = 1e3,burnin = 3e2, prior.cor = "marg.unif")
# BF5test <- BF(cor5)
# test_that("test ordinal correlations multiple groups, marginally uniform prior", {
#   expect_equivalent(
#     round(cor4$meanF,2),c(-0.34,0.20), tolerance = .1
#   )
#   expect_equivalent(
#     round(BF5test$BFtu_exploratory[,1],3),c(1.37,2.35), tolerance = .1
#   )
# })

test_that("correlation test for intervals", {
  skip_on_cran()
  set.seed(1)
  cor_test1 <- cor_test(mtcars[,c('vs','am')])
  BF1 <- BF(x=cor_test1,hypothesis = "am_with_vs = 0; -.1<am_with_vs<.1; .1<am_with_vs<.5; -.5<am_with_vs< -.1")
  expect_equivalent(
    sum(BF1$BFtable_confirmatory[-1,2]), 1, tolerance = .05
  )
  expect_equivalent(
    BF1$BFtu_confirmatory, c(2.9, 2.85, 3.09, .4, 0.05), tolerance = .2
  )
  set.seed(1)
  BF2 <- BF(x=cor_test1,hypothesis = "am_with_vs = 0; -.1<am_with_vs<.1; -.1<am_with_vs<.5")
  expect_equivalent(
    BF2$BFtu_confirmatory, c(2.9, 2.9, 3.03, 0.14), tolerance = .2
  )
})

#' @title Bartlett Test of Homogeneity of Variances
#' @description Performs Bartlett's test of the null that the
#' variances in each of the groups (samples) are the same.
#'
#'@details \code{x} must be a numeric data vector, and \code{g}
#'must be a vector or factor object of the same length as \code{x}
#'giving the group for the corresponding elements of \code{x}.
#'
#'@section Bain t_test:
#'In order to allow users to enjoy the functionality of bain with the familiar
#'stats-function \code{bartlett.test}, we have had to make minor changes to the
#'function \code{bartlett.test.default}. All rights to, and credit for, the
#'function \code{bartlett.test.default}
#'belong to the R Core Team, as indicated in the original license below.
#'We make no claims to copyright and incur no liability with regard to the
#'changes implemented in \code{bartlett_test}.
#'
#'This the original copyright notice by the R core team:
#'File src/library/stats/R/bartlett_test.R
#'Part of the R package, https://www.R-project.org
#'
#'Copyright (C) 1995-2015 The R Core Team
#'
#' This program is free software; you can redistribute it and/or modify
#' it under the terms of the GNU General Public License as published by
#' the Free Software Foundation; either version 2 of the License, or
#' (at your option) any later version.
#'
#' This program is distributed in the hope that it will be useful,
#' but WITHOUT ANY WARRANTY; without even the implied warranty of
#' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#' GNU General Public License for more details.
#'
#' A copy of the GNU General Public License is available at
#' https://www.R-project.org/Licenses/
#'
#'@aliases bartlett_test bartlett_test.default
#'@param x a numeric vector of data values, or a list of
#'numeric data vectors representing the respective samples,
#'or fitted linear model objects (inheriting from class "lm").
#'@param g a vector or factor object giving the group for
#'the corresponding elements of x. Ignored if x is a list.
#'@param ... further arguments to be passed to or from methods.
#'
#'@return A list with class \code{"bartlett_htest"} containing the following
#'components: \item{statistic}{Bartlett's K-squared test statistic.}
#'\item{parameter}{the degrees of freedom of the approximate chi-squared
#'distribution of the test statistic.}
#'\item{p.value}{the p-value of the test.} \item{conf.int}{a confidence
#'interval for the mean appropriate to the specified alternative hypothesis.}
#'\item{method}{the character string "Bartlett test of homogeneity of variances".}
#'\item{data.name}{a character string giving the names of the data.}
#'\item{vars}{the sample variances across groups (samples).}
#'\item{n}{the number of observations per group (sample)}
#'
#'@references Bartlett, M. S. (1937). Properties of sufficiency
#'and statistical tests. Proceedings of the Royal Society of
#'London Series A 160, 268–282. DOI: 10.1098/rspa.1937.0109.
#'
#'@examples
#'require(graphics)
#'
#'plot(count ~ spray, data = InsectSprays)
#'bartlett_test(InsectSprays$count, InsectSprays$spray)
#'
#' @rdname bartlett_test
#' @export
bartlett_test <- function(x, g, ...) UseMethod("bartlett_test", x)


#' @importFrom stats bartlett.test
#' @method bartlett_test default
#' @rdname bartlett_test
#' @export
bartlett_test.default <- function(x, g, ...){
  cl <- match.call()
  cl[[1]] <- as.name("bartlett.test")
  bart <- eval.parent(cl)
  vars <- tapply(x, g, var, na.rm = TRUE)
  n <- table(g)
  names(vars) <- names(n)
  class(bart) <- c("bartlett_htest", class(bart))
  bart$vars <- vars
  bart$n <- n
  bart
}

# require(stats)
# exists("bartlett.test.default") # false
# getS3method("bartlett.test", "default")

#' @importFrom stats rgamma qgamma rchisq
#' @method BF bartlett_htest
#' @export
BF.bartlett_htest <- function(x,
                           hypothesis = NULL,
                           prior.hyp.explo = NULL,
                           prior.hyp.conf = NULL,
                           prior.hyp = NULL,
                           complement = TRUE,
                           log = FALSE,
                           cov.prob = .95,
                           iter = 1e5,
                           ...) {

  if(!(cov.prob>0 & cov.prob<1)){
    stop("The argument 'cov.prob' is a coverage probability for the interval estimates that should lie between 0 and 1. The default is 0.95.")
  }
  CrI_LB <- (1 - cov.prob)/2
  CrI_UB <- 1 - (1 - cov.prob)/2

  get_est <- get_estimates(x)
  nsim <- iter
  s2 <- get_est$estimate
  n <- c(x$n)
  b <- 2/n
  J <- length(n)
  names_coef <- names(get_est$estimate)
  scale.post_group <- s2*(n-1)/2
  shape.post_group <- (n-1)/2
  postestimates <- cbind(scale.post_group/(shape.post_group-1) * (shape.post_group > 1),
                            1/qgamma(.5,      shape=shape.post_group,rate=scale.post_group),
                            1/qgamma(1-CrI_LB,shape=shape.post_group,rate=scale.post_group),
                            1/qgamma(1-CrI_UB,shape=shape.post_group,rate=scale.post_group))
  which.means <- which(shape.post_group>1)
  postestimates[which.means,1] <- scale.post_group[which.means]/(shape.post_group[which.means]-1)
  row.names(postestimates) <- names_coef
  colnames(postestimates) <- c("mean","median",paste0(as.character(round(CrI_LB*100,7)),"%"),
                               paste0(as.character(round(CrI_UB*100,7)),"%"))

  logIN <- log

  # check proper usage of argument 'prior.hyp.conf' and 'prior.hyp.explo'
  if(!is.null(prior.hyp.conf)){
    prior.hyp <- prior.hyp.conf
  }
  prior.hyp.explo <- process.prior.hyp.explo(prior_hyp_explo = prior.hyp.explo, model=x)

  # exploratory BF for equality of variances:
  logmx0 <- - 1 / 2 * sum((1 - b) * n) * log(pi) + 1 / 2 * log(prod(b)) +
    lgamma((sum(n) - J) / 2) - lgamma((sum(b * n) - J) / 2) -
    1 / 2 * (sum(n) - J) * log(sum((n - 1) * s2)) +
    1 / 2 * (sum(b * n) - J) * log(sum(b * (n - 1) * s2))
  logmxu <- - 1 / 2 * sum((1 - b) * n) * log(pi) + 1 / 2 * log(prod(b)) +
    sum(lgamma((n - 1) / 2) - lgamma((b * n - 1) / 2) -
          1 / 2 * (n - 1) * log((n - 1) * s2) +
          1 / 2 * (b * n - 1) * log(b * (n - 1) * s2))
  BF0u <- logmx0 - logmxu

  BFtu_exploratory <- c(BF0u,log(1))
  names(BFtu_exploratory) <- c("homogeneity of variances","no homogeneity of variances")
  norm_BF_explo <- exp(BFtu_exploratory - max(BFtu_exploratory)) * prior.hyp.explo[[1]]
  PHP_exploratory <- norm_BF_explo / sum(norm_BF_explo)

  if (!is.null(hypothesis)){
    parse_hyp <- parse_hypothesis(names_coef, hypothesis)
    parse_hyp$hyp_mat <- do.call(rbind, parse_hyp$hyp_mat)
    RrList <- make_RrList2(parse_hyp)
    RrE <- RrList[[1]]
    RrO <- RrList[[2]]
  }

  if (is.null(hypothesis)) {
    BFmatrix_confirmatory <- PHP_confirmatory <- BFtu_confirmatory <- relfit <-
      relcomp <- hypotheses <- BFtable <- priorprobs <- NULL
  } else if (all(unlist(lapply(append(RrE, RrO), is.null)))) {
    BFmatrix_confirmatory <- PHP_confirmatory <- BFtu_confirmatory <- relfit <-
      relcomp <- hypotheses <- BFtable <- priorprobs <- NULL
  } else { # execute confirmatory Bayes factor test based on hypothesis input
    # check if hypotheses are admissible:
    RrCheck <- do.call(rbind, append(RrE, RrO))
    RrCheck_count <- t(apply(RrCheck[, -ncol(RrCheck), drop = FALSE], 1,
                             function(x) {sapply(list(-1, 1), function(y) {sum(y == x)})}))
    if (any(RrCheck_count != 1) || any(RrCheck[, ncol(RrCheck)] != 0)) {
      stop(paste0("The hypotheses contain inadmissible constraints."))
    }

    Th <- length(RrE)
    logmx <- relfit <- relcomp <- logmxE <- rep(NA, times = Th)
    names(logmx) <- names(relfit) <- names(relcomp) <- names(logmxE) <-
      parse_hyp$original_hypothesis

    for (h in 1:Th) {

      if (is.null(RrE[[h]])) {
        unique_vars <- as.list(1:J)
      } else {
        RrEh <- RrE[[h]][, -ncol(RrE[[h]])]
        if (!is.matrix(RrEh)) {
          RrEh <- t(as.matrix(RrEh))
        }
        RrEh_pos <- t(apply(RrEh, 1, function(x) which(!(x == 0))))
        unique_vars <- list()
        rows <- 1:nrow(RrEh_pos)
        while (length(rows) > 0) {
          equal_vars <- RrEh_pos[min(rows), ]
          row_check <- min(rows)
          for (i in setdiff(rows, row_check)) {
            if (any(equal_vars %in% RrEh_pos[i, ])) {
              equal_vars <- unique(c(equal_vars, RrEh_pos[i, ]))
              row_check <- c(row_check, i)
            }
          }
          unique_vars <- c(unique_vars, list(equal_vars))
          rows <- setdiff(rows, row_check)
        }
        unique_vars <- c(unique_vars, setdiff(1:J, unlist(unique_vars)))
      }
      K <- length(unique_vars)
      Jk <- sapply(unique_vars, length)
      s2list <- lapply(unique_vars, function(x) s2[x])
      nlist  <- lapply(unique_vars, function(x) n[x])
      blist  <- lapply(unique_vars, function(x) b[x])
      df <- dfb <- SS <- SSb <- rep(NA, times = K)
      for (i in 1:K) {
        df[i]  <- sum(nlist[[i]]) - Jk[i]
        dfb[i] <- sum(blist[[i]] * nlist[[i]]) - Jk[i]
        SS[i]  <- sum((nlist[[i]] - 1) * s2list[[i]])
        SSb[i] <- sum(blist[[i]] * (nlist[[i]] - 1) * s2list[[i]])
      }
      logmxE[h] <- - 1 / 2 * sum((1 - unlist(blist)) * unlist(nlist)) * log(pi) +
        1 / 2 * log(prod(unlist(blist))) + sum(lgamma(df / 2) - lgamma(dfb / 2) -
                                                 1 / 2 * df * log(SS) + 1 / 2 * dfb * log(SSb))
      if (is.null(RrO[[h]])) {
        logmx[h] <- logmxE[h]
      } else {
        RrOh <- RrO[[h]][, -ncol(RrO[[h]])]
        if (!is.matrix(RrOh)) {
          RrOh <- t(as.matrix(RrOh))
        }
        RrOh_pos <- t(apply(RrOh, 1, function(x) c(which(x == -1), which(x == 1))))
        unique_vars_order <- cbind(
          apply(as.matrix(RrOh_pos[, 1]), 1, function(x) {
            which(unlist(lapply(unique_vars, function(y) {x %in% y})))}),
          apply(as.matrix(RrOh_pos[, 2]), 1, function(x) {
            which(unlist(lapply(unique_vars, function(y) {x %in% y})))})
        )
        post_samp <- prior_samp <- matrix(NA, nrow = nsim, ncol = K)
        indi_post <- indi_prior <- rep(1, times = nsim)
        for (i in unique(c(unique_vars_order))) {
          post_samp[, i] <- SS[i] / rchisq(nsim, df = df[i])
          prior_samp[, i] <- dfb[i] / rchisq(nsim, df = dfb[i])
        }
        for (i in 1:nrow(unique_vars_order)) {
          indi_post <- indi_post * (post_samp[, unique_vars_order[i, 1]] <
                                      post_samp[, unique_vars_order[i, 2]])
          indi_prior <- indi_prior * (prior_samp[, unique_vars_order[i, 1]] <
                                        prior_samp[, unique_vars_order[i, 2]])
        }
        relfit[h] <- sum(indi_post) / nsim
        relcomp[h] <- sum(indi_prior) / nsim
        logmx[h] <- log(relfit[h] / relcomp[h]) + logmxE[h]
      }
    }
    active.complement <- TRUE
    if(complement==TRUE){
      #compute marginal likelihood for complement hypothesis
      relfit <- inversegamma_prob_Hc(shape1=(n-1)/2,scale1=s2*(n-1)/(2*n),relmeas=relfit,RrE1=RrE,RrO1=RrO)
      relcomp <- inversegamma_prob_Hc(shape1=rep(.5,length(n)),scale1=rep(.5,length(n)),relmeas=relcomp,RrE1=RrE,RrO1=RrO)
      if(length(relfit)>Th){
        logmxE <- c(logmxE,logmxu)
        logmx <- c(logmx,logmxu + log(relfit[Th+1]/relcomp[Th+1]))
        names(logmx)[Th+1] <- "complement"
      }else{
        active.complement <- FALSE
      }
    }
    hypotheses <- names(logmx)
    BFtu_confirmatory <- (logmx - logmxu)
    BFmatrix_confirmatory <- BFtu_confirmatory %*% t(rep(1,length(BFtu_confirmatory))) -
      rep(1,length(BFtu_confirmatory)) %*% t(BFtu_confirmatory)
    diag(BFmatrix_confirmatory) <- log(1)
    names(BFtu_confirmatory) <- row.names(BFmatrix_confirmatory) <-
      colnames(BFmatrix_confirmatory) <- hypotheses

    if(is.null(prior.hyp)){
      priorprobs <- rep(1/length(BFtu_confirmatory),length(BFtu_confirmatory))
    }else{
      if(!is.numeric(prior.hyp) || length(prior.hyp)!=length(BFtu_confirmatory)){
        if(complement == TRUE && active.complement == FALSE && length(prior.hyp)==length(BFtu_confirmatory)+1){
          # the prior prob for the complement hypothesis is given but it is not used as the constrained hypotheses
          # cover the entire parameter space
          priorprobs <- prior.hyp[1:length(BFtu_confirmatory)]
          priorprobs <- priorprobs / sum(priorprobs)
        }else{
          warning(paste0("Argument 'prior.hyp' should be numeric and of length ",as.character(length(BFtu_confirmatory)),
                         ". Equal prior probabilities are used."))
          priorprobs <- rep(1/length(BFtu_confirmatory),length(BFtu_confirmatory))
        }
      }else{
        priorprobs <- prior.hyp
        priorprobs <- priorprobs / sum(priorprobs)
      }
    }

    PHP_confirmatory <- exp(BFtu_confirmatory - max(BFtu_confirmatory)) * priorprobs /
      sum(exp(BFtu_confirmatory - max(BFtu_confirmatory)) * priorprobs)
    relcomp[which(is.na(relcomp))] <- 1
    relfit[which(is.na(relfit))] <- 1
    BF_E <- exp(logmxE - logmxu)
    BFtable <- cbind(rep(NA,length(relfit)),relcomp,rep(NA,length(relfit)),relfit,BF_E,
                     relfit/relcomp,BF_E*relfit/relcomp,PHP_confirmatory)
    row.names(BFtable) <- names(PHP_confirmatory)
    colnames(BFtable) <- c("complex=","complex>","fit=","fit>","BF=","BF>","BF","PHP")

    if(logIN == FALSE){
      BFtu_confirmatory <- exp(BFtu_confirmatory)
      BFmatrix_confirmatory <- exp(BFmatrix_confirmatory)
    }
  }

  if(logIN == FALSE){
    BFtu_exploratory <- exp(BFtu_exploratory)
  }

  BFlm_out <- list(
    BFtu_exploratory=BFtu_exploratory,
    PHP_exploratory=PHP_exploratory,
    BFtu_confirmatory=BFtu_confirmatory,
    PHP_confirmatory=PHP_confirmatory,
    BFmatrix_confirmatory=BFmatrix_confirmatory,
    BFtable_confirmatory=BFtable,
    prior.hyp.explo=prior.hyp.explo,
    prior.hyp.conf=priorprobs,
    hypotheses=hypotheses,
    estimates=postestimates,
    model=x,
    bayesfactor="generalized adjusted fractional Bayes factors",
    parameter="group variances",
    log = logIN,
    call=match.call())

  class(BFlm_out) <- "BF"

  return(BFlm_out)
}


# The function computes the probability of the complement hypothesis
inversegamma_prob_Hc <- function(shape1,scale1,relmeas,RrE1,RrO1,samsize1=1e5){

  numhyp <- length(RrE1)
  whichO <- unlist(lapply(1:numhyp,function(h){is.null(RrE1[[h]])}))
  numO <- sum(whichO)
  numpara <- length(shape1)

  if(numO==length(RrE1)){ # Then the complement is equivalent to the unconstrained hypothesis.
    relmeas <- c(relmeas,1)
    names(relmeas)[numhyp+1] <- "complement"
  }else{ # So there is at least one hypothesis with only order constraints
    if(numO==1){ # There is one hypothesis with only order constraints. Hc is complement of this hypothesis.
      relmeas <- c(relmeas,1-relmeas[whichO])
      names(relmeas)[numhyp+1] <- "complement"
    }else{ # So more than one hypothesis with only order constraints
      randomDraws <- rmvnorm(samsize1,mean=rep(0,numpara),sigma=diag(numpara))
      #get draws that satisfy the constraints of the separate order constrained hypotheses
      checksOC <- lapply(which(whichO),function(h){
        Rorder <- as.matrix(RrO1[[h]][,-(1+numpara)])
        if(ncol(Rorder)==1){
          Rorder <- t(Rorder)
        }
        rorder <- as.matrix(RrO1[[h]][,1+numpara])
        apply(randomDraws%*%t(Rorder) > rep(1,samsize1)%*%t(rorder),1,prod)
      })
      checkOCplus <- Reduce("+",checksOC)

      if(sum(checkOCplus > 0) < samsize1){ #then the joint order constrained hypotheses do not completely cover the parameter space.
        if(sum(checkOCplus>1)==0){ # then order constrained spaces are nonoverlapping
          relmeas <- c(relmeas,1-sum(relmeas[whichO]))
          names(relmeas)[numhyp+1] <- "complement"
        }else{ #the order constrained subspaces at least partly overlap
          randomDraws <- matrix(unlist(lapply(1:numpara,function(par){
            1/rgamma(1e5,shape=shape1[par],rate=scale1[par])
            #rinvgamma(1e5,shape=shape1[par]/2,scale=scale1[par])
          })),ncol=numpara)
          checksOCpost <- lapply(which(whichO),function(h){
            Rorder <- as.matrix(RrO1[[h]][,-(1+numpara)])
            if(ncol(Rorder)==1){
              Rorder <- t(Rorder)
            }
            rorder <- as.matrix(RrO1[[h]][,1+numpara])
            apply(randomDraws%*%t(Rorder) > rep(1,samsize1)%*%t(rorder),1,prod)
          })
          relmeas <- c(relmeas,sum(Reduce("+",checksOCpost) == 0) / samsize1)
          rownames(relmeas)[numhyp+1] <- "complement"
        }
      }
    }
  }
  return(relmeas)
}

set.seed(57)
x <- c(rnorm(20), rnorm(20, sd = .765))
g <- factor(c(rep("V1", 20), rep("V2", 20)))
res_standard <- bartlett.test(x, g)
res <- bartlett_test(x, g)

test_that("bartlett_test same as bartlett.test", {
  expect_equivalent(res_standard, res[1:length(res_standard)])
})

#get_estimates(res)
out <- BF(res, hypothesis = "V1 > V2")

test_that("BF.bartlett returns correct results,", {
  expect_true(out$PHP_confirmatory[1] > .5 & out$PHP_confirmatory[2] < .5)
})

in_count <- c(10, 7, 20, 14, 14, 12, 10, 23, 17, 20, 14, 13, 11, 17, 21,
              11, 16, 14, 17, 17, 19, 21, 7, 13, 0, 1, 7, 2, 3, 1, 2, 1, 3,
              0, 1, 4, 3, 5, 12, 6, 4, 3, 5, 5, 5, 5, 2, 4, 3, 5, 3, 5, 3,
              6, 1, 1, 3, 2, 6, 4, 11, 9, 15, 22, 15, 16, 13, 10, 26, 26, 24,
              13)

in_spray <- structure(c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 2L,
                        2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 3L, 3L, 3L, 3L, 3L,
                        3L, 3L, 3L, 3L, 3L, 3L, 3L, 4L, 4L, 4L, 4L, 4L, 4L, 4L, 4L, 4L,
                        4L, 4L, 4L, 5L, 5L, 5L, 5L, 5L, 5L, 5L, 5L, 5L, 5L, 5L, 5L, 6L,
                        6L, 6L, 6L, 6L, 6L, 6L, 6L, 6L, 6L, 6L, 6L), .Label = c("A",
           "B", "C", "D", "E", "F"), class = "factor")

res_standard <- bartlett.test(in_count, in_spray)
res <- bartlett_test(in_count, in_spray)
test_that("bartlett_test same as bartlett.test", {
  expect_equivalent(res_standard, res[1:length(res_standard)])
})

out <- BF(res, "F> A=B>D>E=C")
test_that("BF.bartlett returns correct results,", {
  expect_equivalent(out$PHP_confirmatory, c(.9958, .00417), tolerance = .001)
})

set.seed(57)
out2 <- BF(res, "F> A=B>D>E=C", log = TRUE)
test_that("BF.bartlett returns correct results,", {
  expect_equivalent(out2$BFmatrix_confirmatory[2,1], -5.452534, tolerance = .01)
})

set.seed(57)
out2b <- BF(res, log = TRUE, prior.hyp.explo = c(1,2))
test_that("BF.bartlett returns correct results with prior.hyp.explo,", {
  expect_equivalent(out2b$PHP_exploratory,
                    exp(out2$BFtu_exploratory) * c(1,2) / sum(exp(out2$BFtu_exploratory) * c(1,2)),
                    tolerance = .01)
})

#test for variances
vtest1 <- bartlett_test(InsectSprays$count, InsectSprays$spray)
hypothesis <- "A=B=F>C=D=E"
set.seed(123)
BF1 <- BF(x=vtest1,hypothesis)
BF11 <- BF(x=vtest1,prior.hyp.explo = 3:4)

#check results exploratory test
test_that("BF.bartlett_htest exploratory hypotheses correctly evaluated", {
  expect_equivalent(
    round(BF1$PHP_exploratory,7),c(0.0044175,0.9955825)
  )
  expect_equivalent(
    unname(round(BF11$PHP_exploratory,6)),
    unname(round(BF1$BFtu_exploratory * (3:4) / sum(BF1$BFtu_exploratory * (3:4)),6))
  )
})
#check results confirmatory test
test_that("BF.bartlett_htest confirmatory hypotheses correctly evaluated", {
expect_equivalent(
  round(BF1$PHP_confirmatory,7),c(0.9911905,0.0088095)
)})

BF1a <- BF(x=vtest1,hypothesis,log=TRUE)
test_that("BF.bartlett_htest exploratory hypotheses correctly evaluated on log scale", {
  expect_equivalent(
    round(exp(BF1a$BFtu_exploratory),5),round(BF1$BFtu_exploratory,5)
  )})

hypothesis <- "A=B=F>C=D=E; A=B=F>C>D>E"
set.seed(123)
BF1 <- BF(x=vtest1,hypothesis,complement = F, log = TRUE)
#check results confirmatory test
test_that("BF.bartlett_htest confirmatory hypotheses correctly evaluated log(BF)", {
  expect_equivalent(
    round(BF1$BFtu_confirmatory,3),c(4.723,4.245)
  )})
test_that("BF.bartlett_htest confirmatory hypotheses correctly evaluated log(BF)", {
  expect_equivalent(
    round(BF1$BFtu_exploratory[1],4),-5.4178
  )})

bartlett <- bartlett_test(x = attention$accuracy, g = attention$group)
hypothesis <- c("Controls=TS<ADHD;
    Controls<TS=ADHD;
    Controls=TS=ADHD")
set.seed(358)
BF_var <- BF(bartlett, hypothesis)
#check posterior probability exploratory test
test_that("bartlett_test works for an exploratory test of equality of variances", {
expect_equivalent(
  round(BF_var$PHP_exploratory[1],7),0.8032805
)})

#check equality part equals NA
test_that("bartlett_test works for an exploratory test of equality of variances", {
expect_true(
  is.na(BF_var$BFtable_confirmatory[1,1])
)
})

#check names hypotheses confimatory test
test_that("names match of output for a bartlett_test", {
expect_equivalent(
  colnames(BF_var$BFmatrix_confirmatory),BF_var$hypotheses
)})
test_that("names match of output for a bartlett_test", {
expect_equivalent(
  row.names(BF_var$BFmatrix_confirmatory),BF_var$hypotheses
)})
test_that("bartlett_test works for a specific conformator test of equality of variances", {
expect_equivalent(
  round(BF_var$PHP_confirmatory,3),c(0.426,0.278,0.238,0.058)
)})

BF_var2 <- BF(bartlett, hypothesis, log = TRUE)
test_that("names match of output for a bartlett_test", {
  expect_equivalent(
    round(BF_var$BFmatrix_confirmatory,2),round(exp(BF_var2$BFmatrix_confirmatory),2),tol=.02
  )})

#BF method for coeftest class objects

#' @method BF coeftest
#' @export
BF.coeftest <- function(x,
                     hypothesis = NULL,
                     prior.hyp.explo = NULL,
                     prior.hyp.conf = NULL,
                     prior.hyp = NULL,
                     complement = TRUE,
                     log = FALSE,
                     cov.prob = 0.95,
                     ...){

  logIN <- log

  Sigma <- diag(x[, 2L]^2)
  n <- attr(x, "nobs")

  if(is.null(n)) stop("'BF.coeftest' only works if 'nobs.coeftest' gives the number of observations.")
  if(!is.null(hypothesis)) warning("constrained hypothesis testing is not supported for objects of class 'coeftest'")
  if(!is.null(prior.hyp)) warning("prior specification via 'prior.hyp' is not supported for objects of class 'coeftest'")
  #if(!exploratory) stop("only exploratory hypothesis testing is supported for objects of class 'coeftest'")

  out <- BF.default(x[, 1L], Sigma = Sigma, n = n, log = logIN, cov.prob = cov.prob, ...)
  out$model <- x
  out$call <- match.call()
  out

}

#' Memory data on health and schizophrenic patients
#'
#' Data set from study assessing differences between schizophrenic patients and
#' healthy control participants in patterns of correlations among 6 verbal memory
#' tasks (Ichinose et al., 2019).
#' \tabular{lll}{
#'    \strong{Im} \tab \code{numeric} \tab Percent correct on immediate recall of 3 word lists\cr
#'    \strong{Del} \tab \code{numeric} \tab Percent correct on delayed recall of 3 word lists\cr
#'    \strong{Wmn} \tab \code{numeric} \tab Number correct on letter-number span test of auditory working memory\cr
#'    \strong{Cat} \tab \code{numeric} \tab Number correct on category fluency task\cr
#'    \strong{Fas} \tab \code{numeric} \tab Number correct on letter fluency task\cr
#'    \strong{Rat} \tab \code{numeric} \tab Number correct on remote associates task\cr
#'    \strong{Group} \tab \code{factor} \tab Participant Group (HC = Healthy Control; SZ = Schizophrenia) \cr
#' }
#' @docType data
#' @keywords datasets
#' @name memory
#' @usage data(memory)
#' @references Ichinose, M.C., Han, G., Polyn, S., Park, S., & Tomarken, A.J. (2019). Verbal memory performance discordance in schizophrenia: A reflection of cognitive dysconnectivity.
#' Unpublished manuscript.
#' @format A data.frame with 40 rows and 8 columns.
NULL

#' 
#' Data come from an experimental study (Rosa, Rosa, Sarner, and Barrett, 1998) 
#' that were also used in Howell (2012, p.196).
#' An experiment was conducted to investigate if Therapeutic Touch practitioners 
#' who were blindfolded can effectively identify which of their hands is below the experimenter¡¯s.
#' Twenty-eight practitioners were involved and tested 10 times in the experiment. 
#' Researchers expected an average of 5 correct answers from each practitioner 
#' as it is the number by chance if they do not outperform others.
#' 
#' \tabular{lll}{
#'    \strong{correct} \tab \code{integer} \tab How many correct answers are from each practitioner)\cr
#' }
#' @docType data
#' @keywords datasets
#' @name therapeutic
#' @usage data(therapeutic)
#' @references Howell, D. (2012). Statistical methods for psychology (8th ed.). Belmont, CA: Cengage Learning.
#' @format A data.frame with 22 rows and 1 column.
NULL

#' Same location event statistic
#'
#' A matrix coding whether senders of events (in the rows) and receivers of events
#' (in the column) have the same location. Related to the 'events' data object,
#' that contains a relational event sequence, and the 'actors' object, that contains
#' information on the 25 actors involved in the relational event sequence.
#'
#'
#' @name same_location
#' @docType data
#' @usage data(same_location)
#' @format dataframe (25 rows, 4 columns)
#'
#' \tabular{lll}{
#'    \strong{same_location} \tab \code{integer} \tab Event statistic. Matrix with senders in the
#'    rows and receivers in the columns. The event statistic is 1 if sender and receiver have
#'    the same location and 0 otherwise. \cr
#' }
#'
#' @keywords datasets
NULL

#' Facial trustworthiness and criminal sentencing
#'
#' Data from a correlational study in which the correlation between ratings of
#' facial trustworthiness of inmates was correlated with whether they had
#' received the death penalty or not (wilson and Rule, 2015). These data were
#' simulated using the R-package \code{synthpop}, based on the characteristics
#' of the original data.
#'
#'
#'
#' \tabular{lll}{
#'    \strong{stim} \tab \code{integer} \tab Stimulus Number\cr
#'    \strong{sent} \tab \code{integer} \tab Sentence: 1 = Death, 0 = Life\cr
#'    \strong{race} \tab \code{integer} \tab Race: 1 = White, -1 = Black\cr
#'    \strong{glasses} \tab \code{integer} \tab Glasses: 1 = Yes, 0 = No\cr
#'    \strong{tattoos} \tab \code{integer} \tab Tattoos: 1 = Yes, 0 = No \cr
#'    \strong{ztrust} \tab \code{numeric} \tab Trustworthiness \cr
#'    \strong{trust_2nd} \tab \code{numeric} \tab Trustworthiness ratings with 2nd control group; Death targets are same as in primary analysis, Life targets are different.\cr
#'    \strong{afro} \tab \code{numeric} \tab raw Afrocentricity ratings.\cr
#'    \strong{zAfro} \tab \code{numeric} \tab Afrocentricity ratings normalized within target race. Analyses in paper were done with this variable.\cr
#'    \strong{attract} \tab \code{numeric} \tab Attractiveness\cr
#'    \strong{fWHR} \tab \code{numeric} \tab facial width-to-height \cr
#'    \strong{afWHR} \tab \code{numeric} \tab fWHR normalized within target race. Analyses in paper were done with this variable \cr
#'    \strong{maturity} \tab \code{numeric} \tab Maturity
#' }
#' @docType data
#' @keywords datasets
#' @name wilson
#' @usage data(wilson)
#' @references Wilson, J. P., & Rule, N. O. (2015). Facial Trustworthiness
#'   Predicts Extreme Criminal-Sentencing Outcomes. Psychological Science,
#'   26(8), 1325–1331. doi: 10.1177/0956797615590992
#' @format A data.frame with 742 rows and 13 columns.
NULL

survreg1 <- survival::survreg(survival::Surv(futime, fustat) ~ ecog.ps + rx, survival::ovarian,
                              dist='weibull', scale=1)
BF1 <- BF(survreg1)
test_that("BF.survreg exploratory hypotheses correctly evaluated", {
  expect_equivalent(
    round(BF1$BFtu_exploratory[,1],5),c(0.00000,3.88378,3.12164)
  )})

library(testthat)
constraints.fmri2 <- "Face_on_Deep = Face_on_Superficial = Face_on_Middle < 0; Face_on_Deep < Face_on_Superficial = Face_on_Middle < 0"


# Missing data ------------------------------------------------------------

fmri_missing <- fmri
set.seed(123)
for(i in 1:6){
  fmri_missing[sample(1:nrow(fmri), 1), sample(1:ncol(fmri), 1)] <- NA
}

fmri_listdel <- fmri_missing[!is.na(apply(fmri_missing, 1, sum)),]
fmri.lm2_listdel <- lm(cbind(Superficial, Middle, Deep) ~ Face + Vehicle, data = fmri_listdel)

res <- BF(fmri.lm2_listdel, hypothesis = constraints.fmri2)
test_that("BF.fmri with missing data runs", {
  expect_true(res$PHP_confirmatory[1]>res$PHP_confirmatory[3]& res$PHP_confirmatory[1]<res$PHP_confirmatory[2]&res$PHP_confirmatory[3]<res$PHP_confirmatory[2])
})


# Without missing data ----------------------------------------------------

fmri_missing <- fmri
fmri_listdel <- fmri_missing[!is.na(apply(fmri_missing, 1, sum)),]
fmri.lm2_listdel <- lm(cbind(Superficial, Middle, Deep) ~ Face + Vehicle, data = fmri_listdel)

res <- BF(fmri.lm2_listdel, hypothesis = constraints.fmri2)
test_that("BF.fmri without missing data works", {
  expect_true(res$PHP_confirmatory[1]>res$PHP_confirmatory[3]& res$PHP_confirmatory[1]<res$PHP_confirmatory[2]&res$PHP_confirmatory[3]<res$PHP_confirmatory[1])
})

aov1 <- aov(price ~ anchor * motivation, data = tvprices)
#names(aov1$coefficients)
BF1 <- BF(aov1, BF.type = "AFBF")
#check number of tests in exploratory ANOVA
test_that("BF.lm works for an exploratory ANOVA test regarding number of tests", {
  expect_equivalent(
    dim(BF1$BFtu_exploratory),c(4,3)
  )
  expect_equivalent(
    dim(BF1$PHP_main),c(2,2)
  )
  expect_equivalent(
    dim(BF1$PHP_interaction),c(1,2)
  )
})
test_that("BF.lm works for a specific exploratory ANOVA test", {
expect_equivalent(
  round(c(BF1$PHP_main[,1],BF1$PHP_interaction[1]),3),c(0,0,.251)
)})

BF2 <- BF(aov1,hypothesis="anchorrounded = anchorrounded:motivationlow < motivationlow;
          (anchorrounded, anchorrounded:motivationlow) < 0 & motivationlow < 0",
          log = TRUE, BF.type = "AFBF")
test_that("log(BF.aov) confirmatory ANOVA test", {
  expect_equivalent(
  round(unname(BF2$BFtu_confirmatory),3),c(-40.2, -25.8, 0.043),tol=.1
)})

vi <- 1:10 #squared s.e.'s
yi <- 1:10 #sample means

### Fit a random-effects model to the data
res <- metafor::rma(yi = yi, vi = vi)

#check results exploratory test
test_that("metafor exploratory test", {
  set.seed(123)
  BF1 <- BF(res,BF.type="correlation",iter=1e3,cov.prob=.9)
  expect_equivalent(
    round(BF1$BFtu_exploratory[,3],3),c(1.9,1.9), tolerance = .05
  )
  expect_equivalent(
    round(BF1$estimates[1:2,3],2),c(-.27,.34), tolerance = .5
  )
})

options(contrasts = c("contr.treatment", "contr.poly"))
house.plr <- MASS::polr(Sat ~ Infl + Type + Cont, weights = Freq, data = MASS::housing)
house.plr

BF1 <- BF(house.plr,hypothesis="Low|Medium<0<Medium|High & TypeApartment=TypeAtrium")

test_that("BF.polr exploratory hypotheses correctly evaluated", {
  expect_equivalent(
    BF1$PHP_confirmatory,c(0.977,0.023)
  )})

library(survival)
# Create the simplest test data set
test1 <- list(time=c(4,3,1,1,2,2,3),
              status=c(1,1,1,0,1,1,0),
              x=c(0,2,1,1,1,0,0),
              sex=c(0,0,0,0,1,1,1))
# Fit a stratified model
fit <- coxph(Surv(time, status) ~ x + strata(sex), test1)

out <- BF(fit, "x > .5")

test_that("BF.coxph returns correct results,", {
  expect_true(out$PHP_confirmatory[1] > .5 & out$PHP_confirmatory[2] < .5)
})


# Create a simple data set for a time-dependent model
test2 <- list(start=c(1,2,5,2,1,7,3,4,8,8),
              stop=c(2,3,6,7,8,9,9,9,14,17),
              event=c(1,1,1,1,1,1,1,0,0,0),
              pred=c(1,0,0,1,0,1,1,1,0,0))
fit <- coxph(Surv(start, stop, event) ~ pred, test2)

out <- BF(fit, hypothesis = "pred > .5; pred = .5")
test_that("BF.coxph returns correct results,", {
  expect_true(out$PHP_confirmatory[1] < .5 & out$PHP_confirmatory[2] > .5)
})

out <- BF(fit, hypothesis = "pred > .5; pred = .5",complement = FALSE)
test_that("BF.coxph returns correct results,", {
  expect_equivalent(
    out$PHP_confirmatory, c(0.194,0.806), tol = .1
  )
})

# test for BF on an ergm object

#check if confirmatory test are the same for ergm object
test_that("BF.ergm tests", {
  skip_on_cran()
  library(ergm)
  seed <- 123
  # florentine data
  data(florentine)
  ergm_fit <- ergm(flomarriage ~ edges +
                     kstar(2) +
                     absdiff("wealth"),
                   control = control.ergm(seed = seed))
  get_estimates(ergm_fit)
  seed <- 123
  BFergm.test <- BF(ergm_fit,
                    hypothesis = "0 = absdiff.wealth > kstar2",
                    main.iters = 500, prior.hyp.explo = c(1,1,1))
  expect_true(
    all.equal(c(0.185,0.815),
              unname(BFergm.test$PHP_confirmatory), tolerance = .2)
  )
  expect_true(
    all.equal(c(0.1,0,0.9),
              unname(BFergm.test$PHP_exploratory[2,]), tolerance = .2)
  )
  seed <- 123
  BFergm.test2 <- BF(ergm_fit,
                     hypothesis = "0 = absdiff.wealth > kstar2",
                     main.iters = 500, prior.hyp.explo = 1:3)
  expect_equivalent(
    unname(BFergm.test2$PHP_exploratory[1,]),
    unname(BFergm.test2$BFtu_exploratory[1,]*(1:3)/sum(BFergm.test2$BFtu_exploratory[1,]*(1:3))),
    tol=.05
  )
})

# same test with bergm
#check if confirmatory test are the same
test_that("BF.bergm one hypotheses correctly evaluated", {
  skip_on_cran()
  # example analysis
  library(Bergm)
  library(ergm)
  data(florentine)
  set.seed(222)
  bergm_fit <- bergm(flomarriage ~ kstar(2) + edges + absdiff("wealth"),
                     seed = 1,main.iters = 500)
  BFbergm.test <- BF(bergm_fit,hypothesis = "0 = theta3 > theta1",main.iters = 500)
  expect_true(
    all.equal(0.17,
              unname(BFbergm.test$PHP_confirmatory)[1], tolerance = .2)
  )
})

fm_zip <- pscl::zeroinfl(art ~ . | 1, data = pscl::bioChemists)
set.seed(752)
BF1 <- BF(fm_zip,prior.hyp.explo=c(1,1,1))
test_that("BF.zeroinfl exploratory hypotheses correctly evaluated", {
  expect_equivalent(
    BF1$PHP_exploratory[,1],c(0.000,0.006,0.674,0.006,0.938,0.000,0.000),tol=.001
)})

set.seed(123)
n <- 40
data.example <- data.frame(group1 = round(runif(n)*2), group2 = round(runif(n)), dv1 = rnorm(n), dv2 = rnorm(n,sd=.5),cov=round(rnorm(n,sd=.5)))
data.example$group1 <- as.factor(data.example$group1)
data.example$group2 <- as.factor(data.example$group2)

m1 <- aov(dv1 ~ 1 + group1, data.example)
BF1 <- BF(m1,prior.hyp.explo = list(c(1,1,1),c(1,1),c(1,1)))
m2 <- aov(dv1 ~ -1 + group1, data.example)
BF2 <- BF(m2)
test_that("BF.aov exploratory ANOVA (1 factor) when including/excluding the intercept", {
  expect_equivalent(
    BF1$BFtu_main,BF2$BFtu_main
  )})

m3 <- aov(dv1 ~ 1 + group1 + group2, data.example)
BF3 <- BF(m3)
m4 <- aov(dv1 ~ -1 + group1 + group2, data.example)
BF4 <- BF(m4)
test_that("BF.aov exploratory ANOVA (2 factors) when including/excluding the intercept", {
  expect_equivalent(
    BF3$BFtu_main,BF4$BFtu_main
  )})

test_that("BF.manova exploratory MANOVA (1 factor)", {
  skip_on_cran()
  m5 <- manova(cbind(dv1,dv2) ~ 1 + group1, data.example)
  m6 <- manova(cbind(dv1,dv2) ~ -1 + group1, data.example)
  set.seed(123)
  BF5 <- BF(m5)
  set.seed(123)
  BF6 <- BF(m6)
  expect_equivalent(
    BF5$BFtu_main,BF6$BFtu_main
  )
})

test_that("BF.manova exploratory MANOVA (2 factors)", {
  skip_on_cran()
  m5 <- manova(cbind(dv1,dv2) ~ 1 + group1 + group2, data.example)
  m6 <- manova(cbind(dv1,dv2) ~ -1 + group1 + group2, data.example)
  set.seed(123)
  BF5 <- BF(m5)
  set.seed(123)
  BF6 <- BF(m6)
  expect_equivalent(
    BF5$BFtu_main,BF6$BFtu_main
  )
})

m7 <- aov(dv1 ~ 1 + group1 * group2, data.example)
BF7 <- BF(m7)
BF7a <- BF(m7,prior.hyp.explo = list(1:3,2:1,3:4))
m8 <- aov(dv1 ~ -1 + group1 * group2, data.example)
BF8 <- BF(m8)
test_that("BF.aov exploratory ANOVA (2 factors w/ interaction) when including/excluding the intercept", {
  expect_equivalent(
    BF7$BFtu_main,BF8$BFtu_main
  )
  expect_equivalent(
    BF7$PHP_interaction,BF8$PHP_interaction
  )
  expect_equivalent(
    BF7a$PHP_exploratory[4,],
    BF7$BFtu_exploratory[4,] * (1:3) / sum(BF7a$BFtu_exploratory[4,] * (1:3))
  )
  expect_equivalent(
    unname(BF7a$PHP_main[2,]),
    unname(BF7$BFtu_main[2,] * (2:1) / sum(BF7$BFtu_main[2,] * (2:1)))
  )
  expect_equivalent(
    unname(BF7a$PHP_interaction[1,]),
    unname(BF7$PHP_interaction[1,] * (3:4) / sum(BF7$PHP_interaction[1,] * (3:4)))
  )
})

test_that("BF.manova exploratory MANOVA (2 factors w/ interaction)", {
  skip_on_cran()
  m5 <- manova(cbind(dv1,dv2) ~ 1 + group1 * group2, data.example)
  m6 <- manova(cbind(dv1,dv2) ~ -1 + group1 * group2, data.example)
  set.seed(123)
  BF5 <- BF(m5)
  set.seed(123)
  BF6 <- BF(m6)
  expect_equivalent(
    BF5$BFtu_main,BF6$BFtu_main
  )
  expect_equivalent(
    BF5$PHP_interaction,BF6$PHP_interaction
  )
})

test_that("BF.manova exploratory MANOVA (2 factors w/ interaction and numeric covariate, group identifier)", {
  skip_on_cran()
  m5 <- manova(cbind(dv1,dv2) ~ 1 + group1 * group2, data.example)
  m6 <- manova(cbind(dv1,dv2) ~ -1 + group1 * group2 + cov, data.example)
  set.seed(123)
  BF5 <- BF(m5)
  set.seed(123)
  BF6 <- BF(m6)
  expect_equivalent(
    BF5$fraction_group_identifier,BF6$fraction_group_identifier
  )
})

# test BF for t test

#####################
# ONE SAMPLE T TEST #
#####################

# test if confirmatory test gives same result exploratory test
ttest1 <- t_test(therapeutic,mu=5)
BF1 <- BF(x=ttest1,hypothesis="mu=5;mu<5",prior.hyp.conf=c(1/2,1/4,1/4),BF.type = "FBF")
test_that("1 sample t test of multiple hypotheses correctly evaluated", {
  expect_true(
    all.equal(unname(c(BF1$PHP_exploratory)),unname(BF1$PHP_confirmatory))
)})

ttest1 <- t_test(therapeutic,mu=5)
BF1 <- BF(x=ttest1,hypothesis="mu=5;mu<5",complement=FALSE, BF.type = "AFBF")
test_that("1 sample t test of multiple hypotheses correctly evaluated", {
  expect_equal(
    c(0.3523759,0.6476241),unname(BF1$PHP_confirmatory),tol=.01
  )})
BF1b <- BF(x=ttest1,hypothesis="mu=5;mu<5",complement=FALSE,prior.hyp.explo=1:3, BF.type = "AFBF")
test_that("1 sample t test of multiple hypotheses correctly evaluated", {
  expect_equal(
    unname(BF1$BFtu_exploratory * (1:3) / sum(BF1$BFtu_exploratory * (1:3))),
    unname(BF1b$PHP_exploratory),
    tol=.01
  )})
ttest1 <- t_test(therapeutic,mu=5)

BF1 <- BF(x=ttest1,hypothesis="mu=5;mu<5",complement=FALSE,log=TRUE, BF.type = "AFBF")
test_that("test log(BF) t test", {
  expect_equal(
    c(0.6086118),BF1$BFmatrix_confirmatory[2,1],tol=.01
  )})
BF1 <- BF(x=ttest1,hypothesis="mu=5; mu<6 & mu>4",complement=TRUE,log=FALSE, BF.type = "AFBF",prior.hyp.conf = c(2,1,1))
#summary(BF1)
#BF1$BFtable_confirmatory

# test if one-sided PMP is same as one-sided p-value
ttest2 <- t_test(therapeutic,mu=5,alternative="less")
BF2 <- BF(x=ttest2,hypothesis="mu>5", BF.type = "AFBF")
test_that("1 sample t test of one-sided hypotheses correctly evaluated", {
expect_equivalent(
  ttest2$p.value,c(BF2$PHP_confirmatory)[1]
)})

######################################
# TWO SAMPLES T TEST EQUAL VARIANCES #
######################################

# check if posterior model probabilities are correct
ttest3 <- t_test(therapeutic,therapeutic*.9+.1,var.equal=TRUE)
BF3 <- BF(ttest3,prior.hyp.explo=c(1,1,1), BF.type = "AFBF")
test_that("2 samples t test of exploratory hypotheses correctly evaluated
          with equal variances", {
  expect_equivalent(
    c(unname(BF3$PHP_exploratory)),c(0.767913,0.04941605,0.1826709)
    ,tolerance = .00001)
})
# check if posterior model probabilities are correct
ttest3a <- t_test(x=therapeutic$correct,y=therapeutic$correct*.9+.1,paired=TRUE)
BF3a <- BF(ttest3a,prior.hyp.explo=c(1,1,1), BF.type = "AFBF",hypothesis="difference=0.3;difference<0.3")
test_that("2 samples t test of exploratory hypotheses correctly evaluated
          with equal variances", {
            expect_equivalent(
              round(c(unname(BF3a$PHP_confirmatory)),2),c(0.59,0.04,0.36)
              ,tolerance = .00001)
          })

ttest3a <- t_test(x=therapeutic$correct,y=therapeutic$correct*.9+.1,paired=TRUE)
BF3a <- BF(ttest3a,prior.hyp.explo=c(1,1,1), BF.type = "FBF",hypothesis="difference < 1;difference > -1")
#BF3a$BFtable_confirmatory

# t test check for testing interval hypotheses
set.seed(123)
ttest3 <- t_test(therapeutic,therapeutic*runif(length(therapeutic),min=.9,max=1.1)+.1,var.equal=TRUE)
BF3 <- BF(ttest3,hypothesis="difference< 0.5 & difference > -0.5; difference > 0.5; difference < -0.5",
          BF.type = "FBF",log = TRUE)
BF3b <- BF(ttest3,BF.type = "FBF",log = TRUE,prior.hyp.explo = 3:1)
test_that("2 samples t test of exploratory hypotheses correctly evaluated
          with equal variances", {
            expect_equivalent(
              c(unname(BF3$BFmatrix_confirmatory[1,])),c(0,3.5,4.1)
              ,tolerance = 1)
            expect_equivalent(
              BF3b$PHP_exploratory,
              exp(BF3$BFtu_exploratory)*(3:1)/sum(exp(BF3$BFtu_exploratory) * (3:1)),
              tolerance = 1)
          })

# test if one-sided PMP is same as one-sided p-value
ttest4 <- t_test(therapeutic,therapeutic*.9+.1,var.equal=TRUE,alternative="greater")
BF4 <- BF(ttest4,"difference<0", BF.type = "AFBF")
test_that("2 samples t test of one-sided hypotheses correctly evaluated
          with equal variances", {
  expect_equivalent(
    ttest4$p.value,c(BF4$PHP_confirmatory)[1]
)})


########################################
# TWO SAMPLES T TEST UNEQUAL VARIANCES #
########################################

# check posterior probabilities for a given data set
test_that("2 samples t test of two-sided hypotheses correctly evaluated
          with unequal variances", {
  skip_on_cran()
  ttest5 <- t_test(therapeutic,therapeutic*.7+2.5,"two.sided",var.equal=FALSE)
  set.seed(123)
  BF5 <- BF(ttest5,hypothesis="difference=0;difference<0",prior.hyp.explo=c(1,1,1))
  set.seed(123)
  BF5a <- BF(ttest5,prior.hyp.explo=0:2)
  expect_equivalent(
    c(unname(BF5$PHP_exploratory)),c(unname(BF5$PHP_confirmatory)),
    tolerance = .05)
  expect_equivalent(
    unname(BF5a$PHP_exploratory),
    unname(BF5$BFtu_exploratory * (0:2) / sum(BF5$BFtu_exploratory * (0:2))),
    tolerance = .05)

  BF5b <- BF(ttest5,hypothesis="difference=0; difference> -1 & difference<1; difference< -1; difference>1",
             BF.type="FBF",log=TRUE)
  expect_equivalent(
    length(unname(BF5b$PHP_confirmatory)),4
    )
  expect_equivalent(
    round(unname(BF5b$PHP_confirmatory),4),c(0.0464, 0.5373, 0.4163, 0.0000), tol=.05
  )
  BF5c <- BF(ttest5,hypothesis="difference=0; difference> -1 & difference<1; difference< -1",BF.type="FBF",log=TRUE)
  expect_equivalent(
    BF5c$BFtu_confirmatory[4],BF5b$BFtu_confirmatory[4], tol=.05
  )
})

# test BF on lm object

lm1 <-  lm(wt ~ disp + drat + hp, mtcars)
BF1a <- BF(lm1,BF.type = "AFBF",prior.hyp.explo = c(1,1,1))
BF1aa <- BF(lm1,BF.type = "AFBF",prior.hyp.explo = 1:3)
#check exploratory test
PHPexplo <- matrix(
  c(0.065,  0.003,  0.931,
  0.001,  0.000,  0.999,
  0.612,  0.342,  0.046,
  0.742,  0.178,  0.080),nrow=4,byrow=T)
test_that("BF.lm exploratory hypotheses correctly evaluated", {
  expect_equivalent(
    PHPexplo,round(BF1a$PHP_exploratory,3)
  )
  expect_equivalent(
    unname(round(BF1aa$PHP_exploratory[3,],3)),
           unname(round(BF1a$BFtu_exploratory[3,] * c(1:3) / sum(BF1a$BFtu_exploratory[3,] * c(1:3)),3))
  )
  expect_equivalent(
    unname(round(BF1aa$PHP_exploratory[4,],3)),
    unname(round(BF1a$BFtu_exploratory[4,] * c(1:3) / sum(BF1a$BFtu_exploratory[4,] * c(1:3)),3))
  )
})

BF1b <- BF(lm1,BF.type = "FBF",prior.hyp.explo=c(1,1,1))
PHPexplo <- matrix(
  c(0.107,  0.006,  0.887,
    0.002,  0.000,  0.998,
    0.646,  0.300,  0.054,
    0.748,  0.168,  0.085),nrow=4,byrow=T)
test_that("BF.lm exploratory hypotheses correctly evaluated", {
  expect_equivalent(
    PHPexplo,round(BF1b$PHP_exploratory,3)
  )})


BF2 <- BF(lm1,hypothesis="disp=drat=0;disp>drat>0;disp>drat=0",BF.type="AFBF")
test_that("BF.lm multiple confirmatory hypotheses correctly evaluated", {
  expect_equivalent(
    round(BF2$PHP_confirmatory,5),c(0.000,0.533,0.402,0.064), tolerance = .01
)})

BF2 <- BF(lm1,hypothesis="disp=drat=0;disp>drat>0;disp>drat=0",complement=FALSE,BF.type="AFBF")
test_that("BF.lm multiple confirmatory hypotheses correctly evaluated", {
  expect_equivalent(
    round(BF2$PHP_confirmatory,5),c(0.00000,0.57003,0.42997)
  )})

BF3 <- BF(lm1,hypothesis="disp=hp=0",BF.type="AFBF")
test_that("BF.lm one equality hypothesis correctly evaluated", {
  expect_equivalent(
    round(BF3$PHP_confirmatory,5),c(0.00003,0.99997)
  )})

BF4 <- BF(lm1,hypothesis="drat<hp<disp",BF.type="AFBF")
test_that("BF.lm one order hypothesis correctly evaluated", {
  expect_equivalent(
    round(BF4$PHP_confirmatory,5),c(0.96872,0.03128)
  )})

BF5 <- BF(lm1,hypothesis="drat<hp=disp",BF.type="FBF")
test_that("BF.lm one equal/order hypothesis correctly evaluated", {
  expect_equivalent(
    round(BF5$PHP_confirmatory,5),c(0.44049,0.55951)
  )})

BF5b <- BF(lm1,hypothesis="drat<hp=disp",BF.type = "FBF")
test_that("BF.lm one equal/order hypothesis correctly evaluated", {
  expect_equivalent(
    round(BF5b$PHP_confirmatory,5),c(0.44049,0.55951)
  )})

test_that("BF.glm with family=gaussian is the same as BF.lm", {
  glm1 <-  glm(wt ~ disp + drat + hp, data=mtcars, family="gaussian")
  BF.glm1 <- BF(glm1)
  lm1 <-  glm(wt ~ disp + drat + hp, data=mtcars)
  BF.lm1 <- BF(lm1)
  expect_equivalent(
    round(BF.glm1$PHP_exploratory[,2],3),round(BF.lm1$PHP_exploratory[,2],3)
  )
})

#several tests if output is the same
set.seed(36)
# testing coefficients in multivariate normal model
lm1 <- lm(cbind(mpg,cyl,hp) ~ disp + wt, data = mtcars)
BF1 <- BF(lm1, BF.type = "AFBF",prior.hyp.explo=c(1,1,1))
PHPexplo <- matrix(
  c(0.000,  0.000,  1.000,
  0.373,  0.604,  0.023,
  0.087,  0.908,  0.004,
  0.000,  0.000,  1.000,
  0.000,  0.000,  1.000,
  0.741,  0.178,  0.082,
  0.284,  0.017,  0.700,
  0.007,  0.000,  0.993,
  0.697,  0.239,  0.064),nrow=9,byrow=T)
test_that("BF.mlm exploratory correctly evaluated for given data sets", {
  expect_equivalent(
    round(BF1$PHP_exploratory,3),PHPexplo, tolerance = .03
  )})

# tests on same predictor on different DVs
set.seed(123)
BF2 <- BF(x=lm1,hypothesis="disp_on_mpg>disp_on_cyl>disp_on_hp>0; disp_on_mpg=disp_on_cyl=disp_on_hp=0")
test_that("BF.mlm two hypotheses for same IVs correctly evaluated", {
  expect_equivalent(
    round(BF2$PHP_confirmatory,3),c(0.010,0.000,0.988), tolerance = .015
)})
# tests on different predictors on same DVs
set.seed(574)
BF3 <- BF(lm1,hypothesis="disp_on_mpg>wt_on_mpg;disp_on_mpg=wt_on_mpg;disp_on_mpg<wt_on_mpg",
          BF.type="AFBF")
test_that("BF.mlm two hypotheses on same DVs correctly evaluated", {
expect_equivalent(
  round(BF3$PHP_confirmatory,3),c(0.902,0.094,0.005)
)})
BF3.log <- BF(lm1,hypothesis="disp_on_mpg>wt_on_mpg;disp_on_mpg=wt_on_mpg;disp_on_mpg<wt_on_mpg",log=TRUE,
              BF.type="AFBF")
test_that("BF.mlm two hypotheses on same DVs correctly evaluated", {
  expect_equivalent(
    round(BF3.log$BFtu_exploratory,3),round(log(BF3$BFtu_exploratory),3),tol=.01
  )})
test_that("BF.mlm two hypotheses different DVs and different IVs correctly evaluated", {
  skip_on_cran()
  # tests on different predictors on different DVs
  set.seed(4768)
  BF4 <- BF(lm1,hypothesis="disp_on_mpg<wt_on_cyl & disp_on_cyl<wt_on_hp; disp_on_mpg=wt_on_cyl & disp_on_cyl=wt_on_hp",
            log = TRUE, BF.type="AFBF")
  expect_equivalent(
    round(BF4$PHP_confirmatory,3),c(0.018,0.919,0.062), tolerance = .005
  )
  expect_equivalent(
    BF4$PHP_confirmatory[1]/BF4$PHP_confirmatory[2],exp(BF4$BFtu_confirmatory[1] -
            BF4$BFtu_confirmatory[2]), tolerance = .01
  )
})

# Frequentist one sample t test at the null point mu=5
ttest1 <- t_test(therapeutic,alternative="greater",mu=5)
# one-sided Bayesian one sample t test
BF1 <- BF(ttest1,hypothesis="mu<5",prior.hyp=c(.5,.5), BF.type = "AFBF")
test_that("bartlett_htest works for a specific one-sided test", {
  expect_equivalent(
  round(ttest1$p.value,3),round(BF1$PHP_confirmatory[1],3)
)})
BF1b <- BF(ttest1,hypothesis="mu=5;mu<5",prior.hyp.explo=c(1,1,1))
test_that("bartlett_htest works the same for confirmatory and explotory setting", {
  expect_equivalent(
    round(BF1b$PHP_exploratory,3),round(BF1b$PHP_confirmatory,3)
  )})

# confirmatory Bayesian one sample t test
BF2 <- BF(ttest1,hypothesis="mu=5;mu>5",prior.hyp=c(.5,.5,0), BF.type = "AFBF")
#check prior output
test_that("bartlett_htest works for a multiple hypothesis test", {
expect_equivalent(
  BF2$prior.hyp.conf,c(.5,.5,0)
)})
#check posterior probabilities
test_that("bartlett_htest works for a multiple hypothesis test", {
expect_equivalent(
  round(BF2$PHP_confirmatory,3),c(0.943,0.057,0.000)
)
})

test_that("t_test prior correct in output", {expect_equivalent(
  BF2$prior.hyp.conf,c(.5,.5,0)
)})
#check posterior probabilities
test_that("t_test PHP correct", {expect_equivalent(
  round(BF2$PHP_confirmatory,3),c(0.943,0.057,0.000)
)})

#Test glm class object
counts <- c(18,17,15,20,10,20,25,13,12)
outcome <- gl(3,1,9)
treatment <- gl(3,3)
glm.D93 <- glm(counts ~ treatment, family = poisson())
set.seed(432)
BF1 <- BF(x=glm.D93, hypothesis = "treatment2 = 0; treatment2 < 0",prior.hyp.explo=c(1,1,1))
#check if confirmatory and exploratory test are the same for Gaussian estimator
test_that("BF.glm two hypotheses correctly evaluated", {
expect_true(
  all.equal(unname(BF1$PHP_exploratory[2,]),
          unname(BF1$PHP_confirmatory), tolerance = .005)
)})

BF1.log <- BF(x=glm.D93, hypothesis = "treatment2 = 0; treatment2 < 0", log = TRUE)
#check if confirmatory and exploratory test are the same for Gaussian estimator
test_that("BF.glm two hypotheses correctly evaluated", {
  expect_true(
    all.equal(BF1$BFmatrix_confirmatory,
              exp(BF1.log$BFmatrix_confirmatory), tolerance = .005)
  )})

# test for correlation test on hetcor object
set.seed(54)
res <- polycor::hetcor(fmri[,3:5])
#BF1 <- BF(res, hypothesis = "Deep_with_Superficial > Middle_with_Superficial")

BF2 <- BF(res,hypothesis="(Superficial_with_Middle,Deep_with_Superficial,Deep_with_Middle) > 0;
          Middle_with_Superficial=Deep_with_Superficial=Deep_with_Middle= 0",prior.hyp.explo=c(1,1,1))
PHPexplo <- matrix(c(0.51,  0.21,  0.27,
0.50,  0.17,  0.33,
0.51,  0.22,  0.26),nrow=3,byrow=T)
test_that("Hetcor exploratory BF gives correct result", {expect_equivalent(
  BF2$PHP_exploratory, PHPexplo, tolerance = .01)})
set.seed(463)

test_that("Hetcor two hypotheses correctly evaluated", {
  expect_equivalent(
    unname(BF2$BFtu_confirmatory),
    c(1.63,10.2,0.91),
    tolerance = .05
  )}
)

set.seed(164)
BF5 <- BF(res,hypothesis="Middle_with_Superficial = Deep_with_Superficial > 0")
test_that("Hetcor one hypothesis with equality and order constraint correctly evaluated", {
  expect_equivalent(
    unname(BF5$PHP_confirmatory),c(0.75,0.24), tolerance = .1
  )})


set.seed(564)
res1 <- polycor::hetcor(fmri[,3:4])
BF3 <- BF(res1,hypothesis="Superficial_with_Middle > .4;Superficial_with_Middle = .4; Superficial_with_Middle < .4")
test_that("Hetcor test BF3 (automatically omit complement)", {
  expect_equivalent(
    log(unname(BF3$BFtu_confirmatory)),c(-1.2,.38,.27), tolerance = .5
  )})

#check results confirmatory test
test_that("lmerMod two hypotheses correctly evaluated", {
  skip_on_cran()
  library(lme4)
  timssICC_subset <- timssICC[(timssICC$groupNL11==1)+(timssICC$groupHR11==1)>0,
                              ][c(1:150,1395+1:150),]
  outlme1 <- lme4::lmer(math ~ -1 + lln +
                          groupNL11 + (0+groupNL11 | schoolID) +
                          groupHR11 + (0+groupHR11 | schoolID),
                        data=timssICC_subset)
  set.seed(123)
  BFicc <- BF(outlme1,hypothesis=
                "groupNL11<groupHR11;
               groupNL11=groupHR11",log=TRUE)
  expect_equivalent(
    round(BFicc$PHP_confirmatory,7),c(0.7,0.25,0.03), tolerance = .1
  )
  expect_equivalent(
    round(BFicc$BFmatrix_confirmatory[,1],7),c(0,-.93,-3), tolerance = .2
  )
  set.seed(123)
  BFicc2 <- BF(outlme1,log=TRUE,prior.hyp.explo = 3:5)
  expect_equivalent(
    round(BFicc2$PHP_exploratory[1,],7),
    round(exp(BFicc2$BFtu_exploratory[1,])*(3:5)/sum(exp(BFicc2$BFtu_exploratory[1,])*(3:5)),7),
    tolerance = .2
  )
})



#this is a smaller subset of the data with unbalanced groups
#check results exploratory test
test_that("lmerMod exploratory unbalanced", {
  skip_on_cran()
  library(lme4)
  timssICC_subset <- timssICC[timssICC$groupNL11==1,][c(1:15,15+1:12,30+1:8,45+1:4,60+1:9),]
  outlme2 <- lme4::lmer(math ~ -1 + gender +
                          (0+groupNL11 | schoolID),
                        data=timssICC_subset)
  set.seed(123)
  BFicc2 <- BF(outlme2)
  expect_equivalent(
    round(BFicc2$BFtu_exploratory,3),c(0,0,1.07), tolerance = .1
  )
})

fit <- glm(sent ~ ztrust + zfWHR + zAfro + glasses + attract + maturity +
             tattoos, family = binomial(), data = wilson)
set.seed(123)
BF_glm <- BF(fit, hypothesis="ztrust > zfWHR > zAfro > 0;
             ztrust > 0 & zfWHR = zAfro = 0",prior.hyp.explo=c(1,1,1))
#check results exploratory test
test_that("glm two hypotheses correctly evaluated via exploratory test", {
expect_equivalent(
  round(BF_glm$PHP_exploratory[,3],3),c(0.133,1.000,0.999,0.004,0.278,0.029,0.011,0.202)
)})
#check results conformatory test
test_that("glm two hypotheses correctly evaluated via confirmatory test", {
  expect_equivalent(
  BF_glm$PHP_confirmatory,c(0.142, 0.002, 0.856), tolerance = .02
)})

#check results conformatory test
BF_glm2 <- BF(fit, hypothesis="ztrust > zfWHR > zAfro > 0;
             ztrust > 0 & zfWHR = zAfro = 0",log=TRUE)
test_that("glm two hypotheses correctly evaluated via confirmatory test", {
  expect_equivalent(
    log(BF_glm$BFmatrix_confirmatory[,1]),BF_glm2$BFmatrix_confirmatory[,1], tolerance = .02
  )})

#check equal prior probabilities
test_that("glm use correct prior probabilities", {
expect_true(
  all.equal(unname(BF_glm$prior.hyp.conf),rep(1,3)/3)
)})

#check that complement is not added the complete parameter space is covered by hypotheses
BF2 <- BF(x=fit, hypothesis="ztrust = zfWHR; ztrust > zfWHR; ztrust < zfWHR")
test_that("glm does not add a complement hypothesis as the test is exhaustive", {
expect_true(
  length(BF2$PHP_confirmatory)==3
)})

# test for BF on rma.uni object
library(metafor)

# ### Generate data
tau2 <- 0.05 # True between-study variance

set.seed(123)
vi <- runif(50, min = 0.01, max = 0.2) # Observed sampling variances
yi <- rnorm(50, mean = 0, sd = sqrt(vi+tau2)) # Observed effect sizes

test_that("exploratory metafor random effects model", {
  skip_on_cran()
  ### Fit a random-effects model to the data
  res <- rma(yi = yi, vi = vi)
  set.seed(123)
  BFmeta1 <- BF(res,BF.type="stand.effect",cov.prob=.9)
  set.seed(123)
  BFmeta1a <- BF(res,BF.type="stand.effect",prior.hyp.explo = c(5:7))
  set.seed(123)
  BFmeta1b <- BF(res,BF.type=metaBMA::prior("norm", c(mean = 0, sd = .5)))
  expect_equivalent(
    round(BFmeta1$PHP_exploratory[1,],3),c(.950,.021,.029), tolerance = .05
  )
  expect_equivalent(
    round(BFmeta1b$PHP_exploratory[1,],3),c(.906,.040,.055), tolerance = .05
  )
  expect_equivalent(
    unname(round(BFmeta1a$PHP_exploratory[2,],3)),
    unname(round(BFmeta1$BFtu_exploratory[2,]*(5:7)/sum(BFmeta1$BFtu_exploratory[2,]*(5:7)),3)),
    tolerance = .05
  )
})

# test fixed effects meta model
test_that("exploratory metafor fixed effects model", {
  set.seed(123)
  res <- rma(yi = yi, vi = vi, method = "EE")
  BFmeta2 <- BF(res,BF.type="stand.effect")
  expect_equivalent(
    round(BFmeta2$PHP_exploratory,3),c(.964,0.019,0.017), tolerance = .05
  )
  set.seed(123)
  res$ni <- rep(10,50)
  BFmeta2b <- BF(res,BF.type="unit.info")
  expect_equivalent(
    round(BFmeta2b$PHP_exploratory,3),c(.957,0.023,0.022), tolerance = .05
  )
})

# Test class coxph with Gaussian_estimator
test2 <- list(start=c(1, 2, 5, 2, 1, 7, 3, 4, 8, 8),
              stop =c(2, 3, 6, 7, 8, 9, 9, 9,14,17),
              event=c(1, 1, 1, 1, 1, 1, 1, 0, 0, 0),
              x    =c(1, 0, 0, 1, 0, 1, 1, 1, 0, 0) )
coxph1 <- survival::coxph(survival::Surv(start, stop, event) ~ x, test2)
BF1 <- BF(coxph1,prior.hyp.explo = c(1,1,1))
test_that("BF.coxph exploratory hypotheses correctly evaluated", {
  expect_true(
    all.equal(round(c(unname(BF1$PHP_exploratory)),3),c(0.569,0.220,0.211))
)})
BF1log <- BF(coxph1,log = TRUE, prior.hyp.explo = c(1,1,1))
test_that("BF.coxph exploratory hypotheses correctly evaluated", {
  expect_true(
    all.equal(exp(c(BF1log$BFtu_exploratory)),c(BF1$BFtu_exploratory))
  )})
test_that("BF.coxph exploratory hypotheses correctly evaluated", {
  expect_true(
    all.equal(c(unname(BF1$PHP_exploratory)),c(unname(BF1log$PHP_exploratory)))
  )})

---
title: "Introduction to BFpack"
author: "Mulder, J., Williams, D. R., Gu, X., Tomarken, A., Boeing-Messing, F., Olsson-Collentine, A., Meijerink, M., Menke, J., van Aert, R., Fox, J.-P., Hoijtink, H., Rosseel, Y., Wagenmakers, E.-J., and van Lissa, C."
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{BFpack_introduction}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```


## Introduction

`BFpack` contains a collection of functions for Bayesian hypothesis testing using Bayes factors and posterior probabilities in R. The main function `BF` needs a fitted model `x` as input argument. Depending on the class of the fitted model, a standard hypothesis test is executed by default. For example, if `x` is a fitted regression model of class `lm` then posterior probabilities are computed of whether each separate coefficient is zero, negative, or positive (assuming equal prior probabilities). If one has specific hypotheses with equality and/or order constraints on the parameters under the fitted model `x` then these can be formulated using the `hypothesis` argument (a character string), possibly together prior probabilities for the hypotheses via the `prior.hyp` argument (default all hypotheses are equally likely a priori), and the `complement` argument which is a logical stating whether the complement hypotheses should be included in the case (`TRUE` by default).

Alternatively, when the model of interest is not of a class that is currently supported, `x` can also be a named numeric vector containing the estimates of the model parameters of interest, together with the error covariance matrix in the argument `Sigma`, and the sample size used to obtain the estimates, to perform an approximate Bayes factor test using large sample theory.

## Reference

The key references for the package are

Mulder, J., Williams, D. R., Gu, X., Tomarken, A., Boeing-Messing, F., Olsson-Collentine, A., Meijerink, M., Menke, J., van Aert, R., Fox, J.-P., Hoijtink, H., Rosseel, Y., Wagenmakers, E.-J., and van Lissa, C. (2021). BFpack: Flexible Bayes Factor Testing of Scientific Theories in R. *Journal of Statistical Software*. <https://www.jstatsoft.org/article/view/v100i18>

Mulder, J., van Lissa, C., Gu, X., Olsson-Collentine, A., Boeing-Messing, F., Williams, D. R., Fox, J.-P., Menke, J., et al. (2021). BFpack: Flexible Bayes Factor Testing of Scientific Expectations. (Version 0.3.2) <https://CRAN.R-project.org/package=BFpack>

## Usage

`BF(x, hypothesis, prior.hyp = NULL, complement = TRUE, ...)`


## Arguments

* `x`, a fitted model object that is obtained using a R-function. The object can be obtained via the following R functions:
  + `t_test` for t testing,
  + `bartlett_test` for testing independent group variances,
  + `aov` for AN(C)OVA testing,
  + `manova` for MAN(C)OVA testing,
  + `lm` for linear regresssion analysis, 
  + `cor_test` for correlation analysis,
  + `lmer` currently for testing intraclass correlations in random intercept models,
  + `glm` for generalized linear models,
  + `coxph` for survival analysis,
  + `survreg` for survival analysis,
  + `polr` for ordinal regression,
  + `zeroinfl` for zero-inflated regression,
  + `rma` for meta-analysis,
  + `ergm` or `bergm` for an exponential random graph,
  + `x` can also be a named vector with estimates of the key parameters.

* `hypothesis`, a character string specifying the hypotheses with equality and/or order constraints on the key parameters of interest.
  + By default `hypothesis = NULL` which executes exploratory hypothesis tests (examples below).
  + The parameter names are based on the names of the estimated key parameters. An overview of the key parameters is given using the function `get_estimates`, e.g., `get_estimates(model1),` where `model1` is a fitted model object.
  + Separate constraints within a hypothesis are separated with an ampersand `&`. Hypotheses are separated using a semi-colon `;`. For example `hypothesis = "weight > height & height > 0; weight = height = 0"` implies that the first hypothesis assumes that the parameter `weight` is larger than the parameter `height` and that the parameter `height` is positive, and the second hypothesis assumes that the two parameters are equal to zero. Note that the first hypothesis could equivalently have been written as `weight > height > 0`.
* `prior.hyp.explo`, a numeric vector specifying the prior probabilities of the hypotheses in the exploratory tests. In the default setting, the null hypothesis has a prior probability of `0.50`, and each of the two one-sided hypotheses both have a prior probability of `0.25`.
* `prior.hyp.conf`, a numeric vector specifying the prior probabilities of the hypotheses of the `hypothesis` argument. The default setting is `prior.hyp = NULL` which sets equal prior probabilities.
* `complement`, a logical value which specified if a complement hypothesis is included in the tested hypotheses specified under `hypothesis`. The default setting is `TRUE`. The complement hypothesis covers the remaining parameters space that is not covered by the constrained hypotheses. For example, if an equality hypothesis and an order hypothesis are formulated, say, `hypothesis = "weight = height = length; weight > height > length"`, the complement hypothesis covers the remaining subspace where neither `"weight = height = length"` holds, nor `"weight > height > length"` holds.
* `log`, a logical specifying whether the Bayes factors should be computed on a log scale. Default is `FALSE`.

Alternatively if one is interested in testing hypotheses under a model class which that is currently not supported, an approximate Bayesian test can be executed with the following (additional) arguments

* `x`, a named numeric vector of the estimates (e.g., MLE) of the parameters of interest where the labels are equal to the names of the parameters which are used for the `hypothesis` argument.
* `Sigma`, the approximate posterior covariance matrix (e.g,. error covariance matrix) of the parameters of interest.
* `n`, the sample size that was used to acquire the estimates and covariance matrix.


## Output

The output is of class `BF`. By running the `print` function on the `BF` object, a short overview of the results are presented. By running the `summary` function on the `BF` object, a comprehensive overview of the results are presented. 


## Example analyses

### Normal linear models (t-testing, linear regression, (M)AN(C)OVA)

For an object of class `t_test` (t testing), `lm` (linear regression), `lml` (multivariate linear regression), `aov` (AN(C)OVA), `manova` (MAN(C)OVA), the evidence and posterior probabilities are computed using the fractional Bayes factor (FBF) methodology [O'Hagan (1995) ](https://rss.onlinelibrary.wiley.com/doi/abs/10.1111/j.2517-6161.1995.tb02017.x). Manual prior specification is avoided by using a minimal fraction of the data to construct a default (fractional) prior while the remaining fraction is used for hypothesis testing. For a simple t-test, there are 2 unknown parameters (the normal mean and the variance), and therefore at least 2 observations are required to identify the model, yielding a minimal fraction of `2/n`, where `n` is the sample size. In this case, the default prior follows a Cauchy distribution. In the case of grouping variables, i.e., variables of class `factor` in R, it is recommended that the minimal fractions depend on the respective group sizes [De Santis and Spezzaferri (2001)](https://doi.org/10.1016/S0378-3758(00)00240-8). For instance, in case of an ANOVA with 2 groups (equivalent to a 2-sample t test with equal variances), there are 3 unknown parameters (the two normal means and the within-groups variance), and thus, at least 3 observations are required to identify the model. Therefore, the minimal fraction for the observations in group 1 and 2 are equal to `1.5/n_1` and `1.5/n_2`, where `n_1` and `n_2` are the sample sizes in group 1 and 2, respectively. In the resulting `BF` object, the element `fraction_groupID_observations` contains the group IDs, which are based on all combinations of the levels of the factors in the linear model. If no grouping variables (`factor`) are included as predictors, the same minimal fraction is used for all observations. Moreover, besides regular fractional Bayes factors (which is the current default), adjusted fractional Bayes factors (AFBF) can also be computed by setting the argument `BF.type = AFBF`. In the AFBF, the default prior is shifted to the null value ([Mulder, 2014](https://doi.org/10.1016/j.csda.2013.07.017); [Mulder and Olsson-Collentine, 2019](https://doi.org/10.3758/s13428-018-01196-9)). The AFBF was specifically designed for testing hypotheses with only one-sided or order constraints. The general methodology for the multivariate linear model is discussed in [Mulder et al. (2021) ](https://www.jstatsoft.org/article/view/v100i18) and [Mulder and Gu (2021)](https://doi.org/10.1080/00273171.2021.1904809). 

#### Univariate t testing

First a classical one sample t test is executed for the test value
\(\mu = 5\) on the therapeutic data

``` r
ttest1 <- bain::t_test(therapeutic, alternative = "greater", mu = 5)
```

The `t_test` function is part of the ***bain*** package. The function is
equivalent to the standard `t.test` function with the addition that the
returned object contains additional output than the standard `t.test`
function.

To see which parameters can be tested on this object run
``` r
get_estimates(ttest1)
```
which shows that the only parameter that can be tested is the population mean which has name `mu`.

To perform an exploratory Bayesian t test of whether the population mean is equal to, smaller than, or larger than the null value (which is `5` here, as specified when defining the `ttest1` object), one needs to run `BF` function on the object.

``` r
library(BFpack)
BF1 <- BF(ttest1)
```

This executes an exploratory ('exhaustive') test around the null value: `H1: mu = 5`
versus `H2: mu < 5` versus `H3: mu > 5` assuming equal prior
probabilities for `H1`, `H2`, and `H3` of 1/3. The output presents the
posterior probabilities for the three hypotheses.

The same test would be executed when the same hypotheses are explicitly
specified using the `hypothesis` argument.

``` r
hypothesis <- "mu = 5; mu < 5; mu > 5"
BF(ttest1, hypothesis = hypothesis)
```

In the above test the complement hypothesis is excluded automatically as the formualted hypothesis under the `hypothesis` argument cover the complete parameter space. Furthermore, when testing hypotheses via the `hypothesis` argument, the output also presents an `Evidence matrix` containing the Bayes factors between the hypotheses formulated in the `hypothesis` argument.

A standard two-sided test around the null value `mu` is executed by setting the hypothesis argument equal to the precise null hypothesis so that the complement hypothesis (which is included by default) corresponds to the hypothesis that assumes that the population mean is anything but the null value

``` r
hypothesis <- "mu = 5"
BF(ttest1, hypothesis = hypothesis)
```

The argument `prior.hyp` can be used to specify different prior probabilities
for the hypotheses. For example, when the left one-tailed hypothesis is not possible
based on prior considerations (e.g., see [Mulder et al. (2021, Section 4.1)](https://www.jstatsoft.org/article/view/v100i18)) while the precise (null) hypothesis and the right
one-tailed hypothesis are equally likely, the argument `prior.hyp` should be a vector
specifying the prior probabilities of the respective hypotheses
``` r
BF(ttest1, hypothesis = "mu = 5; mu < 5; mu > 5", prior.hyp = c(.5,0,.5))
```

For more information about the methodology, we refer the interested reader to [Mulder et al. (2021)](https://www.jstatsoft.org/article/view/v100i18) and [Mulder and Gu (2021)](https://doi.org/10.1080/00273171.2021.1904809).


#### Multivariate t testing

Bayesian multivariate t tests can be executed by first fitting a multivariate (regression) model using the `lm` function, and subsequently, the means of the dependent variables (or other coefficients) in the model can be tested using the `BF()` function. Users have to be aware however that means are modeled using intercepts which are named `(Intercept)` by default by `lm` while the hypothesis argument in `BF()` does not allow effect names that include brackets (i.e., `(` or `)`). To circumvent this, one can create a vector of 1s, with name (say) `ones`, to replace the intercept. For example, let us consider a multivariate normal model for the dependent variables `Superficial`, `Middle`, and `Deep` in the `fmri` data set:

``` r
fmri1 <- cbind(fmri,ones=1)
mlm1 <- lm(cbind(Superficial,Middle,Deep) ~ -1 + ones, data = fmri1)
```

Next, we can (for instance) test whether all means equal 0 (`H1`), whether all means are positive (`H2`), or none of these two hypotheses (`complement`):

``` r
BFmlm1 <- BF(mlm1, hypothesis="ones_on_Superficial=ones_on_Middle=ones_on_Deep=0;
                               (ones_on_Superficial,ones_on_Middle,ones_on_Deep)>0")
```

#### Analysis of variance

First an analysis of variance (ANOVA) model is fitted using the `aov`
fuction in `R`.

``` r
aov1 <- aov(price ~ anchor * motivation, data = tvprices)
```

Next a Bayesian test can be performed on the fitted object. By default
exploratory tests are executed of whether the individual main and interaction effects
are zero or not (corresponding to the full model) (see [Mulder et al. (2021, Section 4.2)](https://www.jstatsoft.org/article/view/v100i18))

``` r
BF(aov1)
```

One can also test for specific equal/order hypotheses based on scientific expectations
such as whether `anchorrounded` is positive, `motivationlow` is negative, and the interaction
effect `anchorrounded:motivationlow` is negative (see [Mulder et al. (2021, Section 4.2)](https://www.jstatsoft.org/article/view/v100i18)) versus null hypothesis versus the complement hypothesis
(which assumes that the constraints of neither two hypotheses hold). This test can be executed
as follows:
``` r
constraints2 <- "anchorrounded > 0 & motivationlow < 0 &
  anchorrounded:motivationlow < 0; anchorrounded = 0 &
  motivationlow = 0 & anchorrounded:motivationlow = 0"
set.seed(1234)
BF(aov1, hypothesis = constraints2)
```


#### Univariate and multivariate multiple regression

For a univariate regression model, by default an exhaustive test is
executed of whether an effect is zero, negative, or postive.

``` r
lm1 <- lm(Superficial ~ Face + Vehicle, data = fmri)
BF1 <- BF(lm1)
print(BF1)
```

Hypotheses can be tested with equality and/or order constraints on the
effects of interest. If prefered the complement hypothesis can be
omitted using the `complement`
argument

``` r
BF2 <- BF(lm1, hypothesis = "Vehicle > 0 & Face < 0; Vehicle = Face = 0",
          complement = FALSE)
print(BF2)
```

In a multivariate regression model hypotheses can be tested on the
effects on the same dependent variable, and on effects across different
dependent variables. The name of an effect is constructed as the name of
the predictor variable and the dependent variable separated by `_on_`.
Testing hypotheses with both constraints within a dependent variable and
across dependent variables makes use of a Monte Carlo estimate which may
take a few seconds.

``` r
lm2 <- lm(cbind(Superficial, Middle, Deep) ~ Face + Vehicle,
              data = fmri)
constraint2 <- "Face_on_Deep = Face_on_Superficial = Face_on_Middle < 0 <
     Vehicle_on_Deep = Vehicle_on_Superficial = Vehicle_on_Middle;
     Face_on_Deep < Face_on_Superficial = Face_on_Middle < 0 < Vehicle_on_Deep =
     Vehicle_on_Superficial = Vehicle_on_Middle"
set.seed(123)
BF3 <- BF(lm2, hypothesis = constraint2)
summary(BF3)
```

For more information about the methodology, we refer the interested reader to [Mulder and Olsson-Collentine (2019)](https://link.springer.com/article/10.3758/s13428-018-01196-9) and [Mulder and Gu (2021)](https://doi.org/10.1080/00273171.2021.1904809)


### Testing independent group variances

First a classical significance test is executed using the `bartlett_test` function, which is part of the ***BFpack*** package. The function is equivalent to the standard `bartlett.test` function with the addition that the returned object contains additional output needed for the test using the `BF` function.

``` r
bartlett1 <- bartlett_test(x = attention$accuracy, g = attention$group)
```

On an object of this class, by default `BF` executes an exploratory test of homogeneity (equality) of variances against an unconstrained (full) model

``` r
BF(bartlett1)
```

The group variances have names `ADHD`, `Controls`, and `TS`. This can be retrieved by running 
``` r
get_estimates(bartlett1)
```
Let's say we want to test whether a hypothesis (H1) which assumes that  group variances of groups `Controls` and `TS` are equal and smaller than the group variance of the `ADHD` group, a hypothesis (H2) which assumes that the group variances of `ADHD` and `TS` are equal and larger than the `Controls` group, a hypothesis (H3) which assumes all group variances are equal, and a complement hypothesis (H4). To do this we run the following:

``` r
hypothesis <- "Controls = TS < ADHD; Controls < TS = ADHD; Controls = TS = ADHD"
set.seed(358)
BF_var <- BF(bartlett1, hypothesis)
```

A comprehensive output of this analysis can be obtained by running:

``` r
summary(BF_var)
```

which presents the results of an exploratory analysis and the results of a confirmatory analysis (based on the hypotheses formulated under the `hypothesis` argument). The exploratory analysis tests a hypothesis which assumes that the variances are equal across groups (homogeneity of variances) versus an alternative unrestricted hypothesis. The output shows that the posterior probabilities of these two hypotheses are approximately 0.803 and 0.197 (assuming equal priori probabilities). Note that the p value in the classical Bartlett test for these data equals 0.1638 which implies that the hypothesis of homogeneity of variances cannot be rejected using common significance levels, such as 0.05 or 0.01. Note however that this p value cannot be used as a measure for the evidence in the data in favor of homogeneity of group variances. This can be done using the proposed Bayes factor test which shows that the probability that the variances are equal is approximately 0.803. Also note that the exploratory test could equivalently tested via the `hypothesis` argument by running `BF(bartlett1, "Controls = TS = ADHD")`.

The confirmatory test shows that H1 receives strongest support from the data, but H2 and H3 are viable competitors. It appears that even the complement H3 cannot be ruled out entirely given a posterior prob- ability of 0.058. To conclude, the results indicate that TS population are as heterogeneous in their attentional performances as the healthy control population in this specific task, but further research would be required to obtain more conclusive evidence.

For more information about the methodology, we refer the interested reader to [Boing-Messing et al. (2017)](https://www.researchgate.net/publication/317418250_Bayesian_evaluation_of_constrained_hypotheses_on_variances_of_multiple_independent_groups)


### Testing correlations

`BFpack` can be used for testing overlapping, nonoverlapping, and independent correlations, between variables of different measurement levels (continuous, dichotomous, ordinal), possibly while correcting for certain covariates. Joint uniform priors are specified for the correlations in the full correlation matrices. This implies that all combinations of correlation values that result in positive definite correlation matrices are equally likely a priori.

By default `BF` performs exhaustive tests of whether the separate
correlations are zero, negative, or positive.

``` r
set.seed(123)
cor1 <- cor_test(memory[,1:3])
BF1 <- BF(cor1)
print(BF1)
```

The names of the
correlations is constructed using the names of the variables separated
by `_with_`:
``` r
get_estimates(cor1)
```

Specific hypotheses based on prior/theoretical considerations can be tested using the 
`hypothesis` argument. As an example we show here how to test whether all
correlations are equal and positive versus its complement. 
``` r
BF2 <- BF(cor1, hypothesis = "Del_with_Im = Wmn_with_Im = Wmn_with_Del > 0")
print(BF2)
```

We can also test equality and order constraints on independent correlations across different groups. As the seventh column of the `memory` object is a group indicator, let us first create different objects for the two different groups, and perform Bayesian estimation on the correlation matrices of the two different groups 
``` r
memoryHC <- subset(memory,Group=="HC")[,-(4:7)]
memorySZ <- subset(memory,Group=="SZ")[,-(4:7)]
set.seed(123)
cor1 <- cor_test(memoryHC,memorySZ)
```

In this case with multiple groups by default exploratory tests are executed of whether the correlations
are zero, negative, or positive for each separate group (e.g., correlations in group `gr1`
are denoted by `_in_gr1` at the end of the name)
``` r
get_estimates(cor1)
```

Next we test the one-sided hypothesis that the respective correlations in the first group (`g1`) are larger than the correlations in the second group (`g2`) via

``` r
set.seed(123)
BF6_cor <- BF(cor1, hypothesis =
  "Del_with_Im_in_g1 > Del_with_Im_in_g2 &
  Del_with_Wmn_in_g1 > Del_with_Wmn_in_g2 &
  Im_with_Wmn_in_g1 > Im_with_Wmn_in_g2")
```

By running `print(BF6_cor)`, the output shows that the one-sided hypothesis received a posterior probability of 0.991 and the alternative received a posterior probability of .009 (assuming equal prior probabilities).

For more information about the methodology, we refer the interested reader to [Mulder (2016)](https://www.researchgate.net/publication/280256902_Bayes_factors_for_testing_order-constrained_hypotheses_on_correlations) and [Mulder and Gelissen (2019)](https://doi.org/10.1080/02664763.2021.1992360)


### Meta-analyses

`BFpack` supports testing the mean effect in a fixed effects meta-analysis model, a random effects model, and a (hybrid) marginalized random effects meta-analysis (marema) model. Depending on the nature of the parameter, different default priors are readily implemented. For testing a standardized effect (set `BF.type = "stand.effect"`), a normal prior with mean 0 and standard deviation 0.5 is specified which implies that, on average, medium (0.5) standardized effect sizes are expected (if the null is false). For testing a log odds (set `BF.type = "log.odds"`), a Student prior with mean 0, scale 2.36, and 13.1 degrees of freedom is used, which corresponds to success probabilities following uniform priors in the interval (0,1). For testing correlations (set `BF.type = "correlation"`), a logistic prior with scale 0.5 is used for the Fisher transformed correlation, which corresponds to a uniform prior for the correlation in the interval (-1,1). To specify a unit-information prior (set `BF.type = "unit.info"`), the total sample size `sum(ni)` is used to construct a minimally informative normal unit-information prior using the sample sizes `ni` of the individual studies from the `rma.uni` object. For a manually specified prior, the argument `BF.type` needs to be an object of class `prior` from the `metaBMA` package. Note that in order for the Bayes factors to be meaningful, the prior should reflect a plausible range of values given the available context. Arbitrarily vague priors (which can be used for estimation) should not be used for Bayes factor testing.

For illustrative purposes, we use hypothetical simulated data for standardized effect sizes:

``` r
set.seed(123)
tau2 <- 0.05
vi <- runif(10, min = 0.01, max = 0.2)
yi <- rnorm(10, mean = 0, sd = sqrt(vi+tau2))
```

where `tau2` denotes the true between-study heterogeneity, `vi` is a vector containing the squared standard errors of 10 studies, and `yi` is a vector containing the estimated effects sizes in the 10 studies. To test the overall effect size and the between-study heterogeneity using `BFpack`, an initial meta-analysis needs to be executed using the `metafor` package. A random effects meta-analysis model is considered:

``` r
res1 <- metafor::rma(yi = yi, vi = vi)
```

Subsequently, the output is plugged into the `BF` function using the default prior for a standardized effect (normal prior with mean of 0 and a sd of 0.5). For a random effects model (`res1$method` does not equal `"EE"` or `"FE"`), `BFpack` computes Bayes factors and posterior probabilities of a zero, negative, and positive mean effect:

``` r
BFmeta1 <- BF(res1, BF.type = "stand.effect")
```

The `summary` output gives the posterior probabilities for a zero, negative, and positive mean effect size `mu` assuming equal prior probabilities:

``` r
summary(BFmeta1)
```

The results indicate support for a zero effect with a posterior probability of approximately 0.75 under both the random effects model and the marema model.

Under the random effects model and the marema model, the posterior mean and median, the lower and upper bound of the 95% Bayesian credible intervals, and the posterior probability that the parameters are positive can be obtained by calling:

``` r
BFmeta1$estimates
```

Note that under the marema model, the between-study heterogeneity `tau2` can attain negative values (indicating support for a fixed effects model). Therefore, the posterior probability that `tau2` is positive under the marema model gives an indication of the posterior support of a random effects model given the available data. In this case, the posterior probability that `tau2` is positive is about 0.95, which indicates clear support for a random effects model.

In order to test the mean effect under a fixed effects meta-analysis model, this needs to be specified when fitting the initial meta-analytic model:

``` r
res2 <- metafor::rma(yi = yi, vi = vi, method = "EE")
BFmeta2 <- BF(res2, BF.type = "stand.effect")
print(BFmeta2)
```

For more information about the methodology, we refer the interested reader to [Van Aert and Mulder (2021)](https://link.springer.com/article/10.3758/s13423-021-01918-9) and [Mulder and van Aert (2025)](https://doi.org/10.1017/rsm.2025.10060).


### Exponential random graph models

`BFpack` can be used for Bayes factor testing of the coefficients in an exponential random graph model (ERGM). As input the `BF()` function requires a fitted ERGM either from the `ergm` package or the `Bergm` package. Bayes factors are then computed using a so-called unit-information prior. For an ERGM, this prior contains the information of single dyad. Moreover, the prior is centered at 0. For Bayes factor testing, this prior is preferred to an arbitrarily vague prior (Mulder, Friel, \& Leifeld, 2024). For this illustration, we use the `florentine` data from the `ergm` package and we fit the following ERGM:

``` r
library(ergm)
seed <- 123
# load florentine data
data(florentine)
ergm_fit <- ergm(flomarriage ~ edges +
  kstar(2) +
  absdiff("wealth"),
  control = control.ergm(seed = seed))
```

A Bayes factor is computed for a traditional two-sided test of whether the absolute difference in wealth between two families affects tie formation.

``` r
set.seed(123)
BFergm <- BF(ergm_fit,
  hypothesis = "absdiff.wealth = 0",
  main.iters = 500)
print(BFergm)
```

The null hypothesis is not supported give the Bayes factor of 9.36 in favor of the two-sided alternative, and given the posterior probability of 0.097 for the null. 

Finally note that one could retrieve the employed unit-information prior from the `BF` object:

``` r
BFergm$prior.parameters
```

For more information about the methodology, we refer the interested reader to [Mulder, Friel, and Leifeld (2024)](https://doi.org/10.1016/j.socnet.2023.11.004).

### Hypothesis testing using normal approximations

### Running `BF` on a named vector

The input for the `BF` function can also be a named vector containing
the estimates of the parameters of interest. In this case the error
covariance matrix of the estimates is also needed via the `Sigma`
argument, as well as the sample size that was used for obtaining the
estimates via the `n` argument. Bayes factors are then computed using
Gaussian approximations of the likelihood (and posterior), similar as in
classical Wald test. Furthermore, a minimally informative default prior is automatically constructed using a minimal fraction of the data, similar as the adjusted fractional Bayes factor. The minimal fraction is constructed by dividing the number of parameters that are tested in the hypothesis by the total sample size `n`. Moreover, the prior is shifted to the null value, resulting in an approximate adjusted fractional Bayes factor. This methodology is used by default in case the input object `x` has class `glm`, `coxph`, `survreg`, 
`polr`, and `zeroinfl`. Below we first show how the function can be used for testing coefficients in a logistic regression model, and then we show how the function can be used for testing coefficients in a poisson regression model and as well an equivalent analysis when `x` is a named vector using the MLEs, error covariance matrix `Sigma`, and sample size `n`.


#### Logistic regression

An example hypothesis test is considered under a logistic regression
model. First a logistic regression model is fitted using the `glm`
function

``` r
fit_glm <- glm(sent ~ ztrust + zfWHR + zAfro + glasses + attract + maturity +
               tattoos, family = binomial(), data = wilson)
```

By default exploratory exhaustive tests are executed of whether the separate
regression coefficients are zero, negative, or positive:
``` r
BF(fit_glm)
```

The names of the regression coefficients on which constrained hypotheses
can be formualted can be extracted using the `get_estimates` function.

``` r
get_estimates(fit_glm)
```
Two different hypotheses are formulated with competing equality and/or
order constraints on the regression coefficients of interest [Mulder et al. (2021, Section 4.4) ](https://www.jstatsoft.org/article/view/v100i18)

``` r
BF_glm <- BF(fit_glm, hypothesis = "ztrust > (zfWHR, zAfro) > 0;
             ztrust > zfWHR = zAfro = 0")
summary(BF_glm)
```

By calling the `summary` function on the output object of class `BF`,
the results of the exploratory tests are presented of whether each
separate parameter is zero, negative, or positive, and the results of
the confirmatory test of the hypotheses under the `hypothesis` argument
are presented. When the hypotheses do not cover the complete parameter
space, by default the complement hypothesis is added which covers the
remaining parameter space that is not covered by the constraints under
the hypotheses of interest. In the above example, the complement
hypothesis covers the parameter space where neither `"ztrust > (zfWHR,
zAfro) > 0"` holds, nor where `"ztrust > zfWHR = zAfro = 0"` holds.

For more information about the methodology, we refer the interested reader to [Gu et al. (2018)](https://bpspsychub.onlinelibrary.wiley.com/doi/full/10.1111/bmsp.12110?casa_token=tjWwoVqLI7QAAAAA%3Ac84DWKNw8pf23ybOgZTs1pzgyZMuCc-BeTaPFj8vnfnTytzHd0gr-D9ymVxYbKj_MTO1ITGFW8tGBZ1J) and [Mulder et al. (2021)](https://www.jstatsoft.org/article/view/v100i18)


#### Poisson regression

We illustrate this for a Poisson regression model

``` r
poisson1 <- glm(formula = breaks ~ wool + tension, data = datasets::warpbreaks,
             family = poisson)
```

The estimates, the error covariance matrix, and the sample size are
extracted from the fitted model

``` r
estimates <- poisson1$coefficients
covmatrix <- vcov(poisson1)
samplesize <- nobs(poisson1)
```

Constrained hypotheses on the parameters `names(estimates)` can then be
tested as follows

``` r
BF1 <- BF(estimates, Sigma = covmatrix, n = samplesize, hypothesis = 
  "woolB > tensionM > tensionH; woolB = tensionM = tensionH")
```

Note that the same hypothesis test would be executed when calling

``` r
BF2 <- BF(poisson1, hypothesis = "woolB > tensionM > tensionH;
          woolB = tensionM = tensionH")
```

because the same Bayes factor is used when running `BF` on an object of
class `glm` (see `Method: Bayes factor using Gaussian approximations`
when calling `print(BF1)` and `print(BF2)`).

For more information about the methodology, we refer the interested reader to [Gu et al. (2018)](https://bpspsychub.onlinelibrary.wiley.com/doi/full/10.1111/bmsp.12110?casa_token=tjWwoVqLI7QAAAAA%3Ac84DWKNw8pf23ybOgZTs1pzgyZMuCc-BeTaPFj8vnfnTytzHd0gr-D9ymVxYbKj_MTO1ITGFW8tGBZ1J)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tvprices.R
\docType{data}
\name{tvprices}
\alias{tvprices}
\title{Precision of the Anchor Influences the Amount of Adjustment}
\format{
A data.frame with 59 rows and 3 columns.
}
\usage{
data(tvprices)
}
\description{
Data from an experimental study where participants have to guess the price
of a plasma tv. There were two experimental conditions.
These data were simulated using the sufficient statistics from Janiszewski &
Uy (2008).
}
\details{
\tabular{lll}{
   \strong{price} \tab \code{numeric} \tab Participant z-scores of price\cr
   \strong{anchor} \tab \code{factor} \tab Participant anchor\cr
   \strong{motivation} \tab \code{factor} \tab motivation to change\cr
}
}
\references{
Janiszewski, C., & Uy, D. (2008). Precision of the anchor influences the amount of adjustment. Psychological Science, 19(2), 121–127. doi:10.1111/j.1467-9280.2008.02057.x
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/memory.R
\docType{data}
\name{memory}
\alias{memory}
\title{Memory data on health and schizophrenic patients}
\format{
A data.frame with 40 rows and 8 columns.
}
\usage{
data(memory)
}
\description{
Data set from study assessing differences between schizophrenic patients and
healthy control participants in patterns of correlations among 6 verbal memory
tasks (Ichinose et al., 2019).
\tabular{lll}{
   \strong{Im} \tab \code{numeric} \tab Percent correct on immediate recall of 3 word lists\cr
   \strong{Del} \tab \code{numeric} \tab Percent correct on delayed recall of 3 word lists\cr
   \strong{Wmn} \tab \code{numeric} \tab Number correct on letter-number span test of auditory working memory\cr
   \strong{Cat} \tab \code{numeric} \tab Number correct on category fluency task\cr
   \strong{Fas} \tab \code{numeric} \tab Number correct on letter fluency task\cr
   \strong{Rat} \tab \code{numeric} \tab Number correct on remote associates task\cr
   \strong{Group} \tab \code{factor} \tab Participant Group (HC = Healthy Control; SZ = Schizophrenia) \cr
}
}
\references{
Ichinose, M.C., Han, G., Polyn, S., Park, S., & Tomarken, A.J. (2019). Verbal memory performance discordance in schizophrenia: A reflection of cognitive dysconnectivity.
Unpublished manuscript.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fmri.R
\docType{data}
\name{fmri}
\alias{fmri}
\title{fMRI data}
\format{
A data.frame with 13 rows and 6 columns.
}
\usage{
data(fmri)
}
\description{
fMRI data assessing relation between individual differences in the ability to recognize
faces and cars and thickness of the superficial, middle, and deep layers of the
fusiform face area, as assessed by high-resolution fMRI recognition (Williams et al, 2019, under review)
}
\details{
\tabular{lll}{
  \strong{Subject}\tab\code{numeric}\tab Particicpant ID number\cr
   \strong{Face} \tab \code{numeric} \tab Standardized score on face recognition battery\cr
   \strong{Vehicle} \tab \code{numeric} \tab Standardized score on vehicle recognition battery\cr
   \strong{Superficial} \tab \code{numeric} \tab Depth in mm of superficial layer of FFA\cr
   \strong{Middle} \tab \code{numeric} \tab Depth in mm of middle layer of FFA\cr
   \strong{Bform} \tab \code{numeric} \tab Depth in mm of deep layer of FFA\cr
}
}
\references{
McGuigin, R.W., Newton, A.T., Tamber-Rosenau, B., Tomarken, A.J, & Gauthier, I. (under review). Thickness of deep layers in the fusiform face area predicts face recognition.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/same_location.R
\docType{data}
\name{same_location}
\alias{same_location}
\title{Same location event statistic}
\format{
dataframe (25 rows, 4 columns)

\tabular{lll}{
   \strong{same_location} \tab \code{integer} \tab Event statistic. Matrix with senders in the
   rows and receivers in the columns. The event statistic is 1 if sender and receiver have
   the same location and 0 otherwise. \cr
}
}
\usage{
data(same_location)
}
\description{
A matrix coding whether senders of events (in the rows) and receivers of events
(in the column) have the same location. Related to the 'events' data object,
that contains a relational event sequence, and the 'actors' object, that contains
information on the 25 actors involved in the relational event sequence.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BF.mvt_test.R
\name{mvt_test}
\alias{mvt_test}
\title{Multivariate Student t test}
\usage{
mvt_test(X, Y, null = NULL, paired = FALSE, ...)
}
\arguments{
\item{X}{a data matrix with the variables in the columns.}

\item{Y}{an optional data matrix with the variables in the columns.}

\item{null}{a vector of the null values of the variables.}

\item{paired}{a logical indicating whether you want a multivariate paired t-test.}

\item{...}{further arguments to be passed to or from methods.}
}
\value{
An object that can be applied to the \code{BF()}.
}
\description{
First step to performs a Bayesian multivariate one sample Student t test using the
(adjusted) fractional Bayes factor using the \code{BF()} function.
}
\details{
\code{X} must be a data matrix and \code{null}
must be a vector of the assumed null values of the variables.
}
\examples{

mvt_fmri <- mvt_test(fmri[,1:2],null = c(0,0))
BF(mvt_fmri)

# the same test can be executed via the lm() function
intercept <- rep(1,nrow(fmri))
lm1 <- lm(cbind(Face,Vehicle) ~ -1 + intercept, data=fmri)
BF(lm1,hypothesis="intercept_on_Face=intercept_on_Vehicle=0")

}
\references{
Mulder, J. and Gu, X. (2023). Bayesian Testing of Scientific
Expectations under Multivariate Normal Linear Models. Multivariate Behavioral
Research, 57, 767-783. DOI: 10.1080/00273171.2021.1904809.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wilson.r
\docType{data}
\name{wilson}
\alias{wilson}
\title{Facial trustworthiness and criminal sentencing}
\format{
A data.frame with 742 rows and 13 columns.
}
\usage{
data(wilson)
}
\description{
Data from a correlational study in which the correlation between ratings of
facial trustworthiness of inmates was correlated with whether they had
received the death penalty or not (wilson and Rule, 2015). These data were
simulated using the R-package \code{synthpop}, based on the characteristics
of the original data.
}
\details{
\tabular{lll}{
   \strong{stim} \tab \code{integer} \tab Stimulus Number\cr
   \strong{sent} \tab \code{integer} \tab Sentence: 1 = Death, 0 = Life\cr
   \strong{race} \tab \code{integer} \tab Race: 1 = White, -1 = Black\cr
   \strong{glasses} \tab \code{integer} \tab Glasses: 1 = Yes, 0 = No\cr
   \strong{tattoos} \tab \code{integer} \tab Tattoos: 1 = Yes, 0 = No \cr
   \strong{ztrust} \tab \code{numeric} \tab Trustworthiness \cr
   \strong{trust_2nd} \tab \code{numeric} \tab Trustworthiness ratings with 2nd control group; Death targets are same as in primary analysis, Life targets are different.\cr
   \strong{afro} \tab \code{numeric} \tab raw Afrocentricity ratings.\cr
   \strong{zAfro} \tab \code{numeric} \tab Afrocentricity ratings normalized within target race. Analyses in paper were done with this variable.\cr
   \strong{attract} \tab \code{numeric} \tab Attractiveness\cr
   \strong{fWHR} \tab \code{numeric} \tab facial width-to-height \cr
   \strong{afWHR} \tab \code{numeric} \tab fWHR normalized within target race. Analyses in paper were done with this variable \cr
   \strong{maturity} \tab \code{numeric} \tab Maturity
}
}
\references{
Wilson, J. P., & Rule, N. O. (2015). Facial Trustworthiness
  Predicts Extreme Criminal-Sentencing Outcomes. Psychological Science,
  26(8), 1325–1331. doi: 10.1177/0956797615590992
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BF.cortest.R
\name{cor_test}
\alias{cor_test}
\title{Bayesian correlation analysis}
\usage{
cor_test(
  ...,
  formula = NULL,
  iter = 10000,
  burnin = 2000,
  method = "LKJ",
  nugget.scale = 0.999
)
}
\arguments{
\item{...}{matrices (or data frames) of dimensions \emph{n} (observations) by  \emph{p} (variables)
for different groups (in case of multiple matrices or data frames).}

\item{formula}{an object of class \code{\link[stats]{formula}}. This allows for including
control variables in the model (e.g., \code{~ education}).}

\item{iter}{total number of iterations from posterior. The total is split across three chains. By default, the total number of iterations is 10000,
implying a rounded integer of 3333 iterations per chain.}

\item{burnin}{number of iterations for burnin (default is 2000).}

\item{method}{\code{"LKJ"} for onion method sampler of correlation matrix proposed by Lewandowski, Kurowicka,
& Joe (2009), which is the default, or \code{"LD"} for 'PX-RPMH' correlation matrix sampler of Liu and Daniels (2009).
\code{"onion"} method tends to mix better for larger correlation matrices.}

\item{nugget.scale}{When using \code{method = "LD"}, a nugget.scale can be used to avoid computational issues due to posterior draws
for the correlations too close to 1 in absolute value. Posterior draws for the correlations are multiplied with this nugget.scale.
So \code{nugget.scale} should be close to 1 (the default is .999). If the traceplots show that draws are stuck
at 1 or -1 too long try a slightly smaller \code{nugget.scale}.}
}
\value{
list of class \code{cor_test}:
\itemize{
\item \code{meanF} posterior means of Fisher transform correlations
\item \code{covmF} posterior covariance matrix of Fisher transformed correlations
\item \code{correstimates} posterior estimates of correlation coefficients
\item \code{corrdraws} list of posterior draws of correlation matrices per group
\item \code{corrnames} names of all correlations
}
}
\description{
Estimate the unconstrained posterior for the correlations using a joint uniform prior (Mulder and Gelissen,
2023) or a marginally uniform prior (Barnard et al., 2000, Mulder, 2016). Correlation matrices are sampled from the posterior
using the MCMC algorithm of Talhouk et al. (2012).
}
\examples{
\donttest{
# Bayesian correlation analysis of the 6 variables in 'memory' object
# we consider a correlation analysis of the first three variable of the memory data.
fit <- cor_test(BFpack::memory[,1:3])

# Bayesian correlation of variables in memory object in BFpack while controlling
# for the Cat variable
fit <- cor_test(BFpack::memory[,c(1:4)],formula = ~ Cat)

# Example of Bayesian estimation of polyserial correlations
memory_example <- memory[,c("Im","Rat")]
memory_example$Rat <- as.ordered(memory_example$Rat)
fit <- cor_test(memory_example)

# Bayesian correlation analysis of first three variables in memory data
# for two different groups
HC <- subset(BFpack::memory[,c(1:3,7)], Group == "HC")[,-4]
SZ <- subset(BFpack::memory[,c(1:3,7)], Group == "SZ")[,-4]
fit <- cor_test(HC,SZ)

}
}
\references{
Barnard, J., McCulloch, R., & Meng, X. L. (2000). Modeling covariance matrices in terms of standard deviations and
correlations, with application to shrinkage. Statistica Sinica, 1281-1311. <https://www.jstor.org/stable/24306780>

Joe, H. (2006). Generating random correlation matrices based on partial correlations. Journal of Multivariate
Analysis, 97(10), 2177-2189. <https://doi.org/10.1016/j.jmva.2005.05.010>

Mulder, J., & Gelissen, J. P. (2023). Bayes factor testing of equality and order constraints on measures of
association in social research. Journal of Applied Statistics, 50(2), 315-351. <https://doi.org/10.1080/02664763.2021.1992360>

Mulder, J. (2016). Bayes factors for testing order-constrained hypotheses on correlations. Journal of Mathematical
Psychology, 72, 104-115. <https://doi.org/10.1016/j.jmp.2014.09.004>

Lewandowski, D., Kurowicka, D., & Joe, H. (2009). Generating random correlation matrices based on vines and extended
onion method. Journal of Multivariate Analysis, 100(9), 1989–2001. <https://doi.org/10.1016/j.jmva.2009.04.008>

Liu, X., & Daniels, M. J. (2006). A new algorithm for simulating a correlation matrix based on parameter expansion and
reparameterization. Journal of Computational and Graphical Statistics, 15(4), 897-914. <https://doi.org/10.1198/106186006X160681>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/relevents.R
\docType{data}
\name{relevents}
\alias{relevents}
\title{A sequence of innovation-related e-mail messages}
\format{
dataframe (247 rows, 3 columns)

\tabular{lll}{
   \strong{relevents$time} \tab \code{numeric} \tab Time of the e-mail message,
   in seconds since onset of the observation \cr
   \strong{relevents$sender} \tab \code{integer} \tab ID of the sender, corresponding to
   the employee IDs in the actors dataframe \cr
   \strong{relevents$receiver} \tab \code{integer} \tab ID of the receiver \cr
}
}
\usage{
data(relevents)
}
\description{
A time-ordered sequence of 247 communication messages between 25 actors.
}
\details{
The related data files 'actors', 'same_location', 'same_culture' contain information
on the actors and three event statistics respectively.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/therapeutic.R
\docType{data}
\name{therapeutic}
\alias{therapeutic}
\title{Data come from an experimental study (Rosa, Rosa, Sarner, and Barrett, 1998) 
that were also used in Howell (2012, p.196).
An experiment was conducted to investigate if Therapeutic Touch practitioners 
who were blindfolded can effectively identify which of their hands is below the experimenter¡¯s.
Twenty-eight practitioners were involved and tested 10 times in the experiment. 
Researchers expected an average of 5 correct answers from each practitioner 
as it is the number by chance if they do not outperform others.}
\format{
A data.frame with 22 rows and 1 column.
}
\usage{
data(therapeutic)
}
\description{
\tabular{lll}{
   \strong{correct} \tab \code{integer} \tab How many correct answers are from each practitioner)\cr
}
}
\references{
Howell, D. (2012). Statistical methods for psychology (8th ed.). Belmont, CA: Cengage Learning.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BFpack-package.R
\docType{package}
\name{BFpack-package}
\alias{BFpack}
\alias{BFpack-package}
\title{BFpack: Flexible Bayes factor testing of scientific expectations}
\description{
The \code{R} package \strong{BFpack} provides tools for exploratory and
confirmatory Bayesian hypothesis testing using Bayes factors and posterior probabilities
under common statistical models. The main function `BF` needs a fitted model `x` as input
argument. Depending on the class of the fitted model, a standard hypothesis test is
executed by default. For example, if `x` is a
fitted regression model of class `lm` then posterior probabilities are computed of whether
each separate coefficient is zero, negative, or positive (assuming equal prior probabilities).
If one has specific hypotheses with
equality and/or order constraints on the parameters under the fitted model `x` then these
can be formulated using the `hypothesis` argument (a character string), possibly together
prior probabilities for the hypotheses via the `prior` argument (default all hypotheses are
equally likely a priori), and the `complement` argument which is a logical stating whether
the complement hypotheses should be included in the case (`TRUE` by default).

Use compilation for Fortran functions
}
\examples{
\dontrun{
# EXAMPLE 1. One-sample t test
ttest1 <- t_test(therapeutic, mu = 5)
print(ttest1)
# confirmatory Bayesian one sample t test
BF1 <- BF(ttest1, hypothesis = "mu = 5")
summary(BF1)
# exploratory Bayesian one sample t test
BF(ttest1)

# EXAMPLE 2. ANOVA
aov1 <- aov(price ~ anchor * motivation,data = tvprices)
BF1 <- BF(aov1, hypothesis = "anchorrounded = motivationlow;
                              anchorrounded < motivationlow")
summary(BF1)

# EXAMPLE 3. Logistic regression
fit <- glm(sent ~ ztrust + zfWHR + zAfro + glasses + attract + maturity +
   tattoos, family = binomial(), data = wilson)
BF1 <- BF(fit, hypothesis = "ztrust > zfWHR > 0;
                             ztrust > 0 & zfWHR = 0")
summary(BF1)

# EXAMPLE 4. Correlation analysis
set.seed(123)
cor1 <- cor_test(memory[1:20,1:3])
BF1 <- BF(cor1)
summary(BF1)
BF2 <- BF(cor1, hypothesis = "Wmn_with_Im > Wmn_with_Del > 0;
                              Wmn_with_Im = Wmn_with_Del = 0")
summary(BF2)
}

}
\references{
Mulder, J., D.R. Williams, Gu, X., A. Tomarken,
F. Böing-Messing, J.A.O.C. Olsson-Collentine, Marlyne Meyerink, J. Menke,
J.-P. Fox, Y. Rosseel, E.J. Wagenmakers, H. Hoijtink., and van Lissa, C.
(submitted). BFpack: Flexible Bayes Factor Testing of Scientific Theories
in R. \url{https://arxiv.org/abs/1911.07728}

Mulder, J., van Lissa, C., Gu, X., Olsson-Collentine, A., Boeing-Messing, F., Williams,
D. R., Fox, J.-P., Menke, J., et al. (2019). BFpack: Flexible Bayes Factor Testing of
Scientific Expectations. (Version 0.2.1) \url{https://CRAN.R-project.org/package=BFpack}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/jomulder/BFpack}
  \item Report bugs at \url{https://github.com/jomulder/BFpack/issues}
}

}
\author{
\strong{Maintainer}: Joris Mulder \email{j.mulder3@tilburguniversity.edu}

Authors:
\itemize{
  \item Caspar van Lissa \email{c.j.vanlissa@uu.nl} [contributor]
  \item Donald R. Williams \email{drwwilliams@ucdavis.edu} [contributor]
  \item Xin Gu \email{guxin57@hotmail.com} [contributor]
  \item Anton Olsson-Collentine \email{J.A.E.OlssonCollentine@tilburguniversity.edu} [contributor]
  \item Florian Boeing-Messing \email{F.Boeing-Messing@uvt.nl} [contributor]
  \item Jean-Paul Fox \email{g.j.a.fox@utwente.nl} [contributor]
}

Other contributors:
\itemize{
  \item Janosch Menke \email{janosch.menke@uni-muenster.de} [contributor]
  \item Robbie van Aert \email{R.C.M.vanAert@tilburguniversity.edu} [contributor]
  \item Barry Brown [contributor]
  \item James Lovato [contributor]
  \item Kathy Russell [contributor]
  \item Lapack 3.8 [contributor]
  \item Jack Dongarra [contributor]
  \item Jim Bunch [contributor]
  \item Cleve Moler [contributor]
  \item Gilbert Stewart [contributor]
  \item John Burkandt [contributor]
  \item Ashwith Rego [contributor]
  \item Alexander Godunov [contributor]
  \item Alan Miller [contributor]
  \item Jean-Pierre Moreau [contributor]
  \item The R Core Team [copyright holder]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/timssICC.r
\docType{data}
\name{timssICC}
\alias{timssICC}
\title{Trends in International Mathematics and Science Study (TIMSS) 2011-2015}
\format{
A data.frame with 16770 rows and 15 columns.
}
\usage{
data(timssICC)
}
\description{
A stratified sample was drawn by country and school to obtain a balanced
sample of p = 15 grade-4 students
per school for each of four countries (The Netherlands (NL), Croatia (HR),
Germany
(DE), and Denmark (DK)) and two measurement occasions (2011, 2015).
Achievement scores
(first plausible value) of overall mathematics were considered. Performances
of fourth
and eight graders from more than 50 participating countries around the world
can be found at (https://www.iea.nl/timss)
The TIMSS achievement scale is centered at 500 and the standard deviation is
equal to 100 scale score points.
The TIMSS data set has a three-level structure, where students are nested
within classrooms/schools, and
the classrooms/schools are nested within countries. Only one classroom was
sampled per school.
Changes in the mathematics achievement can be investigated by examining the
grouping of
students in schools across countries. Changes in country-specific intraclass
correlation coefficient
from 2011 to 2015, representing heterogeneity in mathematic achievements
within and between schools across years,
can be tested. When detecting a decrease in average performance together
with an increase
of the intraclass correlation, a subset of schools performed worse. For a
constant
intraclass correlation across years the drop in performance applied to the
entire population
of schools. For different countries, changes in the intraclass correlation
across years
can be tested concurrently to examine also differences across countries.
}
\details{
\tabular{lll}{
   \strong{math} \tab \code{numeric} \tab math score child\cr
   \strong{groupNL11} \tab \code{numeric} \tab
   Indicator for child from NL in 2011\cr
   \strong{groupNL15} \tab \code{numeric} \tab
   Indicator for child from NL in 2015\cr
   \strong{groupHR11} \tab \code{numeric} \tab
   Indicator for child from HR in 2011\cr
   \strong{groupHR15} \tab \code{numeric} \tab
   Indicator for child from HR in 2015\cr
   \strong{groupDE11} \tab \code{numeric} \tab
   Indicator for child from DE in 2011\cr
   \strong{groupDE15} \tab \code{numeric} \tab
   Indicator for child from DE in 2015\cr
   \strong{groupDR11} \tab \code{numeric} \tab
   Indicator for child from DK in 2011\cr
   \strong{groupDR15} \tab \code{numeric} \tab
   Indicator for child from DK in 2015\cr
   \strong{gender} \tab \code{numeric} \tab Female=0,Male=1 \cr
   \strong{weight} \tab \code{numeric} \tab Child sampling weight \cr
   \strong{yeargender} \tab \code{numeric} \tab
   Interaction for occassion and gender \cr
   \strong{lln} \tab \code{numeric} \tab
   total number of children in school-class \cr
   \strong{groupschool} \tab \code{factor} \tab
   Nested indicator for school in country\cr
   \strong{schoolID} \tab \code{factor} \tab
   Unique indicator for school
}
}
\references{
Mulder, J. & Fox, J.-P. (2019). Bayes factor testing of multiple
intraclass correlations. Bayesian Analysis. 14, 2, p. 521-552.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Fcor.R
\docType{data}
\name{Fcor}
\alias{Fcor}
\title{Student t approximations of Fisher transformed correlations}
\format{
A data.frame with 3 columns.
}
\usage{
data(Fcor)
}
\description{
Approximated degrees of freedom and approximated scale of the Fisher transformed
correlations depending on the dimension of the vector of dependent variables P
based on a joint uniform prior.
}
\details{
\tabular{lll}{
   \strong{nu} \tab \code{numeric} \tab Approximated degrees of freedom\cr
   \strong{sigma} \tab \code{numeric} \tab Approximated scale\cr
   \strong{P} \tab \code{integer} \tab Dimension of vector of dependent variables\cr
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BF.gaussian.R, R/BF.lm.R, R/BF.meta.R,
%   R/BF.methods.R, R/BF.ttest.R
\name{BF.default}
\alias{BF.default}
\alias{BF.lm}
\alias{BF.rma.uni}
\alias{BF}
\alias{BF.t_test}
\title{Bayes factors for Bayesian exploratory and confirmatory hypothesis
testing}
\usage{
\method{BF}{default}(
  x,
  hypothesis = NULL,
  prior.hyp.explo = NULL,
  prior.hyp.conf = NULL,
  prior.hyp = NULL,
  complement = TRUE,
  log = FALSE,
  cov.prob = 0.95,
  Sigma,
  n,
  ...
)

\method{BF}{lm}(
  x,
  hypothesis = NULL,
  prior.hyp.explo = NULL,
  prior.hyp.conf = NULL,
  prior.hyp = NULL,
  complement = TRUE,
  log = FALSE,
  cov.prob = 0.95,
  BF.type = NULL,
  iter = 10000,
  ...
)

\method{BF}{rma.uni}(
  x,
  hypothesis = NULL,
  prior.hyp.explo = NULL,
  prior.hyp.conf = NULL,
  prior.hyp = NULL,
  complement = TRUE,
  log = FALSE,
  cov.prob = 0.95,
  BF.type,
  iter = 20000,
  ...
)

\method{BF}{t_test}(
  x,
  hypothesis = NULL,
  prior.hyp.explo = NULL,
  prior.hyp.conf = NULL,
  prior.hyp = NULL,
  complement = TRUE,
  log = FALSE,
  cov.prob = 0.95,
  BF.type = NULL,
  iter = 1e+06,
  ...
)
}
\arguments{
\item{x}{An R object containing the outcome of a statistical analysis.
An R object containing the outcome of a statistical analysis. Currently, the
following objects can be processed: t_test(), bartlett_test(), lm(), aov(),
manova(), cor_test(), lmer() (only for testing random intercep variances),
glm(), coxph(), survreg(), polr(), zeroinfl(), rma(), ergm(), bergm(), or named
vector objects. In the case \code{x} is a named vector, the arguments \code{Sigma}
and \code{n} are also needed. See vignettes for elaborations.}

\item{hypothesis}{A character string containing the constrained (informative) hypotheses to
evaluate in a confirmatory test. The default is NULL, which will result in standard exploratory testing
under the model \code{x}.}

\item{prior.hyp.explo}{The prior probabilities of the hypotheses in the exploratory tests. Except for
objects of class \code{aov} (for (M)ANOVA, etc.), this argument should be a vector with three
elements reflecting the prior probability of a zero effect, a negative effect, and a positive effect, respectively.
For objects of class \code{aov}, the argument
should be a list where the first element should be a vector of length 3 specifying the prior probabilities of each parameter
being zero, negative, or positive, the second element should be a vector of length 2 specifying the prior probabilities
of a model where is no main effect for a factor and the full model, and the third element should be a vector of length 2
specifying the prior probabilities of a model where is no interaction effect (if present) for two factors and the full model.
The default (\code{NULL}) specifies equal prior probabilities for each hypothesis per exploratory test.}

\item{prior.hyp.conf}{The prior probabilities of the constrained hypotheses in the confirmatory test.}

\item{prior.hyp}{Deprecated. Please use the argument \code{prior.hyp.conf}.}

\item{complement}{a logical specifying whether the complement should be added
to the tested hypothesis under \code{hypothesis}.}

\item{log}{a logical specifying whether the Bayes factors should be computed on a log scale.
Default is \code{FALSE}.}

\item{cov.prob}{coverage probability of the Bayesian credibility interval in the \code{estimates} element of
the output object. The argument is only used for objects for which a Bayesion estimation algorithm is used,
such as a \code{cor_test} or \code{rma.uni} object. The default coverage probability is 0.95.}

\item{Sigma}{An approximate posterior covariance matrix (e.g,. error covariance
matrix) of the parameters of interest. This argument is only required when \code{x}
is a named vector.}

\item{n}{The (effective) sample size that was used to acquire the estimates in the named vector
\code{x} and the error covariance matrix \code{Sigma}. This argument is only required when \code{x}
is a named vector.}

\item{...}{Parameters passed to and from other functions.}

\item{BF.type}{For certain object classes of \code{x}, different types of Bayes factor tests are supported.
This can be specified using this argument. For models of class 'lm' and 't_test', setting this argument to
\code{'FBF'} the fractional Bayes factor (O'Hagan, 1995) is used and setting this argument to
\code{'AFBF'} the adjusted fractional Bayes factor (Mulder, 2014) is used. The default for these model classes
is the fractional Bayes factor. For models of class 'rma.uni' (meta-analyses), 'BF.type' controls the prior
that is used for the test. Certain defaults are provided for a standardized effect (\code{BF.type="stand.effect"}),
a normal prior with mean 0 and sd 0.5 is used, for log odds (\code{BF.type="log.odds"}), a Student prior with mean 0,
scale 2.36, and 13.1 degrees of freedom is used (corresponding to uniform priors on the success probabilities), for
a correlation (\code{BF.type="correlation"}), a logistic prior is used for the Fisher transformed correlation having
scale 0.5 (corresponding to a uniform prior for the correlation in (-1,1)), for a unit-information prior
(\code{BF.type="unit.info"}), the total sample size \code{sum(ni)} is used to scale a normal unit information prior,
and for a manually specified prior \code{BF.type} needs to be an object of class
\code{prior} from the \code{metaBMA} package.}

\item{iter}{Number of iterations that are used to compute the Monte Carlo estimates.
Only used for certain hypothesis tests of class \code{mlm} (multivariate regression) where constraints are formulated
across different dependent variables as well as different predictors.}
}
\value{
The output is an object of class \code{BF}. The object has elements:
\itemize{
\item \code{BFtu_exploratory}: The Bayes factors of the constrained hypotheses against
the unconstrained hypothesis in the exploratory test.
\item \code{BFtu_main} (only for \code{aov} objects with predictors of class \code{factor}):
The Bayes factors of a constrained model where all levels of a \code{factor} are assumed
to have the same effect on the outcome variable versus an unconstrained (full) model with
no constraints.
\item \code{BFtu_interaction} (only for \code{aov} objects with interaction effects with
predictors of class \code{factor}): The Bayes factors of a constrained model where the effect
of the dummy variables corresponding to an interaction effects are assumed to be zero versus
an unconstrained (full) model with no constraints.
\item \code{PHP_exploratory:} The posterior probabilities of the constrained hypotheses
in the exploratory test.
\item \code{PHP_main} (only for \code{aov} objects with predictors of class \code{factor}):
The posterior probabilities a constrained model where all levels of a \code{factor} are assumed
to have the same effect on the outcome variable versus an unconstrained (full) model with
no constraints.
\item \code{PHP_interaction} (only for \code{aov} objects with interaction effects with
predictors of class \code{factor}): The posterior probabilities of a constrained model where the
effect of the dummy variables corresponding to an interaction effects are assumed to be zero versus
an unconstrained (full) model with no constraints.
\item \code{BFtu_confirmatory}: The Bayes factors of the constrained hypotheses against
the unconstrained hypothesis in the confirmatory test using the \code{hypothesis}
argument.
\item \code{PHP_confirmatory}: The posterior probabilities of the constrained hypotheses
in the confirmatory test using the \code{hypothesis} argument.
\item \code{BFmatrix_confirmatory}: The evidence matrix which contains the Bayes factors
between all possible pairs of hypotheses in the confirmatory test.
\item \code{BFtable_confirmatory}: The \code{Specification table} (output when printing the
\code{summary} of a \code{BF} for a confirmatory test) which contains the different
elements of the extended Savage Dickey density ratio where
\itemize{
\item The first column `\code{complex=}' quantifies the relative complexity of the
equality constraints of a hypothesis (the prior density at the equality constraints in the
extended Savage Dickey density ratio).
\item The second column `\code{complex>}' quantifies the relative complexity of the
order constraints of a hypothesis (the prior probability of the order constraints in the extended
Savage Dickey density ratio).
\item The third column `\code{fit=}' quantifies the relative fit of the equality
constraints of a hypothesis (the posterior density at the equality constraints in the extended
Savage Dickey density ratio).
\item The fourth column `\code{fit>}' quantifies the relative fit of the order
constraints of a hypothesis (the posterior probability of the order constraints in the extended
Savage Dickey density ratio)
\item The fifth column `\code{BF=}' contains the Bayes factor of the equality constraints
against the unconstrained hypothesis.
\item The sixth column `\code{BF>}' contains the Bayes factor of the order constraints
against the unconstrained hypothesis.
\item The seventh column `\code{BF}' contains the Bayes factor of the constrained hypothesis
against the unconstrained hypothesis.
\item The eighth column `\code{PHP}' contains the posterior probabilities of the hypotheses.
}
\item \code{prior.hyp.explo}: The prior probabilities of the constrained hypotheses in the exploratory tests.
\item \code{prior.hyp.conf}: The prior probabilities of the constrained hypotheses in the confirmatory test.
\item \code{hypotheses}: The tested constrained hypotheses in a confirmatory test.
\item \code{estimates}: Descriptives of unconstrained estimates based on flat priors (also for \code{rma.uni} objects for
Bayesian meta-analyses).
\item \code{model}: The input model \code{x}.
\item \code{bayesfactor}: The type of Bayes factor that is used for this model.
\item \code{parameter}: The type of parameter that is tested.
\item \code{log}: \code{logical} whether the Bayes factors were reported on a log scale.
\item \code{fraction_number_groupIDs} (only for objects of class \code{lm}): The number of
'group identifiers' that were identified based on the number of unique combinations of \code{level}s
of predictor variables of class \code{factor} in the data. These group identifiers are used to automatically
specify the minimal fractions that are used to compute (adjusted) fractional Bayes factors.
\item \code{fraction_groupID_observations} (only for objects of class \code{lm}): A vector that
specifies to which 'group identifier' an observation belongs. The group identifiers are constructed
based on the unique combination of the \code{levels} based on the predictor variables of class \code{factor}
of the observations.
\item \code{call}: The call of the \code{BF} function.
}
}
\description{
The \code{BF} function can be used for hypothesis testing and
 model selection using the Bayes factor. By default exploratory hypothesis tests are
performed of whether each model parameter equals zero, is negative, or is
 positive. Confirmatory hypothesis tests can be executed by specifying hypotheses with
equality and/or order constraints on the parameters of interest. Depending on the
class of the fitted model different Bayes factors are used as described in the output.
}
\details{
The function requires a fitted modeling object. Current analyses
that are supported: \code{\link[bain]{t_test}},
\code{\link[BFpack]{bartlett_test}},
\code{\link[stats]{aov}}, \code{\link[stats]{manova}},
\code{\link[stats]{lm}}, \code{mlm},
\code{\link[stats]{glm}}, \code{\link[polycor]{hetcor}},
\code{\link[lme4]{lmer}}, \code{\link[survival]{coxph}},
\code{\link[survival]{survreg}}, \code{\link[ergm]{ergm}},
\code{\link[Bergm]{bergm}},
\code{\link[pscl]{zeroinfl}}, \code{\link[metafor]{rma}} and \code{\link[MASS]{polr}}.

For testing parameters from the results of t_test(), lm(), aov(),
manova(), and bartlett_test(), hypothesis testing is done using
adjusted fractional Bayes factors are computed (using minimal fractions).
For testing measures of association (e.g., correlations) via \code{cor_test()},
Bayes factors are computed using joint uniform priors under the correlation
matrices. For testing intraclass correlations (random intercept variances) via
\code{lmer()}, Bayes factors are computed using uniform priors for the intraclass
correlations. For all other tests,  approximate adjusted fractional Bayes factors
(with minimal fractions) are computed using Gaussian approximations, similar as
a classical Wald test.
}
\section{Methods (by class)}{
\itemize{
\item \code{BF(default)}: S3 method for a named vector 'x'

\item \code{BF(lm)}: S3 method for an object of class 'lm'

\item \code{BF(rma.uni)}: BF S3 method for an object of class 'rma.uni'

\item \code{BF(t_test)}: BF S3 method for an object of class 't_test'

}}
\examples{
\donttest{
# EXAMPLE 1. One-sample t test
ttest1 <- t_test(therapeutic, mu = 5)
print(ttest1)
# confirmatory Bayesian one sample t test
BF1 <- BF(ttest1, hypothesis = "mu = 5")
summary(BF1)
# exploratory Bayesian one sample t test
BF(ttest1)

# EXAMPLE 2. ANOVA
aov1 <- aov(price ~ anchor * motivation,data = tvprices)
BF1 <- BF(aov1, hypothesis = "anchorrounded = motivationlow;
                              anchorrounded < motivationlow")
summary(BF1)

# EXAMPLE 3. linear regression
lm1 <- lm(mpg ~ cyl + hp + wt, data = mtcars)
BF(lm1, hypothesis = "wt < cyl < hp = 0")

# EXAMPLE 4. Logistic regression
fit <- glm(sent ~ ztrust + zfWHR + zAfro + glasses + attract + maturity +
   tattoos, family = binomial(), data = wilson)
BF1 <- BF(fit, hypothesis = "ztrust > zfWHR > 0;
                             ztrust > 0 & zfWHR = 0")
summary(BF1)

# EXAMPLE 5. Correlation analysis
set.seed(123)
cor1 <- cor_test(memory[1:20,c(1,2,6)])
BF1 <- BF(cor1)
summary(BF1)
BF2 <- BF(cor1, hypothesis = "Rat_with_Im > Rat_with_Del > 0;
                             Rat_with_Im = Rat_with_Del = 0")
summary(BF2)
# correlations can also be computed between continuous/ordinal variables
memory_test <- memory[1:20,c(1,2,6)]
memory_test[,3] <- as.ordered(memory_test[,3])
cor2 <- cor_test(memory_test)
BF(cor2)

# EXAMPLE 6. Bayes factor testing on a named vector
# A Poisson regression model is used to illustrate the computation
# of Bayes factors with a named vector as input
poisson1 <- glm(formula = breaks ~ wool + tension,
  data = datasets::warpbreaks, family = poisson)
# extract estimates, error covariance matrix, and sample size:
estimates <- poisson1$coefficients
covmatrix <- vcov(poisson1)
samplesize <- nobs(poisson1)
# compute Bayes factors on equal/order constrained hypotheses on coefficients
BF1 <- BF(estimates, Sigma = covmatrix, n = samplesize, hypothesis =
"woolB > tensionM > tensionH; woolB = tensionM = tensionH")
summary(BF1)
}
}
\references{
Mulder, Williams, Gu, Tomarken, Böing-Messing, Olsson-Collentine, Meyerink,
Menke, Fox, Rosseel, Wagenmakers, Hoijtink, and van Lissa (2021). BFpack: Flexible Bayes
Factor Testing of Scientific Theories in R. Journal of Statistical Software, 100.
<https://doi.org/10.18637/jss.v100.i18>

Mulder and Xin (2022). Bayesian Testing of Scientific Expectations under
Multivariate Normal Linear Models. <https://doi.org/10.1080/00273171.2021.1904809>

Mulder and Gelissen (2021). Bayes factor testing of equality and order
constraints on measures of association in social research. Journal of Applied Statistics,
50. <https://doi.org/10.1080/02664763.2021.1992360>

Mulder and Fox (2019). Bayes Factor Testing of Multiple Intraclass Correlations.
Bayesian Analysis, 14. <http://doi.org/10.1214/18-BA1115>

Hoijtink, Mulder, van Lissa, and Gu (2018). A tutorial on testing hypotheses
using the Bayes factor. Psychological Methods, 24(5), 539–556.
<http://doi.org/10.1037/met0000201>

Boeing-Messing, van Assen, Hofman, Hoijtink, and Mulder (2017).
Bayesian evaluation of constrained hypotheses on variances of multiple independent groups.
Psycholological Methods, 22(2), 262-287. <https://doi.org/10.1037/met0000116>

van Aert and Mulder (2021). Bayesian hypothesis testing and estimation under
the marginalized random-effects meta-analysis model. Psychonomic Bulletin and Review, 29,
55–69. <https://doi.org/10.3758/s13423-021-01918-9>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/attention.R
\docType{data}
\name{attention}
\alias{attention}
\title{Multiple Sources of Attentional Dysfunction in Adults With Tourette's Syndrome}
\format{
A data.frame with 51 rows and 2 columns.
}
\usage{
data(attention)
}
\description{
Data from a psychological study comparing attentional performances of
Tourette's syndrome (TS) patients, ADHD patients, and controls.
These data were simulated using the sufficient statistics from Silverstein,
Como, Palumbo, West, and Osborn (1995).
}
\details{
\tabular{lll}{
   \strong{accuracy} \tab \code{numeric} \tab Participant's accuracy in the attentional task\cr
   \strong{group} \tab \code{factor} \tab Participant's group membership (TS patient, ADHD patient, or control)\cr
}
}
\references{
Silverstein, S. M., Como, P. G., Palumbo, D. R., West, L. L., & Osborn, L. M. (1995). Multiple sources of attentional dysfunction in adults with Tourette's syndrome: Comparison with attention deficit-hyperactivity disorder. Neuropsychology, 9(2), 157-164. doi:10.1037/0894-4105.9.2.157
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/actors.R
\docType{data}
\name{actors}
\alias{actors}
\title{Actors from a small hypothetical network}
\format{
dataframe (25 rows, 4 columns)

\tabular{lll}{
   \strong{actors$id} \tab \code{integer} \tab ID of the employee, corresponding to
   the sender and receiver IDs in the events dataframe \cr
   \strong{actors$location} \tab \code{numeric} \tab Location of the actor,
   ranging from 1-4 \cr
   \strong{actors$culture} \tab \code{character} \tab Categorical variable, indicating the
   culture of the employee \cr
}
}
\usage{
data(actors)
}
\description{
The related data files 'events', 'same_location', 'same_culture' contain
information on the  event sequence and the two event statistics respectively.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sivan.R
\docType{data}
\name{sivan}
\alias{sivan}
\title{Wason task performance and morality}
\format{
A data.frame with 887 rows and 12 columns.
}
\usage{
data(sivan)
}
\description{
Data from an experimental study, using the Wason selection task (Wason 1968)
to examine whether humans have cognitive adaptations for detecting violations
of rules in multiple moral domains. Moral domains are operationalized in
terms of the five domains of the Moral Foundations Questionnaire
(Graham et al. 2011).
These data were simulated using the
R-package \code{synthpop}, based on the characteristics of the original data.
}
\details{
\tabular{lll}{
   \strong{sex} \tab \code{factor} \tab Participant sex\cr
   \strong{age} \tab \code{integer} \tab Participant age\cr
   \strong{nationality} \tab \code{factor} \tab Participant nationality\cr
   \strong{politics} \tab \code{integer} \tab How would you define your political opinions? Likert type scale, from 1 (Liberal) to 6 (Conservative)\cr
   \strong{WasonOrder} \tab \code{factor} \tab Was the Wason task presented before, or after the MFQ? \cr
   \strong{Harm} \tab \code{numeric} \tab MFQ harm domain.\cr
   \strong{Fairness} \tab \code{numeric} \tab MFQ fairness domain.\cr
   \strong{Loyalty} \tab \code{numeric} \tab MFQ loyalty domain.\cr
   \strong{Purity} \tab \code{numeric} \tab MFQ purity domain.\cr
   \strong{Tasktype} \tab \code{ordered} \tab How was the Wason task framed?\cr
   \strong{GotRight} \tab \code{factor} \tab Did the participant give the correct answer to the Wason task?
}
}
\references{
Sivan, J., Curry, O. S., & Van Lissa, C. J. (2018). Excavating the Foundations: Cognitive Adaptations for Multiple Moral Domains. Evolutionary Psychological Science, 4(4), 408–419. doi:10.1007/s40806-018-0154-8
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BF.var.R
\name{bartlett_test}
\alias{bartlett_test}
\alias{bartlett_test.default}
\title{Bartlett Test of Homogeneity of Variances}
\usage{
bartlett_test(x, g, ...)

\method{bartlett_test}{default}(x, g, ...)
}
\arguments{
\item{x}{a numeric vector of data values, or a list of
numeric data vectors representing the respective samples,
or fitted linear model objects (inheriting from class "lm").}

\item{g}{a vector or factor object giving the group for
the corresponding elements of x. Ignored if x is a list.}

\item{...}{further arguments to be passed to or from methods.}
}
\value{
A list with class \code{"bartlett_htest"} containing the following
components: \item{statistic}{Bartlett's K-squared test statistic.}
\item{parameter}{the degrees of freedom of the approximate chi-squared
distribution of the test statistic.}
\item{p.value}{the p-value of the test.} \item{conf.int}{a confidence
interval for the mean appropriate to the specified alternative hypothesis.}
\item{method}{the character string "Bartlett test of homogeneity of variances".}
\item{data.name}{a character string giving the names of the data.}
\item{vars}{the sample variances across groups (samples).}
\item{n}{the number of observations per group (sample)}
}
\description{
Performs Bartlett's test of the null that the
variances in each of the groups (samples) are the same.
}
\details{
\code{x} must be a numeric data vector, and \code{g}
must be a vector or factor object of the same length as \code{x}
giving the group for the corresponding elements of \code{x}.
}
\section{Bain t_test}{

In order to allow users to enjoy the functionality of bain with the familiar
stats-function \code{bartlett.test}, we have had to make minor changes to the
function \code{bartlett.test.default}. All rights to, and credit for, the
function \code{bartlett.test.default}
belong to the R Core Team, as indicated in the original license below.
We make no claims to copyright and incur no liability with regard to the
changes implemented in \code{bartlett_test}.

This the original copyright notice by the R core team:
File src/library/stats/R/bartlett_test.R
Part of the R package, https://www.R-project.org

Copyright (C) 1995-2015 The R Core Team

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
https://www.R-project.org/Licenses/
}

\examples{
require(graphics)

plot(count ~ spray, data = InsectSprays)
bartlett_test(InsectSprays$count, InsectSprays$spray)

}
\references{
Bartlett, M. S. (1937). Properties of sufficiency
and statistical tests. Proceedings of the Royal Society of
London Series A 160, 268–282. DOI: 10.1098/rspa.1937.0109.
}
